c     ==============================================================
c              MAIN PROGRAM FOR SELF-CONSISTENT DISK SOLVER
c
c     This program acts as a shell around a set of three independent
c     programs for computing the radiative transfer and vertical
c     structure in an irradiated circumstellar disk. The three 
c     programs are:
c
c      1a) radialrt.F      Transfer of direct starlight into disk
c                          along radial rays from the star to every
c                          location in the disk.
c      1b) radalpha.F      Same functionality as radialrt, but with
c                          different transfer prescription (it uses
c                          the grazing angle prescription).
c      1c) radmonte.F      Same functionality as radalpha, but then 
c                          with scattering included, using a Monte 
c                          Carlo recipe.
c
c      2)  vertrt.F        Vertical radiative transfer of reprocessed
c                          radiation from the disk's surface. 
c
c      3a) vertstruct.F    Vertical integration of the hydrostatic
c                          equations for the disk, from the computed
c                          temperatures from the above RT programs.
ccccc      3b) hydro.F         Vertical hydrodynamics program as a 
ccccc                          replacement of the vertstruct.F
c
c     Each of these programs reads and writes a file with the same
c     format: five colums with R[cm],Z[cm],RHO[g/cm^3],T[K],vz[cm/s].
c     These files are called diskstruct_0.dat ... diskstruct_3.dat,
c     and are the intermediate files between the three programs.
c     
c     The present main program calls the above three programs in 
c     the proper order, and repeats this until convergence is
c     reached.
c     ==============================================================

#include "main.h"

      program disk
      implicit none
c
      doubleprecision convcrit,dum,error,a,b,c,d,e
      integer nr,nz,iformat,itermax,iz,ir,iter,idum
      logical f_ex
      character*800 base
      integer index,srclen,iradrt,ivertrt,time_nr_save
      character*800 ext
      character*800 filename,command,srcdir
      doubleprecision flareindex(FRSIZE_R)
      doubleprecision visc,zcov,hpdr,hpdrmin,hs(FRSIZE_R)
      doubleprecision frkep,damping,time,timemax,dtime,mstar,omk
      doubleprecision timesave,tsave,iterflaredefault
      integer iterflaremax,iterflare,itertot,itodo,itime,itimemax
      integer itsave,isave,irstart,inice
c
#include "common_grid.h"
#include "common_disk.h"
c
c     Remove the disk.success file
c
      call system('rm -f disk.success')
c
c     Source directory
c
      srcdir='/Users/cornelisdullemond/science/'//
     %        'prog/diskstruct/VERSION_RELEASE_AUG_10/'//
     %        'diskstruct_release_28.08.10/'
      srclen=len_trim(srcdir)
c
c     Read the input file
c
      visc = 0.d0
      inice = 0
      iterflaredefault = 4
      inquire(file='disk.inp',exist=f_ex)
      if(f_ex) then 
          open(unit=1,file='disk.inp')
          read(1,*) itodo        ! What to do?
c
c            itodo = 1    -->  Find static solution
c            itodo = 2    -->  Do hydrodynamics, start from diskstruct.inp
c            itodo = 3    -->  Do hydrodynamics, start from diskstruct_3.inp
c
          if(itodo.eq.1) then
c
c             Static solutions
c
              read(1,*) itermax  ! Maximum nr of iterations
              read(1,*) convcrit ! Convergence criterion
              read(1,*) zcov     ! Zcover factor (for vertstruct.F)
              read(1,*) visc     ! Viscosity parameter
              read(1,*) iradrt   ! Type of algorithm for radial rt
c
c                iradrt  = 0   --> [radialrt/]  Real radial transfer       
c                iradrt  = 1,2   : [radalpha/]  Flaring index transfer
c                        = 1   -->                Fixed flare idx
c                        = 2   -->                Iterated flare idx
c                iradrt  = 3,4   : [radmonte/]    Monte-Carlo with Scattering
c                        = 3   -->                Fixed flare idx
c                        = 4   -->                Iterated flare idx
c
              read(1,*) ivertrt  ! Type of algorithm for vertical rt
c
c                ivertrt = 0   --> [vertrt/]    Moment Eqs Mean Opacities
c                        > 0   --> [fullvrt/]   Full Freq-Angle Dep Transfer

              read(1,*,end=44) inice  ! Nice the job?
              read(1,*,end=44) iterflaredefault ! How many flidx iterations?
 44           continue
          endif
          if((itodo.eq.2).or.(itodo.eq.3)) then
             write(*,*) 'HYDRO IS HERE DEACTIVATED...'
             stop
c
c             Hydrodynamic simulations
c
              read(1,*) itimemax ! Max nr of time steps
              read(1,*) timemax  ! Max time span
              read(1,*) irstart  ! Do hydro starting from irstart onwards
              read(1,*) frkep    ! Time step = frkep of Kepler at inner radius
              read(1,*) damping  ! Damping factor
              read(1,*) hpdrmin  ! Stop simulation if Hp/R < hpdrmin
              read(1,*) visc     ! Viscosity parameter
              read(1,*) iradrt   ! Type of algorithm for radial rt
c
c                iradrt  = 0   --> [radialrt/]  Real radial transfer       
c                iradrt  = 1,2   : [radalpha/]  Flaring index transfer
c                        = 1   -->                Fixed flare idx
c                        = 2   -->                Iterated flare idx
c                iradrt  = 3,4   : [radmonte/]    Monte Carlo with Scattering
c                        = 3   -->                Fixed flare idx
c                        = 4   -->                Iterated flare idx
c
              read(1,*) ivertrt  ! Type of algorithm for vertical rt
c
c                ivertrt = 0   --> [vertrt/]    Moment Eqs Mean Opacities
c                        > 0   --> [fullvrt/]   Full Freq-Angle Dep Transfer

              read(1,*) isave    ! Nr of time steps for save
              read(1,*) tsave    ! Time interval for save
              read(1,*,end=45) inice  ! Nice the job?
 45           continue
              timemax = timemax * 31536000.
              tsave   = tsave * 31536000.
          endif
          close(1)
      else
          itodo=1
          itermax=5
          convcrit=0.01
      endif
c
c     Do the 'nice' thing
c
      if(inice.ne.0) then
          srcdir='nice '//srcdir(1:srclen)
          srclen=srclen+5
      endif
c
c     Check if some files are there
c
      if(ivertrt.eq.0) then
          inquire(file='dustopac_1.inp',exist=f_ex)
          if(.not.f_ex) then
              inquire(file='planckopac.inp',exist=f_ex)
              if(.not.f_ex) then
                  write(*,*) 'ERROR: Need planckopac.inp for vertrt.F'
                  stop 13
              endif
              inquire(file='rossopac.inp',exist=f_ex)
              if(.not.f_ex) then
                  write(*,*) 'ERROR: Need rossopac.inp for vertrt.F'
                  stop 13
              endif
          endif
      endif
c
c     Make a warning if the artificial inner wall is included in the
c     computation. It is fine to use this inner wall, but one should
c     be very aware of this, since the ir.eq.1 shell is used for this
c     inner wall, and therefore has a different meaning.
c
      inquire(file='flareindex.inp',exist=f_ex)
      if(f_ex) then 
          open(unit=1,file='flareindex.inp')
          read(1,*) idum
          read(1,*) dum
          if(dum.lt.-1d4) then
              write(*,*) '========================================'
              write(*,*) 'NOTE: ARTIFICIAL INNER WALL ACTIVATED!!!'
              write(*,*) '========================================'
          endif
          close(1)
      endif
c
c     Pass the viscosity parameter to the vertical transfer program
c
      open(unit=1,file='viscosity.inp',status='unknown')
      write(1,*) visc
      close(1)
c
c     Pass the z coverage factor to the vertical hydrostatics
c
c      open(unit=1,file='zcover.inp',status='unknown')
c      write(1,*) zcov
c      close(1)
c
c     Read the starinfo file
c
      open(unit=1,file='starinfo.inp',status='old')
      read(1,*) iformat
      read(1,*) dum
      read(1,*) mstar
      close(1)      
c
c     Read the initial guess for the density
c
      if((itodo.eq.1).or.(itodo.eq.2)) then
          open(unit=1,file='diskstruct.inp',status='old')
      elseif(itodo.eq.3) then
          open(unit=1,file='diskstruct_3.inp',status='old')
      else
          stop 534
      endif
      read(1,*) nr,nz
      if(nr.gt.FRSIZE_R) then
          write(*,*) 'ERROR in diskstruct:'
          write(*,*) 'R grid too large'
          stop 13
      endif
      if(nz.gt.FRSIZE_Z) then
          write(*,*) 'ERROR in diskstruct:'
          write(*,*) 'Z grid too large'
          stop 13
      endif
      do ir=1,nr
          do iz=1,nz
              read(1,*) a,b,c,d,e
              grid_r(ir) = a
              grid_z(iz,ir) = b
              ds_rho(iz,ir) = c
              ds_vz(iz,ir) = e
          enddo
      enddo
      close(1)
      grid_nr = nr
      grid_nz = nz
c
c     Copy this initial guess to the file diskstruct_1.dat
c
      open(unit=1,file='diskstruct_0.dat',status='unknown')
      write(1,*) grid_nr,grid_nz
      write(1,*)
      do ir=1,grid_nr
          do iz=1,grid_nz
              write(1,10) grid_r(ir),grid_z(iz,ir),ds_rho(iz,ir),
     %               0.d0,ds_vz(iz,ir)
 10           format(5(E13.6,1X))
          enddo
      enddo
      close(1)
c
c     Now the main loop
c
      if(itodo.eq.1) then
c
c         Do the static solution....
c
          if((iradrt.eq.2).or.(iradrt.eq.4)) then
              iterflaremax=iterflaredefault
              if(iterflaredefault.eq.4) then ! Default value
                  write(*,*) 'NOTE: Using only 4 flaring ',
     %                     'index iterations!'
              endif
          else
              iterflaremax=1
          endif
c
c         Loop over flaring index
c
          itertot = 1
          do iterflare=1,iterflaremax
c
              if(iterflare.gt.1) then
                  write(*,*) '==================',
     %                         '========================='
                  write(*,*) '    Flaring index recomputed: new run'
                  write(*,*) '==================',
     %                         '========================='
              endif
c
c             Do the iteration over the disk structure, for the current 
c             flaring index array
c     
              do iter=1,itermax
c
c                 Message
c
                  write(*,*) 'Main Iteration ',iter
c
c                 First the radial transfer
c
                  call system('rm -f diskstruct_1.dat')
                  if(iradrt.eq.0) then
                      command = srcdir(1:srclen)//'radialrt/radialrt' 
                      write(*,*) '.....radialrt'
                  elseif(iradrt.le.2) then
                      command = srcdir(1:srclen)//'radalpha/radalpha' 
                      write(*,*) '.....radalpha'
                  else
                      command = srcdir(1:srclen)//'radmonte/radmonte' 
                      write(*,*) '.....radmonte'
                  endif
                  call system(command)
                  inquire(file='diskstruct_1.dat',exist=f_ex)
                  if(.not.f_ex) then
                      write(*,*) '  MAJOR ERROR: Radial transfer failed'
                      stop 13
                  endif
c
c                 Then the vertical transfer
c
                  call system('rm -f diskstruct_2.dat')
                  if(ivertrt.eq.0) then
                      command = srcdir(1:srclen)//'vertrt/vertrt'
                      write(*,*) '.....vertrt'
                      call system(command)
                  else
                      command = srcdir(1:srclen)//'fullvrt/fullvrt'
                      write(*,*) '.....fullvrt'
                      call system(command)
                  endif
                  inquire(file='diskstruct_2.dat',exist=f_ex)
                  if(.not.f_ex) then
                      write(*,*) '  MAJOR ERROR: ',
     %                     'Vertical transfer failed'
                      stop 13
                  endif
c
c                 Then the vertical structure
c
                  call system('rm -f diskstruct_3.dat')
                  command = srcdir(1:srclen)//
     %                'vertstruct/vertstruct'
                  write(*,*) '.....vertstruct'
                  call system(command)
                  inquire(file='diskstruct_3.dat',exist=f_ex)
                  if(.not.f_ex) then
                      write(*,*) '  MAJOR ERROR: ',
     %                     'Vertical pressure balance failed'
                      stop 13
                  endif
c     
c                 Copy thing to history files
c
cc                  base='diskstruct_hist_'
cc                  index=itertot
cc                  ext='.dat'
cc                  call make_indexed_filename(base,index,ext,filename)
cc                  command='cp diskstruct_3.dat '//filename
cc                  call system(command)
                  open(unit=1,file='diskstruct.info',status='unknown')
                  write(1,*) itertot
                  close(1)
                  itertot = itertot+1
c
c                 Now compare the temperatures with the old ones
c
                  call compare(error,ir,iz)
c
c                 Check error
c
                  if(error.lt.convcrit) goto 100
c
c                 Copy latest result to input
c
                  call system('cp diskstruct_3.dat diskstruct_0.dat')
c
              enddo
c
              write(*,*) '********* MAIN DISK LOOP FAILED *********'
              write(*,77) itermax
 77           FORMAT('********* AFTER ',I4,' ITERATIONS *********')
              write(*,78) ir,iz
 78           FORMAT('********* ERROR LOCATION ',I3,I4,' *********')
              goto 120
 100          continue
              write(*,*) '========= MAIN DISK LOOP SUCCESS ========='
 120          continue
c
c             Determine the new flaring index/angle. We must make sure that
c             any perturbation will flow off the grid, so we must use, at
c             radial index ir, the flaring index of ir-1 or ir-2 (odd/even). 
c
              if((iradrt.eq.2).or.(iradrt.eq.4)) then
                  call find_flareindex(flareindex,hs)
                  call write_flareindex(flareindex,grid_nr)
                  call write_hs(hs,grid_nr)
                  base='flareindex_hist_'
                  index=iterflare
                  ext='.dat'
                  call make_indexed_filename(base,index,
     %                 ext,filename)
                  call system('cp -f flareindex.inp '//filename)
              endif
c
          enddo
      else
c
c         Do the hydrodynamics...
c
         write(*,*) 'HYDRO IS HERE DEACTIVATED...'
         stop
          open(2,file='collapse.dat',status='unknown')
          write(2,*) 0,10.d0
          close(2)
c
c         Find the Kepler time of the innermost radius at which
c         the hydrodynamics is done
c
          omk   = sqrt(6.672d-8 * mstar / grid_r(irstart)**3)
          dtime = frkep / omk
          time  = 0.d0
          timesave = dtime
          itsave = 1
          time_nr_save = 1
c
c         Write hydro input file
c
          open(unit=2,file='hydro.inp',status='unknown')
          write(2,*) dtime 
          write(2,*) 1000
          write(2,*) damping
          write(2,*) irstart
          close(2)
c
c         Clear time file
c
          open(unit=1,file='time.dat',status='unknown')
          write(1,*) ' '
          close(1)
          open(unit=1,file='diskstruct.info',status='unknown')
          write(1,*) 0
          close(1)
c
c         Do the loop
c
          do itime=1,itimemax
c
c             Message
c
              write(*,*) 'Time step .... ',itime
c
c             First the radial transfer
c
              if(iradrt.eq.0) then
                  command = srcdir(1:srclen)//'radialrt/radialrt' 
                  write(*,*) '.....radialrt'
              elseif(iradrt.le.2) then
                  write(*,*) 'For hydrodynamics only radialrt allowed'
                  write(*,*) 'for the moment.'
                  stop 888
                  command = srcdir(1:srclen)//'radalpha/radalpha' 
                  write(*,*) '.....radalpha'
              else
                  write(*,*) 'For hydrodynamics only radialrt allowed'
                  write(*,*) 'for the moment.'
                  stop 888
                  command = srcdir(1:srclen)//'radmonte/radmonte' 
                  write(*,*) '.....radmonte'
              endif
              call system(command)
c
c             Then the vertical transfer
c             
              command = srcdir(1:srclen)//'vertrt/vertrt'
              write(*,*) '.....vertrt'
              call system(command)
c             
c             Then the vertical structure
c             
              command = srcdir(1:srclen)//
     %            'hydro/hydro'
              write(*,*) '.....hydro'
              call system(command)
c             
c             Now check if parts of the disk have collapsed
c             
              call check_collapse(mstar,ir,hpdr)
              if(hpdr.lt.hpdrmin) then
                  open(2,file='collapse.dat',status='unknown')
                  write(2,*) ir,hpdr
                  close(2)
                  write(*,*) '******* COLLAPSED *******'
                  goto 20
              endif
c             
c             Copy latest result to input
c             
              call system('cp diskstruct_3.dat diskstruct_0.dat')
c
c             Check the time
c     
              if(time+dtime.gt.timemax) then 
                  dtime = timemax-time
                  if(dtime.lt.0.d0) stop 629
              endif
c
c             Update the time
c
              time = time + dtime
c         
c             Then save when necessary
c
              if((itsave.ge.isave).or.
     %             (timesave.ge.tsave)) then
                  write(*,*) 'Saving data'
                  base='diskstruct_hist_'
                  index=time_nr_save
                  ext='.dat'
                  call make_indexed_filename(base,index,ext,filename)
                  command='cp diskstruct_3.dat '//filename
                  call system(command)
                  open(unit=1,file='time.dat',status='unknown',
     %                      access='append')
                  write(1,*) time
                  close(1)
                  open(unit=1,file='diskstruct.info',status='unknown')
                  write(1,*) time_nr_save
                  close(1)
                  time_nr_save = time_nr_save + 1
                  timesave     = 0.d0
                  itsave       = 0
              endif
              timesave = timesave + dtime
              itsave   = itsave+1
c     
c             Check the time once again
c
              if(time.ge.0.99999d0*timemax) goto 20
c
          enddo
c
c         Problem... Time not reached before nr of iterations finished
c
          write(*,*) '  MAJOR PROBLEM in main routine: '
          write(*,*) '  Nr of time steps exceeds maximum'
          stop 13
c
 20       continue
      endif
c
c     Write the disk.success file
c
      open(unit=1,file='disk.success',status='unknown')
      write(1,*) 1
      close(1)
c     
      end



c     --------------------------------------------------------------
c                           CHECK ERROR 
c     --------------------------------------------------------------
      subroutine compare(error,ir_er,iz_er)
      implicit none
      doubleprecision error
c
#include "common_grid.h"
#include "common_disk.h"
c
      doubleprecision errcur,a,b,c,d,e
      integer ir,iz,nz,nr,iz_er,ir_er
c
      do ir=1,grid_nr
          do iz=1,grid_nz
              ds_rhoold(iz,ir) = ds_rho(iz,ir) 
          enddo
      enddo
c
      open(unit=1,file='diskstruct_3.dat',status='old')
      read(1,*) nr,nz
      if(nr.ne.grid_nr) then
          write(*,*) 'ERROR in diskstruct:'
          write(*,*) 'R grid not consistent'
          stop 13
      endif
      if(nz.ne.grid_nz) then
          write(*,*) 'ERROR in diskstruct:'
          write(*,*) 'Z grid not consistent'
          stop 13
      endif
      do ir=1,grid_nr
          do iz=1,grid_nz
              read(1,*) a,b,c,d,e
              ds_rho(iz,ir) = c
          enddo
      enddo
      close(1)
c
      error = 0.d0
      do ir=1,grid_nr
          do iz=1,grid_nz
              errcur=abs(ds_rhoold(iz,ir)-ds_rho(iz,ir))/
     %             (ds_rhoold(iz,ir)+ds_rho(iz,ir))
              if(errcur.gt.error) then
                  error=errcur
                  ir_er=ir
                  iz_er=iz
              endif
          enddo
      enddo
c
      end



c     --------------------------------------------------------------
c                 CHECK IF SOLUTION HAS COLLAPSED OR NOT
c     --------------------------------------------------------------
      subroutine check_collapse(mstar,irmin,hpdrmin)
      implicit none
      doubleprecision hpdrmin,mstar
      integer irmin
c
#include "common_grid.h"
#include "common_disk.h"
c
      doubleprecision errcur,a,b,c,d,e,mugas,omk2,hpdr
      parameter(mugas=MUMOLHYD)
      integer ir,iz,nz,nr
c
      irmin   = 0 
      hpdrmin = 1.d33
      open(unit=1,file='diskstruct_3.dat',status='old')
      read(1,*) nr,nz
      if(nr.ne.grid_nr) then
          write(*,*) 'ERROR in diskstruct:'
          write(*,*) 'R grid not consistent'
          stop 13
      endif
      if(nz.ne.grid_nz) then
          write(*,*) 'ERROR in diskstruct:'
          write(*,*) 'Z grid not consistent'
          stop 13
      endif
      do ir=1,grid_nr
          do iz=1,grid_nz
              omk2 = 6.672d-8 * mstar / grid_r(ir)**3
              read(1,*) a,b,c,d,e
              if(iz.eq.1) then
                  hpdr = sqrt ( 8.2548d7 * d / ( mugas * omk2 ) )
              endif
          enddo
          if((hpdr.lt.hpdrmin).and.(ir.ge.2)) then
              hpdrmin = hpdr
              irmin   = ir
          endif
      enddo
      close(1)
c
      end



c     --------------------------------------------------------------
c                      MAKE INDEXED FILENAME
c
c     This routine makes filenames of the type dusttemp_4.dat etc.
c     --------------------------------------------------------------
      subroutine make_indexed_filename(base,index,ext,filename)
      implicit none
      character*80 base
      integer index
      character*80 ext
      character*80 filename
      character*12 ch
c     
      if((index.lt.0).or.(index.ge.1000)) then
          write(*,*) 'ERROR in make_indexed_filename()'
          stop 729
      endif
      if(index.lt.10) then
          write(ch,11) index
 11       format(I1)
      elseif(index.lt.100) then
          write(ch,12) index
 12       format(I2)
      elseif(index.lt.1000) then
          write(ch,13) index
 13       format(I3)
      endif
      filename = base(1:lnblnk(base))//
     %           ch(1:lnblnk(ch))//
     %           ext(1:lnblnk(ext))
c
      end



c     --------------------------------------------------------------
c                      FIND THE FLARING INDEX
c     --------------------------------------------------------------
      subroutine find_flareindex(flareindex,ztau)
      implicit none
      doubleprecision flareindex(FRSIZE_R)
      integer ir,iz,nr,nz,m,iswitch,ioffset
      doubleprecision a,b,c,d,dztau,ztau(FRSIZE_R),tmax,ttau
      logical fex
c
#include "common_grid.h"
#include "common_disk.h"      
c
c     Read the flareindex for first two points
c     
      inquire(file='flareindex.inp',exist=fex)
      if(fex) then
c
c         If the file exists, then read the flaring index from this file
c     
          open(unit=1,file='flareindex.inp',status='old')
          read(1,*) nr
          read(1,*) flareindex(1)
          if(nr.ge.2) then
              read(1,*) flareindex(2)
          endif
          if((nr.ge.3).and.(flareindex(1).lt.-1d4)) then
              read(1,*) flareindex(3)
          endif
          close(1)
      else
c
c         Else make an educated guess
c
          flareindex(1) = 2./7.
          if(FRSIZE_R.gt.1) flareindex(2) = 2./7.
      endif
c
c     Read the temperature structure
c
      open(unit=1,file='diskstruct_3.dat',status='old')
      read(1,*) nr,nz
      if(nr.ne.grid_nr) then
          write(*,*) 'ERROR in diskstruct:'
          write(*,*) 'R grid not consistent'
          stop 13
      endif
      if(nz.ne.grid_nz) then
          write(*,*) 'ERROR in diskstruct:'
          write(*,*) 'Z grid not consistent'
          stop 13
      endif
      do ir=1,grid_nr
          do iz=1,grid_nz
              read(1,*) a,b,c,d
              ds_temp(iz,ir) = d
          enddo
      enddo
      close(1)
c
c     Find the tau=1 surface: tau=1 means T=Tmax*0.7788
c     
      m=4
      do ir=1,grid_nr
          tmax = ds_temp(grid_nz,ir)
          ttau = tmax*0.7788
          call hunt(ds_temp(1,ir),grid_nz,ttau,iz)
          if((iz.eq.0).or.(iz.eq.grid_nz)) then
              write(*,*) 'SORRY: Could not find the tau=1 surface'
              write(*,*) '       at radius ir=',ir
              stop 13
          endif
          iz = min(max(iz-(m-1)/2,1),grid_nz+1-m)
          call polint(ds_temp(iz,ir),grid_z(iz,ir),m,ttau,
     %                  ztau(ir),dztau)
      enddo
c
c     Now determine the flaring index, and shift them 2 points, and
c     compute the index only pairwise (see Chiang et al 2001 appendix).
c
      iswitch=0
      ioffset=0
      if(flareindex(1).lt.-1d4) then 
          ioffset=1
      endif
      do ir=2+ioffset,grid_nr-1
          if(iswitch.eq.0) then
              flareindex(ir+1) =
     %           ((ztau(ir)/grid_r(ir)-ztau(ir-1)/grid_r(ir-1))/
     %           (log(grid_r(ir))-log(grid_r(ir-1))))/
     %           (0.5*(ztau(ir)/grid_r(ir)+ztau(ir-1)/grid_r(ir-1)))
              iswitch=1
          else
              flareindex(ir+1) = flareindex(ir)
              iswitch=0
          endif
      enddo
c
c     Check if the results are okay. First we check the first 6 points.
c     A negative flare index can happen as a result of a wrong guess
c     for the initial flaring index (index 1 and 2). So we are a bit
c     forgiving here.
c
      do ir=1+ioffset,min(6,grid_nr)
          if(flareindex(ir).lt.3.d-2) then
              flareindex(ir) = 3.d-2
              write(*,*) '**************************************'
              write(*,*) '  WARNING: NEG FLARING INDEX IR=',ir
              write(*,*) '**************************************'
          endif
      enddo
c
c     Now check the entire thing, and make error when flareindex lt 0
c
      do ir=1+ioffset,grid_nr
          if(flareindex(ir).lt.0.d0) then
              write(*,*) 'ERROR: Flaring index negative!!'
              write(*,*) '    at ir = ',ir
              stop 13
          endif
      enddo
c
      end


c     --------------------------------------------------------------
c                      WRITE THE FLARING INDEX
c     --------------------------------------------------------------
      subroutine write_flareindex(flareindex,nr)
      implicit none
      integer nr
      doubleprecision flareindex(nr)
      integer ir
c     
      open(unit=1,file='flareindex.inp',status='unknown')
      write(1,*) nr
      do ir=1,nr
          write(1,10) flareindex(ir)
 10       format(E13.6)
      enddo
      close(1)
      end


c     --------------------------------------------------------------
c                      WRITE THE HS
c     --------------------------------------------------------------
      subroutine write_hs(hs,nr)
      implicit none
      integer nr
      doubleprecision hs(nr)
      integer ir
c     
      open(unit=1,file='surfheight.dat',status='unknown')
      write(1,*) nr
      do ir=1,nr
          write(1,10) hs(ir)
 10       format(E13.6)
      enddo
      close(1)
      end


c     -------------------------------------------------------------
c               HUNT FOR LOCATION OF A NUMBER IN A TABLE
c     -------------------------------------------------------------
      subroutine hunt(xx,n,x,jlo)
      integer jlo,n
      doubleprecision x,xx(n)
      integer inc,jhi,jm
      logical ascnd
      ascnd=xx(n).gt.xx(1)
      if(jlo.le.0.or.jlo.gt.n)then
        jlo=0
        jhi=n+1
        goto 3
      endif
      inc=1
      if(x.ge.xx(jlo).eqv.ascnd)then
1       jhi=jlo+inc
        if(jhi.gt.n)then
          jhi=n+1
        else if(x.ge.xx(jhi).eqv.ascnd)then
          jlo=jhi
          inc=inc+inc
          goto 1
        endif
      else
        jhi=jlo
2       jlo=jhi-inc
        if(jlo.lt.1)then
          jlo=0
        else if(x.lt.xx(jlo).eqv.ascnd)then
          jhi=jlo
          inc=inc+inc
          goto 2
        endif
      endif
3     if(jhi-jlo.eq.1)return
      jm=(jhi+jlo)/2
      if(x.gt.xx(jm).eqv.ascnd)then
        jlo=jm
      else
        jhi=jm
      endif
      goto 3
      end
C  (C) Copr. 1986-92 Numerical Recipes Software =v1.9"217..

c     -------------------------------------------------------------
c                   POLYNOMIAL INTERPOLATION (NRECIP)
c     -------------------------------------------------------------
      SUBROUTINE polint(xa,ya,n,x,y,dy)
      INTEGER n,NMAX
      DOUBLEPRECISION dy,x,y,xa(n),ya(n)
      PARAMETER (NMAX=10)
      INTEGER i,m,ns
      DOUBLEPRECISION den,dif,dift,ho,hp,w,c(NMAX),d(NMAX)
      ns=1
      dif=abs(x-xa(1))
      do 11 i=1,n
        dift=abs(x-xa(i))
        if (dift.lt.dif) then
          ns=i
          dif=dift
        endif
        c(i)=ya(i)
        d(i)=ya(i)
11    continue
      y=ya(ns)
      ns=ns-1
      do 13 m=1,n-1
        do 12 i=1,n-m
          ho=xa(i)-x
          hp=xa(i+m)-x
          w=c(i+1)-d(i)
          den=ho-hp
          if(den.eq.0.)stop 'failure in polint'
          den=w/den
          d(i)=hp*den
          c(i)=ho*den
12      continue
        if (2*ns.lt.n-m)then
          dy=c(ns+1)
        else
          dy=d(ns)
          ns=ns-1
        endif
        y=y+dy
13    continue
      return
      END
C  (C) Copr. 1986-92 Numerical Recipes Software =v1.9"217..
