c     ==============================================================
c               ADD-ON MODULE CONTAINING ALI ROUTINES
c
c     This module is used by the lambdaiter.F module, and contains
c     routines relevant to the accelerated lambda iteration method.
c
c     ==============================================================

#include "main.h"

#define TAU_EPS 1.d-4

c     --------------------------------------------------------------
c                  SOLVE THE LOCAL DUST-ALI EQUATION
c
c     This is the core routine for the local non-linearized dust
c     ALI accelerator. It works both in the usual mode and in the
c     frequency-integrated mode. This routine is completely self-
c     contained (no common blocks).
c
c     This routine can be called once the complete Lambda Iteration
c     procedure for a single iteration has been completed. At the
c     end of this iteration, the solve_local_dust_ali() can then
c     be solved 
c
c     ARGUMENTS:
c       nf            Nr of frequencies
c       lambdastar    The diagonal approx operator (see calc_ali_accell)
c       kappaabs      The absorption opacity
c       epsilon       The ratio kappa_abs/(kappa_scat+kappa_abs)
c       q             The external source (re-emitted photons)
c       tempold       The temperature before applying LI
c       meanintold    The mean intensity before applying LI
c       temp          The temperature from Lambda Iteration
c       meanint       The mean intensity from Lambda Iteration
c
c     RESULTS:
c       temp          The temperature from Accelerated Lambda Iteration
c       meanint       The mean intensity from Accelerated Lambda Iteration
c       iwarn         The return of dust_warn_zero_temp, i.e. the 
c                     counter of how many error: zero dust temp
c
c     --------------------------------------------------------------
      subroutine solve_local_dust_ali(nf,nu,lambdastar,kappaabs,
     %             epsilon,q,tempold,meanintold,temp,meanint,iwarn)
      implicit none
      integer nf,iwarn
      doubleprecision lambdastar(nf),epsilon(nf),temp,q
      doubleprecision meanint(nf),kappaabs(nf)
      doubleprecision meanintold(nf),tempold
      doubleprecision nu(nf)
c
      common/dustalisolve/dae_meanint,dae_lambdastar,dae_epsilon,
     %       dae_kappaabs,dae_rhs,dae_q,dae_nu
      doubleprecision dae_meanint(FRSIZE_FREQ)
      doubleprecision dae_lambdastar(FRSIZE_FREQ)
      doubleprecision dae_epsilon(FRSIZE_FREQ)
      doubleprecision dae_kappaabs(FRSIZE_FREQ)
      doubleprecision dae_rhs(FRSIZE_FREQ)
      doubleprecision dae_nu(FRSIZE_FREQ)
      doubleprecision dae_q
      common/idustalisolve/dae_nf
      integer dae_nf
c
      doubleprecision sspi,zbrent,dust_ali_equation
      doubleprecision bplanck,dum1,dum2
      parameter(sspi=1.80491254763d-5)     ! ss / pi
      external dust_ali_equation
c
      integer inu
c
c     Copy stuff to common blocks
c
      dae_nf=nf
      do inu=1,nf
          dae_lambdastar(inu) = lambdastar(inu)
          dae_kappaabs(inu) = kappaabs(inu)
          dae_epsilon(inu) = epsilon(inu)
          dae_meanint(inu) = meanint(inu)
          dae_nu(inu)      = nu(inu)
      enddo
c
      if(nf.eq.1) then
c
c         Frequency-integrated mode
c
          dae_rhs(1) = meanint(1)-lambdastar(1)*
     %                  (epsilon(1)*sspi*tempold**4+
     %                   (1.d0-epsilon(1))*meanintold(1))
          dae_q      = q + kappaabs(1)*dae_rhs(1)
      else
c
c         Frequency-dependent mode
c
          do inu=1,nf
              dae_rhs(inu) = meanint(inu)-lambdastar(inu)*
     %                  (epsilon(inu)*bplanck(tempold,nu(inu))+
     %                   (1.d0-epsilon(inu))*meanintold(inu))
          enddo
          dae_q      = q
          do inu=2,nf
              dae_q      = dae_q + 0.5 * 
     %              ( kappaabs(inu) * dae_rhs(inu) +
     %                kappaabs(inu-1) * dae_rhs(inu-1) ) *
     %             abs( nu(inu) - nu(inu-1) )
          enddo
      endif
c
c     Check the behavior of the equation
c
      dum1 = dust_ali_equation(DUST_TEMP_MIN)
      dum2 = dust_ali_equation(DUST_TEMP_MAX)
      if(dum1.gt.0.d0) then
          iwarn = iwarn + 1
          temp = 0.d0
          stop 66
          return
      endif
      if(dum2.lt.0.d0) then
          write(*,*) 'Error in dust-ALI: '
          write(*,*) '  dust temperature > ',DUST_TEMP_MAX,' K'
          stop 1
      endif
c
c     Now use zbrent to solve the system
c
      temp     = zbrent(dust_ali_equation,DUST_TEMP_MIN,
     %                      DUST_TEMP_MAX,1.d-4*DUST_TEMP_MIN)
c
c     Return also the newly computed mean intensity
c
      do inu=1,nf
          meanint(inu) = dae_meanint(inu)
      enddo
c
      end



c     --------------------------------------------------------------
c                HELPER ROUTINE FOR SOLVING LOCAL ALI
c
c     This routine is called by the zbrent routine and returns the
c     thermal equilibrium equation for dust, using the ALI recipe
c     with a local operator. It includes (and solves for) scattering 
c     as well, and this is returned in the common block /dustali/. 
c     --------------------------------------------------------------
      function dust_ali_equation(temp)
      implicit none
      doubleprecision dust_ali_equation,temp
c
      common/dustalisolve/dae_meanint,dae_lambdastar,dae_epsilon,
     %       dae_kappaabs,dae_rhs,dae_q,dae_nu
      doubleprecision dae_meanint(FRSIZE_FREQ)
      doubleprecision dae_lambdastar(FRSIZE_FREQ)
      doubleprecision dae_epsilon(FRSIZE_FREQ)
      doubleprecision dae_kappaabs(FRSIZE_FREQ)
      doubleprecision dae_rhs(FRSIZE_FREQ)
      doubleprecision dae_nu(FRSIZE_FREQ)
      doubleprecision dae_q
      common/idustalisolve/dae_nf
      integer dae_nf
c
      doubleprecision bplanck,alifactor,result,sspi
      doubleprecision intgrand(FRSIZE_FREQ)
      parameter(sspi=1.80491254763d-5)     ! ss / pi
      integer inu
c
      if(dae_nf.eq.1) then
c
c         Frequency-integrated mode
c
          alifactor      = 1.d0/(1.d0-dae_lambdastar(1)*
     %                           (1.d0-dae_epsilon(1)))
          dae_meanint(1) = alifactor * ( dae_rhs(1) +
     %                     dae_lambdastar(1)*dae_epsilon(1)*
     %                     sspi*temp**4 )
          result         = dae_q - dae_kappaabs(1) * ( 
     %                     (1.d0-dae_lambdastar(1)*
     %                     dae_epsilon(1))*sspi*temp**4 -
     %                     dae_lambdastar(1)*
     %                     (1.d0-dae_epsilon(1))*dae_meanint(1) )
      else
c
c         Frequency-dependent mode
c
          do inu=1,dae_nf
              alifactor        = 1.d0/(1.d0-dae_lambdastar(inu)*
     %                                (1.d0-dae_epsilon(inu)))
              dae_meanint(inu) = alifactor * ( dae_rhs(inu) +
     %                           dae_lambdastar(inu)*dae_epsilon(inu)*
     %                           bplanck(temp,dae_nu(inu)) )
              intgrand(inu)    = dae_kappaabs(inu)*((1.d0-
     %                           dae_lambdastar(inu)*dae_epsilon(inu))*
     %                           bplanck(temp,dae_nu(inu)) -
     %                           dae_lambdastar(inu)*(1.d0-
     %                           dae_epsilon(inu))*dae_meanint(inu))
          enddo
          result = -dae_q
          do inu=2,dae_nf
              result           = result + 0.5 * 
     %            ( intgrand(inu) + intgrand(inu-1) ) *
     %             abs( dae_nu(inu) - dae_nu(inu-1) )
          enddo
      endif
c
      dust_ali_equation = result
      return
      end



c     --------------------------------------------------------------
c       COMPUTE THE DIAGONAL OF LAMBDA AND THE ACCELLERATION FACTOR
c
c     This routine computes the accelleration factor for the ALI
c     accelleration at each point on the grid at each frequency.
c     This routine needs to be called only once, before the ALI 
c     iteration is stared, unless the opacities change per iteration,
c     which in general they don't.
c     --------------------------------------------------------------
      subroutine calc_lambdastar(nz,nmu,nf,z,mu,nu,alp,lambdastar)
      implicit none
      integer nz,nmu,nf
      doubleprecision z(FRSIZE_Z),mu(FRSIZE_MU),nu(FRSIZE_FREQ)
      doubleprecision alp(FRSIZE_FREQ,FRSIZE_Z)
      doubleprecision lambdastar(FRSIZE_FREQ,FRSIZE_Z)
c
      doubleprecision dummy,tau
      doubleprecision dzn,dzp,intgr_p(FRSIZE_MU),intgr_m(FRSIZE_MU)
      doubleprecision alp_min(FRSIZE_FREQ),alp_pls(FRSIZE_FREQ)
      integer inu,iz,imu
c
c     Loop over space
c     
      do iz=1,nz
c
c         Find the dz in both directions
c
          if(iz.ne.nz) then
              dzp = z(iz+1) - z(iz)
          else
              dzp = z(iz) - z(iz-1)
          endif
          if(iz.ne.1) then
              dzn = z(iz) - z(iz-1)
          else
              dzn = z(iz+1) - z(iz)
          endif
c
c         Get the mean alpha's in positive and negative direction
c
          if(iz.ne.1) then
              do inu=1,nf
                  alp_min(inu) = 0.5 * ( alp(inu,iz-1) + alp(inu,iz) )
              enddo
          else
              do inu=1,nf
                  alp_min(inu) = alp(inu,1)
              enddo
          endif
          if(iz.ne.nz) then
              do inu=1,nf
                  alp_pls(inu) = 0.5 * ( alp(inu,iz+1) + alp(inu,iz) )
              enddo
          else
              do inu=1,nf
                  alp_pls(inu) = alp(inu,nz)
              enddo
          endif
c
c         Now a loop over frequency
c
          do inu=1,nf
c
c             First the mu>0 direction (dzn,alp_min)
c
              do imu=1,nmu
                  tau = alp_min(inu) * dzn / mu(imu)
                  if(tau.gt.TAU_EPS) then
                      intgr_m(imu) = (tau-1.d0+exp(-tau))/tau
                  elseif(tau.gt.0.d0) then
                      intgr_m(imu) = 0.5d0 * tau
                  endif
              enddo
c
c             Then the mu<0 direction (dzp,alp_pls)
c
              do imu=1,nmu
                  tau = alp_pls(inu) * dzp / mu(imu)
                  if(tau.gt.TAU_EPS) then
                      intgr_p(imu) = (tau-1.d0+exp(-tau))/tau
                  elseif(tau.gt.0.d0) then
                      intgr_p(imu) = 0.5d0 * tau
                  endif
              enddo
c
c             Now integrate this stuff
c
              dummy = 0.d0
              do imu=2,nmu
                  dummy = dummy + 0.5 * 
     %               ( intgr_m(imu) + intgr_m(imu-1) ) *
     %               ( mu(imu) - mu(imu-1) )
              enddo
              dummy = dummy + mu(1) * intgr_m(1)
              do imu=2,nmu
                  dummy = dummy + 0.5 * 
     %               ( intgr_p(imu) + intgr_p(imu-1) ) *
     %               ( mu(imu) - mu(imu-1) )
              enddo
              dummy = dummy + mu(1) * intgr_p(1)
c
c             Now since there is a 1/2 before the integral, 
c             multiply here with 0.5
c
              lambdastar(inu,iz) = 0.5 * dummy
c
c             Safety check....
c
              if(lambdastar(inu,iz).gt.1.d0) then
                  write(*,*) 'Lambda* exceeds 1.0'
                  stop 13
              endif
c#################################
c              write(*,*) 'yes'
c              lambdastar(inu,iz) = 0.0 * lambdastar(inu,iz)
c              write(*,*) 'iz,inu,ls ',iz,inu,lambdastar(inu,iz)
c#################################
c
          enddo
      enddo
c
      end
