c     ==============================================================
c               FULL FREQ-MU-DEPENDENT RADIATIVE TRANSFER
c
c     This module is (almost) the same as the transfer.F module of
c     the dustslab code. Only, 'x' is replaced by 'z'. The only
c     difference is that this module first integrates from z=zmax
c     down to z=0 and then back to z=zmax, with a mirror boundary
c     condition at z=0.
c     ==============================================================

#include "main.h"

c     --------------------------------------------------------------
c                      THE TRANSFER MODULE
c
c     This routine integrates the formal transfer equation for all
c     angles and frequencies, and stores the intensity in the 
c     intensity array. The caller can then compute mean intensities
c     and all sorts of other things.
c
c     ARGUMENTS:
c        z          = The z-grid 
c        mu         = mu-grid (radiation angles)
c        nu         = frequency-grid
c        intens     = The intensity
c        alpha      = The rho*kappa_nu
c        source     = The source term S_nu
c        intbcminus = Bounray condition at iz=zmax: the in-intensity
c
c     NOTE: the mu-grid must be symmetric: for each mu there must 
c            be a -mu. 
c     --------------------------------------------------------------
      subroutine do_feautrier(z,mu,source,alpha,intens,
     %                       int0,nz,nmu,nf)
      implicit none
      integer nz,nmu,nf,intorder
      doubleprecision z(FRSIZE_Z),mu(FRSIZE_MU)
      doubleprecision source(FRSIZE_FREQ,FRSIZE_Z)
      doubleprecision alpha(FRSIZE_FREQ,FRSIZE_Z)
      doubleprecision intens(FRSIZE_FREQ,FRSIZE_MU,FRSIZE_Z)
      doubleprecision int0
c
      doubleprecision tau(FRSIZE_Z),snu(FRSIZE_Z)
      doubleprecision intj(FRSIZE_Z),inth(FRSIZE_Z)
      doubleprecision d(FRSIZE_Z),v(FRSIZE_Z),w(FRSIZE_Z)
      doubleprecision a(FRSIZE_Z),b(FRSIZE_Z),c(FRSIZE_Z),q(FRSIZE_Z)
      integer iz,inu,imu
c
c     Do a safety check
c
      do imu=1,nmu/2
          if(mu(imu).ge.0.d0) stop 612
      enddo
      do imu=nmu/2+1,nmu
          if(mu(imu).le.0.d0) stop 613
      enddo
c
c     Loop over positive mu
c
      do imu=nmu/2+1,nmu
c
c         Loop over frequency
c
          do inu=1,nf
c
c             Make tau grid and S_nu along the grid
c
              tau(1) = 0.d0
              snu(1) = source(inu,nz)
              do iz=nz-1,1,-1
                  tau(nz-iz+1) = tau(nz-iz) + 0.5 * 
     %                ( alpha(inu,iz) + alpha(inu,iz+1) ) *
     %                ( z(iz+1) - z(iz) ) / mu(imu)
                  snu(nz-iz+1) = source(inu,iz)
              enddo
c
c             Solve the transfer along this ray, using Feautrier
c
              call feautrier(nz,tau,snu,int0,intj,inth,
     %                       d,v,w,a,b,c,q)
c
c             Now reconstruct the intensity
c
              do iz=1,nz
################
              enddo
          enddo
      enddo
c
c     Boundary condition at z=zmax
c
      do imu=1,nmu/2
          do inu=1,nf
              intensity(inu,imu,nz) = intbcminus(inu,imu)
          enddo
      enddo
c
c     The radiative transfer in negative direction (i.e.
c     in the direction -z)
c     
      do iz=nz-1,1,-1
          do imu=1,nmu/2
              ds   = ( z(iz) - z(iz+1) ) / mu(imu)
              if(iz.ne.1) then
                  dsd  = ( z(iz-1) - z(iz) ) / mu(imu)
              else
                  dsd  = ds       ! BC at z=0: mirror
              endif
              do inu=1,nf
                  s1   = src(inu,iz+1)
                  alp1 = alp(inu,iz+1)
                  s2   = src(inu,iz)
                  alp2 = alp(inu,iz)
                  if(iz.ne.1) then
                      s3   = src(inu,iz-1)
                      alp3 = alp(inu,iz-1)
                  else
                      s3   = s2             ! BC at z=0: mirror
                      alp3 = alp2
                  endif
                  int0 = intensity(inu,imu,iz+1)
                  if(intorder.eq.1) then
                      intensity(inu,imu,iz) = qdr_src_2(int0,
     %                     s1,alp1,s2,alp2,ds)
                  else
                      intensity(inu,imu,iz) = qdr_src_3(int0,
     %                     s1,alp1,s2,alp2,s3,alp3,ds,dsd)
                  endif
              enddo
          enddo
      enddo
c
c     A mirror BC at the lower boundary (iz=1)
c
      do imu=nmu/2+1,nmu
          do inu=1,nf
              intensity(inu,imu,1) = intensity(inu,nmu-imu+1,1)
          enddo
      enddo
c
c     The radiative transfer in positive direction (i.e.
c     in the direction z)
c     
      do iz=2,nz
          do imu=nmu/2+1,nmu
              ds   = ( z(iz) - z(iz-1) ) / mu(imu)
              if(iz.ne.nz) then
                  dsd  = ( z(iz+1) - z(iz) ) / mu(imu)
              else
                  dsd  = ds       ! BC at z=xmax: extrapolation
              endif
              do inu=1,nf
                  s1   = src(inu,iz-1)
                  alp1 = alp(inu,iz-1)
                  s2   = src(inu,iz)
                  alp2 = alp(inu,iz)
                  if(iz.ne.nz) then
                      s3   = src(inu,iz+1)
                      alp3 = alp(inu,iz+1)
                  else
                      s3   = s2             ! BC at z=zmax: extrapol
                      alp3 = alp2
                  endif
                  int0 = intensity(inu,imu,iz-1)
                  if(intorder.eq.1) then
                      intensity(inu,imu,iz) = qdr_src_2(int0,
     %                       s1,alp1,s2,alp2,ds)
                  else
                      intensity(inu,imu,iz) = qdr_src_3(int0,
     %                     s1,alp1,s2,alp2,s3,alp3,ds,dsd)
                  endif
              enddo
          enddo
      enddo
c
      end




c     --------------------------------------------------------------
c               DO FEAUTRIER TRANSFER ALONG ONE RAY
c
c     Do feutrier along tau, where tau=0 is assumed to be the free
c     boundary (where int0 is given), and tau=taumax is the mirror
c     boundary. The mu points towards smaller tau, and so the inth
c     is also oriented that way (i.e. inth>0 for flux moving towards
c     smaller tau). 
c
c     ARGUMENTS:
c       
c
c     --------------------------------------------------------------
      subroutine feautrier(ntau,tau,source,int0,intj,inth,
     %                     d,v,w,a,b,c,q)
      implicit none
      integer ntau
      doubleprecision tau(ntau),source(ntau),inth(ntau,intj(ntau)
      doubleprecision d(ntau),v(ntau),w(ntau)
      doubleprecision a(ntau),b(ntau),c(ntau),q(ntau)
      doubleprecision int0
c
      integer itau
c
      do itau=1,ntau
          d(itau) = 1.d0
          w(itau) = 1.d0
          v(itau) = 0.d0
      enddo
      call solvede(ntau,tau,intj,v,d,source,w,
     %             -1.d0,1.d0,1.d0,0.d0,int0,0.d0,
     %                     a,b,c,q)
      do itau=1,ntau-1
          a(itau) = ( intj(itau+1) - intj(itau) ) / 
     %              ( tau(itau+1) - tau(itau) )
      enddo
      do itau=2,ntau-1
          inth(ix) = 0.5 * ( a(itau) + a(itau-1) )
      enddo
      inth(1)    = a(1)
      inth(ntau) = a(ntau-1)
c
      end




c     --------------------------------------------------------------
c                    SOLVE A DIFFUSION TYPE EQUATION
c
c     Solve the following DE:
c
c          d /     \    d /      du \
c         -- | u v | - -- | D(x) -- | = S - w(x) u
c         dx \     /   dx \      dx /
c
c     with boundary conditions
c
c         du |            |
c       p -- |      + q u |       = r
c         dx |x=xbc       |x=xbc
c
c     ARGUMENTS:
c       n             Nr of grid points
c       x             The grid
c       u             The current values of u(x)
c       v             The values for v(x)
c       d             The values for D(x)
c       s             The source function (rhs)
c       w             The value before u on rhs
c       pl,pr         The p value for the lhs/rhs of the BC equation
c       ql,qr         The q value for the lhs/rhs of the BC equation
c       rl,rr         The r value for the lhs/rhs of the BC equation
c       a,b,c,q       Dummy arrays used by the solvede() routine
c
c     --------------------------------------------------------------
      subroutine solvede(n,x,u,v,d,s,w,pl,pr,ql,qr,rl,rr,
     %                     a,b,c,q)
      implicit none
      integer n
      doubleprecision pl,pr,ql,qr,rl,rr
      doubleprecision x(n),u(n),v(n),d(n),s(n),w(n),a(n),b(n),c(n),q(n)
c
      integer i
      doubleprecision dum,dpl,dmn,vpl,vmn
c
c     The main body of the equation
c
      do i=2,n-1
          dum  = 2.0/(x(i+1)-x(i-1))
          dpl  = dum*0.5*(d(i+1)+d(i))/(x(i+1)-x(i))
          dmn  = dum*0.5*(d(i)+d(i-1))/(x(i)-x(i-1))
          vpl  = v(i+1)/(x(i+1)-x(i-1))
          vmn  = v(i-1)/(x(i+1)-x(i-1))
          b(i) = dpl+dmn+w(i)
          a(i) = -dmn-vmn
          c(i) = -dpl+vpl
          q(i) = s(i)
      enddo
c
c     The left (i=1) boundary condition
c
      b(1) = ql-pl/(x(2)-x(1))
      c(1) = pl/(x(2)-x(1))
      q(1) = rl
c
c     The right (i=n) boundary condition
c
      b(n) = qr+pr/(x(n)-x(n-1))
      a(n) = -pr/(x(n)-x(n-1))
      q(n) = rr
c
c     Solve the equation
c      
      call tridag(a,b,c,q,u,n)
c
      end
