c     ==============================================================
c           FULL FREQ-MU-DEPENDENT VERTICAL RADIATIVE TRANSFER
c
c     This program is similar in functionality to vertrt.F, but it
c     computes the vertical transfer using complete transfer, 
c     without using any approximations other than discretization.
c     The algorithm does not use Lambda Iteration (although this can
c     be switched on as well), but instead uses the variable 
c     Eddington factors and radiation-mean opacity approach. This
c     allows one to solve a moment equation, with the guarantee
c     that the answer is equal to the correct one for full RT,
c     simply because the eddington factors and mean opacities are
c     tuned to the complete radiation field, and the moment eqs are
c     exact (without approx) integrals of the full RT equation.
c
c     MODIFIED: Included viscous dissipation
c               03.10.05
c     ==============================================================

#include "main.h"

#define GASTODUST (100.d0)

#define PLANCKTAB_NT 100
#define PLANCKTAB_T0 (1.D-1)
#define PLANCKTAB_T1 (1.D+4)

c     --------------------------------------------------------------
c                             MAIN ROUTINE
c     --------------------------------------------------------------
      program main
      implicit none
c
#include "common_grid.h"
#include "common_dust.h"
#include "common_fullvrt.h"
c
      call read_all_data()
      call do_all_radtrans()
      call write_temperatures()
      call write_aux_data()
      call write_fluxes()
c
      end



c     --------------------------------------------------------------
c                         READ ALL THE DATA IN
c     --------------------------------------------------------------
      subroutine read_all_data()
      implicit none
c
#include "common_grid.h"
#include "common_dust.h"
#include "common_fullvrt.h"
c
      doubleprecision rstar,a,b,c,d,e
      integer nr,nz,iz,ir,nrp,nzp,nf,inu,imu,isize,ispec,nnr,nnz
      integer nrspec,nrsize(DUST_SPECIES_MAX)
      integer nnrspec,nnrsize(DUST_SPECIES_MAX)
      integer iformat
      logical f_ex,fexf,fexo,multi
c
c     Read the fullvrt input file
c
      global_init=0
      global_dump_intens=0
      open(unit=1,file='fullvrt.inp',status='old')
      read(1,*) iformat
      read(1,*) global_ir_start
      read(1,*) global_algorithm
      read(1,*) global_intorder
      read(1,*) vet_itermax
      read(1,*) vet_convcrit
      read(1,*) ang_nmu
      read(1,*) global_mu0
      read(1,*) global_idump
      read(1,*,end=100,err=100) global_init
      read(1,*,end=100,err=100) global_dump_intens
      close(1)
 100  continue
c
c     Make mu-grid
c
      if(ang_nmu.gt.FRSIZE_MU) then
          write(*,*) 'ERROR: ang_nmu gt FRSIZE_MU'
          stop 13
      endif
      do imu=1,ang_nmu
          ang_mu(imu) = global_mu0 * (1.d0/global_mu0)**
     %         ((imu-1.d0)/(ang_nmu-1.d0))
      enddo
c
c     Read the primary temperature and the density
c
      open(unit=1,file='diskstruct_1.dat',status='old')
      read(1,*) nr,nz
      nrspec=1
      nrsize(1)=1
      if(nr.lt.1) then
c
c         New format: multiple sizes/species
c
          read(1,*) nr,nz
          read(1,*) nrspec
          if(nrspec.gt.DUST_SPECIES_MAX) then
              write(*,*) 'ERROR: Too many species specified.'
              write(*,*) '       Recompile with larger max specs.'
              stop
          endif
          do ispec=1,nrspec
              read(1,*) nrsize(ispec)
              if(nrsize(ispec).gt.DUST_SIZE_MAX) then
                  write(*,*) 'ERROR: Too many sizes specified.'
                  write(*,*) '       Recompile with larger max size.'
                  stop
              endif
          enddo
          multi=.true.
      else
          nrspec=1
          nrsize(1)=1
          multi=.false.
      endif
      if(nr.gt.FRSIZE_R) then
          write(*,*) 'ERROR in fullvrt:'
          write(*,*) 'file diskstruct_1.dat: R grid too large'
          stop 13
      endif
      if(nz.gt.FRSIZE_Z) then
          write(*,*) 'ERROR in fullvrt:'
          write(*,*) 'file diskstruct_1.dat: Z grid too large'
          stop 13
      endif
      grid_nr = nr
      grid_nz = nz
      do ir=1,nr
          do iz=1,nz
              if(.not.multi) then
                  read(1,*) a,b,c,d,e
                  grid_r(ir)=a
                  grid_z(iz,ir)=b
                  vrt_rhogas(iz,ir) = c
                  vrt_rho(1,iz,ir)=c/GASTODUST
                  vrt_rhotot(iz,ir)=c/GASTODUST
                  vrt_primtemp(1,1,iz,ir)=d
                  vrt_vz(iz,ir)=e
              else
                  read(1,*) a,b,c,e
                  grid_r(ir) = a
                  grid_z(iz,ir) = b
                  vrt_rhogas(iz,ir) = c
                  vrt_vz(iz,ir) = e
                  vrt_rhotot(iz,ir) = 0.d0
                  do ispec=1,nrspec
                      read(1,*) a
                      vrt_rho(ispec,iz,ir) = a * c / GASTODUST
                      vrt_rhotot(iz,ir) = vrt_rhotot(iz,ir) + 
     %                   vrt_rho(ispec,iz,ir)
                  enddo
                  do ispec=1,nrspec
                      do isize=1,nrsize(ispec)
                          read(1,*) d
                          vrt_primtemp(isize,ispec,iz,ir) = d
                      enddo
                  enddo
              endif
          enddo
      enddo
      close(1)
c
c     Make initial guess for the temperature, if not read later
c
      do ir=1,grid_nr
          do iz=1,grid_nz
              do ispec=1,nrspec
                  do isize=1,nrsize(ispec)
                      vrt_temp(isize,ispec,iz,ir) = 
     %                  max(vrt_primtemp(isize,ispec,iz,ir),1.d0)
                  enddo
              enddo
          enddo
      enddo
c
c     Read the initial guess for the temperature
c
      inquire(file='diskstruct_2.dat',exist=f_ex) 
      if(f_ex.and.(global_init.ne.0)) then
          open(unit=1,file='diskstruct_2.dat',status='old')
          read(1,*) nr,nz
          if(nr.lt.1) then
              read(1,*) nr,nz
              read(1,*) nnrspec
              if(nnrspec.gt.DUST_SPECIES_MAX) then
                  write(*,*) 'ERROR: Too many species specified.'
                  write(*,*) '       Recompile with larger max specs.'
                  stop
              endif
              do ispec=1,nnrspec
                  read(1,*) nnrsize(ispec)
              enddo
              multi=.true.
          else
              nnrspec=1
              nnrsize(1)=1
              multi=.false.
          endif
          if(nrspec.ne.nnrspec) then
              write(*,*) 'ERROR: Incompatible files diskstruct_*.dat'
              stop
          endif
          do ispec=1,nrspec
              if(nrsize(ispec).ne.nnrsize(ispec)) then
                  write(*,*) 'ERROR: Incompatible files ',
     %                       'diskstruct_*.dat'
                  stop
              endif
          enddo
          if(nr.gt.FRSIZE_R) goto 200
          if(nz.gt.FRSIZE_Z) goto 200
          do ir=1,nr
              do iz=1,nz
                  if(.not.multi) then
                      read(1,*) a,b,c,d,e
                      vrt_temp(1,1,iz,ir)=d
                  else
                      read(1,*) a,b,c,e
                      do ispec=1,nrspec
                          read(1,*) a
                      enddo
                      do ispec=1,nrspec
                          do isize=1,nrsize(ispec)
                              read(1,*) d
                              vrt_temp(isize,ispec,iz,ir) = d
                          enddo
                      enddo
                  endif
              enddo
          enddo
          write(*,*) '  Initial guess read'
 200      continue
          close(1)
      endif
c
c     Read viscosity data, if present
c
      inquire(file='visc_alpha.inp',exist=f_ex) 
      if(f_ex) then
          accr_alpha_max=0.d0
          open(unit=1,file='visc_alpha.inp',status='old')
          read(1,*) nnr,nnz
          if((nnr.ne.nr).or.(nnz.ne.nz)) then
              write(*,*) 'ERROR: Nr of grid points in visc_alpha.inp'
              write(*,*) '       incompatible with diskstruct_1.inp'
              stop
          endif
          do ir=1,nr
              do iz=1,nz
                  read(1,*) accr_alpha(iz,ir)
                  if(accr_alpha(iz,ir).gt.accr_alpha_max) then
                      accr_alpha_max = accr_alpha(iz,ir)
                  endif
              enddo
          enddo
          close(1) 
          if(accr_alpha_max.gt.0.d0) then
              open(unit=1,file='starinfo.inp',status='old')
              read(1,*) a
              read(1,*) a
              read(1,*) accr_mstar
              close(1)
          endif
      else
          do ir=1,nr
              do iz=1,nz
                  accr_alpha(iz,ir) = 0.d0
              enddo
          enddo
      endif
c      write(*,*) accr_alpha_max,accr_mstar
c
c     Read the frequency array and the opacities
c
      inquire(file='frequency.inp',exist=fexf)
      inquire(file='dustopac.inp',exist=fexo)
      if((.not.fexf).or.(.not.fexo)) then
          write(*,*) 'Need both frequency.inp and dustopac.inp'
          stop 13
      endif
      open(unit=1,file='frequency.inp',status='old')
      read(1,*) nf
      if(nf.gt.FRSIZE_FREQ) then
          write(*,*) 'ERROR in fullvrt:'
          write(*,*) 'Frequency grid too large'
          stop 13
      endif
      do inu=1,nf
          read(1,*) a
          freq_nu(inu) = a
      enddo
      freq_nr = nf
      call read_dustdata()
c
c     Check if nr of species and sizes consistent
c
      if(nrspec.ne.dust_nr_species) then
          write(*,*) 'ERROR: Nr of dust species in dustopac*.inp'
          write(*,*) '   is unequal to the number specified in '
          write(*,*) '   diskstruct_1.dat'
          stop
      endif
      do ispec=1,dust_nr_species
          if(nrsize(ispec).ne.dust_nr_size(ispec)) then
              write(*,*) 'ERROR: Nr of dust sizes in dustopac*.inp'
              write(*,*) '   is unequal to the number specified in '
              write(*,*) '   diskstruct_1.dat'
              stop
          endif
      enddo
c
c     Check if domain.inp exists, else make maximum domain
c
      inquire(file='domain.inp',exist=f_ex)
      if(f_ex) then
          open(unit=1,file='domain.inp',status='old')
          read(1,*) grid_irstart,grid_irend
          close(1)
          if((grid_irstart.lt.1).or.(grid_irend.gt.nr).or.
     %       (grid_irstart.gt.grid_irend)) then
              write(*,*) 'ERROR: The domain specified in domain.inp'
              write(*,*) '       is out of range.'
              stop
          endif
      else
          grid_irstart = global_ir_start
          grid_irend   = nr
      endif
c
      end



c     --------------------------------------------------------------
c                         PERFORM ALL RT
c     --------------------------------------------------------------
      subroutine do_all_radtrans()
      implicit none
c
#include "common_grid.h"
#include "common_dust.h"
#include "common_fullvrt.h"
c
      doubleprecision find_dust_kappa,t0,t1
      integer ir,iz,inu,nt,isize,ispec
      parameter(nt=PLANCKTAB_NT)
      parameter(t0=PLANCKTAB_T0)
      parameter(t1=PLANCKTAB_T1)
c
c     First make a table of opacities, both the freq-dependent
c     ones, and the Planck-mean ones.
c
      if(dust_opacity_tempdep.ne.0) then
          write(*,*) 'Sorry, but fullrt cannot use temperature'
          write(*,*) 'dependent opacities for the moment...'
          stop 13
      endif
      do ispec=1,dust_nr_species
          do isize=1,dust_nr_size(ispec)
              do inu=1,freq_nr
                  optab_kappa(inu,isize,ispec) = 
     %                find_dust_kappa(inu,isize,ispec,t0,1,0)
              enddo
              call compute_planckopac(isize,ispec,nt,t0,t1,
     %                                optab_kappa(1,isize,ispec))
          enddo
      enddo
c
c     Do vertical RT for all other radial shells and check if 
c     energy is conserved
c
      do ir=grid_irstart,grid_irend
          call do_vrt_shell(ir)
          call check_encons(ir)
      enddo
c
      end




c     --------------------------------------------------------------
c                DO THE VERTICAL TRANSFER FOR ONE SHELL
c     --------------------------------------------------------------
      subroutine do_vrt_shell(ir)
      implicit none
      integer ir
c
#include "common_grid.h"
#include "common_dust.h"
#include "common_fullvrt.h"
c
      doubleprecision kapprim,find_planckopac,qv,omk,temp
      doubleprecision viscousheating,GG
      parameter(GG  = 6.672d-8)      ! Gravitational constant
      integer ierror,iz,ing,iali,imethod,intdump,isize,ispec
c
c     Compute the source term from irradiation for the vertical transfer
c
      do iz=1,grid_nz
          do ispec=1,dust_nr_species
              do isize=1,dust_nr_size(ispec)
                  kapprim = find_planckopac(isize,ispec,
     %                          vrt_primtemp(isize,ispec,iz,ir))
                  vrt_qfpi(isize,ispec,iz,ir) = 1.8049d-5 * kapprim *
     %                  vrt_primtemp(isize,ispec,iz,ir)**4
              enddo
          enddo
      enddo
c
c     *** ADDED 03-10-05 ***
c
c     Compute source term from viscous dissipation, but divided by
c     temperature, so that the iteration procedure can multiply
c     itself with the temperature.
c
c     NOTE: Multiplied with GASTODUST because the viscous energy
c     must be radiated away by the dust.
c
      if((accr_alpha_max.ne.0.d0).and.((dust_nr_species.ne.1).or.
     %   (dust_nr_size(1).ne.1))) then 
          write(*,*) 'ERROR: Viscous dissipation only in ',
     %               'single-grain mode'
          stop 
      endif
      if(accr_alpha_max.ne.0.d0) then
          do iz=1,grid_nz
              ispec = 1
              isize = 1
              temp  = 1.d0
              omk   = sqrt(GG*accr_mstar/(grid_r(ir)**3.d0))
              qv    = viscousheating(temp,accr_alpha(iz,ir),
     %             omk)/12.5663706144d0
              vrt_qvdivt(iz,ir) = qv*GASTODUST/temp
          enddo
      else
          do iz=1,grid_nz
              vrt_qvdivt(iz,ir) = 0.d0
          enddo
      endif
c
c     Now do the full transfer for this radial shell
c
      if((global_algorithm.ge.1).and.(global_algorithm.le.4)) then
c
c         Call the (Accelerated) Lambda Iteration module
c
          call check_no_multidust()
          if((global_algorithm.ge.1).and.(global_algorithm.le.2)) then
              write(*,*) '  Vertical transfer with ',
     %               'Lambda Iteration ir = ',ir
              iali = 0
              if(global_algorithm.eq.1) then
                  ing = 0       ! No Ng acceleration
              else
                  ing = 1       ! Yes Ng acceleration
              endif
          else
              write(*,*) '  Vertical transfer with ',
     %               'Accelerated Lambda Iteration ir = ',ir
              iali = 1
              if(global_algorithm.eq.3) then
                  ing = 0       ! No Ng acceleration
              else
                  ing = 1       ! Yes Ng acceleration
              endif
          endif
          call do_lambda_iteration(grid_nz,ang_nmu,freq_nr,grid_z(1,ir),
     %              ang_mu,freq_nu,vrt_rho(1,1,ir),vrt_qfpi(1,1,1,ir),
     %              vet_itermax,vet_convcrit,ing,iali,global_intorder,
     %              global_idump,ierror,
     %              ftr_intj,ftr_inth,optab_kappa,vrt_temp(1,1,1,ir),
     %              vet_jj(1,ir),vet_jnu(1,1,ir))
      elseif((global_algorithm.eq.10).or.(global_algorithm.eq.11)
     %       .or.(global_algorithm.eq.12)
     %       .or.(global_algorithm.eq.-1)
     %       .or.(global_algorithm.eq.-2)) then
c
c         Call the VET transfer module
c
          if(global_algorithm.eq.10) then 
              write(*,*) '  Vertical transfer with ',
     %               'Variable Eddington Factors (old way)'
              imethod = 0
              call check_no_multidust()
          elseif(global_algorithm.eq.-1) then
              
              write(*,*) '  Vertical transfer with ',
     %               'VEF with Ross & Planck means opacs... (APPROX!)'
              imethod = -1
              ing = 1           ! Yes Ng acceleration
              call check_no_multidust()
          elseif(global_algorithm.eq.-2) then
              
              write(*,*) '  Vertical transfer with ',
     %               'Real opacs but Edd approx... (APPROX!)'
              imethod = -2
              ing = 1           ! Yes Ng acceleration
              call check_no_multidust()
          else
              write(*,*) '  Vertical transfer with ',
     %               'Variable Eddington Factors'
              imethod = 1
              if(global_algorithm.eq.11) then
                  ing = 0       ! No Ng acceleration
              else
                  ing = 1       ! Yes Ng acceleration
              endif
          endif
          if(global_dump_intens.eq.ir) then
              intdump=1
          elseif(global_dump_intens.eq.-ir) then
              intdump=-1
          else
              intdump=0
          endif
          call do_vet_transfer(grid_nz,ang_nmu,freq_nr,grid_z(1,ir),
     %             ang_mu,freq_nu,vrt_rho(1,1,ir),vrt_qfpi(1,1,1,ir),
     %             vrt_qvdivt(1,ir),
     %             vet_itermax,vet_convcrit,ing,global_intorder,
     %             global_idump,intdump,imethod,global_init,ierror,
     %             ftr_intj,ftr_inth,optab_kappa,vrt_temp(1,1,1,ir),
     %             vet_jjme(1,ir),vet_hhme(1,ir),
     %             vet_jj(1,ir),vet_hh(1,ir),vet_jnu(1,1,ir),
     %             vet_hnu(1,1,ir),vet_fj(1,ir),vet_fnu(1,1,ir),
     %             vet_kapj(1,1,1,ir),vet_kaph(1,ir),vet_corfact(1,ir))
      else
          write(*,*) 'Problem in fullvrt:'
          write(*,*) '  Sorry, do not know algorithm type ',
     %               global_algorithm
          stop 13
      endif
c
c     Check error message
c
      if(ierror.ne.0) then
          write(*,*) '===== PROBLEM IN FULLRT AT IR=',ir,' ====='
      endif
c
      end




c     --------------------------------------------------------------
c                THE BLACKBODY PLANCK FUNCTION B_nu(T)
c
c     This function computes the Blackbody function 
c
c                    2 h nu^3 / c^2
c        B_nu(T)  = -------------------    [ erg / cm^2 s ster Hz ]
c                   exp(-h nu / kT) - 1
c
c     ARGUMENTS:
c        nu    [Hz]            = Frequency
c        temp  [K]             = Electron temperature
c     --------------------------------------------------------------
      function bplanck(temp,nu)
      implicit none
      doubleprecision temp
      doubleprecision nu
      doubleprecision bplanck
c
      if(temp.eq.0.d0) then 
          bplanck = 0.d0
          return
      endif
c
      bplanck = 1.47455d-47 * nu * nu * nu / 
     %      (exp(4.7989d-11 * nu / temp)-1.d0) + 1.d-290
c
      return
      end


c     --------------------------------------------------------------
c                THE VISCOUS DISSIPATION HEATING RATE
c     
c     This function returns the Q_visc/rho in units of [erg/s/g].
c     At the moment, only a simple alpha-viscosity law is included.
c     This function requires the sd_**** common variables to be set.
c
c     The viscous heating is the heating according to a radial
c     accretion velocity v_R of:
c
c                  3 nu
c        v_R  =  - - --
c                  2 R
c
c     This produces a viscous dissipation of:
c
c               9
c        q_+  = - nu Omega_K^2
c               4
c
c     which is the result that is returned. 
c
c     Here we simply have an alpha-viscosity prescription of the form:
c
c               2        c_s^2
c        nu  =  - alpha -------
c               3       Omega_K
c
c
c     INPUT:
c       temp               The gas temperature       
c
c     OUTPUT:
c       viscousheating     The Q_visc / rho : specific heating rate
c       
c     --------------------------------------------------------------
      function viscousheating(temp,alpha,omk)
      implicit none
      doubleprecision viscousheating,omk,temp,alpha
c
#     include "common_natconst.h"
c
      doubleprecision mugas,viscnu
      parameter(mugas=2.35)
c
      if(alpha.gt.0.d0) then
          viscnu         = (2.d0/3.d0)*alpha*
     %                  (8.3144d7*temp/mugas)/omk
          viscousheating = (9.d0/4.d0)*omk*omk*viscnu
      endif
      return
      end



c     --------------------------------------------------------------
c                   CHECK THE ENERGY CONSERVATION
c
c     This suroutine checks if the radiative transfer solution in
c     fact obeys energy conservation. 
c     --------------------------------------------------------------
      subroutine check_encons(ir)
      implicit none
      integer ir
c
#include "common_grid.h"
#include "common_fullvrt.h"
#include "common_dust.h"
c
      doubleprecision fluxnu(FRSIZE_FREQ),dum(FRSIZE_MU),sum
      integer iz,inu,imu,ispec,isize
c
c     Compute the incoming flux by integrating the source term
c     rho*q/4pi over z
c
      fio_fluxin(ir)  = 0.d0
      do iz=2,grid_nz
          sum = 0.d0
          do ispec=1,dust_nr_species
              do isize=1,dust_nr_size(ispec)
                  sum = sum + 0.5 *
     %                 ( vrt_rho(ispec,iz,ir)*
     %                   vrt_qfpi(isize,ispec,iz,ir) +
     %                   vrt_rho(ispec,iz-1,ir)*
     %                   vrt_qfpi(isize,ispec,iz-1,ir) )
              enddo
          enddo
          fio_fluxin(ir) = fio_fluxin(ir) + sum *
     %          ( grid_z(iz,ir) - grid_z(iz-1,ir) )
      enddo
      fio_fluxin(ir) = 12.5663706144d0 * fio_fluxin(ir)
c
c     Computing the outgoing flux by integrating the intensity
c     at z=zmax
c
      iz=grid_nz
      do inu=1,freq_nr
          do imu=1,ang_nmu
              dum(imu) = ang_mu(imu) * ( ftr_intj(inu,imu,iz) + 
     %                                   ftr_inth(inu,imu,iz) )
          enddo
          fluxnu(inu) = 0.d0
          do imu=2,ang_nmu
              fluxnu(inu) = fluxnu(inu) + 0.5 * 
     %            ( dum(imu) + dum(imu-1) ) *
     %             abs( ang_mu(imu) - ang_mu(imu-1) )
          enddo
      enddo
      fio_fluxout(ir) = 0.d0
      do inu=2,freq_nr
          fio_fluxout(ir) = fio_fluxout(ir) + 0.5 *
     %         ( fluxnu(inu) + fluxnu(inu-1) ) *
     %         abs( freq_nu(inu) - freq_nu(inu-1) )
      enddo
      fio_fluxout(ir) = 6.28318530718d0 * fio_fluxout(ir)
c
c     Check if the two are equal
c
      if(abs(fio_fluxin(ir)/fio_fluxout(ir)-1.d0).gt.5.d-2) then
          write(*,*) 'ERROR: At ir=',ir,'  fullvrt has a flux ',
     %       'problem'
          write(*,*) '   Flux_in   = ',fio_fluxin(ir)
          write(*,*) '   Flux_out  = ',fio_fluxout(ir)
          write(*,*) '   Ratio     = ',fio_fluxout(ir)/fio_fluxin(ir)
          if(global_algorithm.ge.0) then
c              stop 13
          endif
      endif
c
      end


c     --------------------------------------------------------------
c            WRITE THE FLUX IN AND FLUX OUT AS A FUNCTION OF R
c     --------------------------------------------------------------
      subroutine write_fluxes()
      implicit none
c
#include "common_grid.h"
#include "common_fullvrt.h"
c
      integer ir
c
      open(unit=1,file='fullvrt_fluxes.dat',status='unknown')
      write(1,*) grid_nr
      write(1,*)
      do ir=1,grid_nr
          write(1,10) fio_fluxin(ir),fio_fluxout(ir)
      enddo
 10   format(2(E13.6,1X))
c
      end


c     --------------------------------------------------------------
c                    WRITE THE MAIN OUTPUT FILE
c     --------------------------------------------------------------
      subroutine write_temperatures()
      implicit none
c
#include "common_grid.h"
#include "common_fullvrt.h"
#include "common_dust.h"
c
      integer ir,iz,imu,ispec,isize
      logical multi
c
      multi=.false.
      if((dust_nr_species.gt.1).or.(dust_nr_size(1).gt.1)) then
          multi=.true.
      endif
      open(unit=1,file='diskstruct_2.dat',status='unknown')
      if(multi) then
          write(1,*) -1,0
          write(1,*) grid_nr,grid_nz
          write(1,*) dust_nr_species
          do ispec=1,dust_nr_species
              write(1,*) dust_nr_size(ispec)
          enddo
      else
          write(1,*) grid_nr,grid_nz
      endif
      write(1,*)
      do ir=1,grid_nr
          do iz=1,grid_nz
              if(.not.multi) then
                  write(1,10) grid_r(ir),grid_z(iz,ir),
     %                   vrt_rhogas(iz,ir)+1d-98,
     %                   vrt_temp(1,1,iz,ir)+1d-98,vrt_vz(iz,ir)
 10               format(5(E13.6,1X))
              else
                  write(1,20) grid_r(ir),grid_z(iz,ir),
     %                        vrt_rhogas(iz,ir)+1d-98,vrt_vz(iz,ir)
 20               format(3(E13.6,1X))
                  do ispec=1,dust_nr_species
                      write(1,30) vrt_rho(ispec,iz,ir)*GASTODUST/
     %                            vrt_rhogas(iz,ir)+1d-98
                  enddo
                  do ispec=1,dust_nr_species
                      do isize=1,dust_nr_size(ispec)
                          write(1,30) vrt_temp(isize,ispec,iz,ir)+1d-98
                      enddo
                  enddo
 30               format(1(E13.6,1X))
              endif
          enddo
      enddo
      close(1)
c
      open(unit=1,file='angmu.dat',status='unknown')
      write(1,*) ang_nmu
      write(1,*) 
      do imu=1,ang_nmu
          write(1,*) ang_mu(imu)
      enddo
      close(1)
c
      end


c     --------------------------------------------------------------
c                  WRITE SOME OF THE AUXILIARY DATA
c     --------------------------------------------------------------
      subroutine write_aux_data()
      implicit none
c
#include "common_grid.h"
#include "common_fullvrt.h"
#include "common_dust.h"
c
      doubleprecision dummy
      integer ir,iz,isize,ispec
c
      open(unit=1,file='fullvrt.dat',status='unknown')
      write(1,*) grid_nr,grid_nz
      write(1,*)
      do ir=1,grid_nr
          do iz=1,grid_nz
              dummy = 0.d0
              do ispec=1,dust_nr_species
                  do isize=1,dust_nr_size(ispec)
                      dummy = dummy + vrt_rho(ispec,iz,ir) *
     %                          vrt_qfpi(isize,ispec,iz,ir)
                  enddo
              enddo
              dummy = dummy / vrt_rhotot(iz,ir)
              write(1,10) vet_hh(iz,ir),vet_jj(iz,ir),
     %                    vet_fj(iz,ir),dummy,
     %                    vet_corfact(iz,ir)
 10           format(5(E13.6,1X))
          enddo
      enddo
      close(1)
c
      end


c     --------------------------------------------------------------
c                 CHECK THAT NO MULTIPLE SPECIES SIZES
c
c     Since some parts of this code cannot deal with multiple sizes,
c     species, we must often check this.
c     --------------------------------------------------------------
      subroutine check_no_multidust()
      implicit none
c
#include "common_dust.h"
c
      if((dust_nr_species.ne.1).or.(dust_nr_size(1).ne.1)) then
          write(*,*) 'ERROR: Fullvrt, in this mode, cannot deal'
          write(*,*) '       with multiple dust species / sizes.'
          write(*,*) '       Only the mode 12 can do this.'
          stop
      endif
c
      end
