c     ==============================================================
c                MODULE FOR LAMBDA ITERATION TRANSFER
c
c     This module performs frequency-angle-dependent radiative 
c     transfer using the Lambda Iteration method.
c
C     NOTE: This module is meant only to be used for checking the 
c           results of the Variable Eddington Tensor module, which
c           is much more powerful as the lambda iteration module.
c           Lambda iteration is known for its slow convergence,
c           while the VET method converges very fast and is
c           just as accurate as the LI method. But since the VET
c           method is indirect, it might be useful to check the
c           results with the LI code, which is slow but reliable.
c     ==============================================================

#include "main.h"


c     --------------------------------------------------------------
c                 MAIN ROUTINE FOR THE LAMBDA ITERATION
c
c     This routine performs and iterates the full frequency and
c     angle dependent radiative transfer equation until a convergent
c     temperature profile is reached. It uses the Lambda Iteration
c     method.
c
c     ARGUMENTS:
c        z        = Z-grid
c        mu       = mu-grid (radiation angles)
c        nu       = frequency-grid
c        rho      = Density profile
c        itermax  = Maximum nr of iterations
c        convcrit = Convergence criterion (dimless nr, say 1.d-3)
c        intj     = The Feautrier j(mu,nu) = 0.5*(I(mu,nu)+I(-mu,nu))
c        inth     = The Feautrier h(mu,nu) = 0.5*(I(mu,nu)-I(-mu,nu))
c        kappa    = The full (temperature INdependent) opacity
c        intens   = The intensity field array (=workspace)
c        ing      = 0=no Ng acceleration, 1=Ng acceleration
c        idump    = 0=no dump, >=1 = save conv hist every idump steps
c
c     OUTPUT:
c        ierror   = Error signal (0=no error, 1=no convergence)
c        temp     = The gas/dust temperature 
c        jj       = The freq-integrated mean intensity
c        hh       = The freq-integrated eddington flux H = F / 4 pi
c        jnu      = Mean intensity J_nu 
c        Hnu      = Eddington flux H_nu = F_nu / 4 pi
c        fj       = The freq-integrated eddington factor f = K / J
c        fnu      = The eddington factor f_nu = K_nu / J_nu
c        kapt     = The Planck opacity for temperature temp
c        kapj     = The J-mean opacity
c        kaph     = The H-mean opacity
c     --------------------------------------------------------------
      subroutine do_lambda_iteration(nz,nmu,nf,z,mu,nu,
     %              rho,qfpi,itermax,convcrit,ing,iali,intorder,
     %              idump,ierror,intj,inth,kappa,temp,jj,jnu)
      implicit none
c
      integer nz,nmu,nf,ing,idump,intorder,iali
      doubleprecision z(FRSIZE_Z),mu(FRSIZE_MU),nu(FRSIZE_FREQ)
      doubleprecision intj(FRSIZE_FREQ,FRSIZE_MU,FRSIZE_Z)
      doubleprecision inth(FRSIZE_FREQ,FRSIZE_MU,FRSIZE_Z)
      doubleprecision kappa(FRSIZE_FREQ,DUST_SIZE_MAX,DUST_SPECIES_MAX)
      doubleprecision temp(DUST_SIZE_MAX,DUST_SPECIES_MAX,FRSIZE_Z)
      doubleprecision jj(FRSIZE_Z),hh(FRSIZE_Z)
      doubleprecision qfpi(DUST_SIZE_MAX,DUST_SPECIES_MAX,FRSIZE_Z)
      doubleprecision jnu(FRSIZE_FREQ,FRSIZE_Z)
      doubleprecision hnu(FRSIZE_FREQ,FRSIZE_Z)
      doubleprecision rho(DUST_SPECIES_MAX,FRSIZE_Z)
      doubleprecision convcrit,dummy
      doubleprecision error
      doubleprecision tempold(DUST_SIZE_MAX,DUST_SPECIES_MAX,FRSIZE_Z)
      integer itermax,ierror
c
      doubleprecision bplanck
      doubleprecision alp(FRSIZE_FREQ,FRSIZE_Z)
      doubleprecision src(FRSIZE_FREQ,FRSIZE_Z)
      doubleprecision jnubk(FRSIZE_FREQ,FRSIZE_Z)
      doubleprecision tbk(DUST_SIZE_MAX,DUST_SPECIES_MAX,FRSIZE_Z)
      doubleprecision intbcminus(FRSIZE_FREQ,FRSIZE_MU)
      doubleprecision t1(DUST_SIZE_MAX,DUST_SPECIES_MAX,FRSIZE_Z)
      doubleprecision t2(DUST_SIZE_MAX,DUST_SPECIES_MAX,FRSIZE_Z)
      doubleprecision t3(DUST_SIZE_MAX,DUST_SPECIES_MAX,FRSIZE_Z)
      doubleprecision lambdastar(FRSIZE_FREQ,FRSIZE_Z),eps(FRSIZE_FREQ)
      integer iter,iz,inu,imu,itdmp,savenr,icnt_ng,iwarn,isize,ispec
c
#include "common_dust.h"
c
c     Reset error signal
c
      ierror = 0
c
c     Reset dump counter
c
      itdmp  = 0
      savenr = 1
c     
c     Reset the Ng counter. The Ng counter is meant to count to 4
c     and then reset, triggering the Ng accelleration at each 4-th
c     iteration step.
c
      icnt_ng = 0
c
c     Put the Ng stuff to zero for safety
c
      do iz=1,nz
          do ispec=1,dust_nr_species
              do isize=1,dust_nr_size(ispec)
                  t3(isize,ispec,iz) = 0.d0
                  t2(isize,ispec,iz) = 0.d0
                  t1(isize,ispec,iz) = 0.d0
              enddo
          enddo
      enddo
c
c     The temp must be a reasonable initial guess, check this
c
      do iz=1,nz
          do ispec=1,dust_nr_species
              do isize=1,dust_nr_size(ispec)
                  if(temp(isize,ispec,iz).lt.1.d-4) then
                      write(*,*) 'ERROR in VET rad trans:'
                      write(*,*) 'Temperature initial guess ',
     %                           'is unreasonable'
                      stop 13
                  endif
              enddo
          enddo
      enddo
c
c     If dump is active, then clear the convergence history dump file
c
      if(idump.gt.0) then
          open(unit=1,file='convhist.info',status='unknown')
          write(1,*) 0,0
          close(1)
          open(unit=1,file='convhist.dat',status='unknown')
          write(1,*) nz
          close(1)
      endif
c
c     Put the 'old' temperature to zero, so that the iteration will
c     take at least two steps
c
      do iz=1,nz
          do ispec=1,dust_nr_species
              do isize=1,dust_nr_size(ispec)
                  tempold(isize,ispec,iz) = 0.d0
              enddo
          enddo
      enddo
c
c     Compute the extinction already in advance
c
      do iz=1,nz
          do inu=1,nf
              alp(inu,iz) = 0.d0
              do ispec=1,dust_nr_species
                  do isize=1,dust_nr_size(ispec)
                      alp(inu,iz) = alp(inu,iz) +
     %                   rho(ispec,iz) * kappa(inu,isize,ispec)
                  enddo
              enddo
          enddo
      enddo
c
c     Now we assume that the opacities do not change as a function of
c     the dust temperature, so we can compute the dust ALI operator
c     before the iteration loop
c
      if(iali.ne.0) then
          call calc_lambdastar(nz,nmu,nf,z,mu,nu,alp,lambdastar)
      endif
c
c     Clear the jnu
c
      do iz=1,nz
          do inu=1,nf
              jnu(inu,iz) = 0.d0
          enddo
      enddo
c
c     Do the main iteration
c
      do iter=1,itermax
c
c         Message
c     
          write(*,*) 'Iteration ',iter
c
c         Increase the Ng counter
c
          icnt_ng = icnt_ng + 1
c
c         Copy the mean intensities and the dust temperatures for 
c         the ALI algorithm
c
          if(iali.ne.0) then
              do iz=1,nz
                  do inu=1,nf
                      jnubk(inu,iz) = jnu(inu,iz)
                  enddo
                  do ispec=1,dust_nr_species
                      do isize=1,dust_nr_size(ispec)
                          tbk(isize,ispec,iz) = temp(isize,ispec,iz)
                      enddo
                  enddo
              enddo
          endif
c
c         Compute the source terms and extinction terms for the
c         full transfer module
c
          do iz=1,nz
              do inu=1,nf
                  alp(inu,iz) = 0.d0
                  src(inu,iz) = 0.d0
                  do ispec=1,dust_nr_species
                      do isize=1,dust_nr_size(ispec)
                          dummy       = rho(ispec,iz) * 
     %                                  kappa(inu,isize,ispec)
                          alp(inu,iz) = alp(inu,iz) + dummy 
                          src(inu,iz) = src(inu,iz) + dummy * 
     %                        bplanck(temp(isize,ispec,iz),nu(inu))
                      enddo
                  enddo
              enddo
          enddo
c
c         Set the boundary conditions for the full radiative
c         transfer: no incoming radiation
c
          do inu=1,nf
              do imu=1,nmu
                  intbcminus(inu,imu) = 0.d0
              enddo
          enddo
c
c         Now perform the full frequency-mu-dependent transfer
c
          call do_transfer(z,mu,src,alp,intj,inth,
     %                       intbcminus,intorder,nz,nmu,nf)
c
c         Compute the mean intensity
c
          call compute_jnu(nz,nmu,nf,mu,nu,intj,inth,
     %                     jnu,jj,hnu,hh)
c
c         Compute the dust temperature
c
          call compute_tdust(nz,nf,nu,kappa,jnu,qfpi,temp)
c
c         Now do the ALI stuff, if relevant
c
          if(iali.ne.0) then
              do inu=1,nf
                  eps(inu) = 1.d0     ! Only absorption here
              enddo
              do iz=1,nz
c
c                 I AM NOT SURE IF THIS IS THE CORRECT PROCEDURE 
c                 FOR MULTI SPECS/SIZES, SO CHECK
c                  
                  call check_no_multidust()
c
                  do ispec=1,dust_nr_species
                      do isize=1,dust_nr_size(ispec)
                          call solve_local_dust_ali(nf,nu,
     %                      lambdastar(1,iz),kappa(1,isize,ispec),
     %                      eps,qfpi(isize,ispec,iz),
     %                      tbk(isize,ispec,iz),jnubk(1,iz),
     %                      temp(isize,ispec,iz),jnu(1,iz),iwarn)
                      enddo
                  enddo
              enddo
          endif
c
c         Provide for the Ng acceleration
c
          if(ing.ne.0) then
              if(icnt_ng.eq.4) then
                  write(*,*) '   Do the Ng accelleration'
                  call do_ng_accell(nz,temp,t1,t2,t3)
                  icnt_ng = 0
              endif
c    
c             Now memorize the temperature for Ng-accelleration 
c             later. This is because Ng accelleration uses the memory
c             of the previous steps to make its prediction
c             
              do iz=1,nz
                  do ispec=1,dust_nr_species
                      do isize=1,dust_nr_size(ispec)
                          t3(isize,ispec,iz) = t2(isize,ispec,iz)
                          t2(isize,ispec,iz) = t1(isize,ispec,iz)
                          t1(isize,ispec,iz) = temp(isize,ispec,iz)
                      enddo
                  enddo
              enddo
          endif
c
c         If its time, then dump the conv hist
c
          itdmp = itdmp + 1
          if((itdmp.ge.idump).and.(idump.gt.0)) then
              call append_li_convhist(nz,temp,jj,hh,savenr)
              savenr = savenr+1
              itdmp = 0
          endif
c
c         Now check the error
c
          error = 0.d0
          do iz=1,nz
              do ispec=1,dust_nr_species
                  do isize=1,dust_nr_size(ispec)
                      error=max(error,
     %                   abs(tempold(isize,ispec,iz)-
     %                       temp(isize,ispec,iz))/
     %                       (tempold(isize,ispec,iz)+
     %                        temp(isize,ispec,iz)))
                  enddo
              enddo
          enddo
          if(error.lt.convcrit) goto 10
          do iz=1,nz
              do ispec=1,dust_nr_species
                  do isize=1,dust_nr_size(ispec)
                      tempold(isize,ispec,iz)=temp(isize,ispec,iz)
                  enddo
              enddo
          enddo
      enddo
c
c     Not converged...
c
      ierror = 1
      return
 10   continue
c
      end



c     --------------------------------------------------------------
c                   COMPUTE THE MEAN INTENSITY 
c     --------------------------------------------------------------
      subroutine compute_jnu(nz,nmu,nf,mu,nu,intj,inth,jnu,jj,
     %                       hnu,hh)
      implicit none
c
      integer nz,nmu,nf
      doubleprecision mu(FRSIZE_MU),nu(FRSIZE_FREQ)
      doubleprecision intj(FRSIZE_FREQ,FRSIZE_MU,FRSIZE_Z)
      doubleprecision inth(FRSIZE_FREQ,FRSIZE_MU,FRSIZE_Z)
      doubleprecision jnu(FRSIZE_FREQ,FRSIZE_Z),jj(FRSIZE_Z)
      doubleprecision hnu(FRSIZE_FREQ,FRSIZE_Z),hh(FRSIZE_Z)
c
      integer iz,imu,inu
c
      do iz=1,nz
          do inu=1,nf
              jnu(inu,iz) = 0.d0
              hnu(inu,iz) = 0.d0
              do imu=2,nmu
                  jnu(inu,iz) = jnu(inu,iz) + 0.5 * 
     %              ( intj(inu,imu,iz) + intj(inu,imu-1,iz) ) *
     %              ( mu(imu) - mu(imu-1) )
                  hnu(inu,iz) = hnu(inu,iz) + 0.5 * 
     %              ( inth(inu,imu,iz) * mu(imu) + 
     %                inth(inu,imu-1,iz) * mu(imu-1) ) *
     %              ( mu(imu) - mu(imu-1) )
              enddo
              jnu(inu,iz) = jnu(inu,iz) + mu(1) * intj(inu,1,iz)
          enddo
          jj(iz) = 0.d0
          hh(iz) = 0.d0
          do inu=2,nf
              jj(iz) = jj(iz) + 0.5 * ( jnu(inu,iz)+jnu(inu-1,iz) ) *
     %                    abs( nu(inu) - nu(inu-1) )
              hh(iz) = hh(iz) + 0.5 * ( hnu(inu,iz)+hnu(inu-1,iz) ) *
     %                    abs( nu(inu) - nu(inu-1) )
          enddo
      enddo
c
      end



c     --------------------------------------------------------------
c                   COMPUTE THE DUST TEMPERATURE
c
c     This subroutine computes the dust temperature at all z in
c     the grid. It calls a routine that uses zbrent() for solving
c     the temperature.
c     --------------------------------------------------------------
      subroutine compute_tdust(nz,nf,nu,kappa,jnu,qfpi,temp)
      implicit none
c
      integer nz,nf
      doubleprecision nu(FRSIZE_FREQ)
      doubleprecision kappa(FRSIZE_FREQ,DUST_SIZE_MAX,DUST_SPECIES_MAX)
      doubleprecision jnu(FRSIZE_FREQ,FRSIZE_Z)
      doubleprecision temp(DUST_SIZE_MAX,DUST_SPECIES_MAX,FRSIZE_Z)
      doubleprecision qfpi(DUST_SIZE_MAX,DUST_SPECIES_MAX,FRSIZE_Z)
      doubleprecision solve_dust_temperature
c
      integer iz,isize,ispec
c
#include "common_dust.h"
c
      do iz=1,nz
          do ispec=1,dust_nr_species
              do isize=1,dust_nr_size(ispec)
                  temp(isize,ispec,iz) = 
     %                solve_dust_temperature(nf,nu,
     %                  kappa(1,isize,ispec),jnu(1,iz),
     %                  qfpi(isize,ispec,iz))
              enddo
          enddo
      enddo
c
      end



c     --------------------------------------------------------------
c          SOLVE TEMPERATURE OF DUST GRAIN FOR GIVEN MEAN INT
c     --------------------------------------------------------------
      function solve_dust_temperature(nf,nu,kappa,jnu,qfpi)
      implicit none
      integer nf
      doubleprecision solve_dust_temperature
      doubleprecision dust_integral_emisabs,zbrent
      doubleprecision nu(FRSIZE_FREQ),kappa(FRSIZE_FREQ)
      doubleprecision jnu(FRSIZE_FREQ),qfpi
      external dust_integral_emisabs
c
      common/dustintegral/di_jnu,di_nu,di_kappa,di_qfpi
      doubleprecision di_jnu(FRSIZE_FREQ)
      doubleprecision di_nu(FRSIZE_FREQ)
      doubleprecision di_kappa(FRSIZE_FREQ)
      doubleprecision di_qfpi
      common/idustintegral/di_nf
      integer di_nf
c
      doubleprecision dum1,dum2
      integer inu
c
c     Put stuff into the commons
c
      di_nf   = nf
      di_qfpi = qfpi
      do inu=1,nf
          di_nu(inu)    = nu(inu)
          di_kappa(inu) = kappa(inu)
          di_jnu(inu)   = jnu(inu)
      enddo
c
c     Now check if there is indeed sufficient radiation for a
c     solution.
c
      dum1 = dust_integral_emisabs(0.02d0)
      dum2 = dust_integral_emisabs(3000.d0)
      if(dum1.gt.0.d0) then
          write(*,*) 'Error: dust temperature < 0.01 K'
          write(*,*) 'The mean intensities are:'
          write(*,*) (jnu(inu),inu=1,nf)
          stop 13
      endif
      if(dum2.lt.0.d0) then
          write(*,*) 'Error: dust temperature > 3000 K'
          stop 13
      endif
c
c     Now solve for the dust temperature
c
      solve_dust_temperature = zbrent(dust_integral_emisabs,
     %                              2.d-2,3.d3,1.d-1) 
c
      return
      end



c     --------------------------------------------------------------
c                    INTEGRAL OF EMISSION-ABSORPTION
c     --------------------------------------------------------------
      function dust_integral_emisabs(temp)
      implicit none
      doubleprecision dust_integral_emisabs,temp
c
      common/dustintegral/di_jnu,di_nu,di_kappa,di_qfpi
      doubleprecision di_jnu(FRSIZE_FREQ)
      doubleprecision di_nu(FRSIZE_FREQ)
      doubleprecision di_kappa(FRSIZE_FREQ)
      doubleprecision di_qfpi
      common/idustintegral/di_nf
      integer di_nf
c
      integer inu
      doubleprecision dummy,bplanck
      doubleprecision emabs(FRSIZE_FREQ)
c
c     First compute the dust emission - absorption at
c     all frequencies
c
      do inu=1,di_nf
          emabs(inu) = di_kappa(inu) * 
     %         ( bplanck(temp,di_nu(inu)) - di_jnu(inu) )
      enddo
c
c     Now integrate this over frequency
c
      dummy = 0.d0
      do inu=2,di_nf
          dummy = dummy +  0.5 * ( emabs(inu) + emabs(inu-1) ) *
     %                 abs( di_nu(inu) - di_nu(inu-1) )
      enddo
c
c     Add the source term and return
c
      dust_integral_emisabs = dummy - di_qfpi
      return
      end


c     --------------------------------------------------------------
c                           NG ACCELLERATION 
c
c     This routine extrapolates the temperature, given the current
c     temperature and the temperature at three previous iteration
c     steps. This trick should only be applied every 4 iteration
c     steps. And the calling routine should keep track of the 
c     previous temperatures (t1,t2,t3) itself.
c     --------------------------------------------------------------
      subroutine do_ng_accell(nz,temp,t1,t2,t3)
      implicit none
c
c
      integer nz
      doubleprecision temp(DUST_SIZE_MAX,DUST_SPECIES_MAX,FRSIZE_Z)
      doubleprecision t1(DUST_SIZE_MAX,DUST_SPECIES_MAX,FRSIZE_Z)
      doubleprecision t2(DUST_SIZE_MAX,DUST_SPECIES_MAX,FRSIZE_Z)
      doubleprecision t3(DUST_SIZE_MAX,DUST_SPECIES_MAX,FRSIZE_Z)
c
      doubleprecision a1,a2,b1,b2,c1,c2,a,b,c,x0,x1,x2,x3,w
      doubleprecision dum1,dum2,dum3,dummy
      integer iz,isize,ispec
c
#include "common_dust.h"
c
      a1 = 0.d0
      a2 = 0.d0
      b1 = 0.d0
      b2 = 0.d0
      c1 = 0.d0
      c2 = 0.d0
      do iz=1,nz
          do ispec=1,dust_nr_species
              do isize=1,dust_nr_size(ispec)
                  x0   = temp(isize,ispec,iz)**4
                  x1   = t1(isize,ispec,iz)**4
                  x2   = t2(isize,ispec,iz)**4
                  x3   = t3(isize,ispec,iz)**4
                  w    = 1.d0 / temp(isize,ispec,iz)**4
                  dum1 = x0 - 2.d0*x1 + x2
                  dum2 = x0 - x1 - x2 + x3
                  dum3 = x0 - x1
                  a1   = a1 + dum1 * dum1 * w
                  b1   = b1 + dum1 * dum2 * w
                  c1   = c1 + dum1 * dum3 * w
                  a2   = b1
                  b2   = b2 + dum2 * dum2 * w
                  c2   = c2 + dum2 * dum3 * w
              enddo
          enddo
      enddo
      a = ( c1*b2 - c2*b1 ) / ( a1*b2 - a2*b1 ) 
      b = ( c2*a1 - c1*a2 ) / ( a1*b2 - a2*b1 ) 
      c = 1.d0 - a - b
c
c     Write a message
c
      write(*,*) '       Ng : a = ',a,'  b = ',b
c
c     Now perform the accelleration step
c
      do iz=1,nz
          do ispec=1,dust_nr_species
              do isize=1,dust_nr_size(ispec)
                  dummy = c * temp(isize,ispec,iz)**4 + 
     %                 a * t1(isize,ispec,iz)**4 + 
     %                 b * t2(isize,ispec,iz)**4
                  if(dummy.lt.0.d0) then
                      write(*,*) 'Ng ERROR: negative T^4 at iz=',iz
                      write(*,*) '   a,b,c = ',a,b,c
                      write(*,*) 'Skipping this Ng step'
                  else
                      temp(isize,ispec,iz) = sqrt(sqrt(dummy))
                  endif
              enddo
          enddo
      enddo
c
      end



c     --------------------------------------------------------------
c                APPEND THE CONVERGENCE HISTORY FILE
c
c     This subroutine is only for convenience: it saves the stuff
c     as it converges towards a solution. This is handy for debug
c     and demonstration purposes.
c     --------------------------------------------------------------
      subroutine append_li_convhist(nz,temp,jj,hh,savenr)
      implicit none
      integer nz,savenr
      doubleprecision temp(DUST_SIZE_MAX,DUST_SPECIES_MAX,FRSIZE_Z)
      doubleprecision jj(FRSIZE_Z),hh(FRSIZE_Z)
c
      integer iz
c
      open(unit=1,file='convhist.dat',status='old',access='append')
      do iz=1,nz 
          write(1,*) temp(1,1,iz),jj(iz),hh(iz)
      enddo
      write(1,*)
      close(1)
      open(unit=1,file='convhist.info',status='unknown')
      write(1,*) savenr,0
      close(1)
c
      end
