c     ==============================================================
c             MODULE FOR INTEGRATING THE MOMENT EQUATIONS 
c
c     This module integrates the moment equations of radiative 
c     transfer. There are two subroutines one can choose:
c
c      - do_integr_momeq(): This integrates the pure Moment Equation
c                           without using any frequency-dependent
c                           information.
c
c      - do_integr_momnu(): This is a hybrid scheme, in which the
c                           0-th moment equation is done in the
c                           usual way, but the 1-st is done using
c                           the literal integral over H_nu and 
c                           kappa_nu.
c
c     For use in a Variable Eddington Factor code, the second method 
c     is the best, since it can handle cases in which the netto flux
c     H is zero, while H_nu is very non-zero (e.g. the case of a 
c     passive obliquely irradiated disk).
c     
c     ==============================================================

#include "main.h"


c     --------------------------------------------------------------
c                 DO INTEGRATION OF MOMENT EQUATIONS
c
c     This assumes a slab with H=H0 as boundary condition at z=0,
c     and J=H/hj1 at z=zmax. The hj1 must be computed from the 
c     full radiation field. The moment equations need the variable
c     eddington factor and the mean opacities, which also must be
c     computed before hand from the real radiation field. Finally,
c     the source term qfpi = Q / ( 4 pi rho ).
c
c     ARGUMENTS:
c       z          The z-grid
c       qfpitot    The source term per unit gram per sterrad
c       h0         Boundary condition iz=1: the H-flux there
c       hj1        Boundary condition iz=nz: the ratio H/J there
c       fj         The eddington factor (f=0 means isotropic)
c       alphah     The H-mean opacity alpha_H
c
c     OUTPUT:
c       jj         The mean intensity
c       hh         The Eddington flux
c     
c     NOTE: The f is chosen such that K = ( 1/3 + f ) J. 
c     --------------------------------------------------------------
      subroutine do_integr_momeq(nz,z,qfpitot,h0,hj1,fj,alphah,jj,hh)
      implicit none
      integer nz
      doubleprecision z(FRSIZE_Z),qfpitot(FRSIZE_Z),h0,hj1
      doubleprecision fj(FRSIZE_Z)
      doubleprecision alphah(FRSIZE_Z)
      doubleprecision jj(FRSIZE_Z),hh(FRSIZE_Z)
c
      doubleprecision jjff(FRSIZE_Z)
      integer iz
c
c     The inner boundary condition:  H(z=0)=H0
c
      hh(1) = h0
c
c     Integrate H(z) from 0 to zmax
c
      do iz=2,nz
          hh(iz) = hh(iz-1) + 0.5d0 *
     %                 ( qfpitot(iz) + qfpitot(iz-1) ) *
     %                 ( z(iz) - z(iz-1) ) 
      enddo
c
c     The outer boundary condition: J = H / hj1
c     
      jj(nz)   = abs(hh(nz)) / hj1
c
c     Now compute ( 1/3 + f ) J here
c
      jjff(nz) = ( (1.d0/3.d0) + fj(nz) ) * jj(nz)
c     
c     Now integrate J back to the equator
c
      do iz=nz-1,1,-1
          jjff(iz) = jjff(iz+1) + 0.5d0 
     %                * ( hh(iz+1) * alphah(iz+1) 
     %                  + hh(iz) * alphah(iz) )
     %                * ( z(iz+1) - z(iz) ) 
          jj(iz)   = jjff(iz) / ( (1.d0/3.d0) + fj(iz) )
      enddo
c     
      end



c     --------------------------------------------------------------
c       DO INTEGRATION OF MOMENT EQUATIONS USING H_NU AND KAPPA_NU
c
c     This assumes a slab with H=H0 as boundary condition at z=0,
c     and J=H/hj1 at z=zmax. The hj1 must be computed from the 
c     full radiation field. The moment equations need the variable
c     eddington factor and the mean opacities, which also must be
c     computed before hand from the real radiation field. Finally,
c     the source term qfpi = Q / ( 4 pi rho ). 
c
c     Contrary to the normal freq-integr moment equations, the
c     1-st equation (i.e. not the 0-th eq) does not use the H, but
c     instead the integral over H_nu * kappa_nu. In order to avoid
c     numerical problems (e.g. J<0) one also passes the actual
c     value of J(z=zmax) that follows from the full transfer. If
c     this differs from the J(z=zmax) computed here, then the
c     integral H_nu*kappa_nu is scaled accordingly.
c
c     The VET algorithm in the present way turns out to be somewhat
c     unstable, so there are several tricks built in to stabilize
c     the algorithm. To choose this method use istab. When istab=0
c     no correction is applied, when istab=1 it uses only the 
c     scaling factor, when istab=2 it corrects the Hnu so that 
c     int Hnu dnu equals the flux from the moment equations.
c
c     ARGUMENTS:
c       z          The z-grid
c       qfpitot    The source term per sterrad
c       h0         Boundary condition iz=1: the H-flux there
c       hj1        Boundary condition iz=nz: the ratio H/J there
c       fj         The eddington factor (f=0 means isotropic)
c       j1         The value of J(z=zmax)
c       nu         The frequency grid
c       hnu        The freq-dependent flux
c       alphanu    The freq-dependent rho*kappa
c       istab      0=no stab ticks, 1=scale using j1, 2=using dB/dT
c
c     OUTPUT:
c       jj         The mean intensity
c       hh         The Eddington flux
c       corfact    (for debug) How much has the flux been corrected?
c     
c     NOTE: The f is chosen such that K = ( 1/3 + f ) J. 
c     --------------------------------------------------------------
      subroutine do_integr_momnu(nz,nf,z,nu,temp,qfpitot,h0,
     %                hj1,fj,j1,hnu,alphanu,istab,jj,hh,corfact)
      implicit none
      integer nz,nf,istab
      doubleprecision z(FRSIZE_Z),qfpitot(FRSIZE_Z),h0,hj1,j1
      doubleprecision fj(FRSIZE_Z),nu(FRSIZE_FREQ)
      doubleprecision hnu(FRSIZE_FREQ,FRSIZE_Z)
      doubleprecision temp(DUST_SIZE_MAX,DUST_SPECIES_MAX,FRSIZE_Z)
      doubleprecision jj(FRSIZE_Z),hh(FRSIZE_Z)
      doubleprecision alphanu(FRSIZE_FREQ,FRSIZE_Z)
      doubleprecision corfact(FRSIZE_Z)
c
      doubleprecision jjff(FRSIZE_Z),jratio,hkint(FRSIZE_Z)
      doubleprecision tempav,ntcount
      integer iz,inu,isize,ispec
c
#include "common_dust.h"
c
c     The inner boundary condition:  H(z=0)=H0. That is: the flux at z=0
c     is specified here.
c
      hh(1) = h0
c
c     Integrate H(z) from 0 to zmax
c
      do iz=2,nz
          hh(iz) = hh(iz-1) + 0.5d0 *
     %                 ( qfpitot(iz) + qfpitot(iz-1) ) *
     %                 ( z(iz) - z(iz-1) ) 
      enddo
c
c     The outer boundary condition: J = H / hj1. That is: the ratio between
c     the J and H is specified at z=zmax.
c     
      jj(nz)   = abs(hh(nz)) / hj1
c
c     Now compute K = ( 1/3 + f ) J here
c
      jjff(nz) = ( (1.d0/3.d0) + fj(nz) ) * jj(nz)
c
c     Compute the ratio between this jj(nz) and j1 (=the real jj(nz))
c     This is used for a stability trick later. Note however, that this
c     trick is not used in the variable eddington factor method anymore.
c     Instead the stability trick istab=2 is used (see below). The
c     istab.eq.1 mode is just kept for backwards compatibility.
c
      if(istab.eq.1) then 
          jratio   = min(1.d0,jj(nz)/j1)
      else
          jratio   = 1.d0
      endif
c
c     For numerical stability it is important to 'correct' the
c     flux at every z, so that the freq-integrated flux is equal to 
c     the freq-integrated flux computed from the moment equations
c     (i.e. the integral of hnu over nu should equal hh).
c     
      if(istab.eq.2) then
          do iz=1,nz
c
c             First make an average temperature. We do this in a
c             very simplistic way: unweighted. The reason is that
c             when the temperatures deviate strongly, then we are
c             in the optically thin regime, and the correction 
c             factor is not very important. While when we are in
c             the optically thick regime, all temperatures are the
c             same
c
              tempav  = 0.d0
              ntcount = 0.d0
              do ispec=1,dust_nr_species
                  do isize=1,dust_nr_size(ispec)
                      tempav  = tempav + temp(isize,ispec,iz)
                      ntcount = ntcount+1.d0
                  enddo
              enddo
              tempav = tempav / ntcount
              call correct_flux(nf,nu,hnu(1,iz),hh(iz),
     %                 alphanu(1,iz),tempav,corfact(iz))
          enddo
      endif
c
c     Compute the integral H_nu * kappa_nu dnu at all z
c     and normalize with the jratio for numerical stability 
c     
c
      do iz=1,nz
          hkint(iz) = 0.d0
          do inu=2,nf
              hkint(iz) = hkint(iz) + 0.5 * 
     %          ( hnu(inu,iz)*alphanu(inu,iz) + 
     %            hnu(inu-1,iz)*alphanu(inu-1,iz) ) *
     %          abs( nu(inu) - nu(inu-1) )
          enddo
          hkint(iz) = hkint(iz) * jratio
      enddo
c     
c     Now integrate K back to the equator, and evaluate 
c
      do iz=nz-1,1,-1
          jjff(iz) = jjff(iz+1) + 0.5d0 
     %                * ( hkint(iz+1) + hkint(iz) )
     %                * ( z(iz+1) - z(iz) ) 
          jj(iz)   = jjff(iz) / ( (1.d0/3.d0) + fj(iz) )
      enddo
cc#########################
c      write(*,*) 'JJFF = ',jj(1)
cc#########################
c     
      end




c     --------------------------------------------------------------
c           COMPUTE THE DUST TEMPERATURE FROM MEAN INTENSITY
c
c     This routine will solve for all the temperatures, given the
c     mean intensity, the J-mean opacity, and an external source
c     qfpi.
c
c     ARGUMENTS:
c        qfpi         Source term in erg/s/g/ster
c        jj           Mean intensity in erg/s/cm^3/ster
c        kapj         J-mean opacity
c
c     OUTPUT:
c        temp         Dust temperature
c
c     --------------------------------------------------------------
      subroutine compute_dust_temp_from_jj(nz,qfpi,jj,kapj,temp)
      implicit none
      integer nz
      doubleprecision qfpi(DUST_SIZE_MAX,DUST_SPECIES_MAX,FRSIZE_Z)
      doubleprecision kapj(DUST_SIZE_MAX,DUST_SPECIES_MAX,FRSIZE_Z)
      doubleprecision temp(DUST_SIZE_MAX,DUST_SPECIES_MAX,FRSIZE_Z)
      doubleprecision jj(FRSIZE_Z),kappa,ti,tiold,find_planckopac
c
#     include "common_dust.h"
c
      integer iz,i,ispec,isize
c
c     Loop over space
c
      do iz=1,nz
c
c         Loop over species and sizes
c
          do ispec=1,dust_nr_species
              do isize=1,dust_nr_size(ispec)
c
c                 Make a first guess for the temperature
c
                  ti = 15.3421d0 * ( jj(iz) + qfpi(isize,ispec,iz) /
     %                 kapj(isize,ispec,iz) )**0.25
c
c                 Now do a convergence loop for the temperature
c
                  do i=1,80
                      kappa = find_planckopac(isize,ispec,ti)
                      tiold = ti
                      ti    = 15.3421d0 * ( 
     %                  ( jj(iz)*kapj(isize,ispec,iz) + 
     %                    qfpi(isize,ispec,iz) ) 
     %                  / kappa )**0.25
                      if(abs(ti-tiold)/(ti+tiold).lt.TEMPACCUR) goto 10
                  enddo
                  write(*,*) 'No convergence in temperature'
                  write(*,*) iz,ispec,isize,jj(iz),qfpi(isize,ispec,iz),
     %                       kapj(isize,ispec,iz),ti,tiold,TEMPACCUR
                  stop 733
 10               continue
                  temp(isize,ispec,iz) = ti
              enddo
          enddo
      enddo
c
      end


c     --------------------------------------------------------------
c                     COMPUTE THE PLANCK MEAN OPACITY
c     --------------------------------------------------------------
      subroutine compute_planckopac(isize,ispec,nt,t0,t1,kappa)
      implicit none
      integer nt,isize,ispec
      doubleprecision t0,t1
      doubleprecision kappa(FRSIZE_FREQ)
c
#     include "common_grid.h"
#     include "common_fullvrt.h"
#     include "common_dust.h"
c
      doubleprecision enum,denom,bnucur,bnuold,bplanck
      integer it,inu
c
c     Do a loop over the discretized temperature
c
      dopac_nt = nt
      if(dopac_nt.gt.NR_TEMPERATURE) then
          write(*,*) 'Too many temperatures in requested for planck'
          write(*,*) ' Increase NR_TEMPERATURE and recompile'
          stop 13
      endif
      do it=1,dopac_nt
          dopac_temp(it,isize,ispec)  
     %                    = t0 * (t1/t0)**((it-1.d0)/(nt-1.d0))
          bnucur          = bplanck(dopac_temp(it,isize,ispec),
     %                              freq_nu(1))
          enum            = 0.d0
          denom           = 0.d0
          do inu=2,freq_nr
              bnuold  = bnucur
              bnucur  = bplanck(dopac_temp(it,isize,ispec),
     %                          freq_nu(inu))
              enum    = enum + 0.5 * ( kappa(inu) * bnucur +
     %                    kappa(inu-1) * bnuold ) *
     %                    abs( freq_nu(inu) - freq_nu(inu-1) )
              denom   = denom + 0.5 * ( bnucur + bnuold ) *
     %                    abs( freq_nu(inu) - freq_nu(inu-1) )
          enddo
          dopac_kappa(it,isize,ispec) = enum / denom
      enddo
c
      end


c     --------------------------------------------------------------
c                  INTERPOLATE THE PLANCK MEAN OPACITY
c     --------------------------------------------------------------
      function find_planckopac(isize,ispec,temp)
      implicit none
      integer isize,ispec
      doubleprecision find_planckopac,temp
c
#     include "common_grid.h"
#     include "common_fullvrt.h"
#     include "common_dust.h"
c
      doubleprecision dum1,dum2,dt
      integer it
c
      if(dopac_nt.le.1) then 
          write(*,*) 'Must prepare planck opacity table first'
          stop 13
      endif
c
      if(temp.lt.dopac_temp(1,isize,ispec)) then 
          find_planckopac = dopac_kappa(1,isize,ispec)
      elseif(temp.gt.dopac_temp(dopac_nt,isize,ispec)) then
c          find_planckopac = dopac_kappa(dopac_nt)
          write(*,*) 'Temperature exceeds ',
     %             dopac_temp(dopac_nt,isize,ispec)
          write(*,*) 'If you want this to work, then increase'
          write(*,*) 'the upper temperature of planck opac table'
          stop 13
      else
          call hunt(dopac_temp(1,isize,ispec),dopac_nt,temp,it)
          dt = (temp-dopac_temp(it,isize,ispec))/
     %            (dopac_temp(it+1,isize,ispec)-
     %             dopac_temp(it,isize,ispec))
          if((dt.lt.0.d0).or.(dt.gt.1.d0)) stop 552
          find_planckopac = (1.d0-dt)*dopac_kappa(it,isize,ispec)+
     %       dt*dopac_kappa(it+1,isize,ispec)
      endif
c
      return
      end


c     --------------------------------------------------------------
c                CORRECT THE FLUX FOR STABILITY
c
c     This routine applies a small correction to the H_nu flux from
c     the transfer routine, such that the total flux H is equal to 
c     the flux H following from the moment equations. This turns out
c     to be a crucial stabilizing factor for the variable eddington 
c     factor method. 
c
c     ARGUMENTS:
c       nu        Frequency in Hz
c       hnu       Eddington flux in erg/cm^2/s/Hz
c       hh        Eddington flux (freq-integrated) in erg/cm^2/s
c       kappa     Opacity in cm^2/g
c       temp      Temperature in K
c
c     RESULTS:
c       hnu       New (corrected) Eddington flux in erg/cm^2/s/Hz
c       corfact   (for debug) How much the flux has been corrected? 
c
c     --------------------------------------------------------------
      subroutine correct_flux(nf,nu,hnu,hh,alpha,temp,corfact)
      implicit none
      integer nf
      doubleprecision hnu(FRSIZE_FREQ),hh,temp,corfact
      doubleprecision nu(FRSIZE_FREQ),alpha(FRSIZE_FREQ)
c
      doubleprecision htildenu(FRSIZE_FREQ),htilde,bplanckdt
      doubleprecision chi,hnow,h1,h2
      integer inu
c
c     Find out what the total flux is in Hnu
c
      hnow = 0.d0
      do inu=2,nf
          hnow = hnow + 0.5 * ( hnu(inu) + hnu(inu-1) ) *
     %             abs( nu(inu) - nu(inu-1) )
      enddo
c
c     Now make tilde H_nu, i.e. a dummy flux proportional to
c     what a termal flux (1/alphanu)(dBnu/dT)(dT/dz) would be.
c
      do inu=1,nf
          htildenu(inu) = bplanckdt(temp,nu(inu)) / alpha(inu)
      enddo
c
c     Find out the norm of this dummy flux
c
      htilde = 0.d0
      do inu=2,nf
          htilde = htilde + 0.5 * 
     %          ( htildenu(inu) + htildenu(inu-1) ) *
     %            abs( nu(inu) - nu(inu-1) )
      enddo
c
c     Find out how much of this flux is needed to add to Hnu in
c     order to make the freq-integral of the new Hnu equal to
c     the flux from the moment equations
c
      chi = ( hh - hnow ) / htilde
c
c     Now correct the flux to achieve this goal
c
      do inu=1,nf
          hnu(inu) = hnu(inu) + chi * htildenu(inu)
      enddo
c
c     For curiosity, we compute how large this correction was
c     (is not part of the algorithm!)
c
      h1 = 0.d0
      h2 = 0.d0
      do inu=2,nf
          h1 = h1 + 0.5 * abs( htildenu(inu) + htildenu(inu-1) ) *
     %            abs( nu(inu) - nu(inu-1) )
          h2 = h2 + 0.5 * abs( hnu(inu) + hnu(inu-1) ) *
     %            abs( nu(inu) - nu(inu-1) )
      enddo
      corfact = h1*abs(chi)/h2
c
      end


c     --------------------------------------------------------------
c           THE TEMPERATURE DERIVATIVE OF PLANCK FUNCTION 
c
c     This function computes the temperature derivative of the
c     Blackbody function 
c
c        dB_nu(T)     2 h^2 nu^4      exp(h nu / kT)        1 
c        --------   = ---------- ------------------------  ---
c           dT          k c^2    [ exp(h nu / kT) - 1 ]^2  T^2
c
c     ARGUMENTS:
c        nu    [Hz]            = Frequency
c        temp  [K]             = Electron temperature
c     --------------------------------------------------------------
      function bplanckdt(temp,nu)
      implicit none
      doubleprecision temp
      doubleprecision nu
      doubleprecision bplanckdt,theexp
c
#     ifndef NOCHECK
      if(temp.le.0.d0) then
          write(*,*) 'ERROR: bplanckdt()'
          write(*,*) '       temperature = ',temp,' K'
          stop 13
      endif
#     endif
c
      theexp = exp(4.7989d-11*nu/temp)
      if(theexp.lt.1.d33) then
          bplanckdt = 7.07661334104d-58 * nu**4 * 
     %         theexp / ( (theexp-1.d0)**2 * temp**2 )
     %         + 1.d-290
      else
          bplanckdt = 7.07661334104d-58 * nu**4 
     %         / ( theexp * temp**2 )
     %         + 1.d-290
      endif
c     
      return
      end


