c     ==============================================================
c               FULL FREQ-MU-DEPENDENT RADIATIVE TRANSFER
c
c     This module is (almost) the same as the transfer.F module of
c     the dustslab code. Only, 'x' is replaced by 'z'. The only
c     difference is that this module first integrates from z=zmax
c     down to z=0 and then back to z=zmax, with a mirror boundary
c     condition at z=0.
c     ==============================================================

#include "main.h"

c     --------------------------------------------------------------
c                      THE TRANSFER MODULE
c
c     This routine integrates the formal transfer equation for all
c     angles and frequencies, and stores the intensity in the 
c     intensity array. The caller can then compute mean intensities
c     and all sorts of other things.
c
c     ARGUMENTS:
c        z          = The z-grid 
c        mu         = mu-grid (radiation angles) (only positive ones!)
c        nu         = frequency-grid
c        intj       = The Feautrier j(mu,nu) = 0.5*(I(mu,nu)+I(-mu,nu))
c        inth       = The Feautrier h(mu,nu) = 0.5*(I(mu,nu)-I(-mu,nu))
c        alp        = The rho*kappa_nu
c        src        = The j_nu emissivity
c        intbcminus = Bounray condition at iz=zmax: the in-intensity
c
c     NOTE: Only the positive mus are given, from imu=1,nmu they 
c           go from mu=mumin to mu=1.
c     --------------------------------------------------------------
      subroutine do_transfer(z,mu,src,alp,intj,inth,
     %                       intbcminus,intorder,nz,nmu,nf)
      implicit none
      integer nz,nmu,nf,intorder
      doubleprecision z(FRSIZE_Z),mu(FRSIZE_MU)
      doubleprecision src(FRSIZE_FREQ,FRSIZE_Z)
      doubleprecision alp(FRSIZE_FREQ,FRSIZE_Z)
      doubleprecision intj(FRSIZE_FREQ,FRSIZE_MU,FRSIZE_Z)
      doubleprecision inth(FRSIZE_FREQ,FRSIZE_MU,FRSIZE_Z)
      doubleprecision intbcminus(FRSIZE_FREQ,FRSIZE_MU)
c
      doubleprecision qdr_src_2,qdr_src_3
      doubleprecision ds,dsd,s1,s2,s3,alp1,alp2,alp3,int0
      doubleprecision intmin(FRSIZE_Z),intplus(FRSIZE_Z)
      doubleprecision tau(FRSIZE_Z),snu(FRSIZE_Z)
      doubleprecision ij(FRSIZE_Z),ih(FRSIZE_Z)
      doubleprecision d(FRSIZE_Z),v(FRSIZE_Z),w(FRSIZE_Z)
      doubleprecision a(FRSIZE_Z),b(FRSIZE_Z),c(FRSIZE_Z),q(FRSIZE_Z)
      integer iz,inu,imu
c
c     Safety check
c
      do imu=2,nmu
          if((mu(imu).lt.0.d0).or.(mu(imu).le.mu(imu-1))) then 
              write(*,*) 'INTERNAL ERROR'
              stop 23423
          endif
      enddo
c
c     Loop over mu and nu
c
      do imu=1,nmu
          do inu=1,nf
              if(intorder.eq.-1) then
c
c                 *** Do integration implicitly using feautrier ***
c
c                 Make tau grid and S_nu along the grid
c
                  tau(1) = 0.d0
                  snu(1) = src(inu,nz) / alp(inu,nz)
                  do iz=nz-1,1,-1
                      tau(nz-iz+1) = tau(nz-iz) + 0.5 * 
     %                     ( alp(inu,iz) + alp(inu,iz+1) ) *
     %                     ( z(iz+1) - z(iz) ) / mu(imu)
                      snu(nz-iz+1) = src(inu,iz) / alp(inu,iz)
                  enddo
c
c                 Solve the transfer along this ray, using Feautrier
c
                  call feautrier(nz,tau,snu,int0,ij,ih,
     %                       d,v,w,a,b,c,q)
c
c                 Now put things back into the arrays
c
                  do iz=1,nz
                      intj(inu,imu,iz) = ij(nz-iz+1)
                      inth(inu,imu,iz) = ih(nz-iz+1)
                  enddo
              else
c
c                 *** Do integration explicitly using qdr_src_2/3 ***
c
c                 Boundary condition at z=zmax
c
                  intmin(nz) = intbcminus(inu,imu)
c
c                 Integrate downwards
c
                  do iz=nz-1,1,-1
c
c                     Integrate downwards
c
                      ds   = ( z(iz+1) - z(iz) ) / mu(imu)
                      if(iz.ne.1) then
                          dsd  = ( z(iz) - z(iz-1) ) / mu(imu)
                      else
                          dsd  = ds ! BC at z=0: mirror
                      endif
                      s1   = src(inu,iz+1)
                      alp1 = alp(inu,iz+1)
                      s2   = src(inu,iz)
                      alp2 = alp(inu,iz)
                      if(iz.ne.1) then
                          s3   = src(inu,iz-1)
                          alp3 = alp(inu,iz-1)
                      else
                          s3   = s2 ! BC at z=0: mirror
                          alp3 = alp2
                      endif
                      int0 = intmin(iz+1)
                      if(intorder.eq.1) then
                          intmin(iz) = qdr_src_2(int0,
     %                        s1,alp1,s2,alp2,ds)
                      else
                          intmin(iz) = qdr_src_3(int0,
     %                         s1,alp1,s2,alp2,s3,alp3,ds,dsd)
                      endif
                  enddo
c
c                 Boundary condition at z=0
c
                  intplus(1) = intmin(1)
c
c                 Integrate upwards
c
                  do iz=2,nz
                      ds   = ( z(iz) - z(iz-1) ) / mu(imu)
                      if(iz.ne.nz) then
                          dsd  = ( z(iz+1) - z(iz) ) / mu(imu)
                      else
                          dsd  = ds ! BC at z=xmax: extrapolation
                      endif
                      s1   = src(inu,iz-1)
                      alp1 = alp(inu,iz-1)
                      s2   = src(inu,iz)
                      alp2 = alp(inu,iz)
                      if(iz.ne.nz) then
                          s3   = src(inu,iz+1)
                          alp3 = alp(inu,iz+1)
                      else
                          s3   = s2 ! BC at z=zmax: extrapol
                          alp3 = alp2
                      endif
                      int0 = intplus(iz-1)
                      if(intorder.eq.1) then
                          intplus(iz) = qdr_src_2(int0,
     %                       s1,alp1,s2,alp2,ds)
                      else
                          intplus(iz) = qdr_src_3(int0,
     %                         s1,alp1,s2,alp2,s3,alp3,ds,dsd)
                      endif
                  enddo
c
c                 Now compute the relevant intj and inth
c
                  do iz=1,nz
                      intj(inu,imu,iz) = 0.5 * 
     %                      ( intplus(iz) + intmin(iz) )
                      inth(inu,imu,iz) = 0.5 * 
     %                      ( intplus(iz) - intmin(iz) )
                  enddo
              endif
          enddo
      enddo
c
      end



c     --------------------------------------------------------------
c      INTEGRATE THE FORMAL SOLUTION ALONG ONE ELEMENT OF SHORT CHAR
c
c     This function computes the formal integral over a segment of
c     a local characteristic. This will be done first order here.
c
c     This linear integration is not always suitable for extremely 
c     optically thick systems. In that case the 3-point quadratic 
c     integrator, qdr_src_3() can better be used. 
c     --------------------------------------------------------------
      function qdr_src_2(int,js1,alp1,js2,alp2,ds)
      implicit none
      doubleprecision qdr_src_2
      doubleprecision int,js1,alp1,js2,alp2,ds
c
      doubleprecision e0,e1,a,b,dtau1,xp,src1,src2,theomax
c
      dtau1     = 0.5d0 * ( alp1 + alp2 ) * ds
      theomax   = 0.5d0 * ( js1 + js2 ) * ds
      if(dtau1.gt.1.d-6) then
          xp        = exp(-dtau1)
          e0        = 1.d0 - xp
          e1        = dtau1 - e0
          b         = e1 / dtau1
          a         = e0 - b
      else
          a         = 0.5d0 * dtau1
          b         = 0.5d0 * dtau1
          xp        = 1.d0 - dtau1 
      endif
      if(alp1.gt.0.d0) then
          src1 = js1 / alp1
      elseif(alp2.gt.0.d0) then
          src1 = js2 / alp2
      else
          src1 = 0.d0
      endif
      if(alp2.gt.0.d0) then
          src2 = js2 / alp2
      elseif(alp1.gt.0.d0) then
          src2 = js1 / alp1
      else
          src2 = 0.d0
      endif
      if(dtau1.gt.1e-9) then
         qdr_src_2 = a * src1 + b * src2
      else
         qdr_src_2 = 0.5d0 * ( js1 + js2 ) * ds
      endif
c
c     Bracket the total integrated emission between the
c     theoretical maximum and 0.
c
#     ifndef SRCQDR_NO_BRACKET
      qdr_src_2 = min(qdr_src_2,theomax)
#     endif
c
      qdr_src_2 = int * xp + qdr_src_2
      return
c
      write(*,*) 'Ehh, something wrong in dqr_src_2'
      stop 13
      return
      end



c     --------------------------------------------------------------
c      INTEGRATE THE FORMAL SOLUTION ALONG ONE ELEMENT OF SHORT CHAR
c
c     This function computes the formal integral over a segment of
c     a local characteristic. This will be done second order here.
c     This is necessary for optically thick systems, but hardy so
c     for optically thin systems.
c
c     WARNING: Be careful for optical depths close to 1e-6, because
c              this regime might be uncertain.....
c
c     --------------------------------------------------------------
      function qdr_src_3(int,js1,alp1,js2,alp2,js3,alp3,ds1,ds2)
      implicit none
      doubleprecision qdr_src_3
      doubleprecision int,js1,alp1,js2,alp2,js3,alp3,ds1,ds2
c
      doubleprecision e0,e1,e2,a,b,c,dtau1,dtau2,dtsum,xp
      doubleprecision src1,src2,src3,theomax,theomin
c
      dtau1     = 0.5d0 * ( alp1 + alp2 ) * ds1
      dtau2     = 0.5d0 * ( alp2 + alp3 ) * ds2
      dtsum     = dtau1 + dtau2
c
c     Now compute the theoretical maximum and minimum of the
c     quadrature. These are used to bracket the Olson-Kunasz formula
c     which may sometimes go wrong. I did this previously in an 
c     erronous way (until 6 July 1999).
c
      theomax   = max(js1,js2) * ds1
      theomin   = 0.d0        ! min(js1,js2) * ds1 * exp(-dtau1)
c
      if(dtau1.gt.1.d-6) then
          xp        = exp(-dtau1)
          e0        = 1.d0 - xp
          e1        = dtau1 - e0
          e2        = dtau1 * dtau1 - 2.d0 * e1
          if(dtau2.gt.1.d-6) then
              b  = ( dtsum * e1 - e2 ) / ( dtau1 * dtau2 )
              a  = e0 + ( e2 - ( dtsum + dtau1 ) * e1 ) 
     %                    / ( dtau1 * dtsum )
              c  = ( e2 - dtau1 * e1 ) / ( dtau2 * dtsum )
          else  
              b  = e1 / dtau1
              a  = e0 - b
              c  = 0.d0
          endif
          if(alp1.gt.0.d0) then
             src1 = js1 / alp1
          elseif(alp2.gt.0.d0) then
             src1 = js2 / alp2
          else
             src1 = 0.d0
          endif
          if(alp2.gt.0.d0) then
             src2 = js2 / alp2
          elseif(alp1.gt.0.d0) then
             src2 = js1 / alp1
          elseif(alp3.gt.0.d0) then
             src2 = js3 / alp3
          else
             src2 = 0.d0
          endif
          if(alp3.gt.0.d0) then
             src3 = js3 / alp3
          elseif(alp2.gt.0.d0) then
             src3 = js2 / alp2
          else
             src3 = 0.d0
          endif
          qdr_src_3 = a * src1 + b * src2 + c * src3
c
c         Verify if serious overshoot occur (say: factor 2 or more)
c
c#         ifndef SRCQDR_NO_ERRCHECK
c          if(qdr_src_3.gt.2.d0*theomax) then
c              transerr_overshoot =  transerr_overshoot + 1
c          endif
c#         endif
c
c         NEW: Bracket the total integrated emission between the
c              theoretical maximum and minimum.
c
#         ifndef SRCQDR_NO_BRACKET
          qdr_src_3 = min(qdr_src_3,theomax)
          qdr_src_3 = max(qdr_src_3,theomin)
#         endif
c
          qdr_src_3 = int * xp + qdr_src_3
c
      else
          qdr_src_3 = int + 0.5d0 * ( js1 + js2 ) * ds1
      endif
      return
c
      end


c     --------------------------------------------------------------
c               DO FEAUTRIER TRANSFER ALONG ONE RAY
c
c     Do feutrier along tau, where tau=0 is assumed to be the free
c     boundary (where int0 is given), and tau=taumax is the mirror
c     boundary. The mu points towards smaller tau, and so the inth
c     is also oriented that way (i.e. inth>0 for flux moving towards
c     smaller tau). 
c
c     ARGUMENTS:
c       
c
c     --------------------------------------------------------------
      subroutine feautrier(ntau,tau,source,int0,intj,inth,
     %                     d,v,w,a,b,c,q)
      implicit none
      integer ntau
      doubleprecision tau(ntau),source(ntau),inth(ntau),intj(ntau)
      doubleprecision d(ntau),v(ntau),w(ntau)
      doubleprecision a(ntau),b(ntau),c(ntau),q(ntau)
      doubleprecision int0
c
      integer itau
c
      do itau=1,ntau
          d(itau) = 1.d0
          w(itau) = 1.d0
          v(itau) = 0.d0
      enddo
      call solvede(ntau,tau,intj,v,d,source,w,
     %             -1.d0,1.d0,1.d0,0.d0,int0,0.d0,
     %                     a,b,c,q)
      do itau=1,ntau-1
          a(itau) = ( intj(itau+1) - intj(itau) ) / 
     %              ( tau(itau+1) - tau(itau) )
      enddo
      do itau=2,ntau-1
          inth(itau) = 0.5 * ( a(itau) + a(itau-1) )
      enddo
      inth(1)    = a(1)
      inth(ntau) = a(ntau-1)
c
      end




c     --------------------------------------------------------------
c                    SOLVE A DIFFUSION TYPE EQUATION
c
c     Solve the following DE:
c
c          d /     \    d /      du \
c         -- | u v | - -- | D(x) -- | = S - w(x) u
c         dx \     /   dx \      dx /
c
c     with boundary conditions
c
c         du |            |
c       p -- |      + q u |       = r
c         dx |x=xbc       |x=xbc
c
c     ARGUMENTS:
c       n             Nr of grid points
c       x             The grid
c       u             The current values of u(x)
c       v             The values for v(x)
c       d             The values for D(x)
c       s             The source function (rhs)
c       w             The value before u on rhs
c       pl,pr         The p value for the lhs/rhs of the BC equation
c       ql,qr         The q value for the lhs/rhs of the BC equation
c       rl,rr         The r value for the lhs/rhs of the BC equation
c       a,b,c,q       Dummy arrays used by the solvede() routine
c
c     --------------------------------------------------------------
      subroutine solvede(n,x,u,v,d,s,w,pl,pr,ql,qr,rl,rr,
     %                     a,b,c,q)
      implicit none
      integer n
      doubleprecision pl,pr,ql,qr,rl,rr
      doubleprecision x(n),u(n),v(n),d(n),s(n),w(n),a(n),b(n),c(n),q(n)
c
      integer i
      doubleprecision dum,dpl,dmn,vpl,vmn
c
c     The main body of the equation
c
      do i=2,n-1
          dum  = 2.0/(x(i+1)-x(i-1))
          dpl  = dum*0.5*(d(i+1)+d(i))/(x(i+1)-x(i))
          dmn  = dum*0.5*(d(i)+d(i-1))/(x(i)-x(i-1))
          vpl  = v(i+1)/(x(i+1)-x(i-1))
          vmn  = v(i-1)/(x(i+1)-x(i-1))
          b(i) = dpl+dmn+w(i)
          a(i) = -dmn-vmn
          c(i) = -dpl+vpl
          q(i) = s(i)
      enddo
c
c     The left (i=1) boundary condition
c
      b(1) = ql-pl/(x(2)-x(1))
      c(1) = pl/(x(2)-x(1))
      q(1) = rl
c
c     The right (i=n) boundary condition
c
      b(n) = qr+pr/(x(n)-x(n-1))
      a(n) = -pr/(x(n)-x(n-1))
      q(n) = rr
c
c     Solve the equation
c      
      call tridag(a,b,c,q,u,n)
c
      end
