c     ==============================================================
c        MODULE FOR VARIABLE EDDINGTON FACTOR RADIATIVE TRANSFER
c
c     This module performs the radiative transfer using a variable
c     Eddington factor and mean opacity approach. It uses the 
c     transfer.F module for the full freq-mu-dependent transfer 
c     integral, and the radmomeq.F module for solving the moment
c     equation. In this module the transfer.F and radmomeq.F are
c     alternately called until convergence in the temperatures is
c     reached.
c
c     Note: This module assumes no scattering!
c     ==============================================================

#include "main.h"

c
c     The minimal flux allowed in H (relative to J)
c     (to help the H-averaged opacity determination)
c
#ifndef HFLUXMIN
#define HFLUXMIN 1d-10
#endif
c
c     The minimal temperature difference between cells
c     (to help the H-averaged opacity determination)
c
#ifndef TEMPFLAT
#define TEMPFLAT 1.d-4
#endif

c     --------------------------------------------------------------
c            MAIN ROUTINE FOR THE VARIABLE EDD TENSOR TRANSFER
c
c     This routine performs and iterates the full frequency and
c     angle dependent radiative transfer equation until a convergent
c     temperature profile is reached. The algorithm uses the full 
c     transfer equation to compute all the average quantities used
c     for a moment-equation-integrator. By alternatively calling
c     the moment equation subroutine and the formal transfer 
c     subroutine, one has the convergence speed of the moment eqs,
c     but the accuracy of full RT.
c     
c     It is important to note that the moment equation of rank 1
c     (the one with K and H) introduces a difficulty when taking
c     the frequency-integrated form. It requires the definition of
c     the H-averaged kappa, but there are situations in which the
c     H-averaged kappa diverges to infinity. This can happen when
c     for instance a passive disk is irradiated from above, and 
c     the total flux within the disk is zero, while the freq-dep
c     flux is non-zero: at high freq radiation moves inwards, at
c     low freq it moves outwards, perfectly cancelling each other
c     out. The H is then 0, while the integral H_nu*kappa_nu is
c     then non-zero. To avoid problems of this kind one can put
c     the argument imethod to 1, and then the full integral is
c     used. 
c
c     NOTE: Always use imethod=1. NEVER use imethod=0, unless 
c           for demonstration purposes.
c
c     ARGUMENTS:
c        z        = Z-grid
c        mu       = mu-grid (radiation angles)
c        nu       = frequency-grid
c        rho      = Density profile
c        itermax  = Maximum nr of iterations
c        convcrit = Convergence criterion (dimless nr, say 1.d-3)
c        qfpi     = Heat source (divided by 4pi)
c        qvdivt   = Viscous heat source (divided by 4pi and by T)
c        intj     = The Feautrier j(mu,nu) = 0.5*(I(mu,nu)+I(-mu,nu))
c        inth     = The Feautrier h(mu,nu) = 0.5*(I(mu,nu)-I(-mu,nu))
c        kappa    = The full (temperature INdependent) opacity
c        ing      = 0=no Ng acceleration, 1=Ng acceleration
c        idump    = 0=no dump, >=1 = save conv hist every idump steps
c        idint    = 0=no dump, >=1 = dump full intensity array at the end
c        imethod  = 0=use kaph (DONT USE), 1=use full H_nu*kap_nu integral
c
c     OUTPUT:
c        ierror   = Error signal (0=no error, 1=no convergence)
c        temp     = The gas/dust temperature 
c        jjme     = The freq-int mean intensity from moment equation
c        hhme     = The freq-int eddington flux from moment equation
c        jj       = The freq-integrated mean intensity
c        hh       = The freq-integrated eddington flux H = F / 4 pi
c        jnu      = Mean intensity J_nu 
c        Hnu      = Eddington flux H_nu = F_nu / 4 pi
c        fj       = The freq-integrated eddington factor f = K / J
c        fnu      = The eddington factor f_nu = K_nu / J_nu
c        kapj     = The J-mean opacity
c        kaph     = The H-mean opacity
c        corfact  = (for debug) How much has the flux been corrected?
c     --------------------------------------------------------------
      subroutine do_vet_transfer(nz,nmu,nf,z,mu,nu,
     %              rho,qfpi,qvdivt,itermax,convcrit,
     %              ing,intorder,idump,idint,
     %              imethod,init,ierror,intj,inth,kappa,temp,jjme,hhme,
     %              jj,hh,jnu,hnu,fj,fnu,kapj,kaph,corfact)
      implicit none
c
      integer nz,nmu,nf,idump,imethod,ing,intorder,init,idint
      doubleprecision z(FRSIZE_Z),mu(FRSIZE_MU),nu(FRSIZE_FREQ)
      doubleprecision intj(FRSIZE_FREQ,FRSIZE_MU,FRSIZE_Z)
      doubleprecision inth(FRSIZE_FREQ,FRSIZE_MU,FRSIZE_Z)
      doubleprecision kappa(FRSIZE_FREQ,DUST_SIZE_MAX,DUST_SPECIES_MAX)
      doubleprecision temp(DUST_SIZE_MAX,DUST_SPECIES_MAX,FRSIZE_Z)
      doubleprecision jj(FRSIZE_Z),hh(FRSIZE_Z)
      doubleprecision qfpi(DUST_SIZE_MAX,DUST_SPECIES_MAX,FRSIZE_Z)
      doubleprecision qvdivt(FRSIZE_Z)
      doubleprecision jjme(FRSIZE_Z),hhme(FRSIZE_Z)
      doubleprecision jnu(FRSIZE_FREQ,FRSIZE_Z)
      doubleprecision hnu(FRSIZE_FREQ,FRSIZE_Z)
      doubleprecision fj(FRSIZE_Z),dummy
      doubleprecision rho(DUST_SPECIES_MAX,FRSIZE_Z)
      doubleprecision fnu(FRSIZE_FREQ,FRSIZE_Z)
      doubleprecision kapj(DUST_SIZE_MAX,DUST_SPECIES_MAX,FRSIZE_Z)
      doubleprecision kaph(FRSIZE_Z)
      doubleprecision qfpitot0(FRSIZE_Z)
      doubleprecision qfpitot(FRSIZE_Z)
      doubleprecision tempold(DUST_SIZE_MAX,DUST_SPECIES_MAX,FRSIZE_Z)
      doubleprecision convcrit
      doubleprecision error,corfact(FRSIZE_Z)
      integer itermax,ierror
c     >>> BUGFIX 15-11-02 : ierror was doubleprecision... <<<
c
      doubleprecision j1,h0,hj1,find_planckopac,bplanck
      doubleprecision alp(FRSIZE_FREQ,FRSIZE_Z)
      doubleprecision src(FRSIZE_FREQ,FRSIZE_Z)
      doubleprecision alp00(FRSIZE_Z)
      doubleprecision intbcminus(FRSIZE_FREQ,FRSIZE_MU)
      doubleprecision t1(DUST_SIZE_MAX,DUST_SPECIES_MAX,FRSIZE_Z)
      doubleprecision t2(DUST_SIZE_MAX,DUST_SPECIES_MAX,FRSIZE_Z)
      doubleprecision t3(DUST_SIZE_MAX,DUST_SPECIES_MAX,FRSIZE_Z)
      integer iter,iz,inu,imu,itdmp,savenr,icnt_ng,idokaph
      integer isize,ispec
c
#include "common_dust.h"
c
c     If imethod eq 0, only single dust species/size
c
      if(imethod.eq.0) then
          call check_no_multidust()
      endif
c
c     Reset error signal
c
      ierror = 0
c
c     The imethod=0 is used only for demonstration purposes (to show
c     why the kaph method does not work). Normally use imethod=1.
c
      if(imethod.eq.0) then
          write(*,*) '#### Caution: Bad VET method called... ####'
          write(*,*) '####          Better use imethod=1     ####'
          stop 13
      endif
c
c     Reset dump counter
c
      itdmp  = 0
      savenr = 1
c     
c     Reset the Ng counter. The Ng counter is meant to count to 4
c     and then reset, triggering the Ng accelleration at each 4-th
c     iteration step.
c
      icnt_ng = 0
c
c     Put the Ng stuff to zero for safety
c
      do iz=1,nz
          do ispec=1,dust_nr_species
              do isize=1,dust_nr_size(ispec)
                  t3(isize,ispec,iz) = 0.d0
                  t2(isize,ispec,iz) = 0.d0
                  t1(isize,ispec,iz) = 0.d0
              enddo
          enddo
      enddo
c
c     Set the boundary conditions
c
      h0  = 0.d0
      hj1 = 1.d0 / sqrt(3.d0)    ! Initial one
c
c     The temp must be a reasonable initial guess, check this
c
      do iz=1,nz
          do ispec=1,dust_nr_species
              do isize=1,dust_nr_size(ispec)
                  if(temp(isize,ispec,iz).lt.1.d-4) then
                      write(*,*) 'ERROR in VET rad trans:'
                      write(*,*) 'Temperature initial guess ',
     %                           'is unreasonable'
                      write(*,*) temp(isize,ispec,iz),isize,ispec,iz
                      stop 13
                  endif
              enddo
          enddo
      enddo
c
c     If dump is active, then clear the convergence history dump file
c
      if(idump.gt.0) then
          open(unit=1,file='convhist.info',status='unknown')
          write(1,*) 0,1
          close(1)
          open(unit=1,file='convhist.dat',status='unknown')
          write(1,*) nz
          close(1)
      endif
c
c     Put the variable eddington factor to zero as a first guess
c     (zero means isotropic, or in other words: Eddington approximation)
c
      do iz=1,nz
          fj(iz) = 0.d0
      enddo
c
c     Compute the planck mean opacity, and put them as initial guesses 
c     for kapj and kaph
c
      do iz=1,nz
          do ispec=1,dust_nr_species
              do isize=1,dust_nr_size(ispec)
                  kapj(isize,ispec,iz) = 
     %               find_planckopac(isize,ispec,temp(isize,ispec,iz)) ! Initial guess
              enddo
          enddo
          kaph(iz) = kapj(1,1,iz)  ! Initial guess (kaph only for imethod=0)
      enddo
c
c     Put the 'old' temperature to zero, so that the iteration will
c     take at least two steps
c
      do iz=1,nz
          do ispec=1,dust_nr_species
              do isize=1,dust_nr_size(ispec)
                  tempold(isize,ispec,iz) = 0.d0
              enddo
          enddo
      enddo
c
c     Compute the alpha_nu
c
      do iz=1,nz
          do inu=1,nf
              dummy=0.d0
              do ispec=1,dust_nr_species
                  do isize=1,dust_nr_size(ispec)
                      dummy = dummy + rho(ispec,iz) * 
     %                           kappa(inu,isize,ispec)
                  enddo
              enddo
              alp(inu,iz) = dummy
          enddo
      enddo
c
c     Compute the qfpitot0
c
      do iz=1,nz
          dummy=0.d0
          do ispec=1,dust_nr_species
              do isize=1,dust_nr_size(ispec)
                  dummy = dummy + rho(ispec,iz) * 
     %                           qfpi(isize,ispec,iz)
              enddo
          enddo
          qfpitot0(iz) = dummy
      enddo
c
c     Reset the j1 and hnu (necessary for imethod.eq.1)
c
      j1 = 1.d99
      do iz=1,nz
          do inu=1,nf
              jnu(inu,iz) = 0.d0
              hnu(inu,iz) = 0.d0
          enddo
      enddo
c
c     Do the main iteration
c
      do iter=1,itermax
c
c         Message
c     
          write(*,*) 'Iteration ',iter
c
c         If initial guess, then jump right to the transfer at iter.eq.1
c
          if((init.ne.0).and.(iter.eq.1)) goto 300
c
c         Compute the viscous heat release
c
c         NOTE: Only valid in single-dust-species/size mode
c
          do iz=1,nz
              qfpitot(iz) = qfpitot0(iz) + rho(1,iz) * 
     %                      qvdivt(iz)*temp(1,1,iz)
          enddo
c
c         Increase the Ng counter
c
          icnt_ng = icnt_ng + 1
c
c         Integrate the moment equations
c          
          if(imethod.eq.0) then 
c
c             Use the simple approach with kaph (DO NOT USE!)
c     
              do iz=1,nz
                  alp00(iz)=alp(1,iz)
              enddo
              call do_integr_momeq(nz,z,qfpitot,h0,hj1,fj,
     %                             alp00,jjme,hhme)
c
c             And notify the compute_jh_meanopac_and_vet() routine
c             to indeed compute the kaph
c
              idokaph = 1 
c
          elseif(imethod.eq.1) then
c
c             Use the better method with Hnu*kappa_nu
c
              call do_integr_momnu(nz,nf,z,nu,temp,qfpitot,h0,
     %                hj1,fj,j1,hnu,alp,2,jjme,hhme,corfact)
cc########################################
c              write(*,*) iter,jjme(4),fj(50),temp(1,1,1),hj1,j1
cc########################################
c
c             Make sure the kaph is not computed (to prevent NaN)
c
              idokaph = 0
c
          elseif(imethod.eq.-1) then 
c
c             [ONLY FOR EVALUATION]
c             This method uses planck and rosseland mean opacities 
c             but still full angle-dependence.
c
              write(*,*) 'THIS MODE IS DISABLED (imethod=-1)'
              stop
c
cc              write(*,*) 'USING A TEST METHOD'
cc              call compute_meanopacs(nf,nu,kappa,temp,kapj,kaph)
cc              call do_integr_momeq(nz,z,qfpitot,h0,hj1,fj,
cc     %                             alp,jjme,hhme)
cc              idokaph = 0
c
          elseif(imethod.eq.-2) then 
c
c             [ONLY FOR EVALUATION]
c             This method uses the right mean opacities, but sets
c             the fj=0.
c
              write(*,*) 'THIS MODE IS DISABLED (imethod=-2)'
              stop
c     
cc              write(*,*) 'USING A TEST METHOD'
cc              do iz=1,nz
cc                  fj(iz) = 0.d0
cc              enddo
cc              hj1 = 1.d0/sqrt(3.d0)
cc              call do_integr_momnu(nz,nf,z,nu,temp,qfpitot,h0,
cc     %                hj1,fj,j1,hnu,alp,2,jjme,hhme,corfact)
cc              idokaph = 0
c
          else
              write(*,*) 'Do not know VET method ',imethod
              stop 13
          endif
c
c         Compute the temperatures from these equations
c
          call compute_dust_temp_from_jj(nz,qfpi,jjme,kapj,temp)
c
c         Provide for the Ng acceleration
c
          if(ing.ne.0) then
              if(icnt_ng.eq.4) then
                  write(*,*) '   Do the Ng accelleration'
                  call do_ng_accell(nz,temp,t1,t2,t3)
                  icnt_ng = 0
              endif
c    
c             Now memorize the temperature for Ng-accelleration 
c             later. This is because Ng accelleration uses the memory
c             of the previous steps to make its prediction
c             
              do iz=1,nz
                  do ispec=1,dust_nr_species
                      do isize=1,dust_nr_size(ispec)
                          t3(isize,ispec,iz) = t2(isize,ispec,iz)
                          t2(isize,ispec,iz) = t1(isize,ispec,iz)
                          t1(isize,ispec,iz) = temp(isize,ispec,iz)
                      enddo
                  enddo
              enddo
          endif
c
c         Compute the source terms and extinction terms for the
c         full transfer module
c
 300      continue
          do iz=1,nz
              do inu=1,nf
                  alp(inu,iz) = 0.d0
                  src(inu,iz) = 0.d0
                  do ispec=1,dust_nr_species
                      do isize=1,dust_nr_size(ispec)
                          dummy = rho(ispec,iz) * 
     %                         kappa(inu,isize,ispec)
                          alp(inu,iz) = alp(inu,iz) + dummy
                          src(inu,iz) = src(inu,iz) + dummy *
     %                       bplanck(temp(isize,ispec,iz),nu(inu))
                      enddo
                  enddo
              enddo
          enddo
cc########################################
c          write(*,*) '== ',iter,alp(50,50),src(50,50),temp(1,1,50),
c     %                  temp(1,2,50)
cc########################################
c
c         Set the boundary conditions for the full radiative
c         transfer: no incoming radiation
c
          do inu=1,nf
              do imu=1,nmu
                  intbcminus(inu,imu) = 0.d0
              enddo
          enddo
c
c         Now perform the full frequency-mu-dependent transfer
c
          call do_transfer(z,mu,src,alp,intj,inth,
     %                       intbcminus,intorder,nz,nmu,nf)
c
c         Compute all the vet factors etc
c
          call compute_jh_meanopac_and_vet(nz,nmu,nf,
     %              z,mu,nu,intj,inth,kappa,temp,idokaph,
     %              jj,hh,jnu,hnu,fj,fnu,kapj,kaph)
c
c         Compute the H/J ratio at z=zmax
c          
          hj1 = abs(hh(nz)/jj(nz))
c
c         Safety check
c
          if(hj1.lt.1.d-4) stop 555
c
c         Get the j1 value 
c
          j1 = jj(nz)
c
c         If its time, then dump the conv hist
c
          itdmp = itdmp + 1
          if((itdmp.ge.idump).and.(idump.gt.0)) then
              call append_vet_convhist(nz,temp,jjme,hhme,jj,hh,
     %                                 kapj,kaph,fj,savenr)
              savenr = savenr+1
              itdmp = 0
          endif
c
c         Now check the error
c
          error = 0.d0
          do iz=1,nz
              do ispec=1,dust_nr_species
                  do isize=1,dust_nr_size(ispec)
                      error=max(error,
     %                   abs(tempold(isize,ispec,iz)-
     %                       temp(isize,ispec,iz))/
     %                       (tempold(isize,ispec,iz)+
     %                        temp(isize,ispec,iz)))
                  enddo
              enddo
          enddo
          if(error.lt.convcrit) goto 10
          do iz=1,nz
              do ispec=1,dust_nr_species
                  do isize=1,dust_nr_size(ispec)
                      tempold(isize,ispec,iz)=temp(isize,ispec,iz)
                  enddo
              enddo
          enddo
      enddo
c
c     Not converged...
c
      ierror = 1
      if(idint.gt.0) then
          write(*,*) '>>> WRITING INTENSITY ARRAY <<<'
          call write_intensity(z,mu,nu,intj,inth,nz,nmu,nf)
      elseif(idint.lt.0) then
          call write_surfint(z,mu,nu,intj,inth,nz,nmu,nf)
      endif
      return
 10   continue
c
c     Converged!!!
c
      if(idint.gt.0) then
          write(*,*) '>>> WRITING INTENSITY ARRAY <<<'
          call write_intensity(z,mu,nu,intj,inth,nz,nmu,nf)
      elseif(idint.lt.0) then
          call write_surfint(z,mu,nu,intj,inth,nz,nmu,nf)
      endif
      return
c
      end






c     --------------------------------------------------------------
c       COMPUTE J,H, THE MEAN OPACITIES AND THE VARIABLE EDD FACT           
c
c     This subroutine computes all that is necessary for the mean
c     variable eddington radiative transfer. One has to give it the
c     full radiation field I(nu,mu,z), the opacity kappa(nu) and
c     the gas temperature T(z), and it will compute the mean intens
c     J_nu(z), the Edd flux H_nu(z), their freq-integrated versions 
c     J(z) and H(z), the variable eddington factor f_nu(z), and its
c     freq-integrated version f(z), and the mean opacities kappa(T),
c     kappa(J) and kappa(H).
c
c     ARGUMENTS:
c        z        = Z-grid
c        mu       = mu-grid (radiation angles)
c        nu       = frequency-grid
c        intj     = The Feautrier j(mu,nu) = 0.5*(I(mu,nu)+I(-mu,nu))
c        inth     = The Feautrier h(mu,nu) = 0.5*(I(mu,nu)-I(-mu,nu))
c        kappa    = The full (temperature INdependent) opacity
c        temp     = The gas/dust temperature 
c        idokaph  = 0= do not compute kaph, 1= yes, do kaph as well
c
c     OUTPUT:
c        jj       = The freq-integrated mean intensity
c        hh       = The freq-integrated eddington flux H = F / 4 pi
c        jnu      = Mean intensity J_nu 
c        Hnu      = Eddington flux H_nu = F_nu / 4 pi
c        fj       = The freq-integrated eddington factor f = K / J
c        fnu      = The eddington factor f_nu = K_nu / J_nu
c        kapj     = The J-mean opacity
c        kaph     = The H-mean opacity
c 
c     --------------------------------------------------------------
      subroutine compute_jh_meanopac_and_vet(nz,nmu,nf,
     %              z,mu,nu,intj,inth,kappa,temp,idokaph,
     %              jj,hh,jnu,hnu,fj,fnu,kapj,kaph)
      implicit none
c
      integer nz,nmu,nf,idokaph
      doubleprecision z(FRSIZE_Z),mu(FRSIZE_MU),nu(FRSIZE_FREQ)
      doubleprecision intj(FRSIZE_FREQ,FRSIZE_MU,FRSIZE_Z)
      doubleprecision inth(FRSIZE_FREQ,FRSIZE_MU,FRSIZE_Z)
      doubleprecision kappa(FRSIZE_FREQ,DUST_SIZE_MAX,DUST_SPECIES_MAX)
      doubleprecision temp(DUST_SIZE_MAX,DUST_SPECIES_MAX,FRSIZE_Z)
      doubleprecision jj(FRSIZE_Z),hh(FRSIZE_Z)
      doubleprecision jnu(FRSIZE_FREQ,FRSIZE_Z)
      doubleprecision hnu(FRSIZE_FREQ,FRSIZE_Z)
      doubleprecision fj(FRSIZE_Z)
      doubleprecision fnu(FRSIZE_FREQ,FRSIZE_Z)
      doubleprecision kapj(DUST_SIZE_MAX,DUST_SPECIES_MAX,FRSIZE_Z)
      doubleprecision kaph(FRSIZE_Z)
c
      integer hflag(FRSIZE_Z)
      doubleprecision enum,denom,bnuold,bnucur,bplanck
      integer iz,inu,imu,isize,ispec
c
#include "common_dust.h"
c
c     First compute the J_nu, H_nu and their freq-integrated 
c     counterparts J and H
c
      do iz=1,nz
          do inu=1,nf
              jnu(inu,iz) = 0.d0
              hnu(inu,iz) = 0.d0
              do imu=2,nmu
                  jnu(inu,iz) = jnu(inu,iz) + 0.5 * 
     %                 ( intj(inu,imu,iz) + intj(inu,imu-1,iz) ) *
     %                 ( mu(imu) - mu(imu-1) )
                  hnu(inu,iz) = hnu(inu,iz) + 0.5 * 
     %                 ( inth(inu,imu,iz) * mu(imu) + 
     %                   inth(inu,imu-1,iz) * mu(imu-1) ) *
     %                 ( mu(imu) - mu(imu-1) )
              enddo
              jnu(inu,iz) = jnu(inu,iz) + mu(1) * intj(inu,imu,iz)
          enddo
          jj(iz) = 0.d0
          hh(iz) = 0.d0
          do inu=2,nf
              jj(iz) = jj(iz) + 0.5 * ( jnu(inu,iz) + jnu(inu-1,iz) ) *
     %                    abs( nu(inu) - nu(inu-1) ) 
              hh(iz) = hh(iz) + 0.5 * ( hnu(inu,iz) + hnu(inu-1,iz) ) *
     %                    abs( nu(inu) - nu(inu-1) ) 
          enddo
      enddo
c
c     Compute the frequency-dependent Eddington factor (the traceless 
c     one, i.e. isotropic --> fnu=0), and the freq-integrated
c     counterpart f
c
      do iz=1,nz
          do inu=1,nf
              fnu(inu,iz) = 0.d0
              do imu=2,nmu
                  fnu(inu,iz) = fnu(inu,iz) + 0.5 * 
     %                 ( intj(inu,imu,iz) * mu(imu)**2 + 
     %                   intj(inu,imu-1,iz) * mu(imu-1)**2 ) *
     %                 ( mu(imu) - mu(imu-1) )
              enddo
              fnu(inu,iz) = fnu(inu,iz) / jnu(inu,iz) - (1.d0/3.d0)
          enddo
          enum = 0.d0
          do inu=2,nf
              enum = enum + 0.5 * ( fnu(inu,iz) * jnu(inu,iz) + 
     %                  fnu(inu-1,iz) * jnu(inu-1,iz) ) *
     %                  abs( nu(inu) - nu(inu-1) ) 
          enddo
          fj(iz) = enum / jj(iz)
      enddo
c
c     At z=0 the flux is always 0, so artificially make it non-zero
c      
      do inu=1,nf
          hnu(inu,1) = hnu(inu,2)
      enddo
      hh(1) = hh(2)
c
c     Check if |H| <<< J. If so, we need to use the Rosseland 
c     mean opacity for the kaph, otherwise things might blow up
c     (NOTE: only for imethod.eq.0, so normally not used!)
c
      if(idokaph.ne.0) then
          do iz=1,nz
              hflag(iz) = 0
              if(abs(hh(iz)).lt.HFLUXMIN*jj(iz)) then
                  hflag(iz) = 1
                  write(*,*) 'H=0 at ',iz
                  stop 13       ! For the moment we simply stop!
              endif
          enddo
      endif
c
c     Compute the J-mean opacity
c
      do iz=1,nz
          do ispec=1,dust_nr_species
              do isize=1,dust_nr_size(ispec)
                  enum = 0.d0
                  do inu=2,nf
                      enum = enum + 0.5 * 
     %                  ( kappa(inu,isize,ispec) * jnu(inu,iz) +
     %                    kappa(inu-1,isize,ispec) * jnu(inu-1,iz) ) *
     %                    abs( nu(inu) - nu(inu-1) )
                  enddo
                  kapj(isize,ispec,iz) = enum / jj(iz)
              enddo
          enddo
      enddo
c
c     Compute the H-mean opacity
c     (NOTE: only for imethod.eq.0, so normally not used!)
c
      if(idokaph.ne.0) then
          do iz=1,nz
              enum = 0.d0
              do inu=2,nf
                  enum = enum + 0.5 * ( kappa(inu,1,1) * hnu(inu,iz) +
     %                 kappa(inu-1,1,1) * hnu(inu-1,iz) ) *
     %                 abs( nu(inu) - nu(inu-1) )
              enddo
              kaph(iz) = enum / hh(iz)
          enddo
      endif
c
      end




c     --------------------------------------------------------------
c           [EVALUATION] COMPUTE ROSS AND PLANCK MEAN OPACS
c     --------------------------------------------------------------
      subroutine compute_meanopacs(nf,nu,kappa,temp,kapp,kapr)
      implicit none
c
      integer nf,inu
      doubleprecision nu(nf),kappa(nf),temp,kapp,kapr,dum
      doubleprecision bnu(FRSIZE_FREQ),dbdt(FRSIZE_FREQ)
      doubleprecision bplanck,bplanckdt
c
      do inu=1,nf
          bnu(inu)  = bplanck(temp,nu(inu))
          dbdt(inu) = bplanckdt(temp,nu(inu))
      enddo
      kapp = 0.d0
      do inu=2,nf
          kapp = kapp + 0.5 * ( kappa(inu)*bnu(inu) + 
     %                          kappa(inu-1)*bnu(inu) ) *
     %                   abs( nu(inu) - nu(inu-1) )
          dum  = dum  + 0.5 * ( bnu(inu) + bnu(inu-1) ) *
     %                   abs( nu(inu) - nu(inu-1) )
      enddo
      kapp = kapp / dum
      kapr = 0.d0
      do inu=2,nf
          kapr = kapr + 0.5 * ( dbdt(inu)/kappa(inu) + 
     %                          dbdt(inu)/kappa(inu-1) ) *
     %                   abs( nu(inu) - nu(inu-1) )
          dum  = dum  + 0.5 * ( dbdt(inu) + dbdt(inu-1) ) *
     %                   abs( nu(inu) - nu(inu-1) )
      enddo
      kapr= dum / kapr
c
      end



c     --------------------------------------------------------------
c                APPEND THE CONVERGENCE HISTORY FILE
c
c     This subroutine is only for convenience: it saves the stuff
c     as it converges towards a solution. This is handy for debug
c     and demonstration purposes.
c     --------------------------------------------------------------
      subroutine append_vet_convhist(nz,temp,jjme,hhme,jj,hh,
     %                    kapj,kaph,fj,savenr)
      implicit none
      integer nz,savenr
      doubleprecision temp(DUST_SIZE_MAX,DUST_SPECIES_MAX,FRSIZE_Z)
      doubleprecision jj(FRSIZE_Z),hh(FRSIZE_Z)
      doubleprecision kapj(FRSIZE_Z),kaph(FRSIZE_Z)
      doubleprecision fj(FRSIZE_Z),jjme(FRSIZE_Z),hhme(FRSIZE_Z)
c
      integer iz
c
      open(unit=1,file='convhist.dat',status='old',access='append')
      do iz=1,nz 
          write(1,10) temp(1,1,iz),jjme(iz),hhme(iz),jj(iz),hh(iz),
     %              kapj(iz),kaph(iz),fj(iz)
 10       format(9(E13.6,1X))
      enddo
      write(1,*)
      close(1)
      open(unit=1,file='convhist.info',status='unknown')
      write(1,*) savenr,1
      close(1)
c
      end



c     --------------------------------------------------------------
c                  WRITE THE FULL INTENSITY ARRAY
c
c     This subroutine writes the entire intensity array to file. 
c     Note that this is a huge file. It is only useful for very
c     detailed debugging. For most other applications the other
c     subroutine write_surfint() is more useful: only the intensity
c     at the surface is written.
c     --------------------------------------------------------------
      subroutine write_intensity(z,mu,freq,intj,inth,nz,nmu,nf)
      implicit none
      integer nz,nmu,nf
      doubleprecision z(FRSIZE_Z)
      doubleprecision mu(FRSIZE_MU)
      doubleprecision freq(FRSIZE_FREQ)
      doubleprecision intj(FRSIZE_FREQ,FRSIZE_MU,FRSIZE_Z)
      doubleprecision inth(FRSIZE_FREQ,FRSIZE_MU,FRSIZE_Z)
c
      integer inu,imu,iz
c
      open(unit=1,file='intensity.dat',status='unknown')
      write(1,*) nf,nmu,nz
      write(1,*) 
      do iz=1,nz
          write(1,5) z(iz)
 5        format(E13.6)
      enddo
      write(1,*) 
      do imu=1,nmu
          write(1,5) mu(imu)
      enddo
      write(1,*) 
      do inu=1,nf
          write(1,5) freq(inu)
      enddo
      write(1,*) 
      do iz=1,nz
          do imu=1,nmu
              do inu=1,nf
                  if(abs(inth(inu,imu,iz)).gt.1d-98) then
                      write(1,10) intj(inu,imu,iz),
     %                            inth(inu,imu,iz)
                  else
                      write(1,10) intj(inu,imu,iz),0.d0
                  endif
 10               format(2(E13.6,1X))
              enddo
          enddo
      enddo
      close(1)
c     
      end

c     --------------------------------------------------------------
c                  WRITE ONLY THE SURFACE INTENSITY ARRAY
c     --------------------------------------------------------------
      subroutine write_surfint(z,mu,freq,intj,inth,nz,nmu,nf)
      implicit none
      integer nz,nmu,nf
      doubleprecision z(FRSIZE_Z)
      doubleprecision mu(FRSIZE_MU)
      doubleprecision freq(FRSIZE_FREQ)
      doubleprecision intj(FRSIZE_FREQ,FRSIZE_MU,FRSIZE_Z)
      doubleprecision inth(FRSIZE_FREQ,FRSIZE_MU,FRSIZE_Z)
c
      integer inu,imu
c
      open(unit=1,file='surfintens.dat',status='unknown')
 5    format(E13.6)
      write(1,*) nf,nmu
      write(1,*) 
      do imu=1,nmu
          write(1,5) mu(imu)
      enddo
      write(1,*) 
      do inu=1,nf
          write(1,5) freq(inu)
      enddo
      write(1,*) 
      do imu=1,nmu
          do inu=1,nf
              write(1,10) max(intj(inu,imu,nz),1d-99),
     %             max(inth(inu,imu,nz),1d-99)
 10           format(2(E13.6,1X))
          enddo
      enddo
      close(1)
c     
      end
