c     ==============================================================
c        PROGRAM THAT PERFORMS RADIAL RADIATIVE TRANSFER IN DISKS
c                    (C) 2001  by  C.P. Dullemond
c
c     This program reads a density structure rho(r,z) of a dust
c     disk, the specifications of the central star, and the dust
c     opacities. It assumes that scattering is off. It then performs
c     a frequency-dependent extinction radiative transfer along 
c     radial rays. No re-emission is computed, only the extinction.
c     The dust temperature according to this primary stellar 
c     radiation field is then computed (without the contribution
c     from the re-emission), and stored in a file primarytemp.dat.
c     ==============================================================

#include "main.h"

#define GASTODUST (100.d0)

c     --------------------------------------------------------------
c                             MAIN ROUTINE
c     --------------------------------------------------------------
      program main
      implicit none
c
      call read_all_data()
      call make_star_spectrum()
      call do_rt_starlight()
      call write_temperatures()
ccc      call write_meanint()
c
      end




c     --------------------------------------------------------------
c                          READ ALL THE DATA
c     --------------------------------------------------------------
      subroutine read_all_data()
      implicit none
c
#include "common_grid.h"
#include "common_radialrt.h"
#include "common_dust.h"
c
      integer iformat,ispec,isize,i
      integer nr,nz,ir,iz,nf,inu,nrd,nrspec,nrsize(DUST_SPECIES_MAX)
      doubleprecision a,b,c,d,e
      doubleprecision frq,dummy
      logical f_ex,multi
      logical fexf,fexo,fexs
c
c     Read the star information
c
      open(unit=1,file='starinfo.inp',status='old')
      read(1,*) iformat
      read(1,*) star_r
      read(1,*) star_m
      read(1,*) star_t
      close(1)
c     
c     Read the current density structure
c
      open(unit=1,file='diskstruct_0.dat',status='old')
      read(1,*) nr,nz
      nrspec=1
      nrsize(1)=1
      if(nr.lt.1) then
c
c         New format: multiple sizes/species
c
          read(1,*) nr,nz
          read(1,*) nrspec
          if(nrspec.gt.DUST_SPECIES_MAX) then
              write(*,*) 'ERROR: Too many species specified.'
              write(*,*) '       Recompile with larger max specs.'
              stop
          endif
          do ispec=1,nrspec
              read(1,*) nrsize(ispec)
              if(nrsize(ispec).gt.DUST_SIZE_MAX) then
                  write(*,*) 'ERROR: Too many sizes specified.'
                  write(*,*) '       Recompile with larger max size.'
                  stop
              endif
          enddo
          multi=.true.
      else
          nrspec=1
          nrsize(1)=1
          multi=.false.
      endif
      if(nr.gt.FRSIZE_R) then
          write(*,*) 'ERROR in radialrt:'
          write(*,*) 'R grid too large'
          stop 13
      endif
      if(nz.gt.FRSIZE_Z) then
          write(*,*) 'ERROR in radialrt:'
          write(*,*) 'Z grid too large'
          stop 13
      endif
      do ir=1,nr
          do iz=1,nz
              if(.not.multi) then
                  read(1,*) a,b,c,d,e
                  grid_r(ir) = a
                  grid_z(iz,ir) = b
                  rrt_temp(1,1,iz,ir) = d
                  rrt_rho(1,iz,ir) = c / GASTODUST
                  rrt_vz(iz,ir) = e
              else
                  read(1,*) a,b,c,e
                  grid_r(ir) = a
                  grid_z(iz,ir) = b
                  rrt_rhogas(iz,ir) = c
                  rrt_vz(iz,ir) = e
                  do ispec=1,nrspec
                      read(1,*) a
                      rrt_rho(ispec,iz,ir) = a * c / GASTODUST
                  enddo
                  do ispec=1,nrspec
                      do isize=1,nrsize(ispec)
                          read(1,*) d
                          rrt_temp(isize,ispec,iz,ir) = d
                      enddo
                  enddo
              endif
          enddo
      enddo
      close(1)
c
c     Read the flaring index file
c
      inquire(file='flareindex.inp',exist=f_ex)
      if(.not.f_ex) then 
          write(*,*) 'For this alpha-angle version of radial rt'
          write(*,*) 'we need a file flareindex.inp'
          stop 13
      endif
      open(unit=1,file='flareindex.inp',status='old')
      read(1,*) nrd
      if(nrd.ne.nr) then 
          write(*,*) 'File flareindex.inp has wrong dimensions in r'
          stop 13
      endif
      do ir=1,nr
          read(1,*) rrt_flareindex(ir)
          if((rrt_flareindex(ir).gt.1.d0).or.
     %       (rrt_flareindex(ir).le.0.d0)) then
              write(*,*) 'ERROR: Flaring index should be around 2/7'
              write(*,*) '       or so, but not > 1 or <=0'
              write(*,*) '       At ir=',ir,'  flare=',
     %                      rrt_flareindex(ir)
              stop 13
          endif
      enddo
      close(1)
c
c     If present, read the starvis.inp file, which is the file containing
c     numbers between 0 and 1 saying how much of the stellar surface is
c     visible.
c
      inquire(file='starvis.inp',exist=fexf)
      if(fexf) then
          open(unit=1,file='starvis.inp',status='old')
          read(1,*) nrd
          if(nrd.ne.nr) then
              write(*,*) 'File starvis.inp has wrong NR'
              stop 13
          endif
          do ir=1,nr
              read(1,*) rrt_starvis(ir)
          enddo
          close(1)
      else
          write(*,*) '  File starvis.inp not found. Assume starvis=1.0.'
          do ir=1,nr
              rrt_starvis(ir) = 1.d0
          enddo
      endif
c
c     Read the frequency array and the opacities
c
      inquire(file='kappaplanck.inp',exist=fexf)
      if(fexf) then
          write(*,*) 'PROBLEM: kappaplanck.inp is obsolete...'
          write(*,*) 'TIP    : mv kappaplanck.inp planckopac.inp'
          write(*,*) '         and add two colums with 0.d0'
          write(*,*) '         (i.e. kap_scat and g_scat).'
          stop 13
      endif
      inquire(file='planckopac.inp',exist=fexo)
      if(fexo) then
c
c         Planck Opacity file is there, so use it. If the frequency
c         dependent opacities are also there, then they are assumed
c         to be the opacities from which the planck opacities were
c         derived. The frequency-dependent opacitities can then be
c         used to make the SED.
c     
          freq_nr = 1
          call read_planckopac()
c
c         Check if also dust opacs present
c
          inquire(file='frequency.inp',exist=fexf)
          inquire(file='dustopac_1.inp',exist=fexo)
          if(fexf.and.fexo) then
              write(*,*) '   (Read planck opacity from planckopac.inp)'
          endif
c
c         Check for multi species sizes
c
          if((nrspec.gt.1).or.(nrsize(1).gt.1)) then
              write(*,*) 'Mean opacity mode incompatible with multi'
              write(*,*) 'dust species/sizes.'
              stop
          endif
c
c         WARNING since this mode has not been tested for long time
c
          write(*,*) 'Mean opacity mode disactivated!!!'
          stop
      else
c
c         Find the frequency-dependent opacities, and do the radial
c         transfer with these opacitities (do it with frequency
c         dependence).
c
          inquire(file='frequency.inp',exist=fexf)
          inquire(file='dustopac.inp',exist=fexo)
          if((.not.fexf).or.(.not.fexo)) then
              write(*,*) 'Need both frequency.inp and dustopac.inp'
              stop 13
          endif
          open(unit=1,file='frequency.inp',status='old')
          read(1,*) nf
          if(nf.gt.FRSIZE_FREQ) then
              write(*,*) 'ERROR in radialrt:'
              write(*,*) 'Frequency grid too large'
              stop 13
          endif
          do inu=1,nf
              read(1,*) a
              freq_nu(inu) = a
          enddo
          freq_nr = nf
          call read_dustdata()
c
c         Check if nr of species and sizes consistent
c
          if(nrspec.ne.dust_nr_species) then
              write(*,*) 'ERROR: Nr of dust species in dustopac*.inp'
              write(*,*) '   is unequal to the number specified in '
              write(*,*) '   diskstruct_0.dat'
              stop
          endif
          do ispec=1,dust_nr_species
              if(nrsize(ispec).ne.dust_nr_size(ispec)) then
                  write(*,*) 'ERROR: Nr of dust sizes in dustopac*.inp'
                  write(*,*) '   is unequal to the number specified in '
                  write(*,*) '   diskstruct_0.dat'
                  stop
              endif
          enddo
      endif
c
c     Check
c
      if((freq_nr.eq.1).and.((dust_nr_species.gt.1).or.
     %                       (dust_nr_size(1).gt.1))) then
          write(*,*) 'ERROR: Inconsistency with sizes/species'
          stop 
      endif
c
c     Check if domain.inp exists, else make maximum domain
c
      inquire(file='domain.inp',exist=f_ex)
      if(f_ex) then
          open(unit=1,file='domain.inp',status='old')
          read(1,*) grid_irstart,grid_irend
          close(1)
          if((grid_irstart.lt.1).or.(grid_irend.gt.nr).or.
     %       (grid_irstart.gt.grid_irend)) then
              write(*,*) 'ERROR: The domain specified in domain.inp'
              write(*,*) '       is out of range.'
              stop
          endif
      else
          grid_irstart = 1
          grid_irend   = nr
      endif
c
c     Read (if required) the input spectrum
c
      if(star_t.le.0.d0) then
          inquire(file='starspectrum.inp',exist=fexs)
          if(.not.fexs) then
              write(*,*) 'ERROR: Must specify either star_t '
              write(*,*) '       or starspectrum.inp'
              stop
          endif
          write(*,*) 'Reading starspectrum.inp'
          open(unit=1,file='starspectrum.inp')
          read(1,*) i
          if(i.ne.freq_nr) then
              write(*,*) 'Sorry, stellar spectrum in starspectrum.inp'
              write(*,*) 'does not have the same number of freqs as ',
     %             'other input files.'
              stop 13
          endif
          if(i.eq.1) then
              read(1,*) frq,dummy
              star_lum(1) = 1.1965280793d38 * dummy 
          else
              do i=1,freq_nr
                  read(1,*) frq,dummy 
                  if(abs(frq-freq_nu(i))/(frq+freq_nu(i)).gt.1.d-3) then
                      write(*,*) 'PROBLEM: Frequency grid of stellar ',
     %                     'spectrum unequal to frequency.inp'
                      write(*,*) frq,freq_nu(i)
                      stop 13
                  endif
                  star_lum(i) = 1.1965280793d38 * dummy 
              enddo
          endif
          close(1)
      endif
c
      grid_nr = nr
      grid_nz = nz
c
      end




c     --------------------------------------------------------------
c                       MAKE THE STARSPECTRUM
c     --------------------------------------------------------------
      subroutine make_star_spectrum()
      implicit none
c
#     include "common_natconst.h"
#     include "common_grid.h"
#     include "common_radialrt.h"
c
      doubleprecision bplanck
      integer inu
c
      if(freq_nr.eq.1) then
          star_lum(1) = 4 * pi * star_r**2 * ss * star_t**4
      elseif(star_t.gt.0.d0) then
          do inu=1,freq_nr
              star_lum(inu) = 4 * pi * star_r**2 * pi *
     %             bplanck(star_t,freq_nu(inu)) 
          enddo
      else
          star_t = 10.
      endif
c
      end




c     --------------------------------------------------------------
c               DO SIMPLE RADIATIVE TRANSFER FOR STARLIGHT
c     --------------------------------------------------------------
      subroutine do_rt_starlight()
      implicit none
c
#     include "common_grid.h"
#     include "common_radialrt.h"
c
      integer ir,iz
c
c     Do radial RT
c
      do ir=grid_irstart,grid_irend
          do iz=1,grid_nz
              call integrate_radial_rt(ir,iz,rrt_temp(1,1,iz,ir))
          enddo
      enddo
c
      end




c     --------------------------------------------------------------
c               RADIATIVE TRANSFER ALONG RADIAL RAYS
c
c     This subroutine performs RT from the star to a certain point
c     (r,z) = (r_irpt,z_izpt) in the disk. It returns the mean intensity
c     as its last argument. During the calculation, though, it will
c     interpret slum as the luminosity as inferred from a certain point.
c     At then end of the subroutine this will be converted into a
c     mean intensity.
c     --------------------------------------------------------------
      subroutine integrate_radial_rt(irpt,izpt,temp)
      implicit none
      integer irpt,izpt
      doubleprecision temp(DUST_SIZE_MAX,DUST_SPECIES_MAX)
c
      doubleprecision bb,z_coord_lin_ext
c
#     include "common_grid.h"
#     include "common_radialrt.h"
#     include "common_dust.h"
c
      doubleprecision ds,find_dust_kappa,angle,find_planckopac
      doubleprecision solve_dust_temperature,told,t0,tempalt,kapalt
      doubleprecision slum(1:FRSIZE_FREQ),meanint(1:FRSIZE_FREQ)
      doubleprecision alpha,rho(DUST_SPECIES_MAX)
      integer izz,inu,ii,isize,ispec
c
c     Copy the stellar luminosity into first bin
c
      do inu=1,freq_nr
          slum(inu) = star_lum(inu) * rrt_starvis(irpt) 
      enddo
c
c     Main loop
c
      if(izpt.lt.grid_nz) then
          do izz=izpt+1,grid_nz
              angle = 0.4d0 * star_r / grid_r(irpt) + 
     %             rrt_flareindex(irpt) * grid_z(izz,irpt) / 
     %             grid_r(irpt) 
              ds    = grid_z(izz,irpt)-grid_z(izz-1,irpt)
              do ispec=1,dust_nr_species
                  rho(ispec) = 0.5 * ( rrt_rho(ispec,izz,irpt) + 
     %                                 rrt_rho(ispec,izz-1,irpt) )
              enddo
              do inu=1,freq_nr
                  alpha = 0.d0
                  do ispec=1,dust_nr_species
                      do isize=1,dust_nr_size(ispec)
                          alpha = alpha + rho(ispec) * 
     %                       find_dust_kappa(inu,isize,ispec,
     %                       star_t,1,0) / angle
                      enddo
                  enddo
                  slum(inu) = slum(inu) * exp(-ds * alpha)
                  if(slum(inu).lt.1.d-99) then
                      slum(inu) = 0.d0
                  endif
              enddo
          enddo
      endif
c
c     Compute the mean intensity
c
c                  F_nu         L_nu         
c         J_nu  =  ----  =   -----------   
c                  4 pi      16 pi^2 R^2 
c
      do inu=1,freq_nr
          meanint(inu) = 6.332574d-3 * slum(inu) / (grid_r(irpt)**2)
      enddo
c
c     (for debugging)  c###################c
c
      if(freq_nr.gt.1) then
          db_meanint(izpt,irpt) = 0.d0
          do inu=2,freq_nr
              db_meanint(izpt,irpt) = db_meanint(izpt,irpt) +
     %             0.5 * ( meanint(inu) + meanint(inu-1) ) * 
     %             abs(freq_nu(inu)-freq_nu(inu-1))
          enddo
      else
          db_meanint(izpt,irpt) = meanint(1)
      endif
c
c     Compute the dust temperature
c
      if(freq_nr.gt.1) then
          do ispec=1,dust_nr_species
              do isize=1,dust_nr_size(ispec)
                  temp(isize,ispec) = solve_dust_temperature(meanint,
     %                                        isize,ispec)
              enddo
          enddo
      else
c
c         Disactivated for the moment
c
          write(*,*) 'MEAN OPACITY MODE DISABLED'
          stop
c
cc          t0        = (meanint(1)*5.540d4)**0.25d0
cc          temp(1,1) = t0
cc          do ii=1,60
cc              told  = temp(1,1)
cc              temp(1,1)  = t0*(kappa(1) / 
cc     %                     find_planckopac(temp(1,1)))**0.25
cc              if(abs(temp(1,1)-told)/(temp(1,1)+told+1d-4)
cc     %             .lt.TEMPACCUR) goto 20
cc          enddo
cc          write(*,*) 'ERROR: Temperature not converged'
cc          write(*,*) '   temp,told,ts0,kapstar,find_planckopac=',
cc     %         irpt,izpt,
cc     %         temp(1,1),told,meanint(1),kappa(1),
cc     %                 find_planckopac(temp(1,1))
cc          stop 9323
cc 20       continue
      endif
c
      return
      end




c     --------------------------------------------------------------
c                THE BLACKBODY PLANCK FUNCTION B_nu(T)
c
c     This function computes the Blackbody function 
c
c                    2 h nu^3 / c^2
c        B_nu(T)  = -------------------    [ erg / cm^2 s ster Hz ]
c                   exp(-h nu / kT) - 1
c
c     ARGUMENTS:
c        nu    [Hz]            = Frequency
c        temp  [K]             = Electron temperature
c     --------------------------------------------------------------
      function bplanck(temp,nu)
      implicit none
      doubleprecision temp
      doubleprecision nu
      doubleprecision bplanck
c
      if(temp.eq.0.d0) then 
          bplanck = 0.d0
          return
      endif
c
      bplanck = 1.47455d-47 * nu * nu * nu / 
     %      (exp(4.7989d-11 * nu / temp)-1.d0) + 1.d-290
c
      return
      end


c     --------------------------------------------------------------
c                       WRITE THE TEMPERATURES
c     --------------------------------------------------------------
      subroutine write_temperatures()
      implicit none
c
#include "common_grid.h"
#include "common_radialrt.h"
#include "common_dust.h"
c
      integer ir,iz,ispec,isize
      logical multi
c
      multi=.false.
      if((dust_nr_species.gt.1).or.(dust_nr_size(1).gt.1)) then
          multi=.true.
      endif
      open(unit=1,file='diskstruct_1.dat',status='unknown')
      if(multi) then
          write(1,*) -1,0
          write(1,*) grid_nr,grid_nz
          write(1,*) dust_nr_species
          do ispec=1,dust_nr_species
              write(1,*) dust_nr_size(ispec)
          enddo
      else
          write(1,*) grid_nr,grid_nz
      endif
      write(1,*)
      do ir=1,grid_nr
          do iz=1,grid_nz
              if(.not.multi) then
                  write(1,10) grid_r(ir),grid_z(iz,ir),
     %                        rrt_rho(1,iz,ir)*GASTODUST,
     %                        rrt_temp(1,1,iz,ir),rrt_vz(iz,ir)
 10               format(5(E13.6,1X))
              else
                  write(1,20) grid_r(ir),grid_z(iz,ir),
     %                        rrt_rhogas(iz,ir),rrt_vz(iz,ir)
 20               format(3(E13.6,1X))
                  do ispec=1,dust_nr_species
                      write(1,30) rrt_rho(ispec,iz,ir)*GASTODUST/
     %                            rrt_rhogas(iz,ir)
                  enddo
                  do ispec=1,dust_nr_species
                      do isize=1,dust_nr_size(ispec)
                          write(1,30) rrt_temp(isize,ispec,iz,ir)
                      enddo
                  enddo
 30               format(1(E13.6,1X))
              endif
          enddo
      enddo
      close(1)
c
      end


c     --------------------------------------------------------------
c                       WRITE THE MEAN INTENSITY
c     --------------------------------------------------------------
      subroutine write_meanint()
      implicit none
c
#include "common_grid.h"
#include "common_radialrt.h"
c
      integer ir,iz
c
      open(unit=1,file='meanint.dat',status='unknown')
      write(1,*) grid_nr,grid_nz
      write(1,*)
      do ir=1,grid_nr
          do iz=1,grid_nz
              write(1,10) db_meanint(iz,ir)+1d-90
 10           format(E13.6)
          enddo
      enddo
      close(1)
c
      end


c     --------------------------------------------------------------
c                    READ THE PLANCK MEAN OPACITY
c     --------------------------------------------------------------
      subroutine read_planckopac
      implicit none
c
#     include "common_grid.h"
#     include "common_radialrt.h"
c
      doubleprecision dum1,dum2,dum3,dum4
      integer it,nt
      logical fex
c
      open(unit=1,file='planckopac.inp')
      read(1,*) dopac_nt
      if(dopac_nt.gt.NR_TEMPERATURE) then
          write(*,*) 'Too many temperatures in kappaplanck.inp'
          write(*,*) ' Increase NR_TEMPERATURE and recompile'
          stop 13
      endif
      do it=1,dopac_nt
          read(1,*) dum1,dum2,dum3,dum4
          dopac_temp(it)=dum1
          dopac_kappa(it)=dum2
          dopac_kscat(it)=dum3
          dopac_gscat(it)=dum4
      enddo
      close(1)
c
      end



c     --------------------------------------------------------------
c                  INTERPOLATE THE PLANCK MEAN OPACITY
c     --------------------------------------------------------------
      function find_planckopac(temp)
      implicit none
      doubleprecision find_planckopac,temp
c
#     include "common_grid.h"
#     include "common_radialrt.h"
c
      doubleprecision dum1,dum2,dt
      integer it
c
      if(dopac_nt.le.1) then 
          find_planckopac = dopac_kappa(1)
          return
      endif
c
      if(temp.lt.dopac_temp(1)) then 
          find_planckopac = dopac_kappa(1)
      elseif(temp.gt.dopac_temp(dopac_nt)) then
          find_planckopac = dopac_kappa(dopac_nt)
      else
          call hunt(dopac_temp,dopac_nt,temp,it)
          dt = (temp-dopac_temp(it))/
     %            (dopac_temp(it+1)-dopac_temp(it))
          if((dt.lt.0.d0).or.(dt.gt.1.d0)) stop 552
          find_planckopac = (1.d0-dt)*dopac_kappa(it)+
     %       dt*dopac_kappa(it+1)
      endif
c
      return
      end


ccc     --------------------------------------------------------------
ccc                INTERPOLATE THE PLANCK MEAN SCAT OPAC
ccc     --------------------------------------------------------------
cc      subroutine find_planckall(temp,cabs,cscat,g)
cc      implicit none
cc      doubleprecision temp,cabs,cscat,g
ccc
cc#     include "common_grid.h"
cc#     include "common_radialrt.h"
ccc
cc      doubleprecision dum1,dum2,dt
cc      integer it
ccc
cc      if(dopac_nt.le.1) then 
cc          cabs  = dopac_kappa(1)
cc          cscat = dopac_kscat(1)
cc          g     = dopac_gscat(1)
cc          return
cc      endif
ccc
cc      if(temp.lt.dopac_temp(1)) then 
cc          cabs  = dopac_kappa(1)
cc          cscat = dopac_kscat(1)
cc          g     = dopac_gscat(1)
cc      elseif(temp.gt.dopac_temp(dopac_nt)) then
cc          cabs  = dopac_kappa(dopac_nt)
cc          cscat = dopac_kscat(dopac_nt)
cc          g     = dopac_gscat(dopac_nt)
cc      else
cc          call hunt(dopac_temp,dopac_nt,temp,it)
cc          dt = (temp-dopac_temp(it))/
cc     %            (dopac_temp(it+1)-dopac_temp(it))
cc          if((dt.lt.0.d0).or.(dt.gt.1.d0)) stop 552
cc          cabs  = (1.d0-dt)*dopac_kappa(it)+
cc     %       dt*dopac_kappa(it+1)
cc          cscat = (1.d0-dt)*dopac_kscat(it)+
cc     %       dt*dopac_kscat(it+1)
cc          g     = (1.d0-dt)*dopac_gscat(it)+
cc     %       dt*dopac_gscat(it+1)
cc      endif
ccc
cc      return
cc      end

