c     ==============================================================
c              A TRIVIAL 1-D SLAB GEOMETRY MONTE-CARLO CODE
c
c     We can do the MC in two different ways:
c
c       1) Follow photon packages until they are absorbed, then
c          release all their energy in that cell
c
c       2) Follow a photon package as it loses its energy by
c          dE/ds = - rho kappa_abs E, until one reaches E=Emin.
c
c     The two ways are switched by 'mctype'.
c
c     The gridding in tauz goes as follows: tauz goes from 
c     tauz(1)=0.0 to tauz(ntau)=maximum. So there are ntau-1 cells,
c     each cell j being in between tauz(j) and tauz(j+1). The
c     histogram is for each of the ntau cells. The value of the
c     histogram at index j=ntau is zero, because this is not a
c     cell. 
c
c     MODIFIED 25-02-03: From now on one can also have an albedo
c                        that is non-constant.      
c
c     ==============================================================

#include "main.h"

#ifndef MAXNRSCAT
#define MAXNRSCAT 1000
#endif

c     --------------------------------------------------------------
c                       LAUNCH A SERIES OF PHOTONS
c
c     This routine lauches a series of photons at a fixed incident
c     angle, and keeps track of the statistics of how many photons
c     are diffusively reflected and how many are absorbed. It also
c     keeps a histogram of where (how deep in tauz) the photons are
c     absorbed. This histogram will finally be normalized so that
c     the integral over tau is unity. Note that the last element
c     of the histogram is normalized in a different way than the
c     rest: It is simply the N divided by the nphot, i.e. the total
c     remaining photons that managed to get through to this optical
c     depth.
c     --------------------------------------------------------------
      subroutine launch_many_photons(mctype,nphot,nmaxscat,
     %                               enerlimit,g,albedo,mu0,
     %                               fracscat,fracstuck,fracthrough,
     %                               taugrid,tauhist,ntau,iseed) 
      implicit none
      integer nphot,ntau,iseed,nmaxscat,mctype
      doubleprecision g(ntau),albedo(ntau),mu0
      doubleprecision taugrid(ntau),tauhist(ntau)
      doubleprecision fracscat,mu,phi,fracstuck,fracabs,fracthrough
      doubleprecision enerlimit,nscat,nstuck,energy
c
      integer iphot,itau
      doubleprecision tauz
c
c     Check MC algorithm type
c
      if(mctype.gt.1) then
          stop 764
      endif
c
c     Check stuff
c
      if((mu0.lt.-1.d0).or.(mu0.gt.1.d0)) then
          write(*,*) 'Unphysical incident angle'
          stop 13
      endif
c
c     Reset statistics
c
      nscat=0
      nstuck=0
      do itau=1,ntau
          tauhist(itau)=0.d0
      enddo
c
c     Do statistics
c
      if(mctype.eq.0) then
c
c         Do it the most rudimentary way: literal simulation of 
c         scattering or absorption events
c
c         First check if albedo is tau-dependent. If so, then 
c         stop
c
          do itau=2,ntau
              if(abs(albedo(itau)/(albedo(itau-1)+1d-50)-1.d0).gt.
     %               1d-6) then
                  write(*,*) 'ERROR: For simple MC recipe we should'
                  write(*,*) '       have constant albedo.'
                  stop
              endif
          enddo
c
c         Else go on
c
          fracthrough=0
          do iphot=1,nphot
              mu=mu0
              phi=0.d0
              call walk_full_path_simple(nmaxscat,g,albedo(1),
     %                  mu,phi,tauz,iseed)
              if(tauz.ge.1d29) then
                  nstuck=nstuck+1
              elseif(tauz.le.taugrid(1)) then
                  nscat=nscat+1
              elseif(tauz.ge.taugrid(ntau)) then
                  fracthrough = fracthrough + 1
              else
                  call hunt(taugrid,ntau,tauz,itau)
                  tauhist(itau)=tauhist(itau)+1
              endif
          enddo
      elseif(mctype.eq.1) then
c
c         Do it a cleverder way: start with a photon package, and
c         let it dilute by absorption, while simulating the scattering
c         with discrete events. The package leaves its energy as it
c         gets absorbed, until only a certain small fraction of the
c         energy is left
c
          fracthrough=0
          do iphot=1,nphot
              mu=mu0
              phi=0.d0
              call walk_full_path_smooth(nmaxscat,enerlimit,
     %                  g,albedo,mu,phi,tauz,ntau,
     %                  taugrid,tauhist,energy,iseed)
              if(tauz.ge.1d29) then
                  nstuck=nstuck+energy
              elseif(tauz.le.taugrid(1)) then
                  nscat=nscat+energy
              elseif(tauz.ge.taugrid(ntau)) then
                  fracthrough = fracthrough + energy
              endif
          enddo
      else
          write(*,*) 'Do not have any mc method type ',mctype
          stop 13
      endif
c
c     Normalize statistics
c
      fracabs = 0
      do itau=1,ntau-1
          fracabs = fracabs + tauhist(itau)
      enddo
      tauhist(ntau)=tauhist(ntau)/nphot
      if(tauhist(ntau).gt.0.d0) then
          write(*,*) '...Waarom dit?'
          stop 13
      endif
      fracscat    = (1.d0*nscat)/(1.d0*nphot)
      fracstuck   = (1.d0*nstuck)/(1.d0*nphot)
      fracabs     = (1.d0*fracabs)/(1.d0*nphot)
      fracthrough = (1.d0*fracthrough)/(1.d0*nphot)
      if(abs(fracscat+fracabs+fracstuck+fracthrough
     %       -1.d0).gt.1.d-3) then
          write(*,*) 'Something went wrong with statistics...'
          write(*,*) fracscat,fracabs,fracstuck,fracthrough,
     %     fracscat+fracabs+fracstuck+fracthrough-1.d0
          write(*,*) 'Tip: lower the enerlimit...'
c          stop 13
      endif
c
      end



c     --------------------------------------------------------------
c                  DO ALL SCATTERINGS UNTIL ABSORPTION
c
c     This simple routine assumes a constant albedo, and lets the
c     photon enter the slab at a certain angle mu,phi, and let it
c     scatter about until it either gets absorbed or gets scattered.
c     The return values mu,phi are the output mu,phi when the photon
c     manages to escape. 
c
c     This is the original walk_full_path() routine from the old
c     mcslab.F program. Now there is a more sophisticated one called
c     walk_full_path_smooth(), which keeps track of the energy lost
c     by a package, and only stops after leaving the grid, or losing
c     more than some percentage of the energy.
c     --------------------------------------------------------------
      subroutine walk_full_path_simple(nmaxscat,g,albedo,
     %             mu,phi,tauz,iseed)
      implicit none
      doubleprecision mu,phi,albedo,g,ran2,rnr
      integer imax,i,iseed,nmaxscat
      parameter(imax=MAXNRSCAT)
      doubleprecision taupath,tauz,mu_scat,phi_scat,mu_out,phi_out
      common/path/path_tauz,path_mu,path_phi
      doubleprecision path_tauz(imax),path_mu(imax),path_phi(imax)
      common/ipath/ipath_nr
      integer ipath_nr
c
      if(mu.ge.0.d0) then
          write(*,*) 'Problem: cannot start with mu>=0'
          stop 13
      endif
c
      if(nmaxscat.gt.MAXNRSCAT) then
          write(*,*) 'Sorry, cannot do more than ',MAXNRSCAT,
     %         ' scatterings'
          stop 15
      endif
c
      tauz=0.d0
      do i=1,nmaxscat
c
c         Move photon to next scattering/absorption event
c
          taupath = -log(1.d0-ran2(iseed))
          tauz    = tauz-taupath*mu
c
c         Store data in common         
c
          path_tauz(i) = tauz
          path_mu(i)   = mu
          path_phi(i)  = phi
c
c         If escaped, then return
c
          if(tauz.lt.0.d0) then
              goto 10
          endif
c
c         Determine the nature of the event: scattering or absorption?
c
          rnr = ran2(iseed)
          if(rnr.le.albedo) then
c
c             Scattering: find another mu,phi
c
              call henyeygreenstein(g,mu_scat,phi_scat,iseed)
              call rotate(mu,phi,mu_scat,phi_scat,mu_out,phi_out)
              mu  = mu_out
              phi = phi_out
          else
c
c             Absorption: tauz gives the location of the absorption
c
              goto 10              
          endif
      enddo
      tauz=1d30
      return
 10   continue
      ipath_nr = i
c
      end




c     --------------------------------------------------------------
c                 DO ALL SCATTERINGS WHILE ABSORBING
c
c     This routine does the scatterings like walk_full_path_simple,
c     but it keeps track of the energy lost by a package, and only 
c     stops after leaving the grid, or losing more than some 
c     percentage of the energy. This way one can get rid of the 
c     very high noise at very low optical depth.
c     --------------------------------------------------------------
      subroutine walk_full_path_smooth(nmaxscat,enerlimit,g,albedo,
     %                mu,phi,tauz,ntau,taugrid,tauhist,
     %                energy,iseed)
      implicit none
      integer ntau,itauz
      doubleprecision mu,phi,albedo(ntau),g,ran2,rnr,enerlimit
      integer imax,i,iseed,nmaxscat
      parameter(imax=MAXNRSCAT)
      doubleprecision taupath,tauz,mu_scat,phi_scat,deltatauabs
      doubleprecision tauzold,mu_out,phi_out,energy,enold
      doubleprecision taugrid(ntau),tauhist(ntau)
      common/path/path_tauz,path_mu,path_phi
      doubleprecision path_tauz(imax),path_mu(imax),path_phi(imax)
      common/ipath/ipath_nr
      integer ipath_nr
c
      if(mu.ge.0.d0) then
          write(*,*) 'Problem: cannot start with mu>=0'
          stop 13
      endif
c
      if(nmaxscat.gt.MAXNRSCAT) then
          write(*,*) 'Sorry, cannot do more than ',MAXNRSCAT,
     %         ' scatterings'
          stop 15
      endif
c
c     We start at tauz=0.d0
c
c      tauz  = 1.d-50     ! Not exactly tau=0.0, for numerical reasons
      tauz  = 1.d-150     ! Not exactly tau=0.0, for numerical reasons
      itauz = 1
c
c     At tauz=0.d0 we still have the full energy in the photon packet
c
      energy = 1.d0
c
c     Now do the walk
c
      do i=1,nmaxscat
c
c         Memorize current tauz and energy
c
          tauzold = tauz
c
c         Move photon to next scattering (not absorption!) event
c
          taupath     = -log(1.d0-ran2(iseed))
          deltatauabs = abs(taupath)
ccc
ccc          taupath = -log(1.d0-ran2(iseed))
ccc          if(albedo(itauz).gt.0.d0) then
ccc              tauz = tauz-taupath*mu/albedo(itauz)
ccc          else
ccc
          if(mu.lt.0.d0) then
              tauz = 1d20
          else
              tauz = -1d20
          endif
ccc
ccc          endif
ccc
c
c         Now go from cell to cell and leave behind the absorbed 
c         energy, until you reach tauz 
c
          if(mu.ne.0.d0) then
              call walk_cells(ntau,taugrid,tauhist,tauzold,tauz,
     %              deltatauabs,mu,albedo,itauz,energy)
          else
              write(*,*) 'Extreme coincidence!!! Mu=0.d0!!!'
              enold  = energy
              energy = energy * exp(-taupath)
              tauhist(itauz) = tauhist(itauz) + abs(enold-energy)
          endif
c
c         Store data in common         
c
          path_tauz(i) = tauz
          path_mu(i)   = mu
          path_phi(i)  = phi
c
c         If escaped, then return
c
          if((tauz.le.0.d0).or.(tauz.ge.taugrid(ntau))) then
              goto 10
          endif
c
c         If virtually no energy left, then also leave
c
          if(energy.lt.enerlimit) then
              goto 10
          endif
c
c         Scattering event: find another mu,phi
c
          call henyeygreenstein(g,mu_scat,phi_scat,iseed)
          call rotate(mu,phi,mu_scat,phi_scat,mu_out,phi_out)
          mu  = mu_out
          phi = phi_out
c
      enddo
      tauz=1d30
      return
 10   continue
      ipath_nr = i
c
      end


c     --------------------------------------------------------------
c              WALK CELLS AND LEAVE THE ENERGY BEHIND
c     --------------------------------------------------------------
      subroutine walk_cells(ntau,taugrid,tauhist,tauzold,tauznew,
     %            deltatauabs,mu,albedo,itauzold,energy)
      implicit none
      integer ntau,itauzold
      doubleprecision taugrid(ntau),tauhist(ntau),tauzold,tauznew
      doubleprecision mu,energy,albedo(ntau),eps,ee
      doubleprecision encur,tauzcur,deltatauabs,taucnt,dtau
      integer itauz
c
c     Safety checks
c
      if(mu.eq.0.d0) stop 7121
      if(mu*(tauznew-tauzold).gt.0.d0) stop 7123
      if((itauzold.lt.1).or.(itauzold.ge.ntau)) then
          write(*,*) itauzold,ntau
          stop 7124
      endif
      if(taugrid(itauzold).ge.tauzold) stop 7125
      if(taugrid(itauzold+1).lt.tauzold) then
          write(*,*) itauzold+1,ntau,taugrid(itauzold+1),tauzold
          stop 7126
      endif
c
c     Memorize stuff
c
      encur   = energy
      tauzcur = tauzold
c
c     Initialize stuff
c
      taucnt  = 0.d0
c
c     Walk either upwards or downwards
c
      if(mu.gt.0.d0) then
c
c         Upwards (i.e. lower tau):
c          
          do itauz=itauzold,1,-1
c
c             Check endpoint tauznew
c
              if(tauznew.gt.taugrid(itauz)) then 
                  goto 10
              endif
c
c             Compute the absorption optical depth to the next
c             cell boundary
c
              eps     = 1.d0-albedo(itauz)
              dtau    = abs(taugrid(itauz)-tauzcur)/abs(mu)
c
c             Check endpoint scattering event
c
              if(taucnt.gt.deltatauabs-albedo(itauz)*dtau) then
c
c                 Find the new tauznew
c
                  ee      = (deltatauabs-taucnt)/(albedo(itauz)*dtau)
                  if((ee.lt.0.d0).or.(ee.gt.1.d0)) stop 9124
                  tauznew = (1.d0-ee)*tauzcur+
     %                             ee*taugrid(itauz)
                  goto 10
              endif
c
c             Compute the energy dilution within this cell
c
              energy  = energy * exp(-eps*dtau)
c
c             Drop the energy in this cell
c
              tauhist(itauz) = tauhist(itauz) + (encur-energy)
c
c             Update current values
c
              tauzcur = taugrid(itauz)
              encur   = energy
              taucnt  = taucnt + albedo(itauz)*dtau
          enddo
          itauzold = 0    
      else
c
c         Downwards (i.e. higher tau):
c
          do itauz=itauzold,ntau-1
c
c             Check endpoint tauznew
c
              if(tauznew.le.taugrid(itauz+1)) then 
                  goto 10
              endif
c
c             Compute the absorption optical depth to the next
c             cell boundary
c
              eps     = 1.d0-albedo(itauz)
              dtau    = abs(taugrid(itauz+1)-tauzcur)/abs(mu)
c
c             Check endpoint scattering event
c
              if(taucnt.gt.deltatauabs-albedo(itauz)*dtau) then
                  ee      = (deltatauabs-taucnt)/(albedo(itauz)*dtau)
                  if((ee.lt.0.d0).or.(ee.gt.1.d0)) stop 9224
                  tauznew = (1.d0-ee)*tauzcur+
     %                             ee*taugrid(itauz+1)
                  goto 10
              endif
c
c             Compute the energy dilution within this cell
c
              energy  = energy * exp(-eps*dtau)
c
c             Drop the energy in this cell
c
              tauhist(itauz) = tauhist(itauz) + (encur-energy)
c
c             Update current values
c
              tauzcur = taugrid(itauz+1)
              encur   = energy
              taucnt  = taucnt + albedo(itauz)*dtau
          enddo
          itauzold = ntau
      endif
c
c     If simply walked off the grid, then don't do anything anymore
c
      return
 10   continue
c
c     Then update this last cell with energy:
c
      eps     = 1.d0-albedo(itauz)
      dtau    = abs(tauznew-tauzcur)/abs(mu)
c
c     Compute the energy dilution within this cell
c
      energy  = energy * exp(-eps*dtau)
c
c     Drop the energy in this cell
c
      tauhist(itauz) = tauhist(itauz) + (encur-energy)
c
c     Update the itauzold, since this is what goes back to the caller
c
      itauzold = itauz
c
      return            
c
      end



c     --------------------------------------------------------------
c              HENYEY-GREENSTEIN RANDOM SCATTERING
c
c     This routine gives the random mu_scat and phi_scat according
c     to the small-angle-scattering formula of Henyey-Greenstein.
c
c                1         1 - g^2
c        P(mu) = - -------------------------
c                2 (1 + g^2 - 2 g mu )^(3/2)
c
c     We do this by introducing a variable xi
c
c                1 1 - g  /        1 + g                  \
c        xi   := - -----  | ------------------------- - 1 |
c                2   g    \ (1 + g^2 - 2 g mu )^(1/2)     /
c
c     NOTE: The current implementation has been tested in a simple
c           IDL program called hengreen.pro
c     --------------------------------------------------------------
      subroutine henyeygreenstein(g,mu,phi,iseed)
      implicit none
      doubleprecision g,mu,phi,xi,g2,pi,ran2
      parameter(pi=3.1415926535897932385d0)
      integer iseed
c
      if(g.ne.0.d0) then
          g2 = g*g
          xi = ran2(iseed)
          mu = (0.5d0/g)*(1.+g2-((1.-g2)/(1.-g+2*g*xi))**2)
      else
          mu = 2.d0*ran2(iseed)-1.d0
      endif
      phi = 2*pi*ran2(iseed)-0.5*pi
c
      end


c     --------------------------------------------------------------
c             ROTATE SCATTERING DIRECTION INTO REAL DIRECTION
c
c     This is a stupid-man's version of a 3-D rotation. Instead of
c     using standard formulae, I use matrix rotation. 
c     --------------------------------------------------------------
      subroutine rotate(mu_in,phi_in,mu_scat,phi_scat,mu_out,phi_out)
      implicit none
      doubleprecision mu_in,phi_in,mu_scat,phi_scat,mu_out,phi_out
c
      doubleprecision vecscat(3),vecdum(3),veclab(3),dum,pi
      parameter(pi=3.1415926535897932385d0)
c
c     First create the scattering direction in the local particle
c     frame
c
      dum        = sqrt(1.d0-mu_scat**2)
      vecscat(1) = mu_scat
      vecscat(2) = dum * cos(phi_scat)
      vecscat(3) = dum * sin(phi_scat)
c
c     Now rotate vector over mu (=3-axis)
c
      dum        = sqrt(1.d0-mu_in**2)
      vecdum(1)  = mu_in*vecscat(1) - dum*vecscat(2)
      vecdum(2)  = dum*vecscat(1) + mu_in*vecscat(2)
      vecdum(3)  = vecscat(3)
c
c     Now rotate vector around Z (=1) axis, so that the current
c     direction vector lies in the (1-2) plane
c
      veclab(1)  = vecdum(1)
      veclab(2)  = cos(phi_in)*vecdum(2) - sin(phi_in)*vecdum(3)
      veclab(3)  = sin(phi_in)*vecdum(2) + cos(phi_in)*vecdum(3)
c
c     Renormalize
c
      dum        = sqrt(veclab(1)**2+veclab(2)**2+veclab(3)**2)
      veclab(1)  = veclab(1)/dum
      veclab(2)  = veclab(2)/dum
      veclab(3)  = veclab(3)/dum
c
c     Now convert back to mu,phi
c
      mu_out     = veclab(1)
      if(veclab(2).ne.0.d0) then
          phi_out    = atan(veclab(3)/veclab(2))
          if(veclab(2).lt.0.d0) then
              phi_out = phi_out+pi
          endif
      else
          if(veclab(3).gt.0.d0) then 
              phi_out = pi/2.d0
          else
              phi_out = -pi/2.d0
          endif
      endif
c
      end



c     --------------------------------------------------------------
c              NUMERICAL RECIPES RANDOM NUMBER GENERATOR
c     --------------------------------------------------------------
      FUNCTION ran2(idum)
      INTEGER idum,IM1,IM2,IMM1,IA1,IA2,IQ1,IQ2,IR1,IR2,NTAB,NDIV
      doubleprecision ran2,AM,EPS,RNMX
      PARAMETER (IM1=2147483563,IM2=2147483399,AM=1./IM1,IMM1=IM1-1,
     *IA1=40014,IA2=40692,IQ1=53668,IQ2=52774,IR1=12211,IR2=3791,
     *NTAB=32,NDIV=1+IMM1/NTAB,EPS=1.2e-7,RNMX=1.-EPS)
      INTEGER idum2,j,k,iv(NTAB),iy
      SAVE iv,iy,idum2
      DATA idum2/123456789/, iv/NTAB*0/, iy/0/
      if (idum.le.0) then
        idum=max(-idum,1)
        idum2=idum
        do 11 j=NTAB+8,1,-1
          k=idum/IQ1
          idum=IA1*(idum-k*IQ1)-k*IR1
          if (idum.lt.0) idum=idum+IM1
          if (j.le.NTAB) iv(j)=idum
11      continue
        iy=iv(1)
      endif
      k=idum/IQ1
      idum=IA1*(idum-k*IQ1)-k*IR1
      if (idum.lt.0) idum=idum+IM1
      k=idum2/IQ2
      idum2=IA2*(idum2-k*IQ2)-k*IR2
      if (idum2.lt.0) idum2=idum2+IM2
      j=1+iy/NDIV
      iy=iv(j)-idum2
      iv(j)=idum
      if(iy.lt.1)iy=iy+IMM1
      ran2=min(AM*iy,RNMX)
      return
      END
C  (C) Copr. 1986-92 Numerical Recipes Software =v1.9"217..



