
c     ==============================================================
c        PROGRAM THAT PERFORMS RADIAL RADIATIVE TRANSFER IN DISKS
c              USING MONTE-CARLO AND GRAZING-ANGLE RECIPE
c                    (C) 2001  by  C.P. Dullemond
c
c     This program reads a density structure rho(r,z) of a dust
c     disk, the specifications of the central star, and the dust
c     opacities. It then performs a Monte-Carlo simulation of the
c     stellar photons as they penetrate into the disk's surface.
c     This is done at each radius using a "grazing-angle" approach.
c     No re-emission is computed, only the extinction. The dust 
c     temperature according to this primary stellar radiation field 
c     is then computed (without the contribution from the 
c     re-absorbed re-emitted rediation).
c
c     REMARK: The case of tau_rad <= 1 is not very well done: a
c             photon that reaches the equator is assumed to have
c             gone through.
c     ==============================================================

#include "main.h"

#define GASTODUST (100.d0)
#define MAXNOISE 0.3

c     --------------------------------------------------------------
c                             MAIN ROUTINE
c     --------------------------------------------------------------
      program main
      implicit none
c
      call read_all_data()
      call make_star_spectrum()
      call do_rt_starlight()
      call write_montecarlo()
      call write_temperatures()
      call write_aux_data()
c
      end




c     --------------------------------------------------------------
c                          READ ALL THE DATA
c     --------------------------------------------------------------
      subroutine read_all_data()
      implicit none
c
#include "common_grid.h"
#include "common_radialrt.h"
#include "common_dust.h"
#include "common_mc.h"
c
      integer iformat,isize,ispec,i
      integer nr,nz,ir,iz,nf,inu,nrd,nrspec,nrsize(DUST_SPECIES_MAX)
      doubleprecision a,b,c,d,e
      doubleprecision frq,dummy
      logical f_ex,multi
      logical fexf,fexo,fexs
c
c     Read Monte-Carlo parameters
c
      inquire(file='radmonte.inp',exist=f_ex)
      if(.not.f_ex) then
          write(*,*) 'ERROR: Need Monte-Carlo input file'
          write(*,*) '       called radmonte.inp.'
          stop 13
      endif
      open(unit=1,file='radmonte.inp',status='old')
      read(1,*) mc_type
      read(1,*) mc_nphot
      read(1,*) mc_nmaxscat
      read(1,*) mc_enerlimit
      read(1,*) mc_iseed
      close(1)      
c
c     Read the star information
c
      open(unit=1,file='starinfo.inp',status='old')
      read(1,*) iformat
      read(1,*) star_r
      read(1,*) star_m
      read(1,*) star_t
      close(1)
c     
c     Read the current density structure
c
      open(unit=1,file='diskstruct_0.dat',status='old')
      read(1,*) nr,nz
      nrspec=1
      nrsize(1)=1
      if(nr.lt.1) then
c
c         New format: multiple sizes/species
c
          read(1,*) nr,nz
          read(1,*) nrspec
          if(nrspec.gt.DUST_SPECIES_MAX) then
              write(*,*) 'ERROR: Too many species specified.'
              write(*,*) '       Recompile with larger max specs.'
              stop
          endif
          do ispec=1,nrspec
              read(1,*) nrsize(ispec)
              if(nrsize(ispec).gt.DUST_SIZE_MAX) then
                  write(*,*) 'ERROR: Too many sizes specified.'
                  write(*,*) '       Recompile with larger max size.'
                  stop
              endif
          enddo
          multi=.true.
      else
          nrspec=1
          nrsize(1)=1
          multi=.false.
      endif
      if(nr.gt.FRSIZE_R) then
          write(*,*) 'ERROR in radialrt:'
          write(*,*) 'R grid too large'
          stop 13
      endif
      if(nz.gt.FRSIZE_Z) then
          write(*,*) 'ERROR in radialrt:'
          write(*,*) 'Z grid too large'
          stop 13
      endif
      do ir=1,nr
          do iz=1,nz
              if(.not.multi) then
                  read(1,*) a,b,c,d,e
                  grid_r(ir) = a
                  grid_z(iz,ir) = b
                  rrt_temp(1,1,iz,ir) = d
                  rrt_rho(1,iz,ir) = c / GASTODUST
                  rrt_rhotot(iz,ir) = c / GASTODUST
                  rrt_rhogas(iz,ir) = c
                  rrt_vz(iz,ir) = e
              else
                  read(1,*) a,b,c,e
                  grid_r(ir) = a
                  grid_z(iz,ir) = b
                  rrt_rhogas(iz,ir) = c
                  rrt_vz(iz,ir) = e
                  rrt_rhotot(iz,ir) = 0.d0
                  do ispec=1,nrspec
                      read(1,*) a
                      rrt_rho(ispec,iz,ir) = a * c / GASTODUST
                      rrt_rhotot(iz,ir) = rrt_rhotot(iz,ir) +
     %                                    rrt_rho(ispec,iz,ir)
                  enddo
                  do ispec=1,nrspec
                      do isize=1,nrsize(ispec)
                          read(1,*) d
                          rrt_temp(isize,ispec,iz,ir) = d
                      enddo
                  enddo
              endif
          enddo
      enddo
      close(1)
c
c     Read the flaring index file
c
      inquire(file='flareindex.inp',exist=f_ex)
      if(.not.f_ex) then 
          inquire(file='flareangle.inp',exist=f_ex)
          if(.not.f_ex) then 
              write(*,*) 'For this Monte-Carlo version of radial rt'
              write(*,*) 'we need a file flareindex.inp'
              stop 13
          endif
          rrt_useangle=1
          open(unit=1,file='flareangle.inp',status='old')
          read(1,*) nrd
          if(nrd.ne.nr) then 
              write(*,*) 'File flareangle.inp has wrong dimensions in r'
              stop 13
          endif
          do ir=1,nr
              read(1,*) rrt_flareindex(ir)
          enddo
          close(1)
c#############################
c          write(*,*) 'DISABLED FIXED FLARING ANGLE FOR THE MOMENT...'
c          stop
          write(*,*) '!!!!! FIXED FLARING ANGLE MODE !!!!!'
c#############################
      else
          inquire(file='flareangle.inp',exist=f_ex)
          if(f_ex) then 
              write(*,*) 'Cannot use both flareangle.inp and'
              write(*,*) 'flareindex.inp. Aborting...'
              stop 13
          endif
          rrt_useangle=0
          open(unit=1,file='flareindex.inp',status='old')
          read(1,*) nrd
          if(nrd.ne.nr) then 
              write(*,*) 'File flareindex.inp has wrong dimensions in r'
              stop 13
          endif
          do ir=1,nr
              read(1,*) rrt_flareindex(ir)
          enddo
          close(1)
      endif
c
c     If present, read the starvis.inp file, which is the file containing
c     numbers between 0 and 1 saying how much of the stellar surface is
c     visible.
c
      inquire(file='starvis.inp',exist=fexf)
      if(fexf) then
          open(unit=1,file='starvis.inp',status='old')
          read(1,*) nrd
          if(nrd.ne.nr) then
              write(*,*) 'File starvis.inp has wrong NR'
              stop 13
          endif
          do ir=1,nr
              read(1,*) rrt_starvis(ir)
          enddo
          close(1)
      else
          write(*,*) '  File starvis.inp not found. Assume starvis=1.0.'
          do ir=1,nr
              rrt_starvis(ir) = 1.d0
          enddo
      endif
c
c     Read the frequency array
c
      inquire(file='kappaplanck.inp',exist=fexf)
      if(fexf) then
          write(*,*) 'PROBLEM: kappaplanck.inp is obsolete...'
          write(*,*) 'TIP    : mv kappaplanck.inp planckopac.inp'
          write(*,*) '         and add two colums with 0.d0'
          write(*,*) '         (i.e. kap_scat and g_scat).'
          stop 13
      endif
      inquire(file='planckopac.inp',exist=fexo)
      if(fexo) then
c
c         Planck Opacity file is there, so use it. If the frequency
c         dependent opacities are also there, then they are assumed
c         to be the opacities from which the planck opacities were
c         derived. The frequency-dependent opacitities can then be
c         used to make the SED.
c     
          freq_nr = 1
          call read_planckopac()
c
c         Check if also dust opacs present
c
          inquire(file='frequency.inp',exist=fexf)
          inquire(file='dustopac_1.inp',exist=fexo)
          if(fexf.and.fexo) then
              write(*,*) '   (Read planck opacity from planckopac.inp)'
          endif
      else
c
c         Find the frequency-dependent opacities, and do the radial
c         transfer with these opacitities (do it with frequency
c         dependence).
c
          inquire(file='frequency.inp',exist=fexf)
          inquire(file='dustopac.inp',exist=fexo)
          if((.not.fexf).or.(.not.fexo)) then
              write(*,*) 'Need both frequency.inp and dustopac.inp'
              stop 13
          endif
          open(unit=1,file='frequency.inp',status='old')
          read(1,*) nf
          if(nf.gt.FRSIZE_FREQ) then
              write(*,*) 'ERROR in radialrt:'
              write(*,*) 'Frequency grid too large'
              stop 13
          endif
          do inu=1,nf
              read(1,*) a
              freq_nu(inu) = a
          enddo
          freq_nr = nf
          call read_dustdata()
c
c         Check if nr of species and sizes consistent
c
          if(nrspec.ne.dust_nr_species) then
              write(*,*) 'ERROR: Nr of dust species in dustopac*.inp'
              write(*,*) '   is unequal to the number specified in '
              write(*,*) '   diskstruct_0.dat'
              stop
          endif
          do ispec=1,dust_nr_species
              if(nrsize(ispec).ne.dust_nr_size(ispec)) then
                  write(*,*) 'ERROR: Nr of dust sizes in dustopac*.inp'
                  write(*,*) '   is unequal to the number specified in '
                  write(*,*) '   diskstruct_0.dat'
                  stop
              endif
          enddo
      endif
c
c     Check
c
      if((freq_nr.eq.1).and.((dust_nr_species.gt.1).or.
     %                       (dust_nr_size(1).gt.1))) then
          write(*,*) 'ERROR: Inconsistency with sizes/species'
          stop 
      endif
c
c     Check if domain.inp exists, else make maximum domain
c
      inquire(file='domain.inp',exist=f_ex)
      if(f_ex) then
          open(unit=1,file='domain.inp',status='old')
          read(1,*) grid_irstart,grid_irend
          close(1)
          if((grid_irstart.lt.1).or.(grid_irend.gt.nr).or.
     %       (grid_irstart.gt.grid_irend)) then
              write(*,*) 'ERROR: The domain specified in domain.inp'
              write(*,*) '       is out of range.'
              stop
          endif
      else
          grid_irstart = 1
          grid_irend   = nr
      endif
c
c     Read (if required) the input spectrum
c
      if(star_t.le.0.d0) then
          inquire(file='starspectrum.inp',exist=fexs)
          if(.not.fexs) then
              write(*,*) 'ERROR: Must specify either star_t '
              write(*,*) '       or starspectrum.inp'
              stop
          endif
          write(*,*) 'Reading starspectrum.inp'
          open(unit=1,file='starspectrum.inp')
          read(1,*) i
          if(i.ne.freq_nr) then
              write(*,*) 'Sorry, stellar spectrum in starspectrum.inp'
              write(*,*) 'does not have the same number of freqs as ',
     %             'other input files.'
              stop 13
          endif
          if(i.eq.1) then
              read(1,*) frq,dummy
              star_lum(1) = 1.1965280793d38 * dummy 
          else
              do i=1,freq_nr
                  read(1,*) frq,dummy 
                  if(abs(frq-freq_nu(i))/(frq+freq_nu(i)).gt.1.d-3) then
                      write(*,*) 'PROBLEM: Frequency grid of stellar ',
     %                     'spectrum unequal to frequency.inp'
                      write(*,*) frq,freq_nu(i)
                      stop 13
                  endif
                  star_lum(i) = 1.1965280793d38 * dummy 
              enddo
          endif
          close(1)
      endif
c
      grid_nr = nr
      grid_nz = nz
c
      end




c     --------------------------------------------------------------
c                       MAKE THE STARSPECTRUM
c     --------------------------------------------------------------
      subroutine make_star_spectrum()
      implicit none
c
#     include "common_natconst.h"
#     include "common_grid.h"
#     include "common_radialrt.h"
c
      doubleprecision bplanck,lstar,error
      integer inu
c
      if(freq_nr.eq.1) then
          star_lum(1) = 4 * pi * star_r**2 * ss * star_t**4
      elseif(star_t.gt.0.d0) then
          do inu=1,freq_nr
              star_lum(inu) = 4 * pi * star_r**2 * pi *
     %             bplanck(star_t,freq_nu(inu)) 
          enddo
          lstar = 0.d0
          do inu=2,freq_nr
              lstar = lstar + 0.5 * 
     %              ( star_lum(inu) + star_lum(inu-1) ) *
     %              abs( freq_nu(inu) - freq_nu(inu-1) )
          enddo
          error = abs(lstar/
     %           (4 * pi * star_r**2 * ss * star_t**4)-1.d0)
          if(error.gt.0.05) then
              write(*,*) '****************************************'
              write(*,*) 'WARNING: Luminosity error = ',error
              write(*,*) '****************************************'
          endif
      else
          star_t = 10.
      endif
c
      end




c     --------------------------------------------------------------
c               DO SIMPLE RADIATIVE TRANSFER FOR STARLIGHT
c     --------------------------------------------------------------
      subroutine do_rt_starlight()
      implicit none
c
#     include "common_grid.h"
#     include "common_radialrt.h"
c
      integer ir
c
c     Do radial RT
c
      if(freq_nr.eq.1) then
          do ir=grid_irstart,grid_irend
              call do_rt_oneradius_meanopac(ir)
          enddo
      else
          do ir=grid_irstart,grid_irend
              write(*,*) '   ir=',ir
              call do_rt_oneradius_freqdep(ir)
          enddo
      endif
c
      end



c     --------------------------------------------------------------
c        DO THE MONTE CARLO FOR ONE RADIUS (MEAN OPACITY VARIANT)
c     --------------------------------------------------------------
      subroutine do_rt_oneradius_meanopac(ir)
      implicit none
c
      integer ir
c
#include "common_grid.h"
#include "common_radialrt.h"
#include "common_mc.h"
c
      doubleprecision kappa,kappa_abs,kappa_scat,mu0,pi
      doubleprecision t0,temp,told,qstarfpi,fluxirrad,dsigma
      doubleprecision find_planckopac,tauz(FRSIZE_Z),hs,dhs
      doubleprecision flareangle,thetau
      doubleprecision albedo(FRSIZE_Z),g(FRSIZE_Z)
      parameter(pi=3.1415926535897932385d0)
      integer iz,ii,m
c
c     Internal consistency check
c
      if(freq_nr.gt.1) stop 7311
c
c     Find the kappa and the albedo at stellar wavelengths
c
      call find_planckall(star_t,kappa_abs,kappa_scat,g)
      kappa  = kappa_abs + kappa_scat
      do iz=1,grid_nz
          albedo(iz) = kappa_scat / kappa
          g(iz)      = 0.d0
      enddo
c
c     Find the flaring angle
c
      if(rrt_useangle.ne.0) then
c
c         For testing purposes we might want to use a prescribed
c         grazing angle
c
c#############################
C          write(*,*) 'DISABLED FIXED FLARING ANGLE FOR THE MOMENT...'
C          stop
          write(*,*) '!!!!! FIXED FLARING ANGLE MODE !!!!!'
          stop
c#############################
          flareangle = rrt_flareindex(ir)    ! Index is here used as angle
c
      elseif(rrt_flareindex(ir).lt.-1.d1) then
c
c         Only allowed at the inner rim
c
          if(ir.ne.1) stop 8231
c
c         If the flaring index is negative (take it -1d10), then the 
c         disk here is assumed to be facing the star at an angle of 90
c         degrees. This is meant for the inclusion of the inner rim in
c         the model.
c
          flareangle = 1.5707963268d0
c
c         For the SED computation (see sed.F) it is important to 
c         compute the height of the inner rim. An estimate can be
c         given by assuming light to enter the top of the atmosphere
c         with a flaring angle of 1/8 (see Dullemond, Dominik &
c         Natta 2001).
c
c         First make a tau grid for the incident angle of 1/8:
c
          tauz(grid_nz) = 1.d-33
          do iz=grid_nz-1,2,-1
              tauz(iz) = tauz(iz+1) + 8.d0 * 0.5 * kappa *
     %              ( rrt_rho(1,iz,ir) + rrt_rho(1,iz+1,ir) ) *
     %              ( grid_z(iz+1,ir) - grid_z(iz,ir) ) 
          enddo
          tauz(1) = 1d33
          if(tauz(2).lt.1.d0) then 
              write(*,*) 'SORRY: Could not find the tau=1 surface'
              write(*,*) '       at radius ir=',ir
              write(*,*) '       iz=',iz
              stop 13
          endif
c
c         Then find the tau=1 surface, i.e. the Hs
c     
          m=4
          call hunt(tauz,grid_nz,1.d0,iz)
          if((iz.eq.0).or.(iz.eq.grid_nz)) then
              write(*,*) 'SORRY: Could not find the tau=1 surface'
              write(*,*) '       at radius ir=',ir
              write(*,*) '       iz=',iz
              do iz=1,grid_nz
                  write(*,*) iz,grid_z(iz,ir),tauz(iz)
              enddo
              stop 13
          endif
          iz = min(max(iz-(m-1)/2,1),grid_nz+1-m)
          thetau = 1.d0
          call polint(tauz(iz),grid_z(iz,ir),m,thetau,hs,dhs)
c
c         Store Hs and flareangle in global arrays 
c
          rrt_flang(ir) = flareangle
          rrt_hs(ir)    = hs
c
      else
c
c         But normally we _compute_ the flaring angle instead of 
c         assuming something beforehand. This is done here
c
c         First make the grazing tau grid
c
          tauz(grid_nz) = 1.d-33
          do iz=grid_nz-1,2,-1
              tauz(iz) = tauz(iz+1) + 0.5 * kappa *
     %              ( rrt_rho(1,iz,ir) + rrt_rho(1,iz+1,ir) ) *
     %              ( grid_z(iz+1,ir) - grid_z(iz,ir) ) / 
     %              ( rrt_flareindex(ir) * 
     %                grid_z(iz,ir) / grid_r(ir) )
          enddo
          tauz(1) = 1d33
          if(tauz(2).lt.1.d0) then 
              write(*,*) 'SORRY: Could not find the tau=1 surface'
              write(*,*) '       at radius ir=',ir
              write(*,*) '       iz=',iz
              stop 13
          endif
c
c         Find the tau=1 surface, i.e. the Hs
c     
          m=4
          call hunt(tauz,grid_nz,1.d0,iz)
          if((iz.eq.0).or.(iz.eq.grid_nz)) then
              write(*,*) 'SORRY: Could not find the tau=1 surface'
              write(*,*) '       at radius ir=',ir
              write(*,*) '       iz=',iz
              do iz=1,grid_nz
                  write(*,*) iz,grid_z(iz,ir),tauz(iz)
              enddo
              stop 13
          endif
          iz = min(max(iz-(m-1)/2,1),grid_nz+1-m)
          thetau = 1.d0
          call polint(tauz(iz),grid_z(iz,ir),m,thetau,hs,dhs)
c
c         From this H_s compute the flaring angle from the flaring index
c
          flareangle = 0.4d0 * star_r / grid_r(ir) + 
     %                    rrt_flareindex(ir) * hs / grid_r(ir)
c
c         Store Hs and flareangle in global arrays (for debugging)
c
          rrt_flang(ir) = flareangle
          rrt_hs(ir)    = hs
c
      endif
c
c     Debugging message
c
      write(*,*) 'Flaring angle=',flareangle
c
c     Make make the vertical tau-grid
c
      mc_taugrid(1,ir) = 0.d0
      do iz=grid_nz-1,1,-1
          mc_taugrid(grid_nz-iz+1,ir) = mc_taugrid(grid_nz-iz,ir) +
     %        0.5 * ( rrt_rho(1,iz+1,ir) + rrt_rho(1,iz,ir) ) *
     %        kappa * ( grid_z(iz+1,ir) - grid_z(iz,ir) )
      enddo
c
c     Clear the histogram
c     
      do iz=1,grid_nz 
          mc_tauhist(iz,ir) = 0.d0
      enddo
c
c     Now call the Monte-Carlo solver
c
      mu0 = -sin(flareangle)
      call launch_many_photons(mc_type,mc_nphot,mc_nmaxscat,
     %                         mc_enerlimit,g,albedo,mu0,
     %                         mc_fracscat(ir),mc_fracstuck(ir),
     %                         mc_fracthrough(ir),
     %                         mc_taugrid(1,ir),mc_tauhist(1,ir),
     %                         grid_nz,mc_iseed) 
c
c     Check if the fraction that got through or stuck is not too high
c
      if(mc_fracthrough(ir).gt.0.05) then
          write(*,*) 'Problem: more than 5% of the photons'
          write(*,*) '   got through the disk...' 
          stop 13
      endif
      if(mc_fracstuck(ir).gt.0.05) then
          write(*,*) 'Problem: more than 5% of the photons'
          write(*,*) '   got stuck without being absorbed...' 
          stop 13
      endif
c
c     Check the noise
c
      if(mc_type.eq.0) then
          call check_fluctuations(grid_nz,mc_nphot,
     %        mc_taugrid(1,ir),mc_tauhist(1,ir))
      endif
c
c     Compute the flux that the star puts into the disk
c
      fluxirrad = flareangle * rrt_starvis(ir) * star_lum(1) / 
     %            ( 4 * pi * grid_r(ir)**2 )
c
c     The histogram determines how this flux is vertically 
c     distributed into the disk
c
      do iz=grid_nz-1,1,-1
c
c         A slice of disk matter  [g/cm^2]
c
          dsigma   = 0.5 * ( rrt_rho(1,iz,ir) + rrt_rho(1,iz+1,ir) )
     %               * ( grid_z(iz+1,ir) - grid_z(iz,ir) )
c
c         The heating per gram matter div by 4 pi  [erg/s/g/(4*pi)]
c
          qstarfpi = fluxirrad * mc_tauhist(grid_nz-iz,ir) 
     %               / ( 4 * pi * mc_nphot * dsigma )
c
c         The zeroth order dust temperature (grey)
c
c          ss
c          -- kappa_abs T^4 = qstarfpi
c          pi
c
          t0   = (qstarfpi*5.540d4/kappa_abs)**0.25d0
          temp = t0
c
c         Iterate to find the correct temperature
c
          do ii=1,60
              told  = temp
              temp  = t0*(kappa_abs/find_planckopac(temp))**0.25
              if(abs(temp-told)/(temp+told+1d-4).lt.TEMPACCUR) goto 20
          enddo
          write(*,*) 'ERROR: Temperature not converged'
          write(*,*) '   temp,told,ts0,kapstar,find_planckopac=',
     %         ir,iz,
     %         temp,told,qstarfpi,kappa,find_planckopac(temp)
          stop 9323
 20       continue
          rrt_temp(1,1,iz,ir) = temp
      enddo
c
c     Copy last temperature
c
      rrt_temp(1,1,grid_nz,ir) = rrt_temp(1,1,grid_nz-1,ir) 
c
      end



c     --------------------------------------------------------------
c       DO THE MONTE CARLO FOR ONE RADIUS (FREQ-DEPENDENT VARIANT)
c     --------------------------------------------------------------
      subroutine do_rt_oneradius_freqdep(ir)
      implicit none
c
      integer ir
c
#include "common_grid.h"
#include "common_radialrt.h"
#include "common_dust.h"
#include "common_mc.h"
c
      doubleprecision kappa,mu0,pi
      doubleprecision albedo(FRSIZE_Z),g(FRSIZE_Z)
      doubleprecision t0,temp,told,fluxirrad,dsigma,solve_dust_temp
      doubleprecision find_dust_kappa,find_dust_g,hs,dhs,fdst,dummy
      doubleprecision fluxdist(DUST_SIZE_MAX,DUST_SPECIES_MAX,
     %                         FRSIZE_FREQ,FRSIZE_Z)
      doubleprecision flareangle,tauz(FRSIZE_Z),thetau,fluxsc,totmc
      doubleprecision qstarfpicell(DUST_SIZE_MAX,DUST_SPECIES_MAX,
     %                             FRSIZE_Z)
      doubleprecision qstarfpi(DUST_SIZE_MAX,DUST_SPECIES_MAX,
     %                         FRSIZE_Z)
      doubleprecision fracscat(FRSIZE_FREQ),fracstuck(FRSIZE_FREQ)
      doubleprecision fracthrough(FRSIZE_FREQ),wgt,totlum,frac
      doubleprecision alpha(FRSIZE_Z,FRSIZE_FREQ)
      doubleprecision alphaabs(FRSIZE_Z,FRSIZE_FREQ)
      doubleprecision flang_freq(FRSIZE_FREQ),height_freq(FRSIZE_FREQ)
      parameter(pi=3.1415926535897932385d0)
      integer iz,ii,inu,m,isize,ispec
c
c     Internal consistency check
c
      if(freq_nr.eq.1) stop 7312
c
c     Loop over frequency
c
      do inu=1,freq_nr
c
c         Message
c
c          write(*,*) 'Frequency bin ',inu
c
c         Find the alpha at all locations
c
          do iz=1,grid_nz
              alpha(iz,inu) = 0.d0
              alphaabs(iz,inu) = 0.d0
              do ispec=1,dust_nr_species
                  do isize=1,dust_nr_size(ispec)
                      kappa         = find_dust_kappa(inu,
     %                                isize,ispec,star_t,1,1)
                      alpha(iz,inu) = alpha(iz,inu) + 
     %                                kappa * rrt_rho(ispec,iz,ir)
                      kappa         = find_dust_kappa(inu,
     %                                isize,ispec,star_t,1,0)
                      alphaabs(iz,inu) = alphaabs(iz,inu) + 
     %                                kappa * rrt_rho(ispec,iz,ir) 
                  enddo
              enddo
              albedo(iz) = ( alpha(iz,inu) - alphaabs(iz,inu) ) /
     %                     ( alpha(iz,inu) + 1.d-99 )
              if((dust_nr_species.eq.1).and.(dust_nr_size(1).eq.1)) then
                  g(iz) = find_dust_g(inu,1,1,star_t)
              else
                  do ispec=1,dust_nr_species
                      do isize=1,dust_nr_size(ispec)
                          if(dust_gscat(inu,1,isize,ispec).ne.0.d0) then
                              write(*,*) 'PROBLEM: For multiple dust ',
     %                'species and/or sizes we have '
                              write(*,*) '         not yet installed ',
     %                'an average Henyey g-factor...'
                              stop 
                          endif
                      enddo
                  enddo
                  g(iz)      = 0.d0       ! For the moment do not
              endif
          enddo
          do iz=1,grid_nz-1
              alpha(iz,inu) = 0.d0
              do ispec=1,dust_nr_species
                  do isize=1,dust_nr_size(ispec)
                      kappa         = find_dust_kappa(inu,
     %                                isize,ispec,star_t,1,1)
                      alpha(iz,inu) = alpha(iz,inu) + 
     %                                kappa * 0.5d0 * 
     %                                ( rrt_rho(ispec,iz,ir) + 
     %                                  rrt_rho(ispec,iz+1,ir) )
                  enddo
              enddo
          enddo
c
c         Find the flaring angle
c
          if(rrt_useangle.ne.0) then
c
c             For testing purposes we might want to use a prescribed
c             grazing angle
c     
c#############################
C              write(*,*) 'DISABLED FIXED FLARING ANGLE...'
C              stop
          write(*,*) '!!!!! FIXED FLARING ANGLE MODE !!!!!'
c#############################
              flareangle = rrt_flareindex(ir) ! Index is here used as angle
              hs = 0.d0
c
          elseif(rrt_flareindex(ir).lt.-1.d1) then
c
c             Only allowed at the inner rim
c
              if(ir.ne.1) stop 8231
c
c             If the flaring index is negative (take it -1d10), then the 
c             disk here is assumed to be facing the star at an angle of 90
c             degrees. This is meant for the inclusion of the inner rim in
c             the model.
c
              flareangle = 1.5707963268d0
c
c             For the SED computation (see sed.F) it is important to 
c             compute the height of the inner rim. An estimate can be
c             given by assuming light to enter the top of the atmosphere
c             with a flaring angle of 1/8 (see Dullemond, Dominik &
c             Natta 2001).
c
c             First make a tau grid for the incident angle of 1/8:
c
              tauz(grid_nz) = 1.d-33
              do iz=grid_nz-1,2,-1
                  tauz(iz) = tauz(iz+1) + 8.d0 * alpha(iz,inu) *
     %                 ( grid_z(iz+1,ir) - grid_z(iz,ir) ) 
              enddo
              tauz(1) = 1d33
              if(tauz(2).lt.1.d0) then 
                  do iz=1,grid_nz
                      do ispec=1,dust_nr_species
                          do isize=1,dust_nr_size(ispec)
                              fluxdist(isize,ispec,inu,iz) = 0.d0
                          enddo
                      enddo
                  enddo
                  flang_freq(inu)  = 1.5707963268d0
                  height_freq(inu) = 0.d0
                  goto 102
              endif
c
c             Then find the tau=1 surface, i.e. the Hs
c     
              m=4
              call hunt(tauz,grid_nz,1.d0,iz)
              if((iz.eq.0).or.(iz.eq.grid_nz)) then
                  write(*,*) 'SORRY: Could not find the tau=1 surface'
                  write(*,*) '       at radius ir=',ir
                  write(*,*) '       iz=',iz
                  do iz=1,grid_nz
                      write(*,*) iz,grid_z(iz,ir),tauz(iz)
                  enddo
                  stop 13
              endif
              iz = min(max(iz-(m-1)/2,1),grid_nz+1-m)
              thetau = 1.d0
              call polint(tauz(iz),grid_z(iz,ir),m,thetau,hs,dhs)
c
          else
c
c             But normally we _compute_ the flaring angle instead of 
c             assuming something beforehand. This is done here
c
c             First make the grazing tau grid
c 
              tauz(grid_nz) = 1.d-33
              do iz=grid_nz-1,2,-1
                  tauz(iz) = tauz(iz+1) + alpha(iz,inu) *
     %                 ( grid_z(iz+1,ir) - grid_z(iz,ir) ) / 
     %                 ( rrt_flareindex(ir) * 
     %                 grid_z(iz,ir) / grid_r(ir) )
              enddo
              tauz(1) = 1d33
              if(tauz(2).lt.1.d0) then 
                  do iz=1,grid_nz
                      do ispec=1,dust_nr_species
                          do isize=1,dust_nr_size(ispec)
                              fluxdist(isize,ispec,inu,iz) = 0.d0
                          enddo
                      enddo
                  enddo
c   (BUGFIX 25-06-02: Need to reset flang_freq(inu) and height_freq(inu))
                  flang_freq(inu)  = 0.d0
                  height_freq(inu) = 0.d0
                  goto 102
              endif
c
c             Find the tau=1 surface, i.e. the Hs
c     
              m=4
              call hunt(tauz,grid_nz,1.d0,iz)
              if((iz.eq.0).or.(iz.eq.grid_nz)) then
                  write(*,*) 'SORRY: Could not find the tau=1 surface'
                  write(*,*) '       at radius ir=',ir
                  write(*,*) '       iz=',iz
                  stop 13
              endif
              iz = min(max(iz-(m-1)/2,1),grid_nz+1-m)
              thetau = 1.d0
              call polint(tauz(iz),grid_z(iz,ir),m,thetau,hs,dhs)
c
c             From this H_s compute the flaring angle from the flaring index
c
              flareangle = 0.4d0 * star_r / grid_r(ir) + 
     %                     rrt_flareindex(ir) * hs / grid_r(ir)
c
          endif
c
c         Store the flaring angle and surface height into an array
c     
          flang_freq(inu)  = flareangle
          height_freq(inu) = hs
c
c         Make make the vertical tau-grid
c
          mc_taugrid(1,ir) = 0.d0
          do iz=grid_nz-1,1,-1
              mc_taugrid(grid_nz-iz+1,ir) = mc_taugrid(grid_nz-iz,ir) +
     %             alpha(iz,inu) * ( grid_z(iz+1,ir) - grid_z(iz,ir) )
          enddo
c
c         Clear the histogram
c     
          do iz=1,grid_nz 
              mc_tauhist(iz,ir) = 0.d0
          enddo
c
c         Now call the Monte-Carlo solver
c
          mu0 = -sin(flareangle)
          call launch_many_photons(mc_type,mc_nphot,mc_nmaxscat,
     %                         mc_enerlimit,g,albedo,mu0,
     %                         fracscat(inu),fracstuck(inu),
     %                         fracthrough(inu),
     %                         mc_taugrid(1,ir),mc_tauhist(1,ir),
     %                         grid_nz,mc_iseed) 
c
c         Check the noise
c
          if(mc_type.eq.0) then
              call check_fluctuations(grid_nz,mc_nphot,
     %           mc_taugrid(1,ir),mc_tauhist(1,ir))
          endif
c
c         Compute the flux that the star puts into the disk
c
          fluxirrad = flareangle * rrt_starvis(ir) * star_lum(inu) / 
     %            ( 4 * pi * grid_r(ir)**2 )
c
c         The histogram determines how this flux is vertically 
c         distributed into the disk
c     
          do iz=grid_nz-1,1,-1
              dummy    = 0.d0
              do ispec=1,dust_nr_species
                  dsigma   = 0.5 * ( rrt_rho(ispec,iz,ir) + 
     %                               rrt_rho(ispec,iz+1,ir) )
     %                           * ( grid_z(iz+1,ir) - grid_z(iz,ir) )
                  fdst     = fluxirrad * mc_tauhist(grid_nz-iz,ir) /
     %                           ( 4 * pi * mc_nphot * dsigma )
                  do isize=1,dust_nr_size(ispec)
                      kappa   = find_dust_kappa(inu,
     %                                isize,ispec,star_t,1,0)
                      frac    = rrt_rho(ispec,iz,ir) * kappa / 
     %                                alphaabs(iz,inu)
                      dummy   = dummy + frac
                      if(frac.gt.1.d0) then
                          write(*,*) frac
                          stop 1298
                      endif
                      fluxdist(isize,ispec,inu,iz) = frac * fdst
                  enddo
              enddo
              if(abs(dummy-1.d0).gt.1d-5) stop 8021
          enddo 
 102      continue
      enddo
c
c     Compute the fraction scat, stuck and through, averaged over spectrum
c
      totlum             = 0.d0
      mc_fracscat(ir)    = 0.d0
      mc_fracstuck(ir)   = 0.d0
      mc_fracthrough(ir) = 0.d0
      do inu=2,freq_nr
          wgt     = 0.5 * ( star_lum(inu) + star_lum(inu-1) ) *
     %                      abs( freq_nu(inu) - freq_nu(inu-1) )
          totlum             = totlum + wgt
          mc_fracscat(ir)    = mc_fracscat(ir) + wgt * 
     %              0.5 * ( fracscat(inu) + fracscat(inu-1) )
          mc_fracstuck(ir)   = mc_fracstuck(ir) + wgt * 
     %              0.5 * ( fracstuck(inu) + fracstuck(inu-1) )
          mc_fracthrough(ir) = mc_fracthrough(ir) + wgt * 
     %              0.5 * ( fracthrough(inu) + fracthrough(inu-1) )
      enddo
      mc_fracscat(ir)    = mc_fracscat(ir) / totlum
      mc_fracstuck(ir)   = mc_fracstuck(ir) / totlum
      mc_fracthrough(ir) = mc_fracthrough(ir) / totlum
c
c     Compute the q_*/4pi everywhere, cell-centered!
c
      do iz=1,grid_nz-1
          do ispec=1,dust_nr_species
              do isize=1,dust_nr_size(ispec)
                  qstarfpicell(isize,ispec,iz) = 0.d0
                  do inu=2,freq_nr
                      qstarfpicell(isize,ispec,iz) = 
     %                   qstarfpicell(isize,ispec,iz) + 0.5 * 
     %                     ( fluxdist(isize,ispec,inu,iz) + 
     %                       fluxdist(isize,ispec,inu-1,iz) ) *
     %                     abs( freq_nu(inu) - freq_nu(inu-1) )
                  enddo
              enddo
          enddo
      enddo
      do ispec=1,dust_nr_species
          do isize=1,dust_nr_size(ispec)
              qstarfpicell(isize,ispec,grid_nz) = 
     %             qstarfpicell(isize,ispec,grid_nz-1)
          enddo
      enddo
c
c     Now find the grid-point centered qstarfpi. The idea behind this
c     is that we want the integral over rho*q to be exactly equal to
c     the input flux. With rho being a VERY steep function, this could
c     be jeopardized by sloppy averaging. We choose qstarfpi(iz) here
c     such that the sum  0.5*(rho(iz)*qfpi(iz)+rho(iz-1)*qfpi(iz-1))*
c     (z(iz)-z(iz-1)) is exactly correct.
c     
      do iz=2,grid_nz-1
          do ispec=1,dust_nr_species
              do isize=1,dust_nr_size(ispec)
                  qstarfpi(isize,ispec,iz) = 
     %                 (qstarfpicell(isize,ispec,iz-1)*
     %                  (grid_z(iz,ir)-grid_z(iz-1,ir))+
     %                  qstarfpicell(isize,ispec,iz)*
     %                  (grid_z(iz+1,ir)-grid_z(iz,ir))) /
     %                 ( grid_z(iz+1,ir)-grid_z(iz-1,ir) )
              enddo
          enddo
      enddo
      do ispec=1,dust_nr_species
          do isize=1,dust_nr_size(ispec)
              qstarfpi(isize,ispec,1) = qstarfpicell(isize,ispec,1) 
              qstarfpi(isize,ispec,grid_nz) = 
     %             qstarfpicell(isize,ispec,grid_nz-1)
          enddo
      enddo
c
c     Now for secondary purposes, do some computations:
c
c     Find the average flaring angle and the average surface height
c
      flareangle = 0.d0
      hs         = 0.d0
      do inu=2,freq_nr
          wgt = 0.5 * ( star_lum(inu) + star_lum(inu-1) ) *
     %           abs( freq_nu(inu) - freq_nu(inu-1) )
          flareangle = flareangle + 0.5 * wgt *
     %           ( flang_freq(inu) + flang_freq(inu-1) )
          hs         = hs         + 0.5 * wgt *
     %           ( height_freq(inu) + height_freq(inu-1) )
      enddo
      flareangle = flareangle / totlum
c
c     Store stuff in the auxiliary data array (for debugging)
c
      rrt_flang(ir) = flareangle
      rrt_hs(ir)    = hs / totlum
c
c     Self-consistency checks
c
c      write(*,*) 'Checking local flux conservation...'
c
      if(mc_nphot.eq.1) then
c
c         Compute the input flux
c
          fluxirrad = 0.d0
          do inu=2,freq_nr
              fluxirrad = fluxirrad + 0.5 * 
     %             ( star_lum(inu) + star_lum(inu-1) ) *
     %             abs( freq_nu(inu) - freq_nu(inu-1) )
          enddo
          fluxirrad = fluxirrad * rrt_starvis(ir) * flareangle / 
     %         ( 4 * pi * grid_r(ir)**2 ) 
c
c         Then compute the outgoing flux
c
          fluxsc = 0.d0
cccc >>> BUG FIXED 15-11-02 <<<         do iz=1,grid_nz
          do iz=2,grid_nz
              do ispec=1,dust_nr_species
                  do isize=1,dust_nr_size(ispec)
                      fluxsc = fluxsc + 0.5 * 
     %               ( qstarfpi(isize,ispec,iz)*
     %                 rrt_rho(ispec,iz,ir) +
     %                 qstarfpi(isize,ispec,iz-1)*
     %                 rrt_rho(ispec,iz-1,ir) ) *
     %               ( grid_z(iz,ir) - grid_z(iz-1,ir) )
                  enddo
              enddo
          enddo
          fluxsc = fluxsc * 4 * pi
          if(abs(fluxsc/fluxirrad-1.d0).gt.1.d-2) then
              write(*,*) 'RADMONTE: FLUX PROBLEM FOUND!!!!!'
              write(*,*) 'Ratio of the final and should-be flux = ',
     %             fluxsc/fluxirrad
c              stop 13
          endif
      endif
c
c     Now solve the temperature 
c
      do iz=1,grid_nz-1
          do ispec=1,dust_nr_species
              do isize=1,dust_nr_size(ispec)
                  rrt_temp(isize,ispec,iz,ir) = 
     %                solve_dust_temp(isize,ispec,
     %                      qstarfpi(isize,ispec,iz))
              enddo
          enddo
      enddo
c
c     Copy last temperature
c
      do ispec=1,dust_nr_species
          do isize=1,dust_nr_size(ispec)
              rrt_temp(isize,ispec,grid_nz,ir) = 
     %             rrt_temp(isize,ispec,grid_nz-1,ir) 
          enddo
      enddo
c
      end



c     --------------------------------------------------------------
c                 CHECK FLUCTUATIONS OF MONTE-CARLO CODE
c     --------------------------------------------------------------
      subroutine check_fluctuations(ntau,nphot,taugrid,histogram)
      implicit none
      integer ntau,nphot
      doubleprecision taugrid(ntau),histogram(ntau)
      integer itau,itm,i
      doubleprecision maxhist
c
c     First find the maximum of histogram*tau
c
      maxhist=0.d0
      itm=0
      do itau=1,ntau
          if(histogram(itau)*taugrid(itau).gt.maxhist) then
              maxhist = histogram(itau)*taugrid(itau)
              itm     = itau
          endif
      enddo
c
c     If at tau<tau_max (i.e. it.lt.itm) the noise in the histogram
c     becomes more than a certain percentage (see configure.h), then
c     stop.
c
      do itau=3,itm-1 
          if(abs((histogram(itau)*taugrid(itau))/
     %          (histogram(itau-1)*taugrid(itau-1))-1.d0)
     %             .gt.MAXNOISE) then
              write(*,*) 'Noise of MC at low optical depth'
              write(*,*) '   exceeds maximum tolerable:'
              write(*,*) '   at itau=',itau,' noise = ',
     %                 abs((histogram(itau)*taugrid(itau))/
     %          (histogram(itau-1)*taugrid(itau-1))-1.d0)
              write(*,*) '   Tip: increase either tau_min '
              write(*,*) '   or nphot.'
cc############################
c              do i=1,itm
c                  write(*,*) histogram(i)
c              enddo
cc############################
c              stop 13
          endif
      enddo
c
      end



c     --------------------------------------------------------------
c          SOLVE TEMPERATURE OF DUST GRAIN FOR GIVEN Q_*/4PI
c     --------------------------------------------------------------
      function solve_dust_temp(isize,ispec,qstarfpi)
      implicit none
      doubleprecision solve_dust_temp,qstarfpi
      doubleprecision dust_emisabs,zbrent
      integer isize,ispec
      external dust_emisabs
c
#include "common_grid.h"
#include "common_dust.h"
c
      common/dustintegral/di_qstarfpi
      doubleprecision di_qstarfpi
      common/idustintegral/di_isize,di_ispec
      integer di_isize,di_ispec
c
      doubleprecision dum1,dum2
      integer inu
c
c     Put the data into the common
c
      di_qstarfpi = qstarfpi
      di_isize    = isize
      di_ispec    = ispec
c
c     Now check if there is indeed sufficient radiation for a
c     solution.
c
      dum1 = dust_emisabs(0.02d0)
      dum2 = dust_emisabs(10000.d0)
      if(dum1.gt.0.d0) then
          solve_dust_temp=0.d0
          return
      endif
      if(dum2.lt.0.d0) then
          write(*,*) 'Error: dust temperature > 10000 K'
          write(*,*) dum2,qstarfpi
          stop 13
      endif
c
c     Now solve for the dust temperature
c
      solve_dust_temp = zbrent(dust_emisabs,2.d-2,1.d4,1.d-1) 
c
      return
      end



c     --------------------------------------------------------------
c                    INTEGRAL OF EMISSION-ABSORPTION
c     --------------------------------------------------------------
      function dust_emisabs(temp)
      implicit none
      doubleprecision dust_emisabs,temp
c
      common/dustintegral/di_qstarfpi
      doubleprecision di_qstarfpi
      common/idustintegral/di_isize,di_ispec
      integer di_isize,di_ispec
c
#include "common_grid.h"
#include "common_dust.h"
c     
      integer inu
      doubleprecision dummy,kappa,intgrand(FRSIZE_FREQ)
      doubleprecision bplanck,find_dust_kappa
c
      do inu=1,freq_nr
          kappa         = find_dust_kappa(inu,di_isize,di_ispec,
     %                                    temp,-1,0)
          intgrand(inu) = kappa * bplanck(temp,freq_nu(inu))
      enddo
      dummy = 0.d0
      do inu=2,freq_nr
          dummy = dummy + 0.5 * 
     %         ( intgrand(inu) + intgrand(inu-1) ) *
     %         abs( freq_nu(inu) - freq_nu(inu-1) )
      enddo
c
      dust_emisabs = dummy - di_qstarfpi
      return
      end



c     --------------------------------------------------------------
c                THE BLACKBODY PLANCK FUNCTION B_nu(T)
c
c     This function computes the Blackbody function 
c
c                    2 h nu^3 / c^2
c        B_nu(T)  = -------------------    [ erg / cm^2 s ster Hz ]
c                   exp(-h nu / kT) - 1
c
c     ARGUMENTS:
c        nu    [Hz]            = Frequency
c        temp  [K]             = Electron temperature
c     --------------------------------------------------------------
      function bplanck(temp,nu)
      implicit none
      doubleprecision temp
      doubleprecision nu
      doubleprecision bplanck
c
      if(temp.eq.0.d0) then 
          bplanck = 0.d0
          return
      endif
c
      bplanck = 1.47455d-47 * nu * nu * nu / 
     %      (exp(4.7989d-11 * nu / temp)-1.d0) + 1.d-290
c
      return
      end


c     --------------------------------------------------------------
c                       WRITE THE TEMPERATURES
c     --------------------------------------------------------------
      subroutine write_temperatures()
      implicit none
c
#include "common_grid.h"
#include "common_dust.h"
#include "common_radialrt.h"
c
      integer ir,iz,ispec,isize
      logical multi
c
      multi=.false.
      if((dust_nr_species.gt.1).or.(dust_nr_size(1).gt.1)) then
          multi=.true.
      endif
      open(unit=1,file='diskstruct_1.dat',status='unknown')
      if(multi) then
          write(1,*) -1,0
          write(1,*) grid_nr,grid_nz
          write(1,*) dust_nr_species
          do ispec=1,dust_nr_species
              write(1,*) dust_nr_size(ispec)
          enddo
      else
          write(1,*) grid_nr,grid_nz
      endif
      write(1,*)
      do ir=1,grid_nr
          do iz=1,grid_nz
              if(.not.multi) then
                  write(1,10) grid_r(ir),grid_z(iz,ir),
     %                        rrt_rho(1,iz,ir)*GASTODUST+1d-98,
     %                        rrt_temp(1,1,iz,ir)+1d-98,rrt_vz(iz,ir)
 10               format(5(E13.6,1X))
              else
                  write(1,20) grid_r(ir),grid_z(iz,ir),
     %                        rrt_rhogas(iz,ir)+1d-98,rrt_vz(iz,ir)
 20               format(3(E13.6,1X))
                  do ispec=1,dust_nr_species
                      write(1,30) rrt_rho(ispec,iz,ir)*GASTODUST/
     %                            rrt_rhogas(iz,ir)+1d-98
                  enddo
                  do ispec=1,dust_nr_species
                      do isize=1,dust_nr_size(ispec)
                          write(1,30) rrt_temp(isize,ispec,iz,ir)+1d-98
                      enddo
                  enddo
 30               format(1(E13.6,1X))
              endif
          enddo
      enddo
      close(1)
c
      end



c     --------------------------------------------------------------
c                WRITE THE MONTE CARLO RESULTS TO FILE
c     --------------------------------------------------------------
      subroutine write_montecarlo
      implicit none
c
#include "common_grid.h"
#include "common_mc.h"
#include "common_radialrt.h"
c
      integer ir,itau
c
      open(unit=1,file='mcslab.dat')
      write(1,*) grid_nr,grid_nz
      write(1,*) 
      do ir=1,grid_nr
          write(1,*) mc_fracscat(ir)
          write(1,*) mc_fracstuck(ir)
          write(1,*) mc_fracthrough(ir)
          write(1,*) mc_nphot
          write(1,*) grid_nz
          do itau=1,grid_nz
              write(1,20) mc_taugrid(itau,ir),mc_tauhist(itau,ir)
 20           format(2(E13.6,1X))
          enddo
      enddo
      close(1)
c
      end






c     --------------------------------------------------------------
c                    READ THE PLANCK MEAN OPACITY
c     --------------------------------------------------------------
      subroutine read_planckopac
      implicit none
c
#     include "common_grid.h"
#     include "common_radialrt.h"
c
      doubleprecision dum1,dum2,dum3,dum4
      integer it,nt
      logical fex
c
      inquire(file='planckopac.inp',exist=fex)
      if(.not.fex) then
          inquire(file='kappaplanck.inp',exist=fex)
          if(fex) then
              write(*,*) 'The file kappaplanck.inp is obsolete.'
              write(*,*) '  Need file planckopac.inp.'
          else
              write(*,*) 'Need file planckopac.inp or a full '
              write(*,*) '  freq-dependent opacity table.'
          endif
          stop 13
      endif
      open(unit=1,file='planckopac.inp')
      read(1,*) dopac_nt
      if(dopac_nt.gt.NR_TEMPERATURE) then
          write(*,*) 'Too many temperatures in planckopac.inp'
          write(*,*) ' Increase NR_TEMPERATURE and recompile'
          stop 13
      endif
      do it=1,dopac_nt
          read(1,*) dum1,dum2,dum3,dum4
          dopac_temp(it)=dum1
          dopac_kappa(it)=dum2
          dopac_kscat(it)=dum3
          dopac_gscat(it)=dum4
      enddo
      close(1)
c
      end



c     --------------------------------------------------------------
c                  INTERPOLATE THE PLANCK MEAN OPACITY
c     --------------------------------------------------------------
      function find_planckopac(temp)
      implicit none
      doubleprecision find_planckopac,temp
c
#     include "common_grid.h"
#     include "common_radialrt.h"
c
      doubleprecision dum1,dum2,dt
      integer it
c
      if(dopac_nt.le.1) then 
          find_planckopac = dopac_kappa(1)
          return
      endif
c
      if(temp.lt.dopac_temp(1)) then 
          find_planckopac = dopac_kappa(1)
      elseif(temp.gt.dopac_temp(dopac_nt)) then
          find_planckopac = dopac_kappa(dopac_nt)
      else
          call hunt(dopac_temp,dopac_nt,temp,it)
          dt = (temp-dopac_temp(it))/
     %            (dopac_temp(it+1)-dopac_temp(it))
          if((dt.lt.0.d0).or.(dt.gt.1.d0)) stop 552
          find_planckopac = (1.d0-dt)*dopac_kappa(it)+
     %       dt*dopac_kappa(it+1)
      endif
c
      return
      end


c     --------------------------------------------------------------
c                INTERPOLATE THE PLANCK MEAN SCAT OPAC
c     --------------------------------------------------------------
      subroutine find_planckall(temp,cabs,cscat,g)
      implicit none
      doubleprecision temp,cabs,cscat,g
c
#     include "common_grid.h"
#     include "common_radialrt.h"
c
      doubleprecision dum1,dum2,dt
      integer it
c
      if(dopac_nt.le.1) then 
          cabs  = dopac_kappa(1)
          cscat = dopac_kscat(1)
          g     = dopac_gscat(1)
          return
      endif
c
      if(temp.lt.dopac_temp(1)) then 
          cabs  = dopac_kappa(1)
          cscat = dopac_kscat(1)
          g     = dopac_gscat(1)
      elseif(temp.gt.dopac_temp(dopac_nt)) then
          cabs  = dopac_kappa(dopac_nt)
          cscat = dopac_kscat(dopac_nt)
          g     = dopac_gscat(dopac_nt)
      else
          call hunt(dopac_temp,dopac_nt,temp,it)
          dt = (temp-dopac_temp(it))/
     %            (dopac_temp(it+1)-dopac_temp(it))
          if((dt.lt.0.d0).or.(dt.gt.1.d0)) stop 552
          cabs  = (1.d0-dt)*dopac_kappa(it)+
     %       dt*dopac_kappa(it+1)
          cscat = (1.d0-dt)*dopac_kscat(it)+
     %       dt*dopac_kscat(it+1)
          g     = (1.d0-dt)*dopac_gscat(it)+
     %       dt*dopac_gscat(it+1)
      endif
c
      return
      end



c     --------------------------------------------------------------
c                  WRITE SOME OF THE AUXILIARY DATA
c     --------------------------------------------------------------
      subroutine write_aux_data()
      implicit none
c
#include "common_grid.h"
#include "common_radialrt.h"
c
      integer ir,iz
c
      open(unit=1,file='radmonte.dat',status='unknown')
      write(1,*) grid_nr
      write(1,*)
      do ir=1,grid_nr
          write(1,10) grid_r(ir),rrt_hs(ir),rrt_flang(ir)
 10       format(3(E13.6,1X))
      enddo
      close(1)
c
      end
