;========================================================================
;
;                      >>> WELCOME TO CGPLUS <<<
; 
;  A modeling program for dusty circumstellar disks, based on the models 
;  of Chiang & Goldreich (1997) and Dullemond, Dominik & Natta (2001).
; 
;                    Written by C.P. Dullemond
;                      in collaboration with 
;                     C. Dominik and A. Natta
;                            (C) 2002
; 
;    ----->>>> NOTE: This software is _NOT_ public domain. <<<<-----
;    ----->>>>  Any use of this software requires written  <<<<-----
;    ----->>>>         permission from the authors         <<<<-----
; 
;  This version of the code is released for use within the SIRTF Legacy
;  Programme of N. Evans. Within this Programme, usage is permitted
;  under the agreements made per email between the C.P. Dullemond and 
;  N. Evans, E. van Dishoeck and G. Blake. Please inform C.P. Dullemond
;  (dullemon@mpa-garching.mpg.de) of any project being done using this
;  software. Do not distribute to others outside the SIRTF Legacy 
;  Programme.
;
;                    Last update: 14 Oct 2002
;
;------------------------------------------------------------------------
;
; This program solves the structure of a protostellar/protoplanetary
; dusty disk surrounding a T Tauri or Herbig Ae/Be star. The disk is
; assumed to be passive and having a flaring geometry. The structure
; of the disk is purely determined by the irradiation of it's surface
; by the central star. The SED of such a disk can then be determined
; and fitted to observations. 
;
; This program comes with a user-interface written by S. Walch and
; C.P. Dullemond (FITCGPLUS). So the easiest way to use this code is
; to start FITCGPLUS in the following way: start IDL, type .r fitcgplus
; and type fitcgplus. Read the header of fitcgplus.pro for more details.
;
; The equations used in this program are described in:
;
;    Dullemond, Dominik & Natta (2001) ApJ 560, 957
;
; The model is an improvement of the model proposed by:
;
;    Chiang & Goldreich (1997) ApJ 490, 368
;
;------------------------------------------------------------------------
;
; Effects included:
;  - Self-consistent determination of chi=Hs/Hp from opacity table
;  - Self-consistent treatment of vertically optically thin disk regions
;  - Correct determination of dust temperatures with efficiency factors
;  - Inner wall at self-consistent height
;  - Shadowing of CG97 disk by its own puffed-up inner rim
;  - Include 1/(1-Omega) in the formula for Tin (self-irradiation)
;  - Include irradiation of flaring disk by the rim 
;  - Self-consistent determination of flaring index (see Chiang et al 2001). 
;  - Self-consistent determination of the f_r (starvisibility) factor
;  - Shadowed diffusive region is now included properly
;  - Multiple species + evaporation (but in thermal contact, i.e. one temp)
;  - Dust scattering effects on the structure and in the SED
;    >>>> Dust scattering Has been switched off again <<<<
;
; Options included (for switching off certain effects):
;  - Switch for fixed epsilon
;  - Switch for fixed chi
;  - Switch for fixed flaring index
;  - Switch for no near field term in irradiation flux (the 0.4*Rstar/R)
;  - Switch for no psi corrections
;  - Switch for f_r=0.5 or 1.0 (i.e. constant). 
;
; Agenda:
;  - Include the effect of H_e < H_i (i.e. an optically thin emitting ring)
;  - Check out what the optically thin temperature for dust grains in 
;    front of the inner wall is: if higher than 1500, then push out the
;    inner radius even more
;  - Add the scattering effects in the self-irradiation
;
; Notes:
;  - The correct estimate of mugas is important, since it influences 
;    the Hp(R) pretty much (24-01-01)
;  - BUGFIX 12-02-02: Before, the starvis fraction was not included in
;    the computation of Ts. Now this is fixed. But let's see if the
;    code remains numerically stable.
;  - BUGFIX 23-10-02: Introduced the command keyword_really_set(), and
;    used it for the keyword_set of incl and plsigma. This way, also 
;    setting it to 0 still means setting it (keyword_set would return
;    false).
;  - MODIFICATION 15-11-02: I now determine the Ts with the flux 
;    extincted to exp(-1). That works a lot better. You can switch
;    this off by the switch: /tsthin.
;
;========================================================================
;
;------------------------------------------------------------------------
;              FUNCTION: READ THE DUST OPACITY TABLES
;
; This is a function that reads the files "dustopac_1.inp" etc and
; the corresponding "frequency.inp". 
;------------------------------------------------------------------------
function readgrainopac,nrvec,evaptemp=evaptemp,abun=abun,$
                       gastodust=gastodust
close,1
;
; Read the frequency table
;
file='frequency.inp'
dd=findfile(file,count=count)
if(count le 0) then begin
   print,"Could not find frequency.inp. Taking frequency.dat"
   file='frequency.dat'
   dd=findfile(file,count=count)
   if(count le 0) then begin
      print,"Could not find frequency.dat either"
      stop
   endif
endif
openr,1,file
nf=0
readf,1,nf
freq = fltarr(nf)
wave = fltarr(nf)
for k=1,nf do begin
   readf,1,dum
   freq(k-1) = dum
   wave(k-1) = 2.9979d14 / dum
endfor
close,1
;
ng=n_elements(nrvec)
;
; Then read the opacities
;
if n_elements(evaptemp) eq 0 then evaptemp=dblarr(ng)+100000.
if n_elements(evaptemp) ne n_elements(nrvec) then begin
   print,"Nr of dust species not equal to nr of evap temperatures"
   stop 
endif
if n_elements(abun) eq 0 then begin
   if ng eq 1 then abun=[1.0] else begin
      print,"Need to specify abundances of the dust species!!"
      stop 
   endelse
endif
if n_elements(abun) ne n_elements(nrvec) then begin
   print,"Nr of dust species not equal to nr of abundance specifications"
   stop 
endif
abun=abun/total(abun)
cabs = fltarr(nf,ng)
csca = fltarr(nf,ng)
for ig=0,ng-1 do begin
   filename = 'dustopac_'+strcompress(string(nrvec[ig]),/remove_all)+'.inp'
   print,"Reading ",filename
   openr,1,filename
   nnf=0
   nns=0
   readf,1,nnf,nns
   if nns ne 1 then begin
      print,"ERROR: Cannot use multiple sizes yet"
      stop
   endif
   if nnf ne nf then begin
      print,"ERROR: frequency file has different nr of points as dustopac file"
      stop
   endif
   dum  = 0.e0
   for k=1,nf do begin
      readf,1,dum
      cabs(k-1,ig) = dum
   endfor
;  <<added for scattering>>
   for k=1,nf do begin
      readf,1,dum
      csca(k-1,ig) = dum
   endfor
   nrtrange=1
   trange=[0.d0,0.d0]
   close,1
endfor
;
; If gas-to-dust is specified, then divide the dust opacities
; by it
;
if n_elements(gastodust) gt 0 then abun=abun/gastodust
;
;  <<modified for scattering>>
return,{nf:nf,ns:1,ng:ng,freq:freq,wave:wave,cabs:cabs,csca:csca,$
        nrt:nrtrange,trange:trange,evaptemp:evaptemp,abun:abun}
end


;------------------------------------------------------------------------
;                   FUNCTION: SIMPLE INTEGRATOR
;------------------------------------------------------------------------
function integrate,x,f
n=n_elements(x)
if(n_elements(f) ne n) then begin
   print,"Error in function integrate: x and f not equally long"
   stop 
endif
sign=2.d0*(x[n-1] gt x[0])-1.d0
int=0.d0
for i=1,n-1 do begin
   int=int+0.5d0*(f[i]+f[i-1])*(x[i]-x[i-1])
endfor
return,int*sign
end

;------------------------------------------------------------------------
;             FUNCTION: A BETTER KEYWORD_SET COMMAND
;------------------------------------------------------------------------
function keyword_really_set,x
if n_elements(x) eq 0 then begin
   return,(1 eq 0)
endif else begin
   return,(1 eq 1)
endelse
end


;------------------------------------------------------------------------
;         FUNCTION: FIND THE DUST KAPPA TABLE AT GIVEN TEMPERATURE
;
; Sicne some dust species can evaporate, we must find the opacity
; table temperature-dependently. Here's the function for doing so.
;------------------------------------------------------------------------
function findkappa,opac,T,abs=abs,scat=scat
;  <<modified for scattering>>
if n_elements(abs) eq 0 and n_elements(scat) eq 0 then begin
   print,"findkappa: now must specify /abs, /scat or both"
   stop
endif
if n_elements(abs) ne 0 then iabs=1.0 else iabs=0.0
if n_elements(scat) ne 0 then isca=1.0 else isca=0.0
ng=opac.ng
kappa = 0.d0*opac.freq
for i=0,ng-1 do begin
   if T le opac.evaptemp[i] then begin
      kappa = kappa + opac.abun[i]*(iabs*opac.cabs[*,i]+isca*opac.csca[*,i])
   endif
endfor
return,kappa
end


;------------------------------------------------------------------------
;              F_R FUNCTION: HOW MUCH OF STAR IS VISIBLE?
;------------------------------------------------------------------------
function visibfrac,rstar,hi,ri,h,r,intersect=intersect
if h eq 0.d0 then return,0.5d0
eta = ri*(h-hi)/(r-ri)-hi
if keyword_set(intersect) then return,eta
if eta ge rstar then return,1.d0
if eta le -rstar then return,0.d0
y=eta/rstar
return,(1.d0/!dpi)*(y*sqrt(1.d0-y^2)+asin(y))+0.5d0
end
;
;------------------------------------------------------------------------
;            FUNCTION: FIND THE PSI_S/PSI_I VALUE
;
; The psi=psi_s/psi_i value is the dimensionless constant in front
; of the Ti determination. It represents the effects of finite optical
; depth for the case of a frequency-dependent opacity.
;------------------------------------------------------------------------
function findpsi,Sigma,Ti,Ts,opac=opac,psis=psis,psii=psii
;
; Constants
;
pi    = !dpi
ss    = 5.6703d-5
AU    = 1.496d13
GG    = 6.672d-8
Msun  = 1.99d33
Lsun  = 3.86d33
mugas = 2.3d0
mp    = 1.6726d-24
kk    = 1.3807d-16
;
; A factor sqrt(3) in the opacity because we are doing mu-averaged 
; stuff (fluxes)
;
sqthree = sqrt(3.d0)
;
; Now do the stuff
;
if keyword_set(opac) then begin
   ;;
   ;; The freqs
   ;;
   nu=opac.freq
   nf=n_elements(nu)
   ;;
   ;; Find psis and psii
   ;;
   psis  = 0.d0
   psii  = 0.d0
   btots = 0.d0
   btoti = 0.d0
   kappai = sqthree*findkappa(opac,Ti,/abs)
   kappas = sqthree*findkappa(opac,Ts,/abs)
   for inu=1,nf-2 do begin
      tau    = Sigma*kappai[inu]
      xpp    = 1.d0-exp(-tau)
      dnu    = 0.5d0*nu[inu]*(alog(nu[inu+1])-alog(nu[inu-1]))
      bpls   = bplanck(nu[inu],Ts)
      bpli   = bplanck(nu[inu],Ti)
      psis   = psis + bpls[0]*kappas[inu]*xpp*dnu
      psii   = psii + bpli[0]*xpp*dnu
      btots  = btots + bpls[0]*kappas[inu]*dnu
      btoti  = btoti + bpli[0]*dnu
   endfor
   psis = psis / btots
   psii = psii / btoti
   psi  = psis / psii
endif else begin
   print,'Error: no OPAC in findpsi'
   stop
endelse
return,psi[0]
end


;------------------------------------------------------------------------
;       FUNCTION: SOLVE OPTICALLY THIN DUST TEMPERATURE
;
; Solve the equation:
;
;   1   /                       /
;  ---- | F_nu kappa_nu dnu  =  | B_nu(Ts) kappa_nu dnu
;  4 pi /                       /
;
; where F_nu is the sum of the stellar flux and the flux from 
; the rim. 
;
; NOTE: BUGFIX (12.02.02). The real occultation of the central
;       star is more complicated than the simple thing used before.
;       So now you must specify the starvis fraction yourself.
;
;------------------------------------------------------------------------
function solvetthin,opac,tstar,rstar,trim,rrim,hrim,r,z,$
                    starvis=starvis,fredux=fredux
;
; Constants
;
pi    = !dpi
ss    = 5.6703d-5
AU    = 1.496d13
GG    = 6.672d-8
Msun  = 1.99d33
Lsun  = 3.86d33
mugas = 2.3d0
mp    = 1.6726d-24
kk    = 1.3807d-16
;
; If not specified starvis, then cause trouble
;
if n_elements(starvis) eq 0 then begin
   print,'solvetthin must have starvis fraction specified.'
   stop
endif
;
; If not specified fredux, then take it 1
;
if n_elements(fredux) eq 0 then fredux=1.d0
;
; The freqs
;
nu=opac.freq
nf=n_elements(nu)
;
; Compute the delta and the projected surface of the inner rim
;
if trim ne 0.d0 and rrim ne 0.d0 and hrim ne 0.d0 and z ne 0.d0 then begin
   delta  = (r/rrim-1.d0)/(z/hrim-1.d0)
   if delta gt 0.d0 then begin
      cosi = cos(atan((r-rrim)/(z-hrim)))
      if delta lt 1.d0 then begin
         S = 2.d0 * rrim^2 * cosi * ( delta * sqrt(1.d0-delta^2) $ 
                                      + asin(delta) )
      endif else begin
         S = pi  * rrim^2 * cosi
      endelse
   endif else begin
      S = 0.d0
   endelse
endif else begin
   S = 0.d0
endelse
;
; Check
;
if S eq 0.d0 and starvis eq 0.d0 then return,0.d0
;
; The frequency-dependent flux
;
flux = starvis * pi * bplanck(nu,tstar) * rstar^2 / r^2 + $
       S * bplanck(nu,trim) / r^2
;
; Extinct by exp(-1), or not, dependent on the value of fredux
;
flux = flux * fredux
;
; Check
;
if max(flux) eq 0.d0 then begin
   print,"PROBLEM in solvetthin: zero flux"
   stop
endif
;
; Now the total flux, and the initial guess for the temperature
;
kappa=findkappa(opac,0.d0,/abs)
ftot=0.d0
for inu=1,nf-2 do begin
   ftot = ftot + 0.5*flux(inu)*nu(inu)*(alog(nu(inu+1))-alog(nu(inu-1)))
endfor
;
; If flux eq 0 then return 0
;
if ftot eq 0.d0 then return,0.d0
;
; Else find the initial guess for Td
;
Td  = ( ftot / ( 4.0 * ss ) )^0.25
;
; Now compute the flux * kappa integral 
;
fkap=0.d0
for inu=1,nf-2 do begin
   fkap = fkap + 0.5*flux(inu)*kappa(inu)*nu(inu)*(alog(nu(inu+1))-alog(nu(inu-1)))
endfor
;
; Now do the iteration to get the real temperature
;
iloop = 1
repeat begin
   Tdold = Td
   Td  = ( fkap / ( 4.0 * ss * kappaplanck(opac,Td) ) )^0.25
   iloop = iloop+1      
   if iloop gt 100 then begin
      print,'No convergence in solvethin'
      stop
   endif
endrep until abs(Tdold-Td) lt 1.d-4*abs(Tdold+Td)
return,Td
end



;------------------------------------------------------------------------
;               SOLVE THE TI USING THE PSI FUNCTION
;
; This is a way to include the effects of psi into the deterination of
; the interior temperature Ti.
;------------------------------------------------------------------------
function solvetiwithpsi,Tiorig,Ts,sigma,opac,psitot=psitot,psis=psis,psii=psii
psi=1.0
for i=0,30 do begin
   oldpsi = psi
   Ti     = Tiorig * psi^0.25
   psi    = findpsi(Sigma,Ti,Ts,opac=opac,psis=psis,psii=psii)
   if abs(psi-oldpsi) lt 1.d-3*(psi+oldpsi) then begin
      psitot = psi
      return,Ti
   endif
endfor
print,"MAJOR PROBLEM: solve Ti with Psi did not converge!"
stop
end



;------------------------------------------------------------------------
;               SOLVE TIN INCLUDING THE SELF-IRRADIATION
;
; This is a way to include the self-irradiation of the inner wall into
; the determination of the temperature of the inner wall, if the Rin
; is given (use solverinselfirr() instead if you want to give Tin
; instead of Rin).
;------------------------------------------------------------------------
function solvehpinselfirr,rin,chi,Hpin0
Hpin=Hpin0
for i=0,10 do begin
   ohp=Hpin
   Hpin=Hpin0*((1.d0+chi*Hpin/rin)^(0.125))
   if abs(Hpin-ohp) lt 1.d-3*(Hpin+ohp) then begin
      return,Hpin
   endif
endfor
print,"MAJOR PROBLEM: solve Tin with self-irradiation!"
stop
end



;------------------------------------------------------------------------
;               SOLVE RIN INCLUDING THE SELF-IRRADIATION
;
; This is a way to include the self-irradiation of the inner wall into
; the determination of the radius of the inner wall for a given wall 
; temperature (use solvehpinselfirr() instead if you want to give Rin
; instead of Tin). So we give the Hp0 which is calculated without these 
; self irradiation corrections, and then make the corrections. This
; involves an iteration, because as Rin moves outwards, the Sigma(Rin)
; changes (simply because Sigma(R)=Sigma0*(R/Rsig)^plsig, where Rsig
; is NOT the Rin, but a fixed radius). NOTE: The argument 'kap'
; contains already the factor of 8 due to the structure of the inner
; rim (see paper). 
;------------------------------------------------------------------------
function solverinselfirr,rin0,Sigma0,rsig,plsigma,kap,Hp0,$
        fixhs=fixhs,chi=chi,hsback=hsback,chback=chback
rin = rin0
for i=0,10 do begin
   if not keyword_set(chi) then begin  
       ;; 
       ;; Modify Sigma, according to Sigma(R) powerlaw
       ;;
       Sig = Sigma0*(rin/rsig)^plsigma
       chi = findchi(kap,1.d0,Sig)
   endif
   if not keyword_set(fixhs) then begin
       ;;
       ;; Since Tin is fixed, we can use the equation of vertical
       ;; pressure balance to compute Hpin.
       ;;
       Hpin = Hp0 * (rin/rin0)^(1.5)
       Hsin = chi * Hpin
   endif else begin
       Hsin = fixhs
       Hpin = Hsin / chi
   endelse
   ;;
   ;; Now modify Rin by the effect of self-irradiation
   ;;
   rinold = rin
   rin = rin0 * sqrt(1.d0+Hsin/Rin)
   if abs(rin-rinold) lt 1.d-3*(rin+rinold) then begin
      hsback=hsin
      chback=chi
      return,rin
   endif   
endfor
print,"MAJOR PROBLEM: solve Rin with self-irradiation!"
stop
end



;------------------------------------------------------------------------
;          COMPUTE THE dlg(H/R)/dlgR FROM GIVEN STRUCTURE
;
; This function is important when one wishes to iterate the flaring disk
; structure until energy conservation is guaranteed.
;
;------------------------------------------------------------------------
function find_dlgh,R,H
nr=n_elements(R)
r2=R(1:nr-1)
r1=R(0:nr-2)
hr2=H(1:nr-1)/r2+1d-5
hr1=H(0:nr-2)/r1+1d-5

dum=[0.d0,(r2+r1)*(hr2-hr1)/((r2-r1)*(hr2+hr1))]
dum[0]=dum[1]

return,dum
end



;------------------------------------------------------------------------
;             COMPUTE THE ERROR IN ENERGY CONSERVATION 
;
; Find the difference between the real flaring index and the computed one,
; and integrate this over d(H/R), since this is the measure of the covering
; fraction, and hence of the luminosity of the disk. The function returns
; the computed disk luminosity divided by the luminosity what it should have
; been. If this number is 1.d0, then there is no error.
;
;------------------------------------------------------------------------
function find_encons_error,R,H,flareindex=flareindex
dlgh=find_dlgh(R,H)
nr=n_elements(R)
if not keyword_set(flareindex) then flareindex=2.d0/7.d0
flareindex=dblarr(nr)+flareindex
dum1=0.d0
dum2=0.d0
for ir=1,nr-2 do begin
   dum1=dum1+(flareindex[ir]/dlgh[ir])*0.5d0*((H[ir+1]/R[ir+1])-(H[ir-1]/R[ir-1]))
   dum2=dum2+0.5d0*((H[ir+1]/R[ir+1])-(H[ir-1]/R[ir-1]))
endfor
return,dum1/dum2
end



;------------------------------------------------------------------------
;           FUNCTION: MAKE A CHIANG & GOLDREICH DISK
;
; This function makes a flared disk according to the formulae of CG97,
; with some small modifications. In particular, we self-consistently
; compute certain dimensionless tuning parameters such as the value of
; chi=Hs/Hp and the psi-values that modify the equation for Ti in case
; of non-thick vertical optical depths. All these self-consistent
; determinations are done with the full opacity table.
; If diffedge is given, this function is automatically going to add
; the emission from the inner rim to the irradiative flux.
;
; ARGUMENTS:
;  mstarmsun           Stellar mass in units of Msun
;  tstar               Stellar temperature in Kelvin
;  lstarlsun           Stellar luminosity in units Lsun
;  rsigmaau            Radius where Sigma(R)==Sigma0
;  sigma0              See above
;  plsigma             Sigma(R) = Sigma0 * (R/Rsigma)^plsigma  
;                       (usually plsigma.lt.0.d0)
;  rinau               Inner radius in AU
;  routau              Outer radius in AU
;
; KEYWORDS:
;  opac                Opacity table (see readdust.pro for format), 
;                      such that tau_vert(R,nu) = Sigma(R) * opac.cabs(inu)
;  diffedge            Structure from diffedge() subroutine. If given
;                      here, then it will include the effects of self-
;                      shadowing and self-irradiation of the flaring disk
;                      by the inner rim.
;  /noselfirr          If diffedge is included but this is set, then do
;                      not include self-irradiation (but DO include 
;                      self-shadowing). Note that if self-irradiation
;                      is included, then it is included both in the
;                      CG97 part and in the half-shadow. In particular
;                      this means that we DO calculate the half-shadow
;                      solution! But in the SED this is negligible.
;  starvis             notset: Compute f_r self-consistently 
;                      =1    : Star is half visible (as in CG97 Eq.(1))
;                      =2    : Star is fully visible from disk surface
;  flareindex          The flaring index that was 2/7 in CG97. You can give
;                      it as a scalar (to get CG97: use 2./7.), or as an
;                      array. 
;  initflare           The initial guess for the flaring index (at the inner
;                      edge) if the self-consistent procedure is used. If 
;                      not set, then this value is standard 1/7. Note: it
;                      is better to set this value too low than too high.
;  chifix              If set, then use always Hs/Hp=chifix [for testing]
;  epsfix              If set, then use always eps=epsfix   [for testing]
;  /nonear             If set, skip the 0.4*Rstar/R term [for testing]
;  /nopsi              If set, do not include psi-corrections [for testing]
;  /noscat             If set, then do not include the scattering
;  albedo              If set, then use 'albedo' as albedo, else use the
;                      albedo computed from the opacity tables.
;
; NOTE:
;  To get the original CG97 model from this function, call it as:
;    cgdiskfunc(ms,ts,ls,rsig,sig0,pls,rin,rout,nr,$
;               flareindex=(2.d0/7.d0),chifix=4.d0)
;  but note that the original CG97 model is in fact wrong in the sense 
;  that it does not conserve energy. Chiang et al 2001 fixed this (we use
;  their method to fix it here too), but they didn't mention the serious-
;  ness of the problem of their original model. 
;------------------------------------------------------------------------
function cgdiskfunc,mstarmsun,tstar,lstarlsun,rsigmaau,sigma0,plsigma,$
                rinau,routau,nr,opac=opac,chifix=chifix,epsfix=epsfix,$
                diffedge=diffedge,noselfirr=noselfirr,$
                starvis=starvis,flareindex=flidx,nonear=nonear,$
                nopsi=nopsi,initflare=initflare,tsthin=tsthin,$
                noscat=noscat,albedo=albedo,grazang=grazang,$
                scatmethod=scatmethod
;
; Constants
;
pi    = !dpi
ss    = 5.6703d-5
AU    = 1.496d13
GG    = 6.672d-8
Msun  = 1.99d33
Lsun  = 3.86d33
mugas = 2.3d0
mp    = 1.6726d-24
kk    = 1.3807d-16
;
; Check
;
if n_elements(scatmethod) eq 0 then scatmethod=1
;
; Some other stuff
;
if (n_elements(noscat) eq 0) and (n_elements(albedo) ne 0) then begin
   if albedo eq 0.d0 then noscat=1
endif
;
; If /tsthin, then use the original C&G method of determining the
; surface temperature: optically thin approach. Otherwise (if tsthin
; not set) then include the effect of extinction: exp(-1). 
;
if not keyword_set(tsthin) then begin
   fredux=0.5d0
endif else begin
   if tsthin eq 2 then begin
      fredux=0.367879d0 
   endif else begin
      fredux=1.0000d0
      print,'WARNING: Computing Ts in the old way: without exp(-1) correction'
   endelse
endelse
;
; Use psi?
;
if not keyword_set(nopsi) then begin
   usepsi = 1
endif else begin
   usepsi = 0
endelse
;
; Flare index array
;
if not keyword_set(initflare) then initflare=1.d0/7.d0
if not keyword_set(flidx) then flareindex=initflare else flareindex=flidx
if n_elements(flareindex) eq 1 then flareindex=dblarr(nr)+flareindex
if n_elements(flareindex) ne nr then stop
if keyword_set(flidx) then globflare=1 else globflare=0
;
; Warnings & errors
;
if(usepsi ne 1) then begin
   print,"============================"
   print,"Warning: Psi is deactivated!"
   print,"============================"
endif
if not keyword_set(opac) then begin
   print,"===================================="
   print,"Error: MUST specify opac            "
   print,"===================================="
   stop
endif
if keyword_set(nonear) then begin
   near=0.d0
endif else begin
   near=1.d0
endelse
;
; Convert to CGS
;
rin   = rinau*AU
rout  = routau*AU
rsigma= rsigmaau*AU
Lstar = lstarlsun * Lsun
Mstar = mstarmsun * Msun
rstar = sqrt(lstar/(4*pi*ss*tstar^4))
;
; Stuff...
;
if rin lt rstar then rin=rstar
Tc    = GG * Mstar * mugas * mp / ( kk * Rstar )
;
; Now make the radial grid
;
r     = rin*((rout/rin)^(dindgen(nr)/(nr-1.d0)))
; 
; More stuff...
;
irfullshadow = -1
irhalfshadow = -1
;
; Make the arrays
;
Ti      = dblarr(nr)
Ts      = dblarr(nr)
alpha   = dblarr(nr)
fr      = dblarr(nr)
Hs      = dblarr(nr)
Hp      = dblarr(nr)
tausurf = dblarr(nr)
sigsurf = dblarr(nr)
chi     = dblarr(nr)
psi     = dblarr(nr)
psis    = dblarr(nr)
psii    = dblarr(nr)
fscat   = dblarr(nr)
alb     = dblarr(nr)
;
; First make the Sigma array
;
Sigma   = Sigma0*(r/rsigma)^plsigma
;
; Now get some data about the inner rim
;
if keyword_set(diffedge) then begin
   trim     = diffedge.tin
   hrim     = diffedge.Hsin
   rrim     = r[0]
   incledge = 1
endif else begin
   incledge = 0
endelse
;
; Loop from inside outwards
;
flcnt=0
for ir=0,nr-1 do begin
   ;;
   ;; Find the flareindex, but update only every two steps (see appendix
   ;; of paper by Chiang et al. ApJ 547:1077-1089 (2001).
   ;;
   if globflare eq 0 then begin
      if flcnt ge 2 then begin
         hr2 = Hs[ir-2]/r[ir-2]
         hr1 = Hs[ir-1]/r[ir-1]
         if hr1 gt 0.d0 and hr2 gt 0.d0 and irhalfshadow lt ir-2 then begin
            flareindex[ir]=(r[ir-2]+r[ir-1])*(hr2-hr1)/$
                           ((r[ir-2]-r[ir-1])*(hr2+hr1))
         endif else begin
            if ir gt 0 then flareindex[ir]=flareindex[ir-1]
         endelse
         if flareindex[ir] lt 0.05 then begin
            print,'WARNING: Flaring index < 0.05 at ir=',ir
            flareindex[ir]=0.05
         endif
         flcnt=0
      endif else begin
         if ir gt 0 then flareindex[ir]=flareindex[ir-1]
      endelse
   endif
   ;;
   ;; Start with guess of alpha and Hs
   ;;
   if keyword_set(grazang) then alpha[ir]=grazang[ir] else alpha[ir] = 0.2
   if ir eq 0 then Hs[ir] = 0.0 else Hs[ir]=Hs[ir-1]*r[ir]/r[ir-1]
   ;;
   ;; Now work on the CG97 disk strucuture, where we first do things
   ;; without considering the shadowing effect by the inner rim, but
   ;; we do add the emission from the inner rim. 
   ;;
   ;; So iterate on the CG97 equations (incl inner rim emission) 
   ;; until convergence.
   ;;
   iloop = 1
   repeat begin
      ;;
      ;; The fraction of the starlight visible (if R_in=R_*, this is 0.5;
      ;; if R_in>>R_* this is 1.0).
      ;;
      if keyword_set(starvis) then begin
         if starvis eq 1 then begin
            fr[ir]    = 0.5d0
         endif else begin
            fr[ir]    = 1.0d0
         endelse
      endif else begin
         fr[ir]    = visibfrac(rstar,0.d0,r[0],Hs[ir],r[ir])
      endelse
      ;;
      ;; Then determine the temperature of the surface layer, which is not
      ;; dependent on the disk structure. Compute the dust temperature
      ;; either according to a given epsilon efficiency factor (kappa_vis
      ;; / kappa_ir) or self-consistently.
      ;;
      if keyword_set(epsfix) then begin
         Ts[ir]      = sqrt(0.5*Rstar/r[ir]) * Tstar / epsfix^0.25
      endif else begin
         if not keyword_set(noselfirr) and keyword_set(diffedge) then begin
            ;; BUGFIX 12-02-02
            ;; MODIFICATION 15-11-02: Ts at tau=1
            Ts[ir]      = solvetthin(opac,Tstar,Rstar,trim,rrim,$
                           hrim,r[ir],Hs[ir],starvis=fr[ir],fredux=fredux)
            if(ts[ir] eq 0.d0) then stop
         endif else begin
            ;; BUGFIX 12-02-02
            if fr[ir] gt 0.d0 then begin
               ;; MODIFICATION 15-11-02: Ts at tau=1
               Ts[ir]      = solvetthin(opac,Tstar,Rstar,0.d0,0.d0,$
                                        0.d0,r[ir],Hs[ir],$
                                        starvis=fr[ir],fredux=fredux)
               if(ts[ir] eq 0.d0) then stop
            endif else begin
               Ts[ir] = 1.0   ;; Just for numerical stability
            endelse
         endelse
      endelse
      ;;
      ;; The disk interior temperature according to Eq.(12a) of CG97. Note
      ;; that their factor of 1/4 comes from two factors of 1/2. One comes
      ;; from the fact that they assume the disk to go all the way to the 
      ;; star (i.e. fr=0.5). We use the fr (above) for that. The other 
      ;; comes from the fact that half of the emission of the surface layer
      ;; is radiated away and only half reaches the interior. Tistar is,
      ;; by definition, the Ti due to only stellar irradiation flux. First
      ;; we compute Tistar0, which is the value for zero albedo.
      ;;
      Tistar0 = (0.5d0*alpha[ir]*fr[ir])^0.25 * sqrt(Rstar/r[ir]) * Tstar
      ;;
      ;; We postpone the scattering effects on Tistar to a later point.
      ;; 
      ;; Now, if there is an inner rim of the disk, this may add some more
      ;; flux to the irradiation (we included this effect in Ts already but
      ;; not yet in Ti, so we do this here). We assume here that at 2--3 mum
      ;; the albedo is zero, and we only include the thermal emission from 
      ;; the inner rim. Later we can include also the scattered light from
      ;; the inner rim.##############
      ;;
      if keyword_set(diffedge) and not keyword_set(noselfirr) then begin
         delta  = (r[ir]/rrim-1.d0)/(Hs[ir]/hrim-1.d0)
         if delta gt 0.d0 then begin
            costheta   = cos(atan((r[ir]-rrim)/(Hs[ir]-hrim)))
            if delta lt 1.d0 then begin
               S = 2.d0 * rrim^2 * costheta * ( delta * sqrt(1.d0-delta^2) $ 
                                            + asin(delta) )
            endif else begin
               S = pi  * rrim^2 * costheta
            endelse
         endif else begin
            S = 0.d0
         endelse
         Tiself = (0.5d0*alpha[ir]*S/(pi*r[ir]^2))^0.25 * diffedge.Tin
      endif else begin
         Tiself = 0.d0
      endelse
      ;;
      ;;  <<added for scattering>>
      ;;
      ;; Now include all the effects of scattering into the computation
      ;;
      if n_elements(noscat) eq 0 then begin
         ;;
         ;; Modify the Tistar due to scattering away of part of the secondary
         ;; radiation from the surface layer. We do this according to a new
         ;; recipe (new in comparison to the original 2001 scheme). The idea
         ;; is that we _know_ how much radiation will scatter, for an 
         ;; incident beam of radiation with a certain grazing angle. This
         ;; is the analytic H-function stuff of Chandrasekhar. Assuming now
         ;; the two-layer structure, we also know that after the first
         ;; absorption/scattering event 0.5*alb*F_in will be scattered
         ;; away. The remaining (reflect-0.5*alb)*F_in will be due to 
         ;; secondary scattering in the disk interior. Since the total
         ;; downward flux is 0.5*F_in (independent of albedo), we can
         ;; compute that (0.5-reflect+0.5*alb)*F_in is thermalized.
         ;; Normally this is 0.5*F_in, so the reduction factor is
         ;; redux=(1+alb-2*reflect). 
         ;;
         ;; NOTE: in using the function "reflect(albedo)", we assume that the
         ;; disk is vertically optically thick to scattering and also
         ;; effectively optically thick. This may be a wrong assumption, and
         ;; may need to be fixed in the future.
         ;;
         if n_elements(albedo) ne 0 then alb[ir]=albedo else begin
            alb[ir] = albedoplanck(opac,tstar,Ts[ir])
         endelse
         ;;
         case scatmethod of
            1: begin
               ;;
               ;; Original method: difference between reflect(alb) and 0.5*alb.
               ;; >>> SCATMETHOD: 1<<<
               ;;
               refl   = reflect(alb[ir])
               Tistar = Tistar0 * (1.d0+alb[ir]-2.d0*refl)^0.25
            end
            2: begin
               ;;
               ;; Simple method: just 0.5*alb, and no difference in midplane
               ;; >>> SCATMETHOD: 2<<<
               ;;
               refl   = 0.5*alb[ir]
               Tistar = Tistar0 
            end
            3: begin
               ;;
               ;; Simple method: just 0.5*alb, and equal reduction in surface
               ;; and in midplane
               ;; >>> SCATMETHOD: 3<<<
               ;;
               refl   = 0.5*alb[ir]
               Tistar = Tistar0 * (1-0.5*alb[ir])^0.25
            end
            4: begin
               ;;
               ;; Tryout: just 0.25*alb, but for the surface 0.75*alb 
               ;; >>> SCATMETHOD: 4<<<
               ;;
               refl   = 0.5*alb[ir]
               Tistar = Tistar0 * (1-0.25*alb[ir])^0.25
            end
            5: begin
               ;;
               ;; Simple method: just a reduction of sqrt(1-alb), 
               ;; an equal reduction in surface and in midplane
               ;; >>> SCATMETHOD: 5<<<
               ;;
               refl   = 1.d0-sqrt(1.d0-alb[ir])
               Tistar = Tistar0 * (1.d0-alb[ir])^0.125
            end
         endcase
         ;;
         ;; The Tistar0 can be used to remind what the actual irradiation
         ;; flux from the star was: Firrstar=2*ss*Tistar0^4. This holds even
         ;; when scattering is included. Note that one must use Tistar0, and
         ;; not Tistar, because the latter (see below) is corrected for the
         ;; fact that with scattering part of the flux from the disk
         ;; interior is in fact the 'albedo*reflect(albedo)' fraction of the
         ;; input flux that is not thermalized but instead reflected.
         ;;
         ;; Using this Firrstar, and knowing the albedo, we know the total
         ;; flux of the scattered radiation.
         ;;
         Firrstar  = 1.13406d-4 * Tistar0^4
         Fscat[ir] = refl * Firrstar
         ;;
      endif else begin
         Tistar = Tistar0
         alb[ir] = 0.d0
         Fscat[ir] = 0.d0
      endelse
      ;;
      ;; Then add the contributions of the stellar irradiation and the
      ;; self irradiation to obtain the actual internal temperature.
      ;;
      Ti[ir] = ( Tistar^4 + Tiself^4 )^0.25
      ;;
      ;; Now, if the psi-factor should be consistently included, then we 
      ;; call a separate subroutine to modify the Ti according to the value
      ;; of psi. It then becomes an implicit equation for Ti. The psi factor
      ;; is the factor taking into account the fact that the disk interio may
      ;; not be optically thick to its own radiation and/or the radiation 
      ;; the surface layer. The psi factor depends on the Ti and Ts, and
      ;; on the surface density of the interior and of course on the opacity.
      ;; 
      ;; Note on the inclusion of scattering: we assume here that the disk
      ;; is optically thick to scattering!
      ;;
      psi[ir]  = 1.d0
      psis[ir] = 1.d0
      psii[ir] = 1.d0
      if usepsi eq 1 then begin
         psitmp   = 0.d0
         psistmp  = 0.d0
         psiitmp  = 0.d0
         Ti[ir]   = solvetiwithpsi(Ti[ir],Ts[ir],sigma[ir],opac,$
                                  psitot=psitmp,psis=psistmp,psii=psiitmp)
         psi[ir]  = psitmp
         psis[ir] = psistmp
         psii[ir] = psiitmp
      endif
      ;;
      ;; The pressure scale height according to Eq.(7) of CG97, see also
      ;; for rewritten version Eq.(4) Dullemond 2000 A&A 361, L17-20
      ;;
      Hp[ir]    = r[ir]*sqrt(Ti[ir]*r[ir]/(Tc*Rstar))
      ;;
      ;; Find the chi==Hs/Hp self-consistently by using the grazing angle
      ;; alpha and the Planck mean opacity at stellar temperatures. If 
      ;; self irradiation by the rim is also present, another chi is
      ;; computed at the temperature of the rim.
      ;;
      ;; [bugfix 12-12-00: Ti[ir]-->Tstar in call to kappaplanck]
      if not keyword_set(chifix) then begin
          chi[ir]   = findchi(kappaplanck(opac,Tstar),$
                              alpha[ir],sigma[ir])
          if(chi[ir] lt 1.0) then begin
              print,"Optically thin disk at",ir
              chi[ir] = 1.0
          endif
      endif else begin
          chi[ir]   = chifix
      endelse
      Hs[ir]    = Hp[ir] * chi[ir] 
      ;;
      ;; Find the grazing angle alpha according to Eq.(5) of CG97
      ;;
      alphaold  = alpha[ir]
      if not keyword_set(grazang) then begin
         alpha[ir] = 0.4d0 * near * Rstar / r[ir] + $
           flareindex[ir]*Hs[ir]/r[ir]
      endif else begin
         alpha[ir] = grazang[ir]
      endelse
      iloop = iloop+1      
      if iloop gt 100 then begin
         print,'No convergence at ir=',ir
         stop
      endif
   endrep until abs(alphaold-alpha[ir]) lt 1.d-4*abs(alphaold+alpha[ir])
   ;;
   ;; Now that we found the CG97 value of the Hs (incl emission from the
   ;; inner rim, but still without the self-shadowing), we can check 
   ;; whether it lies in the shadow.
   ;;
   if keyword_set(diffedge) then begin
      if(Hs[ir]/r[ir] lt hrim/rrim) then begin
         ;;
         ;; CG97 lies in shadow. So CG cannot exist here. We either simply
         ;; put the temperature, and scale height to zero (if noselfirr is
         ;; explicitly set to prevent this), or we do the self-irradiated
         ;; disk: irradiation by the disk's inner rim
         ;;
         if not keyword_set(noselfirr) then begin
            ;; 
            ;; Switch to the half-shadowed disk: the disk with irradiation
            ;; by the inner rim. The procedure to find the disk in this
            ;; region is roughly the same as
            ;;
            iloop = 1
            repeat begin
                ;;
                ;; Determine the temperature of the surface layer. 
                ;; We don't test for epsfix here.
                ;; 
                ;; BUGFIX 12-02-02
                Ts[ir]      = solvetthin(opac,0.d0,0.d0,trim,rrim,hrim,$
                                         r[ir],Hs[ir],starvis=0.d0)
                ;;
                ;; Now compute the self-irradiation flux. Assume that
                ;; the albedo at 2--3 mum is zero.
                ;;
                delta  = (r[ir]/rrim-1.d0)/(Hs[ir]/hrim-1.d0)
                if delta gt 0.d0 then begin
                    costheta   = cos(atan((r[ir]-rrim)/(Hs[ir]-hrim)))
                    if delta lt 1.d0 then begin
                        S = 2.d0 * rrim^2 * costheta * ( delta * $
                          sqrt(1.d0-delta^2) + asin(delta) )
                    endif else begin
                        S = pi  * rrim^2 * costheta
                    endelse
                endif else begin
                    S = 0.d0
                endelse
                Ti[ir] = (0.5d0*alpha[ir]*S/(pi*r[ir]^2))^0.25 * diffedge.Tin
                ;;
                ;; Do the psi factors again (see above)
                ;;
                if Ti[ir] gt 0.d0 and Ts[ir] gt 0.d0 then begin
                    psi[ir]  = 1.d0
                    psis[ir] = 1.d0
                    psii[ir] = 1.d0
                    if usepsi eq 1 then begin
                        psitmp   = 0.d0
                        psistmp  = 0.d0
                        psiitmp  = 0.d0
                        Ti[ir]   = solvetiwithpsi(Ti[ir],Ts[ir],sigma[ir],opac,$
                                                  psitot=psitmp,psis=psistmp,$
                                                  psii=psiitmp)
                        psi[ir]  = psitmp
                        psis[ir] = psistmp
                        psii[ir] = psiitmp
                    endif
                endif else begin
                    ;;
                    ;; Apparently the disk cannot be sustained
                    ;; 777 is a magic cookie which will be recognized
                    ;; below
                    ;;
                    iloop = 777    
                endelse
                ;;
                ;; The pressure scale height (see above again)
                ;;
                Hp[ir]    = r[ir]*sqrt(Ti[ir]*r[ir]/(Tc*Rstar))
                ;;
                ;; Find the chi==Hs/Hp self-consistently (again see above)
                ;;
                if not keyword_set(chifix) then begin
                    chi[ir]   = findchi(kappaplanck(opac,trim),$
                                        alpha[ir],sigma[ir])
                    if(chi[ir] lt 1.0) then begin
                        print,"Optically thin disk at",ir
                        chi[ir] = 1.0
                    endif
                endif else begin
                    chi[ir]   = chifix
                endelse
                Hs[ir]    = Hp[ir] * chi[ir] 
                ;;
                ;; Find the grazing angle alpha according to Eq.(5) of CG97
                ;;
                alphaold  = alpha[ir]
                if not keyword_set(grazang) then begin
                   alpha[ir] = flareindex[ir]*Hs[ir]/r[ir]
                endif else begin
                   alpha[ir] = grazang[ir]
                endelse
                iloop = iloop+1      
            endrep until abs(alphaold-alpha[ir]) lt $
              1.d-4*abs(alphaold+alpha[ir]) or iloop eq 777 $
              or Hs[ir] eq 0.d0
            if iloop eq 777 or Hs[ir] eq 0.d0 then begin
                ;;
                ;; Apparently no solution was found here, so the disk
                ;; must be in the full shadow here.
                ;;
                Hs[ir] = 0.d0
                Hp[ir] = 0.d0
                Ti[ir] = 0.d0
                Ts[ir] = 0.d0
                irfullshadow = ir
                irhalfshadow = ir
            endif else begin
                ;;
                ;; Apparently the half-shadowed disk can already survive
                ;; here, so we update the halfshadow radius, but leave
                ;; full shadow un-updated.
                ;;
                irhalfshadow = ir
            endelse
        endif else begin
            ;; 
            ;; No half-shadowed disk allowed: just switch off CG97
            ;;
            Hs[ir] = 0.d0
            Hp[ir] = 0.d0
            Ti[ir] = 0.d0
            Ts[ir] = 0.d0
            irfullshadow = ir
            irhalfshadow = ir
        endelse
        ;;
        ;; Some of the previously used internal variables for the non
        ;; self-shadowed part of the disk will be used below. So we 
        ;; set them to the appropriate values now that this
        ;; non-self-shadowed part does not exist at this point.
        ;; 
        Tistar0=0.d0
        Tiself=Ti[ir]
      endif
   endif
   ;;
   ;; The vertical optical depth of the surface layer is fixed according
   ;; to energy conservation: half of the radiation is emitted by the
   ;; surface layer, and half by the interior. With scattering this is
   ;; not that easy anymore, but after a modification we can still use
   ;; the simple equation of the temperature ratios as used in DDN2001.
   ;;
   ;;  <<modified for scattering>>
   ;;
   case scatmethod of
      1: begin
         ;;
         ;; Original method
         ;; >>> SCATMETHOD: 1 <<<
         ;;
         tausurf[ir] = (1.d0-alb[ir])*(Tistar0^4+Tiself^4) $ 
                   / Ts[ir]^4 / ( 2.0 )
      end
      2: begin
         ;;
         ;; >>> SCATMETHOD: 2 <<<
         ;;
         tausurf[ir] = (1.d0-alb[ir])*(Tistar0^4+Tiself^4) $ 
                   / Ts[ir]^4 / ( 2.0 )
      end
      3: begin
         ;;
         ;; Simple method
         ;; >>> SCATMETHOD: 3<<<
         ;;
         tausurf[ir] = (1.d0-0.5*alb[ir])*(Tistar0^4+Tiself^4) $ 
                   / Ts[ir]^4 / ( 2.0 )
      end
      4: begin
         ;;
         ;; Simple method
         ;; >>> SCATMETHOD: 4<<<
         ;;
         tausurf[ir] = (1.d0-0.75*alb[ir])*(Tistar0^4+Tiself^4) $ 
                   / Ts[ir]^4 / ( 2.0 )
      end
      5: begin
         ;;
         ;; >>> SCATMETHOD: 5<<<
         ;;
         tausurf[ir] = sqrt(1.d0-alb[ir])*(Tistar0^4+Tiself^4) $ 
                   / Ts[ir]^4 / ( 2.0 )
      end
   endcase
   ;;
   ;; Compute the corresponding surface density
   ;;
   if(Ts[ir] gt 0.d0) then begin
      sigsurf[ir] = tausurf[ir] / kappaplanck(opac,Ts[ir])
   endif else begin
      sigsurf[ir] = 0.d0
   endelse
   ;;
   ;; Now check if the Hs/R exceeds 1.d0, which must trigger a warning
   ;;
   if(Hs[ir]/r[ir] gt 1.d0) then begin
      print,'Problem at ir=',ir,', R=',r[ir],': Hs/R > 1.d0'
   endif
   ;;
   ;; Increase flare index counter
   ;;
   flcnt=flcnt+1
endfor
;
; Check whether at some radius the Hs/r goes down, in order to be
; able to check energy conservation: which fraction of 4pi is covered
; by the CG disk? NOTE: this is really a hack....
;
irmaxcg=0
for ir=1,nr-1 do begin
   if hs[ir]/r[ir] ge hs[ir-1]/r[ir-1] then irmaxcg=ir
endfor
;
; Return the whole disk structure
;
id='cgdisk'
return,{id:id,r:r,Ti:Ti,Ts:Ts,Hp:Hp,Hs:Hs,fr:fr,alpha:alpha,$
        tausurf:tausurf,Sigma:Sigma,AU:AU,psi:psi,psis:psis,$
        psii:psii,Tstar:Tstar,Rstar:Rstar,Lstar:Lstar,Mstar:Mstar,$
        opac:opac,irhalfshadow:irhalfshadow,sigsurf:sigsurf,albedo:alb,$
        irfullshadow:irfullshadow,irmaxcg:irmaxcg,fscat:fscat,$
        flareindex:flareindex}
end



;------------------------------------------------------------------------
;       FUNCTION: COMPUTE DIFFUSIVE INNER EDGE STRUCTURE
;------------------------------------------------------------------------
function diffedgefunc,mstarmsun,tstar,lstarlsun,rsigmaau,sigma0,plsigma,$
                routau,nr,rinau=rinau,tin=tin,epss=epss,surf=surf,$ 
                opac=opac,nocghalf=nocghalf,chiwall=chiwall,$
                fixhs=fixhs,noselfirr=noselfirr,noscat=noscat,albedo=albedo
;
; Message
;
;print,"Diffusive inner edge solver"
;print,"  WARNING: The diffusion is still without 1/r^n dilution"
;
; Some defaults
;
Tinsf   = 0.d0
tauinsf = 0.d0
if keyword_set(surf) then begin
   print,"Surface layer for inner wall is deactivated!"
   stop
endif
;
; Constants
;
pi    = !dpi
ss    = 5.6703d-5
AU    = 1.496d13
GG    = 6.672d-8
Msun  = 1.99d33
Lsun  = 3.86d33
mugas = 2.3d0
mp    = 1.6726d-24
kk    = 1.3807d-16
;
; Convert to CGS
;
rout  = routau*AU
rsigma= rsigmaau*AU
Lstar = lstarlsun * Lsun
Mstar = mstarmsun * Msun
rstar = sqrt(lstar/(4*pi*ss*tstar^4))
;
; Warnings & errors
;
if keyword_set(rinau) then begin
    if rinau*AU lt rstar then begin
        print, "================================="
        print, "Warning: Rinau smaller that Rstar"
        print, "================================="
        rinAU = rstar/AU
    endif
endif 
if not keyword_set(opac) then begin
   print,"===================================="
   print,"Error: MUST specify opac "
   print,"===================================="
   stop
endif
if not keyword_set(lstarlsun) then begin
    print, "==============================================="
    print, "Must specify the stellar luminosity (lstarlsun)"
    print, "==============================================="
    stop
endif 
;
; If neither rinau nor tin are set, then use tin=1500 as standard
;
if not keyword_set(tin) and not keyword_set(rinau) then begin
   print,'Setting Tin=1500 K by default!'
   tin = 1500.d0
endif 
;
; Find the effective kappa for radial outward rays. The factor of 8
; comes in because of the expected structure of the diffusive disk
; behind the inner rim. See paper.
;
kap8     = 8.0 * kappaplanck(opac,tstar)
;
; From here there are two possible ways to determine the inner edge
;
if keyword_set(tin) then begin
   ;;
   ;; Determine the inner radius from the temperature tin. Include the
   ;; self-irradiation in this computation, unless /noselfirr. 
   ;; First guess of Rin, based on non-self-irradiating inner rim
   ;;
   rinau = sqrt(Lsun*lstarlsun/(4.*pi*ss*Tin^4))/AU
   rin   = rinau * AU
   rsig  = rsigmaau * AU
   Hpin  = sqrt(kk*Tin*rin^3/(mugas*mp*GG*Mstar))
   ;;
   ;; Now include the effects of self-irradiation
   ;;
   if not keyword_set(noselfirr) then begin
      ;; print,'Rin    (no self-irr)         = ',rinau,' AU'
      rin0   = rin
      Hp0    = Hpin
      hsback = 0.d0
      chback = 0.d0
      rin    = solverinselfirr(rin0,Sigma0,rsig,plsigma,$
                    kap8,Hp0,fixhs=fixhs,chi=chiwall,$
                    hsback=hsback,chback=chback)
      rinau  = rin / AU
      Hpin   = sqrt(kk*Tin*rin^3/(mugas*mp*GG*Mstar))
      Hsin   = Hsback
      chi    = chback
      selfirr=1
   endif else begin
      Sig    = Sigma0*(rin/rsig)^plsigma
      chi    = findchi(kap8,1.d0,Sig)
      if(chi lt 1.0) then begin
          print,"Optically thin disk at inner radius"
          stop
      endif
      if not keyword_set(fixhs) then begin
         Hsin   = Hpin * chi
      endif else begin
         Hsin   = fixhs
      endelse
      print,'Self-irradiation of inner wall not included!'
      selfirr=0
   endelse
   ;;print,'Radius of inner rim          = ',rinau,' AU'
   ;;
   ;; Done determining rin from tin...
   ;;
endif else begin
   ;;
   ;; The inner radius is fixed, either to be R_* or rinau as given
   ;;
   rin  = rinau * AU
   Sig  = Sigma0*(rinau/rsigmaau)^plsigma
   ;;
   ;; Determine the chi=hs/hp for determining the height of the inner rim
   ;;
   if keyword_set(chiwall) then begin
      chi     = chiwall
   endif else begin
      chi     = findchi(kap8,1.d0,Sig)
   endelse
   ;;
   ;; Determine temperature of the inner rim and the vertical height 
   ;; together. If the wall-self-irradiation iss included, then do this
   ;; by an iterative procedure.
   ;;
   Tin      = (Lstar/(4.*pi*rin^2*ss))^0.25
   if not keyword_set(fixhs) then begin
      Hpin  = sqrt(kk*Tin*rin^3/(mugas*mp*GG*Mstar))
      Hsin  = chi * Hpin
   endif else begin
      Hsin  = fixhs
      Hpin  = Hsin / chi
   endelse
   if not keyword_set(noselfirr) then begin
      if not keyword_set(fixhs) then begin
         Hpin0 = Hpin
         Hpin  = solvehpinselfirr(rin,chi,Hpin0)
         Hsin  = chi * Hpin
         ;; print,'Inner rim self-irradiation correction for Hp: ',Hpin/Hpin0
      endif
      Tin = Tin * ( 1.d0 + Hsin/rin )^0.25
      selfirr=1
   endif else begin
      print,'Self-irradiation of inner wall not included!'
      selfirr=0
   endelse
   ;;
   ;; Done determining Tin from Rin
   ;;
endelse
;
; The scattering stuff
;
if n_elements(noscat) ne 0 then albedowall=0.d0 else begin
   if n_elements(albedo) ne 0 then albedowall=albedo else begin
      albedowall = albedoplanck(opac,tstar,tin)
   endelse
endelse
;
; Original method:
;
;reflwall   = reflect(albedowall)
;
; Simple method:
;
reflwall   = 0.5*albedowall
;
; Stuff...
;
if rin lt rstar then begin
   print,'Inner radius of disk within stellar radius... Aborting...'
   stop
endif
;
; Now make the radial grid
;
r     = rin*((rout/rin)^(dindgen(nr)/(nr-1.d0)))
;
; Then make the Sigma array
;
Sigma   = Sigma0*(r/rsigma)^plsigma
;
; Then compute the T(R) and H(R) according to the simple outward
; diffusion recipe from the paper.
;
Hsdiff = Hsin-(r-rin)/8.d0
Hsdiff = 0.5d0*(abs(Hsdiff)+Hsdiff)
Hpdiff = Hsdiff/chi
Tdiff  = Tin * (Hsdiff/Hsin)^2
Hshadow= Hsin * (r/r(0))
;
;  Note: If we want to use the Hsdiff for the SED: we make have to
;        use another chi
;
id='diffwall'
return,{id:id,r:r,Sigma:Sigma,AU:AU,$
        Tstar:Tstar,Rstar:Rstar,Lstar:Lstar,Mstar:Mstar,$
        Tin:Tin,Hpin:Hpin,Hsin:Hsin,Hpdiff:Hpdiff,Hsdiff:Hsdiff,$
        Tdiff:Tdiff,Hshadow:Hshadow,opac:opac,$
        Tinsf:Tinsf,tauinsf:tauinsf,chirim:chi,$
        selfirr:selfirr,reflwall:reflwall,albedowall:albedowall}
end


;------------------------------------------------------------------------
;                  FIND SIGMA0 FROM THE MASS
;
; This routine calculates the sigma0 from the given mass. This routine
; takes into account the dependency of the position of the inner rim
; on the Sigma0 itself.
;------------------------------------------------------------------------
function sigmazero_from_mass,massdustmsun,mstarmsun,tstar,lstarlsun,rsigmaau,$
                plsigma,routau,tin=tin,rinau=rinau,opac=opac,_extra=_extra
;
; Constants
;
pi    = !dpi
ss    = 5.6703d-5
AU    = 1.496d13
GG    = 6.672d-8
Msun  = 1.99d33
Lsun  = 3.86d33
mugas = 2.3d0
mp    = 1.6726d-24
kk    = 1.3807d-16
cc    = 2.9979d10
hh    = 6.6262d-27
;
; Some checks
;
if not keyword_set(tin) and not keyword_set(rinau) then begin
   print,"sigmazero_from_mass: Must specify either tin or rin"
   stop
endif
if keyword_set(tin) and keyword_set(rinau) then begin
   print,"sigmazero_from_mass: Must specify either tin or rin"
   print,"  but not both!"
   stop
endif
;
; Convert some stuff
;
rout  = routau * AU
rsig  = rsigmaau * AU
mdust = massdustmsun * Msun
if keyword_set(rinau) then rin = rinau * AU
;
; If tin is set
;
if keyword_set(tin) then begin
   ;;
   ;; First guess of rin
   ;;
   rin = sqrt(Lsun*lstarlsun/(4.*pi*ss*Tin^4))
   ;;
   ;; Iterate rin
   ;;
   rinold = rin
   conv = 0
   for ii=1,10 do begin
      if conv eq 0 then begin
         ;;
         ;; Find the Sigma0 from the mass, for the current value of rin
         ;;  
         if plsigma ne -2.0 then begin
            sigma0 = mdust*(plsigma+2.)*rsig^plsigma/(2.*!dpi*$
                          (rout^(plsigma+2.)-rin^(plsigma+2.)))
         endif else begin
            sigma0 = mdust*rsig^plsigma/(2.*!dpi*$
                          (alog(rout)-alog(rin)))
         endelse
         ;;
         ;; Call diffedgefunc
         ;;
         sig0=sigma0
         d=diffedgefunc(mstarmsun,tstar,lstarlsun,rsigmaau,sig0,plsigma,$
                        routau,3,tin=tin,opac=opac,_extra=_extra)
         rin = d.r[0]
         ;;
         ;; Check for convergence
         ;;
         if abs(rin/rinold-1.d0) lt 1.d-2 then conv=1
         rinold=rin
         ;;
      endif
   endfor
   if conv eq 0 then begin
      print,"Finding Sigma0 from Mass: no convergence"
      stop
   endif
endif else begin
   ;;
   ;; Find the Sigma0 from the mass, for the current value of rin
   ;;  
   if plsigma ne -2.0 then begin
      sigma0 = mdust*(plsigma+2.)*rsig^plsigma/(2.*!dpi*$
                          (rout^(plsigma+2.)-rin^(plsigma+2.)))
   endif else begin
      sigma0 = mdust*rsig^plsigma/(2.*!dpi*$
                          (alog(rout)-alog(rin)))
   endelse
endelse
return,sigma0
end



;------------------------------------------------------------------------
;                FUNCTION: MAKE AN ACCRETION DISK
;
; This function makes a simple accretion disk model with grey opacity.
;------------------------------------------------------------------------
function accrdiskfunc,mstarmsun,tstar,lstarlsun,mdotmsyr,routau,nr,$
                      rinau=rinau,alpha=alpha,opac=opac,$
                      kappa0=kappa0
;
; Constants
;
pi    = !dpi
ss    = 5.6703d-5
AU    = 1.496d13
GG    = 6.672d-8
Msun  = 1.99d33
Lsun  = 3.86d33
Rsun  = 6.96d10
mugas = 2.3d0
mp    = 1.6726d-24
kk    = 1.3807d-16
yr    = 3.1536d7
gtd   = 1.d2        ; Gas / dust ratio
;
; Convert to CGS
;
rout  = routau*AU
Mstar = mstarmsun * Msun
Mdot  = mdotmsyr * Msun / yr
Lstar = lstarlsun * Lsun
Mstar = mstarmsun * Msun
Rstar = sqrt(lstar/(4*pi*ss*tstar^4))
;
; Keywords
;
if not keyword_set(rinau) then rin=Rstar else rin=rinau*AU
if not keyword_set(alpha) then alpha=0.01
if not keyword_set(kappa0) then kappa0=gtd*1.d4
;
; Warnings & errors
;
if not keyword_set(opac) and not $
  keyword_set(kappa0) then begin
   print,"=============================================="
   print,"Error: MUST specify opac or kappa0"
   print,"=============================================="
   stop
endif
print,'WARNING: BUG IN CODE: SHOULD INCLUDE (1-sqrt(Rin/R)) correction)'
;
; Kappa
;
if keyword_set(kappa0) then begin
    gotopac=0
    opac=0
endif else begin
    gotopac=1
endelse
;
; Now make the radial grid
;
r     = rin*((rout/rin)^(dindgen(nr)/(nr-1.d0)))
;
; Compute the stuff for the output spectrum (Teff is effective surface
; temperature)
;
Omk   = sqrt(GG*Mstar/r^3)
Qplus = (3./(4.*pi))*Mdot*Omk^2
;;;Teff  = (Qplus/(2.*ss))^0.25
Teff  = (Qplus*(1.d0-sqrt(rin/r))/(2.*ss))^0.25
;;; BUG HERE: Must be Teff  = (Qplus*(1.d0-sqrt(rin/r))/(2.*ss))^0.25
;;; But no time to be sure...
;
; Compute the stuff for finding Tc(R) (midplane temperature) and Sigma(R)
;
Tc    = dblarr(nr)
Sigma = dblarr(nr)
for ir=0,nr-1 do begin
   if gotopac eq 0 then begin
      kappa = kappa0
   endif else begin
      kappa = kappaplanck(opac,Teff[ir])
   endelse
   Tc[ir]    = (Qplus[ir]*mugas*mp*kappa*Mdot*Omk[ir]/$
                  (8*pi*ss*kk*alpha))^0.2
   if Tc[ir] lt Teff[ir] then begin
      print,'Optically thin accretion disk.... ir=',ir
      stop
   endif
   Sigma[ir] = mugas*mp*Mdot*Omk[ir]/(2.*pi*kk*alpha*Tc[ir])
endfor
;
print,'Rstar = ',Rstar/Rsun,' Rsun'
;
id='accrdisk'
return,{id:id,r:r,Sigma:Sigma,AU:AU,$
        Tstar:Tstar,Rstar:Rstar,Lstar:Lstar,Mstar:Mstar,$
        Teff:Teff,Tc:Tc,opac:opac}
end



;------------------------------------------------------------------------
;      GLUE THE DIFFUSIVE INNER RIM AND OUTER FLARING DISK TOGETHER
;
; This routine is meant to create the proper diffusive shadowed region
; in between the flaring part and the inner rim. This includes the
; diffusion from the inner rim outwards (which is already done in the
; diffedgefunc()), the diffusion from the flaring disk inwards, and a
; lower limit to the temperature. In later versions we may add effects
; of irradiation from the top of the innr rim (the 'ring source'), or
; other effects. For now we do it in this simple way.
;------------------------------------------------------------------------
function diffglue,cg,de
;
; Constants
;
pi    = !dpi
ss    = 5.6703d-5
AU    = 1.496d13
GG    = 6.672d-8
Msun  = 1.99d33
Lsun  = 3.86d33
mugas = 2.3d0
mp    = 1.6726d-24
kk    = 1.3807d-16
;
r=cg.r
nr=n_elements(r)
irhalfshadow=cg.irhalfshadow
irfullshadow=cg.irfullshadow
hscg=dblarr(nr)
if irfullshadow lt nr-1 then hscg(irfullshadow+1:*)=cg.hs(irfullshadow+1:*)
if irhalfshadow lt nr-1 then begin
   for ir=irhalfshadow,0,-1 do begin
      dum = (hscg[ir+1]/r[ir+1]-(alog(r[ir+1])-alog(r[ir]))/8.d0)*r[ir]
      if dum lt 0.d0 then dum=0.d0
      hscg[ir] = ( dum^8 + hscg[ir]^8)^(1./8.)
   endfor
endif
hs=(hscg^8+de.hsdiff^8)^(1./8.)
hp=dblarr(nr)
Ti=dblarr(nr)
Ts=dblarr(nr)
if irhalfshadow lt nr-1 then begin
   hp(irhalfshadow+1:*)=cg.hp(irhalfshadow+1:*)   
   Ti(irhalfshadow+1:*)=cg.ti(irhalfshadow+1:*)   
   Ts(irhalfshadow+1:*)=cg.ts(irhalfshadow+1:*)   
   alpha=cg.alpha[irhalfshadow+1]
endif else alpha=cg.alpha[nr-1]
Tc = GG * cg.Mstar * mugas * mp / ( kk * cg.Rstar )
for ir=min([cg.irhalfshadow,nr-1]),0,-1 do begin
   chi=findchi(kappaplanck(cg.opac,cg.Tstar),alpha,cg.sigma[ir])
   hp[ir]=max([hs[ir]/chi,de.hpdiff[ir]])
   Ti[ir]=Hp[ir]^2*Tc*cg.Rstar/r[ir]^3
   Ts[ir]=max([cg.Ts[ir],Ti[ir]])
endfor
return,{Ti:Ti,Ts:Ts,Hp:Hp,Hs:Hs}
end



;------------------------------------------------------------------------
;                   FUNCTION: MAKE AN SED
;
; incl             Make the SED as seen at inclination incl.
; notrans          If set: do not take the transparancy of the disk
;                  into account (which is wrong, but which could be
;                  interesting to see what the effect is)
;------------------------------------------------------------------------
function sed,a,lmin=lmin,lmax=lmax,nlambda=nlambda,incl=incl,$
                   notrans=notrans,nostar=nostar,noscat=noscat,$
                   image_surf=image_surf,$
                   image_mid=image_mid,image_scat=image_scat
;
; Constants
;
pi    = !dpi
ss    = 5.6703d-5
AU    = 1.496d13
GG    = 6.672d-8
Msun  = 1.99d33
Lsun  = 3.86d33
mugas = 2.3d0
mp    = 1.6726d-24
kk    = 1.3807d-16
cc    = 2.9979d10
hh    = 6.6262d-27
;
if n_elements(incl) eq 0 then begin
   print,'In this version of cgplus:sed() one MUST specify the inclination'
   stop
endif
;
if a.id ne 'accrdisk' then begin
   nu=a.opac.freq
   lambda=cc/nu
   nlambda=n_elements(lambda)
endif else begin
   nlambda=200
   lambda=1d-4*0.1*10^((4.d0*dindgen(nlambda))/(nlambda-1.d0))
   nu=cc/lambda
endelse
;
; The cosine of the inclination
;
incrad   = 0.01745329252*incl
cosi     = cos(incrad)
;
; Make separate spectral contributions
;
lnu_star = dblarr(nlambda)
lnu_mid  = dblarr(nlambda)
lnu_surf = dblarr(nlambda)
lnu_scat = dblarr(nlambda)
lnu_wall = dblarr(nlambda)
lnu_accr = dblarr(nlambda)
lnu      = dblarr(nlambda)
;
; Make the star spectrum
;
lnu_star = pi * bplanck(nu,a.Tstar) * 4 * pi * a.Rstar^2
;
; Preset the lnu_scat
;
lnu_scat = 0.d0*lnu_star
;
; Copy some info
;
r=a.r
nr=n_elements(r)
;
; Which disk are we dealing with?
;
cg=0
dw=0
ac=0
id=a.id
irmincg = 0
irmaxcg = 0
if id eq 'cgdisk' then begin
   cg      = 1
   irmincg = 0
   ;;  irmaxcg = nr-1
   irmaxcg = a.irmaxcg
endif
if id eq 'diffwall' then begin
   dw      = 1
endif
if id eq 'combi' then begin
   cg      = 1
   dw      = 1
   irmincg = a.cgstart
   ;;  irmaxcg = nr-1
   irmaxcg = a.irmaxcg
endif
if id eq 'accrdisk' then begin
   ac      = 1
endif
;
; Do preparations for scattering
;
scatspec = lnu_star / integrate(nu,lnu_star)
;
; Make the SEDs for CG 
;
if cg ne 0 and irmincg lt irmaxcg then begin
   ;;
   ;; Loop over shells for the interior
   ;;
   nr   = n_elements(a.r)
   ;;
   ;; Make the image
   ;;
   image_mid=dblarr(nlambda,nr)
   image_surf=dblarr(nlambda,nr)
   image_scat=dblarr(nlambda,nr)
   for ir=irmincg,irmaxcg do begin
      if not keyword_set(notrans) then begin
         dil     = 1.d0 - exp(-a.sigma[ir]*a.opac.cabs/cosi)
         transpf = 1.d0 + exp(-a.sigma[ir]*a.opac.cabs/cosi)
      endif else begin
         dil     = 1.d0
         transpf = 1.d0
      endelse
      sigsurf       = a.tausurf[ir] / kappaplanck(a.opac,a.Ts[ir])
      taulay        = transpf * sigsurf * a.opac.cabs / cosi
      srfdil        = 1.d0 - exp(-taulay)
      image_mid[*,ir] = dil * bplanck(nu,a.Ti[ir]) 
      image_surf[*,ir] = srfdil * bplanck(nu,a.Ts[ir]) 
      image_scat[*,ir] = a.Fscat[ir]*scatspec/!dpi
   endfor
   ;;
   ;; Integrate the image to obtain a spectrum
   ;;
   dum = 0.d0
   for ir=irmincg+1,irmaxcg do begin
      dum = dum + 4.0*pi^2 * 0.5d0 * $
             ( image_mid[*,ir] + image_mid[*,ir-1] ) * $
             ( r[ir]^2 - r[ir-1]^2 ) * cosi
   endfor
   lnu_mid = lnu_mid + dum
   dum = 0.d0
   for ir=irmincg+1,irmaxcg do begin
      dum = dum + 4.0*pi^2 * 0.5d0 * $
             ( image_surf[*,ir] + image_surf[*,ir-1] ) * $
             ( r[ir]^2 - r[ir-1]^2 ) * cosi
   endfor
   lnu_surf = lnu_surf + dum
   dum = 0.d0
   for ir=irmincg+1,irmaxcg do begin
      dum = dum + 4.0*pi^2 * 0.5d0 * $
             ( image_scat[*,ir] + image_scat[*,ir-1] ) * $
             ( r[ir]^2 - r[ir-1]^2 ) * cosi
   endfor
   lnu_scat = lnu_scat + dum
endif
;
; Make the SEDs for diff wall
;
if dw ne 0 then begin
   if keyword_set(incl) then begin
      lnu_wall = 16.0 * pi * a.r(0) * a.Hsin * sin(incrad) * bplanck(nu,a.Tin)
   endif else begin
      lnu_wall = 4.0 * pi^2 * a.r(0) * a.Hsin * bplanck(nu,a.Tin)
      if(a.selfirr ne 0) then begin
         ;;
         ;; If self-irradiation: part of the flux of the wall is irradiating
         ;; itself. The wall is therefore hotter than normal, with a local
         ;; flux that is (1+Hs/R) times larger. But the total flux must be
         ;; still the same. So since we have no incl set, we must compute
         ;; the total flux, meaning we divide by (1+Hs/R). In the SED this
         ;; means that the total flux remains the same, but it shifted to
         ;; higher frequencies. The same reasoning one can do for the 
         ;; irradiation of the flaring disk by the inner wall: part of the
         ;; wall's radiation is intercepted by the flaring disk. This means
         ;; that the flaring disk will be slightly more luminous, while 
         ;; the total rim luminosity must be decreased because less fraction
         ;; of 4*pi of the wall radiation reaches infinity.
         ;;
         print,"CHECK THIS"
         fac_wall = 1.d0 / ( 1.d0 + a.Hsin/a.r(0) )
         if a.id eq 'cgdisk' or a.id eq 'combi' then begin
            fac_cg   = ( 1.d0 - a.Hs(a.irmaxcg)/a.r(a.irmaxcg) )
         endif else begin
            fac_cg   = 1.d0
         endelse
         factor   = min([fac_wall,fac_cg])
         lnu_wall = lnu_wall * factor
      endif
   endelse
endif
;
; If accretion disk
;
if ac ne 0 then begin
   nr=n_elements(a.r)
   if keyword_set(incl) then begin
      ;;
      ;; Flux, converted into luminosity
      ;;
      for ir=1,nr-2 do begin
         lnu_accr = lnu_accr + bplanck(nu,a.Teff[ir]) * 4.0*pi^2 * $
           ( R[ir]*R[ir+1] - R[ir]*R[ir-1] ) * cosi
      endfor
      ir = 0
      lnu_accr = lnu_accr + bplanck(nu,a.Teff[ir]) * 4.0*pi^2 * $
        ( R[ir]*R[ir+1] - R[ir]*R[ir] ) * cosi
      ir = nr-1
      lnu_accr = lnu_accr + bplanck(nu,a.Teff[ir]) * 4.0*pi^2 * $
        ( R[ir]*R[ir] - R[ir]*R[ir-1] ) * cosi
   endif else begin
      ;;
      ;; Total luminosity
      ;;
      for ir=1,nr-2 do begin
         lnu_accr = lnu_accr + bplanck(nu,a.Teff[ir]) * 2.0*pi^2 * $
           ( R[ir]*R[ir+1] - R[ir]*R[ir-1] )
      endfor
      ir = 0
      lnu_accr = lnu_accr + bplanck(nu,a.Teff[ir]) * 2.0*pi^2 * $
        ( R[ir]*R[ir+1] - R[ir]*R[ir] )
      ir = nr-1
      lnu_accr = lnu_accr + bplanck(nu,a.Teff[ir]) * 2.0*pi^2 * $
        ( R[ir]*R[ir] - R[ir]*R[ir-1] )
   endelse   
endif
;
; Switch some components off, if requested
;
if not keyword_set(nostar) then inclstar = 1.d0 else inclstar = 0.d0
if not keyword_set(noscat) then inclscat = 1.d0 else inclscat = 0.d0
;
; Add all spectra
;
lnu = inclstar*lnu_star + lnu_mid + lnu_surf + inclscat*lnu_scat + $
      lnu_wall + lnu_accr
;
; Find all luminosities
;
l_star = 0.d0
l_mid  = 0.d0
l_surf = 0.d0
l_scat = 0.d0
l_wall = 0.d0
l_accr = 0.d0
for inu=1,nlambda-2 do begin
   dnu    = 0.5d0 * nu(inu) * ( alog(nu(inu+1)) - alog(nu(inu-1)) )
   l_star = l_star + lnu_star(inu) * dnu
   l_mid  = l_mid  + lnu_mid(inu)  * dnu
   l_surf = l_surf + lnu_surf(inu) * dnu
   l_scat = l_scat + lnu_scat(inu) * dnu
   l_wall = l_wall + lnu_wall(inu) * dnu
   l_accr = l_accr + lnu_accr(inu) * dnu
endfor
;
; Print this information
;
if keyword_set(lstot) and not keyword_set(incl) then begin
   l_tot = l_star+l_mid+l_surf+l_scat+l_wall+l_accr
endif else begin 
   l_tot = l_star
endelse
print,"----------------------------------------------"
print,"Full luminosity          = ",l_tot/Lsun," Lsun"
print,"Relative luminosities:"
print,"  L_wall/L_tot           = ",l_wall/l_tot
print,"  L_mid/L_tot            = ",l_mid/l_tot
print,"  L_surf/L_tot           = ",l_surf/l_tot
print,"  L_scat/L_tot           = ",l_scat/l_tot
print,"  L_accr/L_tot           = ",l_accr/l_tot
;
return,{nu:nu,lambda:lambda,lnu:lnu,lnu_star:lnu_star,$
        lnu_mid:lnu_mid,lnu_surf:lnu_surf,lnu_wall:lnu_wall,$
        lnu_accr:lnu_accr,lnu_scat:lnu_scat,l_accr:l_accr,l_scat:l_scat,$
        l_star:l_star,l_mid:l_mid,l_surf:l_surf,l_wall:l_wall}
end



;------------------------------------------------------------------------
;                   FUNCTION: MAKE A SED (OLD VERSION!!!)
;
; incl             If defined: make the SED as seen at inclination incl.
;                  If not set: make the SED for the outgoing luminosity.
; lstot            If set: for the stellar luminosity: include the
;                  effect that only part of the stellar radiation
;                  reaches infinity. This affects only the spectra
;                  when incl is not set, i.e. when the total lums are
;                  plotted.
; notrans          If set: do not take the transparancy of the disk
;                  into account (which is wrong, but which could be
;                  interesting to see what the effect is)
;------------------------------------------------------------------------
function sed_old,a,lmin=lmin,lmax=lmax,nlambda=nlambda,incl=incl,lstot=lstot,$
                   notrans=notrans
;
; Constants
;
pi    = !dpi
ss    = 5.6703d-5
AU    = 1.496d13
GG    = 6.672d-8
Msun  = 1.99d33
Lsun  = 3.86d33
mugas = 2.3d0
mp    = 1.6726d-24
kk    = 1.3807d-16
cc    = 2.9979d10
hh    = 6.6262d-27
;
if a.id ne 'accrdisk' then begin
   nu=a.opac.freq
   lambda=cc/nu
   nlambda=n_elements(lambda)
endif else begin
   nlambda=200
   lambda=1d-4*0.1*10^((4.d0*dindgen(nlambda))/(nlambda-1.d0))
   nu=cc/lambda
endelse
;
if keyword_really_set(incl) then begin 
   incrad = 0.01745329252*incl
   cosi   = cos(incrad)
endif else begin
   cosi   = 1.d0/sqrt(3.d0)
endelse
;
; Make separate spectral contributions
;
lnu_star = dblarr(nlambda)
lnu_mid  = dblarr(nlambda)
lnu_surf = dblarr(nlambda)
lnu_wall = dblarr(nlambda)
lnu_accr = dblarr(nlambda)
lnu_scat = dblarr(nlambda)
lnu      = dblarr(nlambda)
;
; Make the star spectrum
;
lnu_star = pi * bplanck(nu,a.Tstar) * 4 * pi * a.Rstar^2
;
; Preset the lnu_scat
;
lnu_scat = 0.d0*lnu_star
;
; Now, if keyword lstot is set, then multiply the stellar luminosity
; with 1-Omega, meaning: the total lum is of course smaller than that
; of a bare star, since part of its luminosity is intercepted. Normally
; this is not included, since it would confuse the user. But if one 
; wishes to be precise, then it should be included.
;
if keyword_set(lstot) and not keyword_really_set(incl) then begin
   if a.id eq 'cgdisk' or a.id eq 'combi' then begin
      nr = a.irmaxcg+1
   endif
   if a.id eq 'diffwall' then begin
      nr = 1
   endif
   if a.id eq 'accrdisk' then begin
      nr = n_elements(a.r)
   endif
   lnu_star = lnu_star * ( 1.d0 - a.Hs[nr-1]/a.r[nr-1] )
endif
;
; Copy some info
;
r=a.r
nr=n_elements(r)
;
; Which disk are we dealing with?
;
cg=0
dw=0
ac=0
id=a.id
irmincg = 0
irmaxcg = 0
if id eq 'cgdisk' then begin
   cg      = 1
   irmincg = 0
   ;;  irmaxcg = nr-1
   irmaxcg = a.irmaxcg
endif
if id eq 'diffwall' then begin
   dw      = 1
endif
if id eq 'combi' then begin
   cg      = 1
   dw      = 1
   irmincg = a.cgstart
   ;;  irmaxcg = nr-1
   irmaxcg = a.irmaxcg
endif
if id eq 'accrdisk' then begin
   ac      = 1
endif
;
; Make the SEDs for CG 
;
if cg ne 0 and irmincg lt irmaxcg then begin
   ;;
   ;; Loop over shells for interior
   ;;
   nr   = n_elements(a.r)
   if keyword_really_set(incl) then begin
      ;;
      ;; Flux, converted into luminosity
      ;;
      for ir=irmincg+1,irmaxcg-1 do begin
         kappai=findkappa(a.opac,a.Ti[ir],/abs)
         if not keyword_set(notrans) then begin
            dil  = 1.d0 - exp(-a.sigma[ir]*kappai/cosi)
         endif else begin
            dil  = 1.d0
         endelse
         lnu_mid = lnu_mid + dil * bplanck(nu,a.Ti[ir]) * 4.0*pi^2 * $
           ( R[ir]*R[ir+1] - R[ir]*R[ir-1] ) * cosi
      endfor
      ir = irmincg
      if not keyword_set(notrans) then begin
         kappai=findkappa(a.opac,a.Ti[ir],/abs)
         dil  = 1.d0 - exp(-a.sigma[ir]*kappai/cosi)
      endif else begin
         dil  = 1.d0
      endelse
      lnu_mid = lnu_mid + dil * bplanck(nu,a.Ti[ir]) * 4.0*pi^2 * $
        ( R[ir]*R[ir+1] - R[ir]*R[ir] ) * cosi
      ir = irmaxcg
      if not keyword_set(notrans) then begin
         kappai=findkappa(a.opac,a.Ti[ir],/abs)
         dil  = 1.d0 - exp(-a.sigma[ir]*kappai/cosi)
      endif else begin
         dil  = 1.d0
      endelse
      lnu_mid = lnu_mid + dil * bplanck(nu,a.Ti[ir]) * 4.0*pi^2 * $
        ( R[ir]*R[ir] - R[ir]*R[ir-1] ) * cosi
   endif else begin
      ;;
      ;; Total luminosity
      ;;
      for ir=irmincg+1,irmaxcg-1 do begin
         if not keyword_set(notrans) then begin
            kappai=findkappa(a.opac,a.Ti[ir],/abs)
            dil  = 1.d0 - exp(-a.sigma[ir]*kappai/cosi)
         endif else begin
            dil  = 1.d0
         endelse
         lnu_mid = lnu_mid + dil * bplanck(nu,a.Ti[ir]) * 2.0*pi^2 * $
           ( R[ir]*R[ir+1] - R[ir]*R[ir-1] )
      endfor
      ir = irmincg
      if not keyword_set(notrans) then begin
         kappai=findkappa(a.opac,a.Ti[ir],/abs)
         dil  = 1.d0 - exp(-a.sigma[ir]*kappai/cosi)
      endif else begin
         dil  = 1.d0
      endelse
      lnu_mid = lnu_mid + dil * bplanck(nu,a.Ti[ir]) * 2.0*pi^2 * $
        ( R[ir]*R[ir+1] - R[ir]*R[ir] )
      ir = irmaxcg
      if not keyword_set(notrans) then begin
         kappai=findkappa(a.opac,a.Ti[ir],/abs)
         dil  = 1.d0 - exp(-a.sigma[ir]*kappai/cosi)
      endif else begin
         dil  = 1.d0
      endelse
      lnu_mid = lnu_mid + dil * bplanck(nu,a.Ti[ir]) * 2.0*pi^2 * $
        ( R[ir]*R[ir] - R[ir]*R[ir-1] )
   endelse
   ;;
   ;; Loop over shells for surface layer: thermal emission
   ;;
   ;; This includes the possible effects of transparency of the disk 
   ;; interior, so that the hot layer on the other side can be seen
   ;; (factor transpf). 
   ;;
   nr=n_elements(a.r)
   for ir=irmincg+1,irmaxcg-1 do begin
       if not keyword_set(notrans) then begin
           kappai   = findkappa(a.opac,a.Ti[ir],/abs)
           transpf  = 1.d0 + exp(-a.sigma[ir]*kappai/cosi)
       endif else begin
           transpf  = 1.d0
       endelse
       sigsurf  = a.tausurf[ir] / kappaplanck(a.opac,a.Ts[ir])
       kappas   = findkappa(a.opac,a.Ts[ir],/abs)
       lnu_surf = lnu_surf + 4.0 * pi^2 * sigsurf * kappas * $
         bplanck(nu,a.Ts[ir]) * $
         ( R[ir]*R[ir+1] - R[ir]*R[ir-1] ) * transpf
   endfor
   ir = irmincg
   if not keyword_set(notrans) then begin
       kappai   = findkappa(a.opac,a.Ti[ir],/abs)
       transpf  = 1.d0 + exp(-a.sigma[ir]*kappai/cosi)
   endif else begin
       transpf  = 1.d0
   endelse
   sigsurf  = a.tausurf[ir] / kappaplanck(a.opac,a.Ts[ir])
   kappas   = findkappa(a.opac,a.Ts[ir],/abs)
   lnu_surf = lnu_surf + 4.0 * pi^2 * sigsurf * kappas * $
     bplanck(nu,a.Ts[ir]) * $
     ( R[ir]*R[ir+1] - R[ir]*R[ir] ) * transpf
   ir = irmaxcg
   if not keyword_set(notrans) then begin
       kappai   = findkappa(a.opac,a.Ti[ir],/abs)
       transpf  = 1.d0 + exp(-a.sigma[ir]*kappai/cosi)
   endif else begin
       transpf  = 1.d0
   endelse
   sigsurf  = a.tausurf[ir] / kappaplanck(a.opac,a.Ts[ir])
   kappas   = findkappa(a.opac,a.Ts[ir],/abs)
   lnu_surf = lnu_surf + 4.0 * pi^2 * sigsurf * kappas * $
     bplanck(nu,a.Ts[ir]) * $
     ( R[ir]*R[ir] - R[ir]*R[ir-1] ) * transpf
   ;;
   ;; Loop over shells for surface layer: scattering
   ;;
   ;;  NOTE: here we don't take into account the fact that the
   ;;        disk could become optically thin to scattered light.
   ;;
   rscat    = R[irmincg:irmaxcg]
   ffsc     = 2*!dpi*a.Fscat[irmincg:irmaxcg]*rscat
   lumscat  = integrate(rscat,ffsc)
   lnu_scat = lnu_scat + (lumscat/a.lstar)*lnu_star
endif
;
; Make the SEDs for diff wall
;
if dw ne 0 then begin
   if keyword_really_set(incl) then begin
      lnu_wall = 16.0 * pi * a.r(0) * a.Hsin * sin(incrad) * bplanck(nu,a.Tin)
   endif else begin
      lnu_wall = 4.0 * pi^2 * a.r(0) * a.Hsin * bplanck(nu,a.Tin)
      if(a.selfirr ne 0) then begin
         ;;
         ;; If self-irradiation: part of the flux of the wall is irradiating
         ;; itself. The wall is therefore hotter than normal, with a local
         ;; flux that is (1+Hs/R) times larger. But the total flux must be
         ;; still the same. So since we have no incl set, we must compute
         ;; the total flux, meaning we divide by (1+Hs/R). In the SED this
         ;; means that the total flux remains the same, but it shifted to
         ;; higher frequencies. The same reasoning one can do for the 
         ;; irradiation of the flaring disk by the inner wall: part of the
         ;; wall's radiation is intercepted by the flaring disk. This means
         ;; that the flaring disk will be slightly more luminous, while 
         ;; the total rim luminosity must be decreased because less fraction
         ;; of 4*pi of the wall radiation reaches infinity.
         ;;
         print,"CHECK THIS"
         fac_wall = 1.d0 / ( 1.d0 + a.Hsin/a.r(0) )
         if a.id eq 'cgdisk' or a.id eq 'combi' then begin
            fac_cg   = ( 1.d0 - a.Hs(a.irmaxcg)/a.r(a.irmaxcg) )
         endif else begin
            fac_cg   = 1.d0
         endelse
         factor   = min([fac_wall,fac_cg])
         lnu_wall = lnu_wall * factor
      endif
   endelse
   ;;
   ;; Now include the effects of scattering
   ;;
   if a.reflwall gt 0.d0 then begin
      ;print,a.reflwall
      lnu_wall = lnu_wall * (1.d0-a.reflwall)
      lnu_scat = lnu_scat + a.reflwall*integrate(nu,lnu_wall)*lnu_star/a.lstar
   endif
endif
;
; If accretion disk
;
if ac ne 0 then begin
   nr=n_elements(a.r)
   if keyword_really_set(incl) then begin
      ;;
      ;; Flux, converted into luminosity
      ;;
      for ir=1,nr-2 do begin
         lnu_accr = lnu_accr + bplanck(nu,a.Teff[ir]) * 4.0*pi^2 * $
           ( R[ir]*R[ir+1] - R[ir]*R[ir-1] ) * cosi
      endfor
      ir = 0
      lnu_accr = lnu_accr + bplanck(nu,a.Teff[ir]) * 4.0*pi^2 * $
        ( R[ir]*R[ir+1] - R[ir]*R[ir] ) * cosi
      ir = nr-1
      lnu_accr = lnu_accr + bplanck(nu,a.Teff[ir]) * 4.0*pi^2 * $
        ( R[ir]*R[ir] - R[ir]*R[ir-1] ) * cosi
   endif else begin
      ;;
      ;; Total luminosity
      ;;
      for ir=1,nr-2 do begin
         lnu_accr = lnu_accr + bplanck(nu,a.Teff[ir]) * 2.0*pi^2 * $
           ( R[ir]*R[ir+1] - R[ir]*R[ir-1] )
      endfor
      ir = 0
      lnu_accr = lnu_accr + bplanck(nu,a.Teff[ir]) * 2.0*pi^2 * $
        ( R[ir]*R[ir+1] - R[ir]*R[ir] )
      ir = nr-1
      lnu_accr = lnu_accr + bplanck(nu,a.Teff[ir]) * 2.0*pi^2 * $
        ( R[ir]*R[ir] - R[ir]*R[ir-1] )
   endelse   
endif
;
; Add all spectra
;
lnu = lnu_star + lnu_mid + lnu_surf + lnu_wall + lnu_accr + lnu_scat
;
; Find all luminosities
;
l_star = 0.d0
l_mid  = 0.d0
l_surf = 0.d0
l_wall = 0.d0
l_accr = 0.d0
l_scat = 0.d0
for inu=1,nlambda-2 do begin
   dnu    = 0.5d0 * nu(inu) * ( alog(nu(inu+1)) - alog(nu(inu-1)) )
   l_star = l_star + lnu_star(inu) * dnu
   l_mid  = l_mid  + lnu_mid(inu)  * dnu
   l_surf = l_surf + lnu_surf(inu) * dnu
   l_wall = l_wall + lnu_wall(inu) * dnu
   l_accr = l_accr + lnu_accr(inu) * dnu
   l_scat = l_scat + lnu_scat(inu) * dnu
endfor
;
; Print this information
;
if keyword_set(lstot) and not keyword_really_set(incl) then begin
   l_tot = l_star+l_mid+l_surf+l_wall+l_accr+l_scat
endif else begin 
   l_tot = l_star
endelse
print,"----------------------------------------------"
print,"Full luminosity          = ",l_tot/Lsun," Lsun"
print,"Relative luminosities:"
print,"  L_wall/L_tot           = ",l_wall/l_tot
print,"  L_mid/L_tot            = ",l_mid/l_tot
print,"  L_surf/L_tot           = ",l_surf/l_tot
print,"  L_scat/L_tot           = ",l_scat/l_tot
print,"  L_accr/L_tot           = ",l_accr/l_tot
;
return,{nu:nu,lambda:lambda,lnu:lnu,lnu_star:lnu_star,$
        lnu_mid:lnu_mid,lnu_surf:lnu_surf,lnu_wall:lnu_wall,$
        lnu_scat:lnu_scat,lnu_accr:lnu_accr,l_accr:l_accr,l_scat:l_scat,$
        l_star:l_star,l_mid:l_mid,l_surf:l_surf,l_wall:l_wall}
end



;------------------------------------------------------------------------
;                       FUNCTION: PLANCK
;------------------------------------------------------------------------
function bplanck,nnu,TT
nu=nnu*1.d0
T=TT*1.d0
cc=2.9979d10
hh=6.6262d-27
kk=1.3807e-16
n=n_elements(nu)
bpl=dblarr(n)
if TT eq 0.d0 then return,dblarr(n)
for i=0,n-1 do begin
  x=hh*nu[i]/(kk*T)
  ;print,x
  if x gt 100.0 then b=0.d0
  if x gt 1.d-3 then b=(2.d0*hh*nu[i]^3/cc^2)/(exp(x)-1.d0) $
    else b=2.0*nu[i]^2*kk*T/cc^2
  bpl[i]=b
endfor
return,bpl
end

;------------------------------------------------------------------------
;                      RUN THE WHOLE CG SHIT
;------------------------------------------------------------------------
function cgdisk,mstar=mstar,lstar=lstar,tstar=tstar,rin=rin,rout=rout,$
    sigma0=sigma0,rsigma=rsigma,plsigma=plsigma,nr=nr,nopsi=nopsi,$
    opac=opac,tin=tin,chifix=chifix,starvis=starvis,$
    nonear=nonear,flareindex=flareindex,initflare=initflare,$
    albedo=albedo,noscat=noscat,tsthin=tsthin,scatmethod=scatmethod,$
    mdust=mdust,mdisk=mdisk,gastodust=gastodust,grazang=grazang
;
pi    = !dpi
ss    = 5.6703d-5
AU    = 1.496d13
GG    = 6.672d-8
Msun  = 1.99d33
Lsun  = 3.86d33
mugas = 2.3d0
mp    = 1.6726d-24
kk    = 1.3807d-16
;
if not keyword_set(mstar) then begin
   print,"CGPLUS ERROR: Must specify Mstar [in Msun]"
   stop
endif
if not keyword_set(lstar) then begin
   print,"CGPLUS ERROR: Must specify Lstar [in Lsun]"
   stop
endif
if not keyword_set(tstar) then begin
   print,"CGPLUS ERROR: Must specify Tstar [in K]"
   stop
endif
if not keyword_set(rout) then rout=100.0
if not keyword_set(rsigma) then rsigma=1.0
if not keyword_really_set(plsigma) then plsigma=-1.5
if not keyword_set(nr) then nr=20
if not keyword_set(gastodust) then gastodust=99.d0
;
if not keyword_set(tin) then begin
   if not keyword_set(rin) then rin=0.0
endif else begin
   if not keyword_set(lstar) then begin
      print,'Cannot compute Rin from Tin is Lstar is not specified'
      stop
   endif
   rin=sqrt(Lsun*lstar/(4.*pi*ss*Tin^4))/AU
   print,'Radius of inner rim          = ',rin,' AU'
endelse
;
i=keyword_set(sigma0)
j=keyword_set(mdust)
k=keyword_set(mdisk)
if i+j+k ne 1 then begin
   print,"CGPLUS ERROR: Must specify sigma0 OR mdust OR mdisk"
   stop
endif
if keyword_set(mdisk) then mdust=mdisk/(gastodust+1.d0)
if keyword_set(mdust) then begin
   sigma0=sigmazero_from_mass(mdust,mstar,tstar,lstar,rsigma,$
                plsigma,rout,rinau=rin,opac=opac,$ 
                surf=surf,albedo=albedo,noscat=noscat,$
                chiwall=chiwall,noselfirr=noselfirr)
endif
;
a=cgdiskfunc(mstar,tstar,lstar,rsigma,sigma0,plsigma,rin,rout,nr,$
             opac=opac,chifix=chifix,nonear=nonear,$
             nopsi=nopsi,flareindex=flareindex,initflare=initflare,$
             starvis=starvis,tsthin=tsthin,albedo=albedo,$
             noscat=noscat,grazang=grazang,scatmethod=scatmethod)
;
; Now some printouts
;
radius=a.r
dummy=2*!dpi*a.sigma*a.r
dustmass=integrate(radius,dummy)
print,'Mass of the disk (dust)      = ',dustmass/MSun,' Msun'
print,'Mass of the disk (dus+gas)   = ',(gastodust+1.d0)*dustmass/MSun,' Msun'
if a.irmaxcg ne nr-1 then begin
   print,'Note that hs/r has inflection at r=',a.r[a.irmaxcg]/AU,' AU'
   print,'     Truncating the disk there'
endif
print,'Covering fraction by CG disk = ',a.Hs[a.irmaxcg]/a.r[a.irmaxcg]
for ir=1,nr-1 do begin
   if a.psii[ir] lt 0.95 and a.psii[ir-1] ge 0.95 then begin
      print,'Disk optthin to own rad at  R =',a.r[ir]/AU,' AU'
   endif
   if a.psis[ir] lt 0.95 and a.psis[ir-1] ge 0.95 then begin
      print,'Disk optthin to surf rad at R =',a.r[ir]/AU,' AU'
   endif
endfor
print,'Energy conservation error     = ',find_encons_error(a.r,a.Hs,$
        flareindex=a.flareindex)-1.d0
;
; Now return the stuff
;
return,a
end

;------------------------------------------------------------------------
;                   RUN THE WHOLE DIFFUSIVE SHIT
;------------------------------------------------------------------------
function diffedge,mstar=mstar,lstar=lstar,tstar=tstar,rin=rin,rout=rout,$
    sigma0=sigma0,rsigma=rsigma,plsigma=plsigma,nr=nr,opac=opac,$
    noselfirr=noselfirr,albedo=albedo,noscat=noscat
if not keyword_set(rin) then rin=1.d-30
if not keyword_set(rout) then rout=100.0
if not keyword_set(rsigma) then rsigma=1.0
if not keyword_really_set(plsigma) then plsigma=-1.5
if not keyword_set(nr) then nr=20
return,diffedgefunc(mstar,tstar,lstar,rsigma,sigma0,plsigma,$
                rout,nr,rinau=rin,noselfirr=noselfirr,$
                albedo=albedo,noscat=noscat)
end

;------------------------------------------------------------------------
;            FUNCTION: COMBINE THE DIFFERENT COMPONENTS
;------------------------------------------------------------------------
function cgplus,mstar=mstar,lstar=lstar,tstar=tstar,rin=rin,rout=rout,$
    sigma0=sigma0,rsigma=rsigma,plsigma=plsigma,nr=nr,starvis=starvis,$
    notrunc=notrunc,opac=opac,chifix=chifix,albedo=albedo,noscat=noscat,$
    surf=surf,chiwall=chiwall,nopuff=nopuff,tin=tin,noselfirr=noselfirr,$
    flareindex=flareindex,initflare=initflare,noglue=noglue,$
    mdust=mdust,mdisk=mdisk,gastodust=gastodust,tsthin=tsthin
;
pi    = !dpi
ss    = 5.6703d-5
AU    = 1.496d13
GG    = 6.672d-8
Msun  = 1.99d33
Lsun  = 3.86d33
mugas = 2.3d0
mp    = 1.6726d-24
kk    = 1.3807d-16
;
if not keyword_set(mstar) then begin
   print,"CGPLUS ERROR: Must specify Mstar [in Msun]"
   stop
endif
if not keyword_set(lstar) then begin
   print,"CGPLUS ERROR: Must specify Lstar [in Lsun]"
   stop
endif
if not keyword_set(tstar) then begin
   print,"CGPLUS ERROR: Must specify Tstar [in K]"
   stop
endif
if not keyword_set(rout) then rout=100.0
if not keyword_set(rsigma) then rsigma=1.0
if not keyword_really_set(plsigma) then plsigma=-1.5
if not keyword_set(nr) then nr=20
if not keyword_set(gastodust) then gastodust=99.d0
;
i=keyword_set(sigma0)
j=keyword_set(mdust)
k=keyword_set(mdisk)
if i+j+k ne 1 then begin
   print,"CGPLUS ERROR: Must specify sigma0 OR mdust OR mdisk"
   stop
endif
if keyword_set(mdisk) then mdust=mdisk/(gastodust+1.d0)
if keyword_set(mdust) then begin
   sigma0=sigmazero_from_mass(mdust,mstar,tstar,lstar,rsigma,$
                plsigma,rout,tin=tin,rinau=rin,opac=opac,$ 
                surf=surf,albedo=albedo,noscat=noscat,$
                chiwall=chiwall,noselfirr=noselfirr)
endif
;
if not keyword_set(nopuff) then begin
   b=diffedgefunc(mstar,tstar,lstar,rsigma,sigma0,plsigma,rout,nr,rinau=rin,$
                  opac=opac,surf=surf,albedo=albedo,noscat=noscat,$
                  chiwall=chiwall,tin=tin,noselfirr=noselfirr)
   if not keyword_set(notrunc) then begin
      a=cgdiskfunc(mstar,tstar,lstar,rsigma,sigma0,plsigma,rin,rout,nr,$
                   opac=opac,chifix=chifix,albedo=albedo,$
                   diffedge=b,noselfirr=noselfirr,flareindex=flareindex,$
                   nonear=nonear,nopsi=nopsi,initflare=initflare,$
                   starvis=starvis,noscat=noscat,tsthin=tsthin,$
                   scatmethod=scatmethod)
   endif else begin
      a=cgdiskfunc(mstar,tstar,lstar,rsigma,sigma0,plsigma,rin,rout,nr,$
                   opac=opac,chifix=chifix,albedo=albedo,$
                   nonear=nonear,nopsi=nopsi,flareindex=flareindex,$
                   initflare=initflare,starvis=starvis,noscat=noscat,$
                   tsthin=tsthin,scatmethod=scatmethod)
   endelse
endif else begin
   b=diffedgefunc(mstar,tstar,lstar,rsigma,sigma0,plsigma,rout,nr,rinau=rin,$
                  opac=opac,surf=surf,tin=tin,noselfirr=1,$
                  albedo=albedo,noscat=noscat)
   a=cgdiskfunc(mstar,tstar,lstar,rsigma,sigma0,plsigma,rin,rout,nr,$
                opac=opac,noselfirr=noselfirr,chifix=chifix,albedo=albedo,$
                flareindex=flareindex,nonear=nonear,nopsi=nopsi,$
                initflare=initflare,starvis=starvis,noscat=noscat,$
                tsthin=tsthin,scatmethod=scatmethod)
   b=diffedgefunc(mstar,tstar,lstar,rsigma,sigma0,plsigma,rout,nr,rinau=rin,$
                  opac=opac,surf=surf,albedo=albedo,noscat=noscat,$
                  fixhs=a.Hs[0],tin=tin,noselfirr=noselfirr)
   a=cgdiskfunc(mstar,tstar,lstar,rsigma,sigma0,plsigma,rin,rout,nr,$
                opac=opac,diffedge=b,chifix=chifix,albedo=albedo,$
                noselfirr=noselfirr,flareindex=flareindex,$
                nonear=nonear,nopsi=nopsi,initflare=initflare,$
                starvis=starvis,noscat=noscat,tsthin=tsthin,$
                scatmethod=scatmethod)
endelse
;
; Now remake the shadowed region in a proper way (okay, this is a bit
; a hack, but it works)
;
if not keyword_set(noglue) then begin
   c=diffglue(a,b)
endif else begin
   c=a
endelse
;
; Now some printouts
;
radius=a.r
dummy=2*!dpi*a.sigma*a.r
dustmass=integrate(radius,dummy)
print,'Mass of the disk (dust)      = ',dustmass/MSun,' Msun'
print,'Mass of the disk (dus+gas)   = ',(gastodust+1.d0)*dustmass/MSun,' Msun'
print,'Radius of inner rim          = ',a.r[0]/AU,' AU'
print,'Covering fraction by Wall    = ',b.Hsin/b.r[0],' * 4 pi'
print,'Inner wall Hs/Hp             = ',b.chirim
if a.irmaxcg ne nr-1 then begin
   print,'Note that hs/r has inflection at r=',a.r[a.irmaxcg]/AU,' AU'
   print,'     Truncating the disk there'
endif
print,'Covering fraction by CG disk = ',a.Hs[a.irmaxcg]/a.r[a.irmaxcg]
if a.irfullshadow ge 0 then $
  print,'Full shadow at ir=',strcompress(string(a.irfullshadow)),'       R =',a.r[a.irfullshadow]/AU,' AU'
if not keyword_set(noselfirr) then begin
   if a.irhalfshadow ge 0 then $
     print,'Half shadow at ir=',strcompress(string(a.irhalfshadow)),'       R =',a.r[a.irhalfshadow]/AU,' AU'
endif
for ir=1,nr-1 do begin
   if a.psii[ir] lt 0.95 and a.psii[ir-1] ge 0.95 then begin
      print,'Disk optthin to own rad at  R =',a.r[ir]/AU,' AU'
   endif
   if a.psis[ir] lt 0.95 and a.psis[ir-1] ge 0.95 then begin
      print,'Disk optthin to surf rad at R =',a.r[ir]/AU,' AU'
   endif
endfor
if max(a.albedo) gt 0.d0 then begin
   print,'Albedo (average)              = ',rebin(a.albedo,1)
endif
if a.irhalfshadow+1 lt nr-1 then begin
   print,'Energy conservation error     = ',find_encons_error($
        a.r(a.irhalfshadow+1:*),a.Hs(a.irhalfshadow+1:*),$
        flareindex=a.flareindex(a.irhalfshadow+1:*))-1.d0
endif
;
; Find the new diffusive structure incl diffusion from unshadowed CG97 part
; [not yet done]
;
id='combi'
return,{id:id,r:a.r,Ti:c.Ti,Ts:c.Ts,Hp:c.Hp,Hs:c.Hs,fr:a.fr,alpha:a.alpha,$
        tausurf:a.tausurf,Sigma:a.Sigma,AU:a.AU,psi:a.psi,$
        psis:a.psis,psii:a.psii,Tstar:a.Tstar,Rstar:a.Rstar,$
        Lstar:a.Lstar,Mstar:a.Mstar,fscat:a.fscat,albedo:a.albedo,$
        Tin:b.Tin,Hpin:b.Hpin,Hsin:b.Hsin,Hpdiff:b.Hpdiff,Hsdiff:b.Hsdiff,$
        Tdiff:b.Tdiff,Hshadow:b.Hshadow,cgstart:a.irhalfshadow+1,opac:a.opac,$
        Tinsf:b.Tinsf,tauinsf:b.tauinsf,sigsurf:a.sigsurf,$
        irmaxcg:a.irmaxcg,selfirr:b.selfirr,irhalfshadow:a.irhalfshadow,$
        irfullshadow:a.irfullshadow,flareindex:a.flareindex,$
        reflwall:b.reflwall,albedowall:b.albedowall}
end


;------------------------------------------------------------------------
;                 RUN THE WHOLE ACCRETION DISK SHIT
;------------------------------------------------------------------------
function accrdisk,mstar=mstar,lstar=lstar,tstar=tstar,rin=rin,rout=rout,$
    mdot=mdot,nr=nr,alpha=alpha,kappa0=kappa0,opac=opac
if not keyword_set(rout) then rout=100.0
if not keyword_set(nr) then nr=20
return,accrdiskfunc(mstar,tstar,lstar,mdot,rout,nr,$
                    rinau=rin,alpha=alpha,kappa0=kappa0,opac=opac)
end



;------------------------------------------------------------------------
;                   PROCEDURE: PLOT FULL SED
;------------------------------------------------------------------------
pro plotsed,s,dpc=dpc,lsun=lsun,nocomp=nocomp,notot=notot,$
            oplot=oplot,ylin=ylin,_extra=_extra
pi=!dpi
pc=3.08572d18
if not keyword_set(dpc) then begin
   if not keyword_set(lsun) then begin
       red=1.0
       ytitle='!4m!xL!D!4m!X!N (erg/s)'
    endif else begin
       red=2.60d-34
       ytitle='!4m!xL!D!4m!X!N/L!D!9n!X!N'       
    endelse
endif else begin
   red=1.d0/(4.d0*pi*(dpc*pc)^2)
   ytitle='!4m!xF!D!4m!X!N (erg/s/cm!A2!N)'
endelse
if keyword_set(notot) then begin
   nodata=1 
endif else begin
   nodata=0
endelse
if not keyword_set(oplot) then begin
   if not keyword_set(ylin) then begin
      plot_OO,s.lambda*1e4,red*s.nu*s.lnu,xtitle='!4k!X (!4l!Xm)',$
        ytitle=ytitle,nodata=nodata,_extra=_extra
   endif else begin
      plot_OI,s.lambda*1e4,red*s.nu*s.lnu,xtitle='!4k!X (!4l!Xm)',$
        ytitle=ytitle,nodata=nodata,_extra=_extra
   endelse
endif else begin
   if not keyword_set(notot) then begin
      oplot,s.lambda*1e4,red*s.nu*s.lnu,_extra=_extra
   endif
endelse
if not keyword_set(nocomp) then begin
   oplot,s.lambda*1e4,red*s.nu*s.lnu_star,line=3,_extra=_extra
   oplot,s.lambda*1e4,red*s.nu*s.lnu_mid,line=2,_extra=_extra
   oplot,s.lambda*1e4,red*s.nu*s.lnu_surf,line=1,_extra=_extra
   oplot,s.lambda*1e4,red*s.nu*s.lnu_scat,line=4,_extra=_extra
   oplot,s.lambda*1e4,red*s.nu*s.lnu_wall,line=5,_extra=_extra
   oplot,s.lambda*1e4,red*s.nu*s.lnu_accr,line=1,_extra=_extra
endif
end


;------------------------------------------------------------------------
;              FUNCTION: FIND THE PLANCK MEAN KAPPA 
;------------------------------------------------------------------------
function kappaplanck,opac,temp
;
; Constants
;
pi    = !dpi
ss    = 5.6703d-5
AU    = 1.496d13
GG    = 6.672d-8
Msun  = 1.99d33
Lsun  = 3.86d33
mugas = 2.3d0
mp    = 1.6726d-24
kk    = 1.3807d-16
cc    = 2.9979d10
hh    = 6.6262d-27
;
; First make a planck spectrum
;
nu    = opac.freq
bnu   = bplanck(nu,temp)
;
; Get the kappa table for this temperature
;
kappa = findkappa(opac,temp,/abs)
;
; Now make the average
;
nf=n_elements(nu)
aa=0.d0
bb=0.d0
for i=1,nf-2 do begin
   dnu = 0.5 * nu[i] * (alog(nu[i+1])-alog(nu[i-1]))
   aa  = aa + bnu(i)*kappa[i]*dnu
   bb  = bb + bnu(i)*dnu
endfor
kappa = aa/bb
;
; Find the corresponding kappa
;
return,kappa
;
end


;------------------------------------------------------------------------
;              FUNCTION: FIND THE SURFACE HEIGHT
;
; Solve the equation:
;
;                            2 * alpha
;   1 - erf(chi/sqrt(2)) = -------------
;                          Sigma * kappa
;
; where Sigma is the total surface density from z=-infty..infty.
; The solver goes down to chi=1.0, below which it will simply 
; return 0.d0.
;------------------------------------------------------------------------
function findchi,kappa,alpha,sigma
pi    = !dpi
rhs   = 2.d0 * alpha / ( kappa * sigma )
chi   = 4.0
iloop = 1
if 0.5*sigma*kappa/alpha lt 3.0 then begin
   return,0.d0
endif
lim = 0.1
repeat begin
   chiold = chi
   dum   = -2.d0*alog(rhs*exp(-chi^2/2.d0)/(1.d0-errorf(chi*sqrt(0.5d0))))
   chi   = sqrt(abs(dum))
   iloop = iloop+1      
   if iloop gt 20 then begin
      print,'No convergence in Hs solver'
      stop
   endif
endrep until abs(chi-chiold) lt 1.d-4*abs(chi+chiold) 
return,chi
end



;------------------------------------------------------------------------
;       FUNCTION: COMPUTE THE TOTAL LUMINOSITIES
;------------------------------------------------------------------------
function totlum,a
nf=n_elements(a.nu)
nu=a.nu
lum_star=0.d0
lum_mid=0.d0
lum_surf=0.d0
lum_wall=0.d0
for inu=1,nf-2 do begin
   lum_star = lum_star + 0.5*a.lnu_star(inu)*nu(inu)*(alog(nu(inu+1))-alog(nu(inu-1)))
   lum_mid  = lum_mid  + 0.5*a.lnu_mid(inu)*nu(inu)*(alog(nu(inu+1))-alog(nu(inu-1)))
   lum_surf = lum_surf + 0.5*a.lnu_surf(inu)*nu(inu)*(alog(nu(inu+1))-alog(nu(inu-1)))
   lum_wall = lum_wall + 0.5*a.lnu_wall(inu)*nu(inu)*(alog(nu(inu+1))-alog(nu(inu-1)))
endfor
return,{lstar:lum_star,lmid:lum_mid,lsurf:lum_surf,lwall:lum_wall}
end



;------------------------------------------------------------------------
;         FUNCTION: FIND RADII OF TAU=1 AT ALL FREQUENCIES
;------------------------------------------------------------------------
function findtauradii,opac,sigma,r,Td,tauone=tauone,reverse=reverse
if n_elements(tauone) eq 0 then tauone=1.d0
nr=n_elements(r)
nf=n_elements(opac.freq)
tau=dblarr(nr,nf)
rtau=dblarr(nf)
z0=sigma[0:nr-2]
z1=sigma[1:nr-1]
zz=z1-z0
if(min(zz)*max(zz) lt 0.d0) then begin
   print,"Error in findtauradii: Sigma is not monotonic!"
   stop 
endif
if(min(zz)*max(zz) eq 0.d0) then begin
   return,rtau
endif
for ir=0,nr-1 do begin
   kappa     = findkappa(opac,Td[ir],/abs,/sca)
   tau[ir,*] = sigma[ir]*kappa
endfor
if sigma[nr-1] gt sigma[0] then begin
   rmin=r[0]
   rmax=r[nr-1]
   reverse=0
endif else begin
   rmin=r[nr-1]
   rmax=r[0]
   reverse=1
endelse
for inu=0,nf-1 do begin
   if(min(tau[*,inu]) gt tauone) then begin
      rtau[inu]=rmin
   endif else begin
      if(max(tau[*,inu]) lt tauone) then begin
         rtau[inu]=rmax
      endif else begin
         rtau[inu] = interpol(r,tau(*,inu),tauone)
      endelse
   endelse
endfor
return,rtau
end

;------------------------------------------------------------------------
;              PRO: PLOT THE TAU=0.1, 1.0 ETC RADII
;------------------------------------------------------------------------
pro plottauradii,a
AU    = 1.496d13
z=findtauradii(a.opac,a.sigma,a.r,a.Ti,tauone=1.d0,reverse=reverse)
plot_OO,2.9979d14/a.opac.freq,z/AU,xtitle='!4k!X (!4l!Xm)',$
     ytitle='Radius for !4s!X=1 (AU)',thick=2
end


;  <<added for scattering>>
;------------------------------------------------------------------------
;      FUNCTION: FIND THE REFLECTION RATIO FOR GIVEN ALBEDO
;
; Using Chandrasekhar's radiative transfer theory, a table of reflection
; ratios has been computed as a function of albedo. This table is 
; appropriate for input intensity with a mu-dependence going as 1/mu,
; which is the kind of flux emitted by an optically thin layer. This
; subroutine interpolates within this table.
;------------------------------------------------------------------------
function reflect,albedo
alb=findgen(101)/100.
ref=[0.000E+00,2.488E-03,5.001E-03,7.540E-03,$
     1.011E-02,1.270E-02,1.532E-02,1.797E-02,$
     2.064E-02,2.335E-02,2.608E-02,2.885E-02,$
     3.165E-02,3.447E-02,3.733E-02,4.023E-02,$
     4.315E-02,4.612E-02,4.911E-02,5.214E-02,$
     5.521E-02,5.832E-02,6.146E-02,6.464E-02,$
     6.786E-02,7.113E-02,7.449E-02,7.785E-02,$
     8.125E-02,8.470E-02,8.820E-02,9.174E-02,$
     9.534E-02,9.898E-02,1.027E-01,1.064E-01,$
     1.102E-01,1.141E-01,1.180E-01,1.220E-01,$
     1.261E-01,1.302E-01,1.344E-01,1.386E-01,$
     1.430E-01,1.474E-01,1.518E-01,1.564E-01,$
     1.610E-01,1.658E-01,1.706E-01,1.755E-01,$
     1.802E-01,1.852E-01,1.904E-01,1.956E-01,$
     2.010E-01,2.064E-01,2.120E-01,2.177E-01,$
     2.235E-01,2.295E-01,2.356E-01,2.418E-01,$
     2.482E-01,2.547E-01,2.614E-01,2.682E-01,$
     2.752E-01,2.833E-01,2.908E-01,2.986E-01,$
     3.066E-01,3.148E-01,3.234E-01,3.322E-01,$
     3.413E-01,3.508E-01,3.607E-01,3.710E-01,$
     3.798E-01,3.905E-01,4.017E-01,4.134E-01,$
     4.256E-01,4.383E-01,4.546E-01,4.695E-01,$
     4.853E-01,5.024E-01,5.164E-01,5.348E-01,$
     5.590E-01,5.825E-01,6.027E-01,6.356E-01,$
     6.622E-01,7.003E-01,7.459E-01,8.229E-01,$
     1.000E+00]
return,interpol(ref,alb,albedo)
end


;------------------------------------------------------------------------
;              FUNCTION: FIND THE PLANCK AVERAGED ALBEDO
;------------------------------------------------------------------------
function albedoplanck,opac,trad,tdust
;
; Constants
;
pi    = !dpi
ss    = 5.6703d-5
AU    = 1.496d13
GG    = 6.672d-8
Msun  = 1.99d33
Lsun  = 3.86d33
mugas = 2.3d0
mp    = 1.6726d-24
kk    = 1.3807d-16
cc    = 2.9979d10
hh    = 6.6262d-27
;
; First make a planck spectrum
;
nu    = opac.freq
bnu   = bplanck(nu,trad)
;
; Get the kappa table for this temperature
;
kabs  = findkappa(opac,tdust,/abs)
ksca  = findkappa(opac,tdust,/scat)
;
; Now make the average
;
nf=n_elements(nu)
aa=0.d0
bb=0.d0
for i=1,nf-2 do begin
   dnu = 0.5 * nu[i] * (alog(nu[i+1])-alog(nu[i-1]))
   aa  = aa + bnu(i)*ksca[i]*dnu
   bb  = bb + bnu(i)*kabs[i]*dnu
endfor
albedo = aa/(bb+aa)
;
; Find the corresponding kappa
;
return,albedo
;
end
