;========================================================================
;           PROBLEM GENERATOR MODULE FOR THE RADIALRT PROGRAM
;
; This module is meant for creating the setup for the 1+1-D radiative 
; transfer problem.
;========================================================================

;------------------------------------------------------------------------
;            HELPER ROUTINE FOR MAIN PROBLEM GENERATOR
;------------------------------------------------------------------------
function makedens,r,hp,sigma,nz,coverz
nr=n_elements(r)
rho=fltarr(nr,nz)
z=rho
for ir=0,nr-1 do begin
   z(ir,*)   = coverz*hp[ir]*findgen(nz)/(nz-1.d0)
   rho(ir,*) = 0.39894228d0 * sigma[ir] * $
              exp(-0.5d0*(z[ir,*]/hp[ir])^2) / hp[ir]
endfor
return,{rho:rho,z:z}
end


;------------------------------------------------------------------------
;                  PROBLEM MAIN FUNCTION
;
; This function first calls the cgdisk() function, and from that creates
; the 2-D density distribution.
;------------------------------------------------------------------------
function problem,nz,coverz,nlevr=nlevr,nspanr=nspanr,nstepr=nstepr,$
                 hrlg=hrlg,cg=cg,gastodust=gastodust,_extra=_extra
;
; Gas-to-dust ratio
;
if not keyword_set(gastodust) then gastodust=100.d0
;
; First create the CG97 disk 
;
if not keyword_set(cg) then begin
   cg=cgdisk(_extra=_extra)
endif
;
; Get some info
;
sigma=cg.sigma
hp   =cg.hp
;
; R-grid
;
r=cg.r
nr=n_elements(r)
;
; R-grid refinement
;
if keyword_set(nlevr) then begin
   if not keyword_set(nspanr) then nspanr=1
   if not keyword_set(nstepr) then nstepr=1
   nextra = nspanr*nlevr*(2^nstepr-1)
   nr = nr + nextra
   for ilev=1,nlevr do begin
      for ispan=nspanr,1,-1 do begin
         rins=fltarr(2^nstepr-1)
         fr=(r(ispan)/r(ispan-1))^(0.5^nstepr)
         for i=1,2^nstepr-1 do rins(i-1)=r(ispan-1)*fr^i
         r=[r(0:ispan-1),rins,r(ispan:*)]
      endfor
   endfor
   if nextra gt 0 then begin
      hp   =[hp[0],dblarr(nextra),hp[1:*]]
      sigma=[sigma[0],dblarr(nextra),sigma[1:*]]
      for i=1,nextra do begin
         x = (r[i]-cg.r[0])/(cg.r[1]-cg.r[0])
         hp[i]=(1-x)*cg.hp[0]+x*cg.hp[1]
         sigma[i]=(1-x)*cg.sigma[0]+x*cg.sigma[1]
      endfor
   endif
endif
;
; The density distribution
;
a=makedens(r,hp,sigma,nz,coverz)
rho=gastodust*a.rho
sigma=gastodust*sigma
z=a.z
;
rr=rebin(r,nr,nz)
;
return,{r:r,z:z,rr:rr,rho:rho,hp:hp,sigma:sigma}
end




;------------------------------------------------------------------------
;                     MAKE OPTICAL DEPTHS
;------------------------------------------------------------------------
function maketau,a,kappa
sz=size(a.rho)
nr=sz(1)
nz=sz(2)
taur=fltarr(nr,nz)
tauz=fltarr(nr,nz)
for iz=0,nz-1 do begin
    taur(0,iz) = 0.d0
    for ir=1,nr-1 do begin
        taur(ir,iz) = taur(ir-1,iz) + $
            0.5*(a.rho(ir,iz)+a.rho(ir-1,iz))*kappa * $
            (a.r(ir)-a.r(ir-1))
    endfor
 endfor
for ir=0,nr-1 do begin
    tauz(ir,0) = 0.d0
    for iz=nz-1,1,-1 do begin
        tauz(ir,iz-1) = tauz(ir,iz) + $
            0.5*(a.rho(ir,iz)+a.rho(ir,iz-1))*kappa * $
            (a.z(ir,iz)-a.z(ir,iz-1))
    endfor
endfor
return,{taur:taur,tauz:tauz}
end



;------------------------------------------------------------------------
;                 WRITE PROBLEM FOR RADICAL 
;------------------------------------------------------------------------
pro write_problem_rz,a,iir=iir
sz=size(a.z)
nr=sz(1)
nz=sz(2)
if n_elements(iir) eq 0 then begin
   openw,1,'diskstruct.inp'
   printf,1,nr,nz
   printf,1,' '
   for ir=1,nr do begin
      for iz=1,nz do begin
         printf,1,a.r[ir-1],a.z[ir-1,iz-1],a.rho[ir-1,iz-1],0.d0,0.d0,$
           format='(5(E13.6,1X))'
      endfor
   endfor
   close,1
endif else begin
   openw,1,'diskstruct.inp'
   printf,1,1,nz
   printf,1,' '
   ir=iir
   for iz=1,nz do begin
      printf,1,a.r[ir-1],a.z[ir-1,iz-1],a.rho[ir-1,iz-1],0.d0,0.d0,$
        format='(5(E13.6,1X))'
   endfor
   close,1
endelse
end






;-----------------------------------------------------------------
;                 FIND A FREQUENCY IN A TABLE
;-----------------------------------------------------------------
function find_freq,freq,nu
nf=n_elements(freq)
idx=-100
for inu=0,nf-2 do begin
   if (nu ge freq(inu) and nu le freq(inu+1)) or $
     (nu ge freq(inu+1) and nu le freq(inu)) then begin
      idx = inu
   endif
endfor
if idx lt 0 then begin
   print,"ERROR: nu out of range"
   stop
endif
return,idx
end
