c     ==============================================================
c              PROGRAM THAT PRODUCES A SIMPLE FACE-ON SED
c                    (C) 2001  by  C.P. Dullemond
c
c     This program reads a density and temperature structure 
c     rho(r,z), temp(r,z) of a dust disk, the specifications of the 
c     central star, and the dust opacities, and computes the SED
c     (face-on) of the disk. The scattered radiation is for the 
c     moment ignored. Also the inner rim is not included.
c     ==============================================================

#include "main.h"

#define GASTODUST (100.d0)

c     --------------------------------------------------------------
c                             MAIN ROUTINE
c     --------------------------------------------------------------
      program main
      implicit none
      integer nr
      doubleprecision flidx
      logical fex
c
#include "common_sed.h"
c
c     Default
c     
      glob_itrace=0
      glob_inclrim=0
c
c     First check if the inner rim should be included. This is the
c     case when the first index of the flaring index is negative.
c     Or if the user wants to add it as a blackbody by hand (see 
c     read_all_data() below).
c
      inquire(file='flareindex.inp',exist=fex)
      if(fex) then 
          open(unit=1,file='flareindex.inp',status='old')
          read(1,*) nr
          read(1,*) flidx
          close(1)
          if(flidx.lt.-1.d1) then
              glob_inclrim=1
              write(*,*) 'Including inner rim by special treatment'
              write(*,*) '     of the inner annulus (turning it 90'
              write(*,*) '     degrees.'
          endif
      endif
c
c     Read the data and set things up
c
      call read_all_data()
      call make_star_spectrum()
c
c     Message
c
      if(glob_inclrim.lt.0) then
         write(*,*) 'Including inner rim as a blackbody.'
      endif
c
c     Now do the stuff
c
      if(glob_command.eq.0) then
c
c         Make the simple SED
c
          call do_raytracing(3.08572d18,glob_inclination,
     %                       glob_inclrim,glob_itrace)
          call write_sed()
          call write_image()
          call write_tauimage()
      elseif(glob_command.eq.-1) then
c
c         (Debugging:) Check the energy conservation
c     
          write(*,*) 'WARNING: SED IN DEBUGGING MODE!!!!'
          call do_raytracing(3.08572d18,glob_inclination,0,glob_itrace)
          call write_sed()
          call compute_flux_error()
          call write_flux_error()
      elseif(glob_command.eq.-2) then
c
c         (Debugging:) As "-1", but now make fluxout from Bnu(T=T_{z=0})
c     
          write(*,*) 'WARNING: SED IN DEBUGGING MODE!!!!'
          call make_trivial_sed(3.08572d18,glob_inclination)
          call write_sed()
          call compute_flux_error()
          call write_flux_error()
      else
          write(*,*) 'Dont know command ',glob_command
          stop 13
      endif
c
c     If itrace gt 0 then write the intensity 
c
      if(glob_itrace.gt.0) then
          call write_trace(glob_itrace)
      endif
c
      end




c     --------------------------------------------------------------
c                          READ ALL THE DATA
c     --------------------------------------------------------------
      subroutine read_all_data()
      implicit none
c
#include "common_grid.h"
#include "common_sed.h"
#include "common_dust.h"
c
      integer iformat
      integer nr,nz,ir,iz,nf,inu,nrd,nnr,isize,ispec
      integer nrspec,nrsize(DUST_SPECIES_MAX)
      doubleprecision a,b,c,d,e
      logical f_ex
      logical fexf,fexo,multi
c
c     Look for sed.inp, if present
c
      inquire(file='sed.inp',exist=fexf)
      if(fexf) then
          open(unit=1,file='sed.inp',status='old')
          read(1,*) glob_command
          read(1,*) glob_inclstar
          read(1,*) glob_inclination
          write(*,*) 'SED at inclination ',glob_inclination,
     %              ' degrees'
          glob_inclination = glob_inclination * 0.01745329252d0
          read(1,*,end=55) glob_itrace
          if(glob_inclrim.eq.0) then
             read(1,*,end=55) glob_inclrim
             if(glob_inclrim.ne.0) then
                glob_inclrim=-abs(glob_inclrim)
             endif
          endif
 55       continue
          close(1)
      else
          glob_command=0          ! Make spectrum
          glob_inclstar=1         ! Include the stellar spectrum
          glob_inclination=0.d0   ! Inclination is 0
      endif
c
c     Read the star information
c
      open(unit=1,file='starinfo.inp',status='old')
      read(1,*) iformat
      read(1,*) star_r
      read(1,*) star_m
      read(1,*) star_t
      close(1)
c     
c     Read the density and temperature structure from the
c     second stage of radiative transfer
c
      open(unit=1,file='diskstruct_2.dat',status='old')
      read(1,*) nr,nz
      nrspec=1
      nrsize(1)=1
      if(nr.lt.1) then
c
c         New format: multiple sizes/species
c
          read(1,*) nr,nz
          read(1,*) nrspec
          if(nrspec.gt.DUST_SPECIES_MAX) then
              write(*,*) 'ERROR: Too many species specified.'
              write(*,*) '       Recompile with larger max specs.'
              stop
          endif
          do ispec=1,nrspec
              read(1,*) nrsize(ispec)
              if(nrsize(ispec).gt.DUST_SIZE_MAX) then
                  write(*,*) 'ERROR: Too many sizes specified.'
                  write(*,*) '       Recompile with larger max size.'
                  stop
              endif
          enddo
          multi=.true.
      else
          nrspec=1
          nrsize(1)=1
          multi=.false.
      endif
      if(nr.gt.FRSIZE_R) then
          write(*,*) 'ERROR in sed:'
          write(*,*) 'R grid too large'
          stop 13
      endif
      if(nz.gt.FRSIZE_Z) then
          write(*,*) 'ERROR in sed:'
          write(*,*) 'Z grid too large'
          stop 13
      endif
      do ir=1,nr
          do iz=1,nz
              if(.not.multi) then
                  read(1,*) a,b,c,d,e
                  grid_r(ir) = a
                  grid_z(iz,ir) = b
                  rrt_rhogas(iz,ir) = c
                  rrt_rho(1,iz,ir) = c / GASTODUST
                  rrt_rhotot(iz,ir) = c / GASTODUST
                  rrt_temp(1,1,iz,ir) = d
                  rrt_vz(iz,ir) = e
              else
                  read(1,*) a,b,c,e
                  grid_r(ir) = a
                  grid_z(iz,ir) = b
                  rrt_rhogas(iz,ir) = c
                  rrt_vz(iz,ir) = e
                  rrt_rhotot(iz,ir) = 0.d0
                  do ispec=1,nrspec
                      read(1,*) a
                      rrt_rho(ispec,iz,ir) = a * c / GASTODUST
                      rrt_rhotot(iz,ir) = rrt_rhotot(iz,ir) + 
     %                   rrt_rho(ispec,iz,ir)
                  enddo
                  do ispec=1,nrspec
                      do isize=1,nrsize(ispec)
                          read(1,*) d
                          rrt_temp(isize,ispec,iz,ir) = d
                      enddo
                  enddo
              endif
          enddo
      enddo
      close(1)
c
c     Find the frequency-dependent opacities, and do the radial
c     transfer with these opacitities (do it with frequency
c     dependence).
c
      inquire(file='frequency.inp',exist=fexf)
      inquire(file='dustopac.inp',exist=fexo)
      if((.not.fexf).or.(.not.fexo)) then
          write(*,*) 'Need both frequency.inp and dustopac.inp'
          stop 13
      endif
      open(unit=1,file='frequency.inp',status='old')
      read(1,*) nf
      if(nf.gt.FRSIZE_FREQ) then
          write(*,*) 'ERROR in sed:'
          write(*,*) 'Frequency grid too large'
          stop 13
      endif
      do inu=1,nf
          read(1,*) a
          freq_nu(inu) = a
      enddo
      freq_nr = nf
      call read_dustdata()
c
c     Check if nr of species and sizes consistent
c
      if(nrspec.ne.dust_nr_species) then
          write(*,*) 'ERROR: Nr of dust species in dustopac*.inp'
          write(*,*) '   is unequal to the number specified in '
          write(*,*) '   diskstruct_1.dat'
          stop
      endif
      do ispec=1,dust_nr_species
          if(nrsize(ispec).ne.dust_nr_size(ispec)) then
              write(*,*) 'ERROR: Nr of dust sizes in dustopac*.inp'
              write(*,*) '   is unequal to the number specified in '
              write(*,*) '   diskstruct_1.dat'
              stop
          endif
      enddo
c
c     If the glob_inclrim is set, then the inner rim must be included
c     (in an approximative way). Need to read the 'radmonte.dat' file 
c     to obtain the surface height as a function of radius (to determine
c     the extent of the shadow).
c
      if(glob_inclrim.ne.0) then
          inquire(file='radmonte.dat',exist=f_ex)
          if(.not.f_ex) then
              write(*,*) 'PROBLEM in sed.F: Need file radmonte.dat'
              stop 13
          endif
          open(unit=1,file='radmonte.dat',status='old')
          read(1,*) nnr
          if(nnr.ne.nr) then
              write(*,*) 'PROBLEM in sed.F: radmonte.dat nr wrong.'
              stop 13
          endif
          do ir=1,nr
              read(1,*) a,b,d
              rrt_height(ir)=b
          enddo
          close(1)
      endif
c
c     If only one radial point (for debugging this is handy!) then 
c     we must read the extrar.inp file)
c
      if(nr.eq.1) then
          inquire(file='extrar.inp',exist=f_ex)
          if(.not.f_ex) then
              write(*,*) '  When only one grid point in R, then'
              write(*,*) '  I need rinout.inp to find inner and '
              write(*,*) '  outer radius.'
              stop 13
          endif
          open(unit=1,file='extrar.inp',status='old')
          read(1,*) glob_extrarin
          read(1,*) glob_extrarout
          close(1)
      endif
c
      grid_nr = nr
      grid_nz = nz
c
      end




c     --------------------------------------------------------------
c                       MAKE THE STARSPECTRUM
c     --------------------------------------------------------------
      subroutine make_star_spectrum()
      implicit none
c
#     include "common_grid.h"
#     include "common_sed.h"
c
      doubleprecision bplanck,pi,ss
      parameter(pi  = 3.1415926535897932385d0)
      parameter(ss  = 5.6703d-5)     ! Stefan-Boltzmann const  [erg/cm^2/K^4/s]      
      integer inu
c
      if(freq_nr.eq.1) then
          star_lum(1) = 4 * pi * star_r**2 * ss * star_t**4
      else
          do inu=1,freq_nr
              star_lum(inu) = 4 * pi * star_r**2 * pi *
     %             bplanck(star_t,freq_nu(inu)) 
          enddo
      endif
c
      end




c     --------------------------------------------------------------
c                THE BLACKBODY PLANCK FUNCTION B_nu(T)
c
c     This function computes the Blackbody function 
c
c                    2 h nu^3 / c^2
c        B_nu(T)  = -------------------    [ erg / cm^2 s ster Hz ]
c                   exp(-h nu / kT) - 1
c
c     ARGUMENTS:
c        nu    [Hz]            = Frequency
c        temp  [K]             = Temperature
c     --------------------------------------------------------------
      function bplanck(temp,nu)
      implicit none
      doubleprecision temp
      doubleprecision nu
      doubleprecision bplanck
c
      if(temp.eq.0.d0) then 
          bplanck = 0.d0
          return
      endif
c
      bplanck = 1.47455d-47 * nu * nu * nu / 
     %      (exp(4.7989d-11 * nu / temp)-1.d0) + 1.d-290
c
      return
      end



c     --------------------------------------------------------------
c                     DO THE VERTICAL RAY-TRACING
c
c     This integrates the formal transfer equation vertically 
c     through the disk. It ignores scattering!
c
c     NOTE: To approximately include the inner rim, we treat the
c           radius ir.eq.1 separately IF the flag "inrim" is 1.
c           If it is -1, then use a blackbody for this.
c     --------------------------------------------------------------
      subroutine do_raytracing(dist,incl,inrim,itrace)
      implicit none
      doubleprecision dist,incl
      integer inrim,itrace
c
#include "common_grid.h"
#include "common_sed.h"
#include "common_dust.h"
c
      integer inu,ir,iz,ir0
      doubleprecision find_dust_kappa,tempdum,cosi,sini,int0,alpha
      doubleprecision pi,distsq,nu,bplanck,intensity,qdr_src_2
      doubleprecision alp1,alp2,js1,js2,ds,dtauabs,tautot,incfact
      doubleprecision temprim
      doubleprecision kapabs(DUST_SIZE_MAX,DUST_SPECIES_MAX)
      parameter(pi  = 3.1415926535897932385d0)
      integer ispec,isize
c     
      if(dust_opacity_tempdep.ne.0) then
          write(*,*) 'Temperature dependent opacities not allowed!'
          stop 13
      endif
      tempdum = 100.   
      distsq  = dist**2
      cosi    = cos(incl)
      sini    = sin(incl)
c
c     Do the loop over the frequencies
c
      do inu=1,freq_nr
c
c         Find the opacity at this frequency
c
          do ispec=1,dust_nr_species
              do isize=1,dust_nr_size(ispec)
                  kapabs(isize,ispec) = 
     %              find_dust_kappa(inu,isize,ispec,tempdum,1,0)
              enddo
          enddo
          sed_fnu(inu) = 0.d0
          nu           = freq_nu(inu)
c
c         Do a loop over ir
c
          do ir=1,grid_nr
c
c             Find the inclination factor
c
              incfact = 1.d0 / cosi
              if((ir.eq.1).and.(inrim.gt.0)) then
                  incfact = 1.d0/(abs(sini)+1d-10)
              endif
c
c             Do the vertical tracing integral
c
              intensity = 0.d0
              tautot    = 0.d0
              do iz=grid_nz-1,1,-1
                  alp1 = 0.d0
                  alp2 = 0.d0
                  js1  = 0.d0
                  js2  = 0.d0
                  do ispec=1,dust_nr_species
                      do isize=1,dust_nr_size(ispec)
                          alpha = rrt_rho(ispec,iz+1,ir)*
     %                            kapabs(isize,ispec)
                          alp1  = alp1 + alpha
                          js1   = js1  + alpha*bplanck(
     %                            rrt_temp(isize,ispec,iz+1,ir),nu)
                          alpha = rrt_rho(ispec,iz,ir)*
     $                            kapabs(isize,ispec)
                          alp2  = alp2  + alpha
                          js2   = js2   + alpha*bplanck(
     %                            rrt_temp(isize,ispec,iz,ir),nu)
                      enddo
                  enddo
                  ds        = (grid_z(iz+1,ir)-grid_z(iz,ir))*incfact
                  dtauabs   = 0.5 * ds * (alp1+alp2)
                  tautot    = tautot+dtauabs
                  int0      = intensity
                  intensity = qdr_src_2(int0,
     %                        js1,alp1,js2,alp2,ds)
                  if(itrace.eq.ir) then
                      trace_intensm(inu,iz)=intensity
                      trace_tautotm(inu,iz)=tautot
                      trace_sourcef(inu,iz)=js2/(alp2+1d-99)
                  endif
              enddo
              do iz=2,grid_nz
                  alp1 = 0.d0
                  alp2 = 0.d0
                  js1  = 0.d0
                  js2  = 0.d0
                  do ispec=1,dust_nr_species
                      do isize=1,dust_nr_size(ispec)
                          alpha = rrt_rho(ispec,iz-1,ir)*
     %                            kapabs(isize,ispec)
                          alp1  = alp1 + alpha 
                          js1   = js1  + alpha*bplanck(
     %                            rrt_temp(isize,ispec,iz-1,ir),nu)
                          alpha = rrt_rho(ispec,iz,ir)*
     %                            kapabs(isize,ispec)
                          alp2  = alp2 + alpha
                          js2   = js2  + alpha*bplanck(
     %                            rrt_temp(isize,ispec,iz,ir),nu)
                      enddo
                  enddo
                  ds        = (grid_z(iz,ir)-grid_z(iz-1,ir))*incfact
                  dtauabs   = 0.5 * ds * (alp1+alp2)
                  tautot    = tautot+dtauabs
                  int0      = intensity
                  intensity = qdr_src_2(int0,
     %                        js1,alp1,js2,alp2,ds)
                  if(itrace.eq.ir) then
                      trace_intensp(inu,iz)=intensity
                      trace_tautotp(inu,iz)=tautot
                      trace_sourcef(inu,iz)=js2/(alp2+1d-99)
                  endif
              enddo
              sed_intensity(ir,inu) = intensity
              sed_tauim(ir,inu)     = tautot
          enddo
c     
c         Integrate this image to obtain a flux at distance 'dist'
c
          if(grid_nr.gt.1) then
c
c             If no inner rim annulus, then start integrating from ir=1
c     
              ir0 = 1
c
c             If the inner rim is included, then compute the
c             extent of the shadow. Always start at ir=3, since
c             otherwise the ir=1 point is included in the integration
c             of the intensity...
c
              if(inrim.ne.0) then
                  ir0 = 3
                  do ir=3,grid_nr-2
                      if(rrt_height(ir)/grid_r(ir).lt.
     %                   rrt_height(1)/grid_r(1)) then
                          ir0=ir+1
                      endif
                  enddo
              endif
c
c             Now do the integration of the flaring disk, at least
c             that part of the disk that is not in the shadow of 
c             the inner rim (if this inner rim is included).
c
              do ir=ir0,grid_nr-1
                  sed_fnu(inu) = sed_fnu(inu) + 0.5 * cosi * 
     %                 ( sed_intensity(ir,inu) + 
     %                   sed_intensity(ir+1,inu) )
     %                   * pi * ( grid_r(ir+1)**2 - 
     %                   grid_r(ir)**2 ) / distsq
              enddo
c
c             Add the contribution of the inner rim, if needed
c
              if(inrim.gt.0) then
c
c                Inner rim treated using first annulus, tilted 90 deg.
c
                 sed_fnu(inu) = sed_fnu(inu) + 4.0 * sini * 
     %                rrt_height(1) * grid_r(1) * 
     %                sed_intensity(1,inu) / distsq
              elseif(inrim.lt.0) then
c
c                Inner rim treated by simply adding a blackbody
c
                 temprim = star_t * sqrt(star_r/grid_r(1))
                 sed_fnu(inu) = sed_fnu(inu) + 4.0 * sini * 
     %                rrt_height(1) * grid_r(1) * 
     %                bplanck(temprim,freq_nu(inu)) / distsq
                 
              endif
          else
              sed_fnu(inu) = sed_fnu(inu) + cosi * 
     %             sed_intensity(1,inu) *
     %             pi * abs( glob_extrarout**2 - 
     %             glob_extrarin**2 ) / distsq
          endif
c
c         If required, add the star
c
          if(glob_inclstar.eq.1) then
              sed_fnu(inu) = sed_fnu(inu) + 
     %          star_lum(inu) / ( 4. * pi * distsq )
          endif
      enddo
c
      end


c     --------------------------------------------------------------
c      INTEGRATE THE FORMAL SOLUTION ALONG ONE ELEMENT OF SHORT CHAR
c
c     This function computes the formal integral over a segment of
c     a local characteristic. This will be done first order here.
c
c     This linear integration is not always suitable for extremely 
c     optically thick systems. In that case the 3-point quadratic 
c     integrator, qdr_src_3() can better be used. 
c     --------------------------------------------------------------
      function qdr_src_2(int,js1,alp1,js2,alp2,ds)
      implicit none
      doubleprecision qdr_src_2
      doubleprecision int,js1,alp1,js2,alp2,ds
c
      doubleprecision e0,e1,a,b,dtau1,xp,src1,src2,theomax
c
      dtau1     = 0.5d0 * ( alp1 + alp2 ) * ds
      theomax   = 0.5d0 * ( js1 + js2 ) * ds
      if(dtau1.gt.1.d-6) then
          xp        = exp(-dtau1)
          e0        = 1.d0 - xp
          e1        = dtau1 - e0
          b         = e1 / dtau1
          a         = e0 - b
      else
          a         = 0.5d0 * dtau1
          b         = 0.5d0 * dtau1
          xp        = 1.d0 - dtau1 
      endif
      if(alp1.gt.0.d0) then
          src1 = js1 / alp1
      elseif(alp2.gt.0.d0) then
          src1 = js2 / alp2
      else
          src1 = 0.d0
      endif
      if(alp2.gt.0.d0) then
          src2 = js2 / alp2
      elseif(alp1.gt.0.d0) then
          src2 = js1 / alp1
      else
          src2 = 0.d0
      endif
      if(dtau1.gt.1e-9) then
         qdr_src_2 = a * src1 + b * src2
      else
         qdr_src_2 = 0.5d0 * ( js1 + js2 ) * ds
      endif
c
c     Bracket the total integrated emission between the
c     theoretical maximum and 0.
c
#     ifndef SRCQDR_NO_BRACKET
      qdr_src_2 = min(qdr_src_2,theomax)
#     endif
c
      qdr_src_2 = int * xp + qdr_src_2
      return
c
      write(*,*) 'Ehh, something wrong in dqr_src_2'
      stop 13
      return
      end



c     --------------------------------------------------------------
c                    MAKE TRIVIAL SED: TAKE T=T(Z=0)
c     --------------------------------------------------------------
      subroutine make_trivial_sed(dist,incl)
      implicit none
      doubleprecision dist,incl
c
#include "common_grid.h"
#include "common_sed.h"
#include "common_dust.h"
c
      integer inu,ir,iz
      doubleprecision tempdum,cosi
      doubleprecision pi,distsq,nu,bplanck,intensity
      parameter(pi  = 3.1415926535897932385d0)
c     
      if(dust_opacity_tempdep.ne.0) then
          write(*,*) 'Temperature dependent opacities not allowed!'
          stop 13
      endif
      tempdum = 100.   
      distsq  = dist**2
      cosi    = cos(incl)
c
c     Do the loop over the frequencies
c
      do inu=1,freq_nr
c
c         Find the opacity at this frequency, and adapt this for the
c         inclination
c
          sed_fnu(inu) = 0.d0
          nu           = freq_nu(inu)
c
c         Do a loop over ir
c
          do ir=1,grid_nr
              sed_intensity(ir,inu) = bplanck(rrt_temp(1,1,1,ir),nu)
          enddo
c     
c         Integrate this image to obtain a flux at distance 'dist'
c
          if(grid_nr.gt.1) then
              do ir=1,grid_nr-1
                  sed_fnu(inu) = sed_fnu(inu) + 0.5 * cosi * 
     %                 ( sed_intensity(ir,inu) + 
     %                   sed_intensity(ir+1,inu) )
     %                   * pi * ( grid_r(ir+1)**2 - 
     %                   grid_r(ir)**2 ) / distsq
              enddo
          else
              sed_fnu(inu) = sed_fnu(inu) + cosi * 
     %             sed_intensity(1,inu) *
     %             pi * abs( glob_extrarout**2 - 
     %             glob_extrarin**2 ) / distsq
          endif
c
c         If required, add the star
c
          if(glob_inclstar.eq.1) then
              sed_fnu(inu) = sed_fnu(inu) + 
     %          star_lum(inu) / ( 4. * pi * distsq )
          endif
      enddo
c
      end



c     --------------------------------------------------------------
c                   COMPUTE THE LOCAL FLUX ERROR
c
c     Since the flaring angle recipe gives us precisely the flux
c     that enters the disk, we can compute for each radius what
c     is the error in the flux that goes out. Note, however, that
c     the image contains the intensities, not the fluxes
c     --------------------------------------------------------------
      subroutine compute_flux_error()
      implicit none
c
#include "common_grid.h"
#include "common_sed.h"
#include "common_dust.h"
c
      doubleprecision flux,pi
      parameter(pi  = 3.1415926535897932385d0)
      integer inu,ir,nnr
      logical fexi,fexa
c
c     Compute the surface height
c
      call find_surface_height()
c
c     Read the flaring angle or flaring index
c
      inquire(file='flareindex.inp',exist=fexi)
      inquire(file='flareangle.inp',exist=fexa)
      if(fexi.and.fexa) then
          write(*,*) 'Found both the flaring index and ',
     %                'the flaring angle files'
          write(*,*) 'Using flaring index...'
          fexa=.false.
      endif
      if(fexi) then
c
c         Read the flareing index
c     
          open(unit=1,file='flareindex.inp',status='old')
          read(1,*) nnr
          do ir=1,grid_nr
              read(1,*) rrt_flareindex(ir)
          enddo
          close(1)
c
c         Convert this flaring index to a flaring angle
c
          do ir=1,grid_nr
              rrt_flareangle(ir) = 0.4d0 * star_r / grid_r(ir) +
     %                 rrt_flareindex(ir) * 
     %                 rrt_height(ir) / grid_r(ir)
          enddo
      elseif(fexa) then
c
c         Read the flaring angle directly
c      
          open(unit=1,file='flareangle.inp',status='old')
          read(1,*) nnr
          do ir=1,grid_nr
              read(1,*) rrt_flareangle(ir)
          enddo
          close(1)
      else
          write(*,*) 'Flaring index nor flaring angle file found...'
          stop 13
      endif
c
c     Compute total stellar flux
c     
      star_ltot = 0.d0
      do inu=2,freq_nr
          star_ltot = star_ltot + 0.5 * 
     %         ( star_lum(inu) + star_lum(inu-1) ) 
     %         * abs( freq_nu(inu) - freq_nu(inu-1) )
      enddo
c
c     Now compute the flux that goes in
c
      do ir=1,grid_nr
          flx_fluxin(ir) = star_ltot * rrt_flareangle(ir) 
     %                      / ( 4*pi*grid_r(ir)**2 )
      enddo
c
c     Now the flux that goes out
c
      do ir=1,grid_nr
          flux = 0.d0
          do inu=2,freq_nr
              flux = flux + pi * ( sed_intensity(ir,inu) +
     %                sed_intensity(ir,inu-1) ) 
     %              * abs( freq_nu(inu) - freq_nu(inu-1) )
          enddo
c          flux = flux / sqrt(3.0)
          flx_fluxout(ir) = flux * 0.577350269189d0
      enddo
c
      end



c     --------------------------------------------------------------
c                       FIND THE SURFACE HEIGHT
c     --------------------------------------------------------------
      subroutine find_surface_height()
      implicit none
c
#include "common_grid.h"
#include "common_sed.h"
#include "common_dust.h"
c
      doubleprecision tmax,ttau,ztau,dztau
      integer ir,iz,m
c
c     Find the tau=1 surface: tau=1 means T=Tmax*0.7788
c     
      m=4
      do ir=1,grid_nr
          tmax = rrt_temp(1,1,grid_nz,ir)
          ttau = tmax*0.7788
          call hunt(rrt_temp(1,1,1,ir),grid_nz,ttau,iz)
          if((iz.eq.0).or.(iz.eq.grid_nz)) then
              write(*,*) 'SORRY: Could not find the tau=1 surface'
              write(*,*) '       at radius ir=',ir
              stop 13
          endif
          iz = min(max(iz-(m-1)/2,1),grid_nz+1-m)
          call polint(rrt_temp(1,1,iz,ir),grid_z(iz,ir),m,ttau,
     %                  ztau,dztau)
          rrt_height(ir) = ztau
      enddo
c
      end



c     --------------------------------------------------------------
c                      WRITE THE SED TO FILE
c     --------------------------------------------------------------
      subroutine write_sed()
      implicit none
c
#include "common_grid.h"
#include "common_sed.h"
#include "common_dust.h"
c
      integer inu
c
      open(unit=1,file='spectrum.dat',status='unknown')
      write(1,*) freq_nr
      write(1,*)
      do inu=1,freq_nr
          write(1,10) freq_nu(inu),sed_fnu(inu)+1d-98
 10       format(2(E13.6))
      enddo
      close(1)
      end


c     --------------------------------------------------------------
c                      WRITE THE IMAGE TO FILE
c     --------------------------------------------------------------
      subroutine write_image()
      implicit none
c
#include "common_grid.h"
#include "common_sed.h"
#include "common_dust.h"
c
      integer inu,ir
c
      open(unit=1,file='image.dat',status='unknown')
      write(1,*) freq_nr,grid_nr
      write(1,*)
      do inu=1,freq_nr
          write(1,10) freq_nu(inu)
      enddo
      write(1,*)
      do ir=1,grid_nr
          write(1,10) grid_r(ir)
      enddo
      write(1,*)
      do inu=1,freq_nr
          do ir=1,grid_nr
              write(1,10) max(sed_intensity(ir,inu),1d-99)
          enddo
          write(1,*)
      enddo
 10   format(2(E13.6))
      close(1)
c
      end


c     --------------------------------------------------------------
c                      WRITE THE TAU-IMAGE TO FILE
c     --------------------------------------------------------------
      subroutine write_tauimage()
      implicit none
c
#include "common_grid.h"
#include "common_sed.h"
#include "common_dust.h"
c
      integer inu,ir
c
      open(unit=1,file='tauimage.dat',status='unknown')
      write(1,*) freq_nr,grid_nr
      write(1,*)
      do inu=1,freq_nr
          write(1,10) freq_nu(inu)
      enddo
      write(1,*)
      do ir=1,grid_nr
          write(1,10) grid_r(ir)
      enddo
      write(1,*)
      do inu=1,freq_nr
          do ir=1,grid_nr
              write(1,10) max(sed_tauim(ir,inu),1d-99)
          enddo
          write(1,*)
      enddo
 10   format(2(E13.6))
      close(1)
c
      end


c     --------------------------------------------------------------
c                     WRITE THE FLUX INPUT AND OUTPUT
c     --------------------------------------------------------------
      subroutine write_flux_error()
      implicit none
c
#include "common_grid.h"
#include "common_sed.h"
#include "common_dust.h"
c
      integer ir
c
      open(unit=1,file='fluxerror.dat',status='unknown')
      write(1,*) grid_nr
      write(1,*)
      write(1,*) star_ltot
      write(1,*)
      do ir=1,grid_nr
          write(1,10) grid_r(ir),rrt_flareangle(ir),
     %         rrt_height(ir),flx_fluxin(ir),flx_fluxout(ir)
      enddo
 10   format(5(E13.6))
      close(1)
      
c
      end


c     --------------------------------------------------------------
c                      WRITE THE TRACE ARRAY
c     --------------------------------------------------------------
      subroutine write_trace(ir)
      implicit none
      integer ir
c
#include "common_sed.h"
#include "common_grid.h"
c
      integer inu,iz
c
      write(*,*) '!!!!!!!!!!!!! WRITING TRACE !!!!!!!!!!!!!!!'
      write(*,*) '!!!!!!! TAKES LOTS OF DISK SPACE !!!!!!!!!!'
c
      open(unit=1,file='trace.dat',status='unknown')
      write(1,*) ir,grid_nz,freq_nr
      write(1,*)
      write(1,*) grid_r(ir)
      write(1,*)
      do inu=1,freq_nr
          write(1,*) freq_nu(inu)
      enddo
      write(1,*)
      do iz=1,grid_nz
          write(1,*) grid_z(iz,ir)
      enddo
      write(1,*)
      do iz=1,grid_nz
          do inu=1,freq_nr
              write(1,10) trace_intensm(inu,iz)+1d-99,
     %                    trace_intensp(inu,iz)+1d-99,
     %                    trace_tautotm(inu,iz)+1d-99,
     %                    trace_tautotp(inu,iz)+1d-99,
     %                    trace_sourcef(inu,iz)+1d-99
 10           format(5(1X,E13.6))
          enddo
      enddo
      close(1)
c
      end
