c     ==============================================================
c                       PROTOSTELLAR DISK MODEL
c              MODULE: MEAN OPACITIES FROM FULL OPACITIES
c
c                  FULLY REVISED EDITION : 23-04-00
c                         MODIFIED A BIT : 02-08-00
c                     MAJOR BUGS REMOVED : 13-08-00
c                         MINOR ADDITION : 10-01-01
c
c     The main opacity routines are given in the first section:
c
c       rosseland_mean_dust_opacity()
c       planck_mean_dust_kappa()
c       planck_integral_dust_kappa()
c       spectrum_mean_dust_kappa()         [added 10-01-01]
c
c     Since these routines are not always very quick, there are
c     ways to store the planck mean and planck integrated opacities
c     per species,size in a database. Under special conditions also
c     the Rosseland mean can be stored, but under generic conditions
c     this is impossible. The database routines are:
c
c       create_mean_opacity_arrays()
c       lgkapbnuint()
c       planckmeankappa()
c       totalrosselandmeanalpha()
c
c     These routines will automatically call the above ones, if the
c     database either does not exist, or is incompatible with the
c     input.
c     
c     Then there is a routine for solving the dust temperatures 
c     from the mean intensity:
c
c       solvedtemp()
c
c     And finally there is a function that allows you to test
c     whether the database storage for the rosseland mean is 
c     possible:
c
c       is_global_rosseland_possible()
c
c     ==============================================================
#include "main.h"

c     --------------------------------------------------------------
c                    THE ROSSELAND-MEAN OPACITY
c
c     This routine returns the Rosseland mean opacity
c
c         1        /    1     dB_nu(T)       / / dB_nu(T)            
c      -------  =  | -------- -------- dnu  /  | -------- dnu 
c      alpha_R     / alpha_nu    dT        /   /    dT     
c
c     for all species and sizes together, at a specified temperature 
c     (radtemp), and either for pure absorption (iabs=1), pure scattering 
c     (iscat=1) or both. The dust temperatures are still necessary to
c     be specified, because of the possible temperature-dependence 
c     of the dust opacities.
c     --------------------------------------------------------------
      function rosseland_mean_dust_opacity(dustrho,dusttemp,
     %                                     radtemp,iabs,iscat)
      implicit none
      doubleprecision rosseland_mean_dust_opacity
      doubleprecision dustrho(DUST_SPECIES_MAX)
      doubleprecision dusttemp(DUST_SIZE_MAX,DUST_SPECIES_MAX)
      doubleprecision radtemp,bpldt
      integer iabs,iscat
c
#     include "common_grid.h"
#     include "common_dust.h"
c
      doubleprecision find_dust_kappa,alpha,bplanckdt,enumer,denom
      integer inu,ispec,isize
c
c     Perform the integrations of enumerator and denominator
c
      enumer = 0.d0
      denom  = 0.d0
      do inu=1,freq_nr
c
c         Find dB_nu(T)/dT
c
          bpldt = bplanckdt(radtemp,freq_nu(inu))
c
c         Find the alpha at current temperature
c
          alpha = 0.d0
          do ispec=1,dust_nr_species
              do isize=1,dust_nr_size(ispec)
                  alpha = alpha + dustrho(ispec) *
     %                 find_dust_kappa(inu,isize,ispec,
     %                 dusttemp(isize,ispec),iabs,iscat)
              enddo
          enddo
c
c         Now add to integral of enumerator
c
          enumer = enumer + dust_freq_wgt(inu) * bpldt / alpha
c
c         and the denominator
c
          denom  = denom + dust_freq_wgt(inu) * bpldt
c
      enddo
c
c     Then divide, and obtain the Rosseland mean
c
      rosseland_mean_dust_opacity = denom / enumer
c
      return
      end




c     --------------------------------------------------------------
c                 THE PLANCK-MEAN KAPPA PER SPECIES,SIZE
c
c     This routine returns the Planck mean opacity
c
c                  /                        / /             
c      kappa_P  =  | kappa_nu B_nu(T) dnu  /  | B_nu(T) dnu 
c                  /                      /   /             
c
c     for a single dust species (ispec) and grain size (isize), at
c     a specified temperature (temp), and either for pure absorption 
c     (iabs=1), pure scattering (iscat=1) or both.
c     --------------------------------------------------------------
      function planck_mean_dust_kappa(temp,ispec,isize,iabs,iscat)
      implicit none
      doubleprecision temp,planck_mean_dust_kappa
      integer ispec,isize,iabs,iscat
c
#     include "common_grid.h"
#     include "common_dust.h"
c
      doubleprecision find_dust_kappa,kappa,bplanck,enumer,denom
      integer inu
c
c     First the enumerator
c
      enumer = 0.d0
      do inu=1,freq_nr
c
c         Find the kappa at current temperature
c
          kappa = find_dust_kappa(inu,isize,ispec,
     %         temp,iabs,iscat)
c
c         Now add to integral
c
          enumer = enumer + kappa * dust_freq_wgt(inu) 
     %         * bplanck(temp,freq_nu(inu))
c
      enddo
c
c     Then the denominator
c      
      denom = 0.d0
      do inu=1,freq_nr
          denom = denom + dust_freq_wgt(inu) 
     %             * bplanck(temp,freq_nu(inu))
      enddo
c
c     Then divide, and obtain the Planck mean
c
      planck_mean_dust_kappa = enumer / denom
c
      return
      end





c     --------------------------------------------------------------
c               THE PLANCK-INTEGRAL KAPPA PER SPECIES,SIZE
c
c     This routine returns the Planck-integral (NOT MEAN!) opacity
c
c                  /                     
c      kappa_B  =  | kappa_nu B_nu(T) dnu
c                  /                     
c
c     for a single dust species (ispec) and grain size (isize), at
c     a specified temperature (temp), and either for pure absorption 
c     (iabs=1), pure scattering (iscat=1) or both.
c
c     This value is necessary for the determination of the dust
c     temperatures.
c     --------------------------------------------------------------
      function planck_integral_dust_kappa(temp,ispec,isize,iabs,iscat)
      implicit none
      doubleprecision temp,planck_integral_dust_kappa
      integer ispec,isize,iabs,iscat
c
#     include "common_grid.h"
#     include "common_dust.h"
c
      doubleprecision find_dust_kappa,kappa,bplanck,enumer
      integer inu
c
      enumer = 0.d0
      do inu=1,freq_nr
c
c         Find the kappa at current temperature
c
          kappa = find_dust_kappa(inu,isize,ispec,
     %         temp,iabs,iscat)
c
c         Now add to integral
c
          enumer = enumer + kappa * dust_freq_wgt(inu) 
     %         * bplanck(temp,freq_nu(inu))
c
      enddo
c
c     The answer
c
      planck_integral_dust_kappa = enumer
c
      return
      end





c     --------------------------------------------------------------
c           THE TEMPERATURE DERIVATIVE OF PLANCK FUNCTION 
c
c     This function computes the temperature derivative of the
c     Blackbody function 
c
c        dB_nu(T)     2 h^2 nu^4      exp(h nu / kT)        1 
c        --------   = ---------- ------------------------  ---
c           dT          k c^2    [ exp(h nu / kT) - 1 ]^2  T^2
c
c     ARGUMENTS:
c        nu    [Hz]            = Frequency
c        temp  [K]             = Electron temperature
c     --------------------------------------------------------------
      function bplanckdt(temp,nu)
      implicit none
      doubleprecision temp
      doubleprecision nu
      doubleprecision bplanckdt,theexp
c
#     ifndef NOCHECK
      if(temp.le.0.d0) then
          write(*,*) 'ERROR: bplanckdt()'
          write(*,*) '       temperature = ',temp,' K'
          stop 13
      endif
#     endif
c
      theexp = exp(4.7989d-11*nu/temp)
      if(theexp.lt.1.d33) then
          bplanckdt = 7.07661334104d-58 * nu**4 * 
     %         theexp / ( (theexp-1.d0)**2 * temp**2 )
     %         + 1.d-290
      else
          bplanckdt = 7.07661334104d-58 * nu**4 
     %         / ( theexp * temp**2 )
     %         + 1.d-290
      endif
c     
      return
      end



c     --------------------------------------------------------------
c             THE SPECTRUM-MEAN KAPPA PER SPECIES,SIZE
c
c     This routine returns the mean opacity weighted with a given 
c     spectrum. 
c
c                  /                     / /             
c      kappa_J  =  | kappa_nu J_nu dnu  /  | J_nu dnu 
c                  /                   /   /             
c
c     for a single dust species (ispec) and grain size (isize), 
c     and either for pure absorption (iabs=1), pure scattering 
c     (iscat=1) or both.
c     --------------------------------------------------------------
      function spectrum_mean_dust_kappa(meanint,temp,
     %               ispec,isize,iabs,iscat)
      implicit none
      doubleprecision temp,meanint(*),spectrum_mean_dust_kappa
      integer ispec,isize,iabs,iscat
c
#     include "common_grid.h"
#     include "common_dust.h"
c
      doubleprecision find_dust_kappa,kappa,bplanck,enumer,denom
      integer inu
c
c     First the enumerator
c
      enumer = 0.d0
      do inu=1,freq_nr
c
c         Find the kappa at current temperature
c
          kappa = find_dust_kappa(inu,isize,ispec,temp,iabs,iscat)
c
c         Now add to integral
c
          enumer = enumer + kappa * dust_freq_wgt(inu) * meanint(inu)
c
      enddo
c
c     Then the denominator
c      
      denom = 0.d0
      do inu=1,freq_nr
          denom = denom + dust_freq_wgt(inu) * meanint(inu)
      enddo
c
c     Then divide, and obtain the Planck mean
c
      spectrum_mean_dust_kappa = enumer / denom
c
      return
      end






c     ==============================================================
c                        OPACITY-DATABASE ROUTINES      
c     ==============================================================


c     --------------------------------------------------------------
c             CREATE ARRAYS OF PLANCK-MEAN, ROSSELAND-MEAN 
c                    AND PLANCK-INTEGRAL OPACITY
c
c     Since the determination of the mean and integral opacities is
c     a bit CPU-time consuming, we make pre-defined arrays here.
c     Note that for multi-species/size systems the rosseland opacity
c     may cause trouble. The mean rosseland opacities can only be
c     precomputed here if 
c       1) The dust relative abundances stay constant
c       2) All dust grains (size/spec) have same temperature
c          or the user knows a way to find a suitable average 
c          temperature (using Planck mean kappa as a weight for
c          example).
c     If the abundance array in the common block will contain 
c     non-zero numbers, then the mo_rosselandmean_opacity array
c     will be filled and the flag mo_rossarray_done set. Else the
c     arrays will be kept zero.
c     --------------------------------------------------------------
      subroutine create_mean_opacity_arrays(iabs,iscat)
      implicit none
c
      integer iabs,iscat
c
#     include "common_grid.h"
#     include "common_dust.h"
#     include "common_meankappa.h"
c
      common/dustabun/dustabun
      doubleprecision dustabun(DUST_SPECIES_MAX)
c
      integer inu,itemp,itrange,ispec,isize
      doubleprecision tmin,tmax,dummy,bplanck,find_dust_kappa
      doubleprecision planck_integral_dust_kappa
      doubleprecision planck_mean_dust_kappa
      doubleprecision rosseland_mean_dust_opacity
      doubleprecision totabun
      doubleprecision dusttemp(DUST_SIZE_MAX,DUST_SPECIES_MAX)
      logical doross
c
      write(*,*) 'Creating mean opacity database'
      write(*,*)
c
c     Save the flags
c
      mo_iabs  = iabs
      mo_iscat = iscat
c
c     First do a check
c
      if(dust_done_read.eq.0) then
          write(*,*) 'Sorry, I cannot create the mean opacity'
          write(*,*) '   arrays if the dust data has not been'
          write(*,*) '   read yet.'
          stop 13
      endif
c
c     Then check whether there are (global) dust abundances 
c     specified for the Rosseland mean kappa. Or if not,
c     whether there may be only one grain spec/size, so that
c     a global Rosseland kappa is possible as well.
c     
      doross = .false.
      do ispec=1,dust_nr_species
          if(dustabun(ispec).ne.0.d0) then
              doross = .true.
          endif
      enddo
      if((dust_nr_species.eq.1).and.(dust_nr_size(1).eq.1)) then
          dustabun(1) = 1.d0
          doross = .true.
      endif
c
c     Check whether the condensation temperatures are the same
c     (else the global relative abundances are impossible)
c
      if(doross) then
          write(*,*) '...Creating Rosseland opacity table...'
          if(dust_nr_species.gt.1) then
              do ispec=2,dust_nr_species
                  if((abs(dust_tmin(ispec)-dust_tmin(1)).gt.1.d0)
     %                 .or.(abs(dust_tmax(ispec)-dust_tmax(1))
     %                 .gt.1.d-2*dust_tmax(1))) then
                      write(*,*) 'WARNING: By passing non-zero ',
     %                     'abundances to create_mean_opacity_arrays()'
                      write(*,*) '         you indicated that you ',
     %                     'wish to have te Rosseland opacities ',
     %                     'pre-computed.'
                      write(*,*) '         But since the evapration ',
     %                     'temperatures are non the same for all ',
     %                     'species,'
                      write(*,*) '         this is intrinsically ',
     %                     'impossible.'
                      write(*,*) '         Switching back to ',
     %                     'on-the-fly calculation of the Rosseland ',
     %                     'mean.'
                      doross=.false.
                  endif
              enddo      
          endif
          if((dust_nr_species.gt.1).or.(dust_nr_size(1).gt.1)) then
              write(*,*) 'Beware: when you want to use pre-computed ',
     %             'Rosseland opacities'
              write(*,*) '        with multiple dust species/sizes, ',
     %             'you are making the '
              write(*,*) '        approximation that (for the ',
     %             'Rosseland mean) all dust '
              write(*,*) '        temperatures are the same. This ',
     %             'could be dangerous.'
          endif
      endif
c
c     Find the total abundance
c
      totabun = 0.d0
      do ispec=1,dust_nr_species
          totabun = totabun + dustabun(ispec)
      enddo      
c
c     Hard settings
c
      mo_ntemp = INTALPBNU_NRTEMP
      tmin     = INTALPBNU_TMIN
      tmax     = INTALPBNU_TMAX
cc      iabs     = 1
cc      iscat    = 0
c
c     Loop over temperature
c
      do itemp=1,mo_ntemp
c
c         First make the actual temperature
c
          mo_temp(itemp) = tmin*exp((itemp-1)*
     %         (log(tmax)-log(tmin))/(mo_ntemp-1.d0))
c
c         If doross, then find the Rosseland opacity and store it
c
          if(doross) then
              do ispec=1,dust_nr_species
                  do isize=1,dust_nr_size(ispec)
                      dusttemp(isize,ispec) = mo_temp(itemp)
                  enddo
              enddo
              dummy = rosseland_mean_dust_opacity(
     %          dustabun,dusttemp,mo_temp(itemp),iabs,iscat)
              mo_rosselandmean_kappa(itemp) = dummy  / totabun
          else
              mo_rosselandmean_kappa(itemp) = -1.2345d20   ! Self check
          endif
c
c         Loop over species, and size
c
          do ispec=1,dust_nr_species
              do isize=1,dust_nr_size(ispec)
c     
c                 First get the condensation value
c
                  mo_condense(itemp,isize,ispec) = 
     %                 find_dust_kappa(1,isize,ispec,
     %                 mo_temp(itemp),iabs,iscat) 
     %                 / find_dust_kappa(1,isize,ispec,
     %                 mo_temp(itemp),-iabs,-iscat) 
c
c                 Then the Planck-integral opacity
c
                  dummy = planck_integral_dust_kappa(
     %                 mo_temp(itemp),ispec,isize,-iabs,-iscat)
                  mo_lgintalpnu(itemp,isize,ispec) = log(dummy)
                  if(dummy.le.0.d0) then
                      write(*,*) 'ERROR: Negative or zero Planck ',
     %                          'integral opacity'
                      stop 13
                  endif
c
c                 Then the Planck-mean opacity
c
                  dummy = planck_mean_dust_kappa(
     %                 mo_temp(itemp),ispec,isize,-iabs,-iscat)
                  mo_planckmean_kappa(itemp,isize,ispec) = dummy
                  if(dummy.le.0.d0) then
                      write(*,*) 'ERROR: Negative or zero Planck ',
     %                          'mean opacity'
                      stop 13
                  endif
c
              enddo
          enddo
      enddo
c
c     Now tell the rest of the program that the mean opacities
c     have been computed
c
      mo_array_done = 1
      if(doross) then
          mo_rossarray_done = 1
      endif
c
      end



c     --------------------------------------------------------------
c             LINEAR INTERPOL OF LOG(INT KAPPA_NU B_NU DNU )
c
c     This routine finds the value of:
c
c       /                      
c       | kappa_nu(T_dust) B_nu(T) dnu 
c       /                      
c
c     either by interpolation from the pre-computed database
c     (see create_mean_opacity_arrays()), or - if this database
c     is not available or inappropriate - from scratch. 
c
c     Usually you would use this function with T=T_dust. But for
c     possible future use, we distinguish the two. The T_dust is 
c     only to determine which opacity to take, since at different
c     temperatures one has different dust species that survive the
c     heat.
c     --------------------------------------------------------------
      function lgkapbnuint(temp,tdust,ispec,isize,iabs,iscat)
      implicit none
      doubleprecision lgkapbnuint
c
#     include "common_grid.h"
#     include "common_dust.h"
#     include "common_meankappa.h"
c
      integer itemp,ispec,isize,iabs,iscat
      doubleprecision tmin,tmax,temp,wgt,t0,t1,tt,tdust
      doubleprecision cnd0,cnd1,planck_integral_dust_kappa
c
c     First check whether we can use the database. If not, then
c     compute on-the-fly.
c
      if((mo_array_done.eq.0).or.(iabs.ne.mo_iabs)
     %   .or.(iscat.ne.mo_iscat)) then
          lgkapbnuint = log(planck_integral_dust_kappa(temp,
     %                    ispec,isize,iabs,iscat))
          return
      endif
c
c     Use the database...
c
c     Find the index of the emission temperature
c
      if(abs(temp).lt.mo_temp(1)) then
          write(*,*) 'Sorry, temperature too low for kapbnu array'
          write(*,*) '  T = ',abs(temp)
          stop 155
      endif
      if(abs(temp).gt.mo_temp(mo_ntemp)) then
          write(*,*) 'Sorry, temperature too high for kapbnu array'
          write(*,*) '  T = ',abs(temp)
          stop 155
      endif
      call hunt(mo_temp(1),mo_ntemp,abs(temp),itemp)
      if(itemp.gt.INTALPBNU_NRTEMP) then
          stop 667
      endif
      if(itemp.eq.mo_ntemp) then
          itemp=mo_ntemp-1
      endif
c
c     Get the condensation/evaporation stuff, unless (by temp<0) 
c     the caller indicated it does not want it.
c
      if(temp.ge.0.d0) then
          cnd0 = mo_condense(itemp,isize,ispec)
          cnd1 = mo_condense(itemp+1,isize,ispec)
      else
          cnd0 = 1.d0
          cnd1 = 1.d0
      endif
c
c     And now return the answer
c
      t0        = log(mo_temp(itemp))
      t1        = log(mo_temp(itemp+1))
      tt        = log(abs(temp))
      wgt       = ( tt - t0 ) / ( t1 - t0 ) 
      lgkapbnuint = (1.d0-wgt)*mo_lgintalpnu(itemp,isize,ispec)*cnd0 
     %                   + wgt*mo_lgintalpnu(itemp+1,isize,ispec)*cnd1
c
      return
      end





c     --------------------------------------------------------------
c        LINEAR INTERPOL OF PLANCK MEAN KAPPA (WITH BNU AS WEIGHT)
c     --------------------------------------------------------------
      function planckmeankappa(temp,ispec,isize,iabs,iscat)
      implicit none
      doubleprecision planckmeankappa
c
#     include "common_grid.h"
#     include "common_dust.h"
#     include "common_meankappa.h"
c
      integer itemp,ispec,isize,iabs,iscat
      doubleprecision tmin,tmax,temp,wgt,t0,t1,tt,tdust
      doubleprecision cnd0,cnd1,planck_mean_dust_kappa
c
c     First check whether we can use the database. If not, then
c     compute on-the-fly.
c
      if((mo_array_done.eq.0).or.(iabs.ne.mo_iabs)
     %   .or.(iscat.ne.mo_iscat)) then
          planckmeankappa = planck_mean_dust_kappa(temp,
     %                    ispec,isize,iabs,iscat)
          return
      endif
c
c     Use the database...
c
c     Find the index of the emission temperature
c
      if(abs(temp).lt.mo_temp(1)) then
          write(*,*) 'Sorry, temperature too low for ',
     %                 'planck mean opacity array'
          write(*,*) '  T    = ',abs(temp)
          write(*,*) '  Tmin = ',mo_temp(1)
          stop 155
      endif
      if(abs(temp).gt.mo_temp(mo_ntemp)) then
          write(*,*) 'Sorry, temperature too high for ',
     %                 'planck mean opacity array'
          write(*,*) '  T    = ',abs(temp)
          write(*,*) '  Tmax = ',mo_temp(mo_ntemp)
          stop 155
      endif
      call hunt(mo_temp(1),mo_ntemp,abs(temp),itemp)
      if(itemp.gt.INTALPBNU_NRTEMP) then
          stop 667
      endif
      if(itemp.eq.mo_ntemp) then
          itemp=mo_ntemp-1
      endif
c
c     Get the condensation/evaporation stuff, unless (by temp<0) 
c     the caller indicated it does not want it.
c
      if(temp.ge.0.d0) then
          cnd0 = mo_condense(itemp,isize,ispec)
          cnd1 = mo_condense(itemp+1,isize,ispec)
      else
          cnd0 = 1.d0
          cnd1 = 1.d0
      endif
c
c     And now return the answer
c
      t0        = log(mo_temp(itemp))
      t1        = log(mo_temp(itemp+1))
      tt        = log(abs(temp))
      wgt       = ( tt - t0 ) / ( t1 - t0 ) 
      planckmeankappa = 
     %     (1.d0-wgt)*mo_planckmean_kappa(itemp,isize,ispec)*cnd0 
     %          + wgt*mo_planckmean_kappa(itemp+1,isize,ispec)*cnd1 
c
      return
      end






c     --------------------------------------------------------------
c                 FIND THE TOTAL ROSSELAND EXTINCTION
c
c     For the radiative transfer one needs the total Rosseland 
c     extinction function, i.e. the kappa's times the densities
c     of the individual species, summed over all species and
c     grain sizes. In short, this function computes:
c                     ___
c                     \
c       alpha_R_tot =  >  rho_i \kappa_R_i
c                     /__
c                    i=1,nspec*nsize
c
c     REMARK: In this routine the radiation temperature (which 
c             determines the position of the peak of the Planck
c             curve) is ASSUMED EQUAL to the temperature of the
c             individual dust grains.
c     --------------------------------------------------------------
      function totalrosselandmeanalpha(dustrho,dusttemp,iabs,iscat)
      implicit none
      integer iabs,iscat
      doubleprecision totalrosselandmeanalpha
      doubleprecision dustrho(DUST_SPECIES_MAX)
      doubleprecision dusttemp(DUST_SIZE_MAX,DUST_SPECIES_MAX)
c
#     include "common_dust.h"
#     include "common_meankappa.h"
c
      doubleprecision dummy,lgkapbnuint
      doubleprecision rosseland_mean_dust_opacity
      doubleprecision rho,temp,sum,dt,t0,t1,tt,wgt
      integer ispec,isize,itemp
c
c     First find a good temperature average and total density
c     The weight-function for the average is the Planck 
c     integrated kappa of each species,size. 
c     
      rho  = 0.d0
      temp = 0.d0
      sum  = 0.d0
      do ispec=1,dust_nr_species
          rho = rho + dustrho(ispec)
          do isize=1,dust_nr_size(ispec)
              dt    = dusttemp(isize,ispec)
              dummy = exp(lgkapbnuint(dt,dt,ispec,isize,iabs,iscat)) 
              temp  = temp + dusttemp(isize,ispec) * dummy
              sum   = sum + dummy
          enddo
      enddo
      temp = temp / sum
c
c     If necessary, limit the temperatuer
c     
#     ifdef TOTMEANKAPPA_NO_LIMITS
      if(temp.lt.INTALPBNU_TMIN) then
          temp=INTALPBNU_TMIN*(1.d0+1.d-3)
      endif
      if(temp.gt.INTALPBNU_TMAX) then
          temp=INTALPBNU_TMAX*(1.d0-1.d-3)
      endif
#     endif
c
c     Now, depending on whether there exist rosseland kappa arrays
c     we use the arrays or we compute the rosseland mean on the fly
c
      if((mo_rossarray_done.eq.0).or.(iabs.ne.mo_iabs)
     %   .or.(iscat.ne.mo_iscat)) then
c
c         If no Rosseland database exists, or it is imcompatible 
c         with current input, then we must compute the Rosseland 
c         mean opacity on-the-fly. 
c
          totalrosselandmeanalpha = rosseland_mean_dust_opacity(
     %             dustrho,dusttemp,temp,iabs,iscat)
          return
      endif
c
c     Rosseland arrays exist, so we use these. 
c
c     NOTE: This should only be the case when either there is 
c           only one dust species, or when the abundances are 
c           global and the different species have the same 
c           condensation/evaporation temperatures. Make this 
c           sure by calling the function:
c               is_global_rosseland_possible()
c
c     Find the index of the emission temperature
c
      if(abs(temp).lt.mo_temp(1)) then
          write(*,*) 'Sorry, temperature too low for ',
     %            'rosseland mean opacity array'
          write(*,*) '  T    = ',abs(temp)
          write(*,*) '  Tmin = ',mo_temp(1)
          stop 155
      endif
      if(abs(temp).gt.mo_temp(mo_ntemp)) then
          write(*,*) 'Sorry, temperature too high for ',
     %            'rosseland mean opacity array'
          write(*,*) '  T    = ',abs(temp)
          write(*,*) '  Tmax = ',mo_temp(mo_ntemp)
          stop 155
      endif
      call hunt(mo_temp(1),mo_ntemp,abs(temp),itemp)
      if(itemp.gt.INTALPBNU_NRTEMP) then
          stop 667
      endif
      if(itemp.eq.mo_ntemp) then
          itemp=mo_ntemp-1
      endif
c
c     And now return the answer
c
      t0        = log(mo_temp(itemp))
      t1        = log(mo_temp(itemp+1))
      tt        = log(abs(temp))
      wgt       = ( tt - t0 ) / ( t1 - t0 ) 
      totalrosselandmeanalpha = rho * (
     %     (1.d0-wgt)*mo_rosselandmean_kappa(itemp)
     %          + wgt*mo_rosselandmean_kappa(itemp+1) )
c
      return
      end








c     ==============================================================
c         ROUTINES FOR SOLVING T_DUST FROM INTEGRATED INTENSITY
c
c     Using the mean opacities one can do integrated intensity 
c     radiative transfer. From the integrated mean intensity one can
c     then (using the mean kappas) find back the dust temperatures.
c     Here are some routines for doing this.
c     ==============================================================

c     --------------------------------------------------------------
c            SOLVE DUST TEMPERATURE FROM THE MEAN INTENSITY J
c
c     The dust/gas temperature is computed from a given value of
c     the integrated mean intensity J. This is done by solving 
c     iteratively the following equation:
c
c         /
c         | kappa_nu B_nu(T) dnu  = ( q_* + q_visc ) / 4pi +  meankappa J
c         /
c
c     The meankappa is computed on-the-fly, and so will the viscous
c     heating. These values are returned. In principle the mean opacity
c     should be dependent on J_nu, rather than B_nu(T). Later we will 
c     add a correction factor, something like: 
c     factor=realmeankappa(J_nu)/meankappa(T), 
c     which can be computed from full radiative transfer.
c
c     ARGUMENTS:
c       qstarfpi          q^* / 4 pi = Heating by stellar radiation
c       mint              Frequency-integrated mean intensity J
c       ispec             Species for which to solve
c       isize             Grain size index of this species for which to solve
c
c     OTHER INPUT:
c       sd_mugas
c       sd_alpha
c       sd_omk
c       ..dust opacities 
c
c     RESULTS:
c       solvedtemp        The temperature to be solved 
c       meankappa         Planck-average opacity
c       qviscfpi          q^visc / 4 pi = Viscous heating
c       viscnu            The viscosity 'nu'
c
c     --------------------------------------------------------------
      function solvedtemp(qstarfpi,mint,ispec,isize,
     %                    meankappa,qviscfpi,viscnu)
      implicit none
      doubleprecision solvedtemp
      doubleprecision solvedtemp_func,zbrent
      doubleprecision qstarfpi,mint,meankappa,qviscfpi,viscnu
      integer ispec,isize
      external solvedtemp_func
c
#     include "common_grid.h"
#     include "common_dust.h"
c
      common/solvedtempint/sdt_qstarfpi,sdt_mint,sdt_meankappa,
     %                     sdt_qviscfpi,sdt_viscnu
      doubleprecision sdt_qstarfpi,sdt_mint,sdt_meankappa
      doubleprecision sdt_qviscfpi,sdt_viscnu
      common/isolvedtempint/sdt_ispec,sdt_isize,sdt_tradj
      integer sdt_ispec,sdt_isize,sdt_tradj
c
      doubleprecision dum1,dum2,tbot,ttop
      integer i
c
      tbot = INTALPBNU_TMIN * (1.d0+1.d-3)
      ttop = INTALPBNU_TMAX * (1.d0-1.d-3)
c
c     Put the data in a common block
c
      sdt_qstarfpi  = qstarfpi
      sdt_mint      = mint
      sdt_meankappa = meankappa
      sdt_ispec     = ispec
      sdt_isize     = isize
      sdt_tradj     = SOLVEDT_USE_TRAD
c
c     Now check if there is indeed sufficient radiation for a
c     solution.
c
      dum1 = solvedtemp_func(0.02d0)
      dum2 = solvedtemp_func(3000.d0)
      if((dum1.gt.0.d0).or.(dum2.lt.0.d0)) then
          goto 100
      endif
c
c     Now solve for the dust temperature
c
      solvedtemp = zbrent(solvedtemp_func,2.d-2,3.d3,1.d-4) 
c
      goto 400
c
c     Try again with huge domain in T, simply to avoid crash when the
c     solution temporarily gets beyond the sublimation temp or below
c     0.01K
c
 100  continue
      dum1 = solvedtemp_func(tbot)
      dum2 = solvedtemp_func(ttop)
      if(dum1.gt.0.d0) then
          write(*,*) 'Error: dust temperature < ',tbot,' K'
          write(*,*) '   Dust temperature       = ',dum1
          stop 1
      endif
      if(dum2.lt.0.d0) then
          write(*,*) 'Error: dust temperature > ',ttop,' K'
          write(*,*) '   Dust temperature       = ',dum2
          write(*,*) '   q_*/4pi                = ',sdt_qstarfpi
          write(*,*) '   Diffuse mean intensity = ',sdt_mint
          write(*,*) '   Mean kappa             = ',sdt_meankappa
          write(*,*) '   isize,ispec            = ',isize,ispec
          stop 1
      endif
c
c     Now solve for the dust temperature
c
      solvedtemp = zbrent(solvedtemp_func,tbot,ttop,1.d-4) 
c
 400  continue
c
c     And return the mean kappa and the viscous heating rate
c
      meankappa = sdt_meankappa
      qviscfpi  = sdt_qviscfpi
      viscnu    = sdt_viscnu
c
      return
      end




c     --------------------------------------------------------------
c         INTEGRAL OF EMISSION-ABSORPTION, USING AVERAGED J
c     
c     NOTE: Be sure to set the sdt_ variables!
c
c     NOTE: A similar function is hanging around in dust.F, but 
c           that one is particularly for (ir,iz). The present one
c           here is independent of that. But it allows only one
c           dust species and size.
c     --------------------------------------------------------------
      function solvedtemp_func(temp)
      implicit none
c
      common/solvedtempint/sdt_qstarfpi,sdt_mint,sdt_meankappa,
     %                     sdt_qviscfpi,sdt_viscnu
      doubleprecision sdt_qstarfpi,sdt_mint,sdt_meankappa
      doubleprecision sdt_viscnu,sdt_qviscfpi
      common/isolvedtempint/sdt_ispec,sdt_isize,sdt_tradj
      integer sdt_ispec,sdt_isize,sdt_tradj
c
#     include "common_grid.h"
#     include "common_dust.h"   
c     
      integer inu
      doubleprecision planckmeankappa,viscousheating,lgkapbnuint
      doubleprecision temp,solvedtemp_func,bplanck,qtotfpi,temprad
c
c     Compute first the viscous heating...
c        (bugfix: 4pi  [01.02.00])
c
#     ifdef INCLUDE_VISCOUS_HEATING
      sdt_qviscfpi    = viscousheating(temp,sdt_viscnu) 
     %                     * 7.95774715459d-2
#     else
      sdt_qviscfpi    = 0.d0
#     endif
c
c     ...and the total viscous + irradiative heating
c
      qtotfpi         = sdt_qviscfpi + sdt_qstarfpi
c
c     Make a guess of the temperature of the radiation field itself
c
      if(sdt_tradj.eq.0) then
c
c         T_rad    = T
c
          temprad  = temp
c
      elseif(sdt_tradj.eq.1) then
c
c         T_rad^4  = (pi/sigma) J
c
          temprad  = (5.5404346d4*sdt_mint)**0.25
c
      else
          stop 55
      endif
c
c     Find kappa from Trad and/or possible other factors that determine
c     the Planck mean kappa. Give -temprad to get kappa without condensation
c     factor.
c     
      sdt_meankappa   = planckmeankappa(-temprad,
     %                     sdt_ispec,sdt_isize,1,0)
c
c     Return the value of the equation
c     
      solvedtemp_func = lgkapbnuint(-temp,-temp,
     %                     sdt_ispec,sdt_isize,1,0) - 
     %                     log(qtotfpi+sdt_meankappa*sdt_mint)
      return
      end






c     ==============================================================
c                           SUPPORT ROUTINES
c     ==============================================================



c     --------------------------------------------------------------
c                          EEN CHECK FUNCTION
c
c     This function checks whether the relative abundances of the
c     density arrays are global, and whether the evaporation 
c     temperatures are equal for all species. If all of this is
c     true, then it returns .true., else .false.
c     --------------------------------------------------------------
      function is_global_rosseland_possible(dustrho,nnx,nny,nx,ny)
      implicit none
      integer nnx,nny,nx,ny
      doubleprecision dustrho(DUST_SPECIES_MAX,nnx,nny)
      logical is_global_rosseland_possible
c
c#     include "common_grid.h"
#     include "common_dust.h"
c
      doubleprecision abun(DUST_SPECIES_MAX),abun00(DUST_SPECIES_MAX)
      doubleprecision dummy
      integer ispec,isize,ix,iy
c
c     If only one species, then no check necessary
c     
      if(dust_nr_species.eq.1) then
          is_global_rosseland_possible = .true.
          return
      endif
c
c     First check the condensation/evaporation temperatures
c
      if(dust_nr_species.gt.1) then
          do ispec=2,dust_nr_species
              if((abs(dust_tmin(ispec)-dust_tmin(1)).gt.1.d0)
     %             .or.(abs(dust_tmax(ispec)-dust_tmax(1))
     %             .gt.1.d-2*dust_tmax(1))) then
                 is_global_rosseland_possible = .false.
                 return
              endif
          enddo
      endif
c
c     Then the relative abundances
c
      dummy = 0.d0
      do ispec=1,dust_nr_species
          abun00(ispec) = dustrho(ispec,1,1)
          dummy         = dummy + abun00(ispec)
      enddo
      if(dummy.eq.0.d0) then
          write(*,*) 'Sorry, cannot check for abundances if the '
          write(*,*) 'Density at (1,1) is zero...'
          stop 13
      endif
      do ispec=1,dust_nr_species
          abun00(ispec) = abun(ispec) / dummy
      enddo      
      do iy=1,ny
          do ix=1,nx
              dummy = 0.d0
              do ispec=1,dust_nr_species
                  abun(ispec) = dustrho(ispec,1,1)
                  dummy       = dummy + abun(ispec)
              enddo
              do ispec=1,dust_nr_species
                  abun(ispec) = abun(ispec) / dummy
                  if(abs(abun(ispec)-abun00(ispec)).gt.1.d-2) then
                      is_global_rosseland_possible = .false.
                      return
                  endif
              enddo      
          enddo
      enddo
c
      is_global_rosseland_possible = .true.
      return      
      end
