c     ==============================================================
c           PROGRAM THAT PERFORMS MOMENT-METHOD VERTICAL RT
c
c     This program reads the dust temperatures resulting from 
c     primary photons. These temperatures (and the corresponding
c     radial radiative transfer) should be performed and written
c     by an external radiative transfer program. This will be the
c     Monte-Carlo program by Gerd-Jan van Zadelhoff. Using these
c     temperatures the present program will perform vertical
c     radiative transfer to find the dust temperature with the 
c     inclusion of secondary radiation. This is essential for the
c     computation of the midplane temperatures of irradiated 
c     disks.
c     ==============================================================

#include "main.h"

#define GASTODUST (100.d0)

c     --------------------------------------------------------------
c                             MAIN ROUTINE
c     --------------------------------------------------------------
      program main
      implicit none
c
#include "common_grid.h"
#include "common_dust.h"
#include "common_vertrt.h"
c
      call read_all_data()
      if(freq_nr.gt.1) then
          call create_mean_opacity_arrays(1,0)
      endif
      call do_all_radtrans()
      call write_temperatures()
      if(global_idump.ne.0) then
          call write_aux_data()
      endif
c
      end



c     --------------------------------------------------------------
c                         READ ALL THE DATA IN
c     --------------------------------------------------------------
      subroutine read_all_data()
      implicit none
c
#include "common_grid.h"
#include "common_dust.h"
#include "common_vertrt.h"
c
      doubleprecision rstar,a,b,c,d,e
      integer nr,nz,iz,ir,nrp,nzp,nf,inu
      integer iformat
      logical f_ex,fexf,fexo
c
c     Default value for starting parameter and for the tautrick switch
c
      global_ir_start = 2
      global_tautrick = 0
      global_idump    = 0
c
c     Is there an input file?
c
      inquire(file='vertrt.inp',exist=f_ex)
      if(f_ex) then
c          write(*,*) 'Found vertrt.inp!'
          open(unit=1,file='vertrt.inp',status='old')
          read(1,*,end=44) global_ir_start
          read(1,*,end=44) global_idump
          read(1,*,end=44) global_tautrick 
 44       continue
          close(1)
      endif
c
c     For viscosity, read star data
c
      inquire(file='viscosity.inp',exist=f_ex) 
      if(f_ex) then
          open(unit=1,file='viscosity.inp',status='old')
          read(1,*,end=20) accr_alpha
          close(1) 
          if(accr_alpha.gt.0.d0) then
              open(unit=1,file='starinfo.inp',status='old')
              read(1,*) a
              read(1,*) a
              read(1,*) accr_mstar
              close(1)
          endif
 20       continue
      endif
c##############################
      if(accr_alpha.gt.0.d0) then
         write(*,*) 'VISCOUS HEATING NOT YET TESTED.'
         write(*,*) 'PROBABLY A BUG IN THERE. DO NOT USE, UNLESS YOU'
         write(*,*) 'DEBUG IT YOURSELF!'
         stop
      endif
c##############################
c      write(*,*) accr_alpha,accr_mstar
c
c     Read the primary temperature in
c
      open(unit=1,file='diskstruct_1.dat',status='old')
      read(1,*) nr,nz
      if(nr.gt.FRSIZE_R) then
          write(*,*) 'ERROR in vertrt:'
          write(*,*) 'file diskstruct_1.dat: R grid too large'
          stop 13
      endif
      if(nz.gt.FRSIZE_Z) then
          write(*,*) 'ERROR in vertrt:'
          write(*,*) 'file diskstruct_1.dat: Z grid too large'
          stop 13
      endif
      grid_nr = nr
      grid_nz = nz
      do ir=1,nr
          do iz=1,nz
              read(1,*) a,b,c,d,e
              grid_r(ir)=a
              grid_z(iz,ir)=b
              vrt_rho(iz,ir)=c/GASTODUST
              vrt_primtemp(iz,ir)=d
              vrt_vz(iz,ir)=e
          enddo
      enddo
      close(1)
c
c     Put global_ir_start to 1 if only 1 or 2 shells
c
      if(nr.le.2) then
          global_ir_start = 1
      endif
c
c     Read the frequency array and the opacities
c
      inquire(file='kappaplanck.inp',exist=fexf)
      if(fexf) then
          write(*,*) 'PROBLEM: kappaplanck.inp is obsolete...'
          write(*,*) 'TIP    : mv kappaplanck.inp planckopac.inp'
          write(*,*) '         and add two colums with 0.d0'
          write(*,*) '         (i.e. kap_scat and g_scat).'
          stop 13
      endif
      inquire(file='planckopac.inp',exist=fexo)
      if(fexo) then
c
c         Planck Opacity file is there, so use it. If the frequency
c         dependent opacities are also there, then they are assumed
c         to be the opacities from which the planck opacities were
c         derived. The frequency-dependent opacitities can then be
c         used to make the SED.
c     
          freq_nr = 1
          call read_planckopac()
          call read_rossopac()
c
c         Check if also dust opacs present
c
          inquire(file='frequency.inp',exist=fexf)
          inquire(file='dustopac_1.inp',exist=fexo)
          if(fexf.and.fexo) then
              write(*,*) '   (Read planck opacity from planckopac.inp)'
          endif
      else
c
c         Find the frequency-dependent opacities, and do the radial
c         transfer with these opacitities (do it with frequency
c         dependence).
c
          inquire(file='frequency.inp',exist=fexf)
          inquire(file='dustopac_1.inp',exist=fexo)
          if((.not.fexf).or.(.not.fexo)) then
              write(*,*) 'Need both frequency.inp and dustopac_1.inp'
              stop 13
          endif
          open(unit=1,file='frequency.inp',status='old')
          read(1,*) nf
          if(nf.gt.FRSIZE_FREQ) then
              write(*,*) 'ERROR in vertrt:'
              write(*,*) 'Frequency grid too large'
              stop 13
          endif
          do inu=1,nf
              read(1,*) a
              freq_nu(inu) = a
          enddo
          dust_setup_nrspecies=1
          dust_setup_nrsizes(1)=1
          freq_nr = nf
          call read_dustdata()
      endif
c
      end



c     --------------------------------------------------------------
c                         PERFORM ALL RT
c     --------------------------------------------------------------
      subroutine do_all_radtrans()
      implicit none
c
#include "common_grid.h"
#include "common_natconst.h"
#include "common_vertrt.h"
c
      doubleprecision totalrosselandmeanalpha,planckmeankappa
      doubleprecision findkappabnu,find_temperature,ts
      doubleprecision find_rossopac
      doubleprecision qstarfpi(FRSIZE_Z),primpower(FRSIZE_Z),a,b,c
      doubleprecision viscousheating,omk,temp,qviscfpi(FRSIZE_Z)
      doubleprecision tau(FRSIZE_Z),kappaav,kapprim(FRSIZE_Z)
      doubleprecision tempold(FRSIZE_Z),error,find_planckopac,tauzero
      integer ir,iz,iter,niter,ii
c
      write(*,*) '.....vertrt'
c
c     For debugging, tauzero may be non-zero
c
      if(global_tautrick.ne.0) then
          tauzero = 0.d0
          write(*,*) '   WARNING: Using old-fashioned tau-trick!'
      else
          tauzero = 1d3
      endif
c
c     Copy the primary temperature to the real temperature
c
      niter=10
      do iz=1,grid_nz
          vrt_temp(iz,1) = vrt_primtemp(iz,1)
      enddo
c
c     Now do the main spatial loop
c     
      if(global_ir_start.eq.0) stop 934
      do ir=global_ir_start,grid_nr
c
c         First we put the primary temperature as a first guess
c         to the final temperature
c
          do iz=1,grid_nz
              vrt_temp(iz,ir) = vrt_primtemp(iz,ir)
          enddo
c
c         Clear the backup dust temperatures
c
          do iz=1,grid_nz
              tempold(iz) = 0.d0
          enddo
c
c         If no full opacities are present, we use the Planck opacity
c         for the integration. We compute a first guess here...
c         At the same time compute the kappa planck at the primary
c         temperature
c
          if(freq_nr.le.1) then
              do iz=1,grid_nz
                  kapprim(iz) = find_planckopac(vrt_primtemp(iz,ir))
                  vrt_meanalpha(iz,ir) = vrt_rho(iz,ir) *
     %                 kapprim(iz)
              enddo
          endif
c
c         Compute the qstarfpi, i.e. the source term for the vertical
c         radiative transfer
c
          if(freq_nr.gt.1) then
              do iz=1,grid_nz
                  qstarfpi(iz) = findkappabnu(vrt_primtemp(iz,ir))
              enddo
          else
              do iz=1,grid_nz
                  qstarfpi(iz) = 1.8049d-5 * kapprim(iz)
     %                             * vrt_primtemp(iz,ir)**4
              enddo
          endif
c
c         At each radius we now do an iteration of radiative
c         transfer calculations. This iteration is necessary because
c         the mean opacities will change as the dust temperature changes
c         
          do iter=1,niter
c
c             We compute here the Rosseland opacities
c
              if(freq_nr.eq.1) then
                  do iz=1,grid_nz
                      vrt_meanalpha(iz,ir) = vrt_rho(iz,ir) * 
     %                     find_rossopac(vrt_temp(iz,ir))
                  enddo
              else
                  do iz=1,grid_nz
                      vrt_meanalpha(iz,ir) = totalrosselandmeanalpha(
     %                     vrt_rho(iz,ir),max(INTALPBNU_TMIN,
     %                     vrt_temp(iz,ir)),1,0)
                  enddo
              endif
c
c             Compute the power absorbed by the dust from the
c             viscous heating
c
              do iz=1,grid_nz
                  if(accr_alpha.gt.0.d0) then
                      temp = vrt_temp(iz,ir)
                      omk  = sqrt(GG*accr_mstar/(grid_r(ir)**3.d0))
                      qviscfpi(iz) = viscousheating(temp,
     %                    accr_alpha,omk)/12.5663706144d0
                  else
                      qviscfpi(iz) = 0.d0
                  endif
                  primpower(iz) = ( qstarfpi(iz) + qviscfpi(iz) ) * 
     %                                vrt_rho(iz,ir) 
              enddo
c
c             Next integrate Hstar(z) from 0 to zmax
c
              vrt_hflux(1,ir) = 0.d0
              do iz=2,grid_nz
                  vrt_hflux(iz,ir) = vrt_hflux(iz-1,ir) + 
     %                 0.5d0 * ( grid_z(iz,ir) - grid_z(iz-1,ir) ) 
     %                 * ( primpower(iz) + primpower(iz-1) )
              enddo
c
c             Set the J at z=zmax to H * sqrt(3) 
c     
              vrt_meanint(grid_nz,ir) = abs(vrt_hflux(grid_nz,ir)) 
     %            * sqrt(3.d0)
c
c             Set the optical depth to zero here (tauzero)
c
              tau(grid_nz) = tauzero
c
c             Now integrate J back to the equator, and also compute
c             the optical depth tau
c
              do iz=grid_nz-1,1,-1
                  vrt_meanint(iz,ir) = vrt_meanint(iz+1,ir) + 
     %                 3.d0 * 0.5d0 
     %                * ( vrt_hflux(iz+1,ir) * vrt_meanalpha(iz+1,ir) 
     %                  + vrt_hflux(iz,ir)   * vrt_meanalpha(iz,ir)  )
     %                * ( grid_z(iz+1,ir) - grid_z(iz,ir) ) 
                  tau(iz) = tau(iz+1) + 0.5*(vrt_meanalpha(iz+1,ir)+
     %                    vrt_meanalpha(iz,ir)) 
     %                   * ( grid_z(iz+1,ir) - grid_z(iz,ir) ) 
              enddo
c
c             Determine the temperature from the mean intensity
c
              ts = vrt_primtemp(grid_nz,ir)
              do iz=1,grid_nz
                  if(vrt_meanint(iz,ir).lt.0.d0) then
                      stop 554
                  endif
c         
c                 Solve the grain temperature
c
                  vrt_temp(iz,ir) = find_temperature(
     %                 vrt_meanint(iz,ir),ts,tau(iz),
     %                 qstarfpi(iz)+qviscfpi(iz),kappaav)
c
c                 Put the kappaav into the new meanalpha
c
                  if(freq_nr.le.1) then
                      vrt_meanalpha(iz,ir) = kappaav * vrt_rho(iz,ir)
                  endif
              enddo
c
c             Now check the error
c
              error = 0.d0
              do iz=1,grid_nz
                  error=max(error,abs(tempold(iz)-vrt_temp(iz,ir)))
              enddo
              if(error.lt.1.d-2) goto 33
              do iz=1,grid_nz
                  tempold(iz)=vrt_temp(iz,ir)
              enddo
c
          enddo ! End iteration
 33       continue
      enddo ! End vertical transfer at this radius
c
      end



c     --------------------------------------------------------------
c                THE BLACKBODY PLANCK FUNCTION B_nu(T)
c
c     This function computes the Blackbody function 
c
c                    2 h nu^3 / c^2
c        B_nu(T)  = -------------------    [ erg / cm^2 s ster Hz ]
c                   exp(-h nu / kT) - 1
c
c     ARGUMENTS:
c        nu    [Hz]            = Frequency
c        temp  [K]             = Electron temperature
c     --------------------------------------------------------------
      function bplanck(temp,nu)
      implicit none
      doubleprecision temp
      doubleprecision nu
      doubleprecision bplanck
c
      if(temp.eq.0.d0) then 
          bplanck = 0.d0
          return
      endif
c
      bplanck = 1.47455d-47 * nu * nu * nu / 
     %      (exp(4.7989d-11 * nu / temp)-1.d0) + 1.d-290
c
      return
      end




c     --------------------------------------------------------------
c                COMPUTE THE ABSORBED PRIMARY POWER
c     --------------------------------------------------------------
      function findkappabnu(temp)
      implicit none
      doubleprecision findkappabnu,temp
c
#include "common_grid.h"
#include "common_dust.h"
c
      doubleprecision dummy,kappa,kapprev,kapcur,find_dust_kappa
      doubleprecision nu,bplanck
      integer inu
c
      dummy   = 0.d0
      kapcur = find_dust_kappa(1,1,1,temp,1,0)
      do inu=2,freq_nr
          kapprev = kapcur
          kapcur  = find_dust_kappa(inu,1,1,temp,1,0)
          nu      = 0.5*(freq_nu(inu)+freq_nu(inu-1))
          kappa   = 0.5*(kapprev+kapcur)
          dummy   = dummy + kappa * bplanck(temp,nu)
     %              * (freq_nu(inu)-freq_nu(inu-1))
      enddo
      findkappabnu = dummy
      return
c
      end



c     --------------------------------------------------------------
c                THE VISCOUS DISSIPATION HEATING RATE
c     
c     This function returns the Q_visc/rho in units of [erg/s/g].
c     At the moment, only a simple alpha-viscosity law is included.
c     This function requires the sd_**** common variables to be set.
c
c     The viscous heating is the heating according to a radial
c     accretion velocity v_R of:
c
c                  3 nu
c        v_R  =  - - --
c                  2 R
c
c     This produces a viscous dissipation of:
c
c               9
c        q_+  = - nu Omega_K^2
c               4
c
c     which is the result that is returned. 
c
c     Here we simply have an alpha-viscosity prescription of the form:
c
c               2        c_s^2
c        nu  =  - alpha -------
c               3       Omega_K
c
c
c     INPUT:
c       temp               The gas temperature       
c
c     OUTPUT:
c       viscousheating     The Q_visc / rho : specific heating rate
c       
c##### IS HERE A BUG? #####
c# I notice that the gastodust ratio is not implemented in the viscous
c# heating. I find this strange... This should be checked.
c# Not yet fixed.
c# 3-10-09
c     --------------------------------------------------------------
      function viscousheating(temp,alpha,omk)
      implicit none
      doubleprecision viscousheating,omk,temp,alpha
c
#     include "common_natconst.h"
c
      doubleprecision mugas,viscnu
      parameter(mugas=2.35)
c
      if(alpha.gt.0.d0) then
          viscnu         = (2.d0/3.d0)*alpha*
     %                  (8.3144d7*temp/mugas)/omk
          viscousheating = (9.d0/4.d0)*omk*omk*viscnu
      endif
      return
      end


c     --------------------------------------------------------------
c                    WRITE THE MAIN OUTPUT FILE
c     --------------------------------------------------------------
      subroutine write_temperatures()
      implicit none
c
#include "common_grid.h"
#include "common_vertrt.h"
c
      integer ir,iz
c
      open(unit=1,file='diskstruct_2.dat',status='unknown')
      write(1,*) grid_nr,grid_nz
      write(1,*)
      do ir=1,grid_nr
          do iz=1,grid_nz
              write(1,10) grid_r(ir),grid_z(iz,ir),
     %                   vrt_rho(iz,ir)*GASTODUST,
     %                   vrt_temp(iz,ir),vrt_vz(iz,ir)
 10           format(5(E13.6,1X))
          enddo
      enddo
      close(1)
c
      end


c     --------------------------------------------------------------
c                  WRITE SOME OF THE AUXILIARY DATA
c     --------------------------------------------------------------
      subroutine write_aux_data()
      implicit none
c
#include "common_grid.h"
#include "common_vertrt.h"
c
      integer ir,iz
c
      open(unit=1,file='vertrt.dat',status='unknown')
      write(1,*) grid_nr,grid_nz
      write(1,*)
      do ir=1,grid_nr
          do iz=1,grid_nz
              write(1,10) vrt_hflux(iz,ir),vrt_meanint(iz,ir)
 10           format(2(E13.6,1X))
          enddo
      enddo
      close(1)
c
      end



c     --------------------------------------------------------------
c                    READ THE PLANCK MEAN OPACITY
c     --------------------------------------------------------------
      subroutine read_planckopac
      implicit none
c
#     include "common_grid.h"
#     include "common_vertrt.h"
c
      doubleprecision dum1,dum2,dum3,dum4
      integer it,nt
      logical fex
c
      open(unit=1,file='planckopac.inp')
      read(1,*) dopac_nt
      if(dopac_nt.gt.NR_TEMPERATURE) then
          write(*,*) 'Too many temperatures in planckopac.inp'
          write(*,*) ' Increase NR_TEMPERATURE and recompile'
          stop 13
      endif
      do it=1,dopac_nt
          read(1,*) dum1,dum2,dum3,dum4
          dopac_temp(it)=dum1
          dopac_kappa(it)=dum2
          dopac_kscat(it)=dum3
          dopac_gscat(it)=dum4
      enddo
      close(1)
c
      end



c     --------------------------------------------------------------
c                  INTERPOLATE THE PLANCK MEAN OPACITY
c     --------------------------------------------------------------
      function find_planckopac(temp)
      implicit none
      doubleprecision find_planckopac,temp
c
#     include "common_grid.h"
#     include "common_vertrt.h"
c
      doubleprecision dum1,dum2,dt
      integer it
c
      if(dopac_nt.le.1) then 
          find_planckopac = dopac_kappa(1)
          return
      endif
c
      if(temp.lt.dopac_temp(1)) then 
          find_planckopac = dopac_kappa(1)
      elseif(temp.gt.dopac_temp(dopac_nt)) then
          find_planckopac = dopac_kappa(dopac_nt)
      else
          call hunt(dopac_temp,dopac_nt,temp,it)
          dt = (temp-dopac_temp(it))/
     %            (dopac_temp(it+1)-dopac_temp(it))
          if((dt.lt.0.d0).or.(dt.gt.1.d0)) stop 552
          find_planckopac = (1.d0-dt)*dopac_kappa(it)+
     %       dt*dopac_kappa(it+1)
      endif
c
      return
      end


c     --------------------------------------------------------------
c                INTERPOLATE THE PLANCK MEAN SCAT OPAC
c     --------------------------------------------------------------
      subroutine find_planckall(temp,cabs,cscat,g)
      implicit none
      doubleprecision temp,cabs,cscat,g
c
#     include "common_grid.h"
#     include "common_vertrt.h"
c
      doubleprecision dum1,dum2,dt
      integer it
c
      if(dopac_nt.le.1) then 
          cabs  = dopac_kappa(1)
          cscat = dopac_kscat(1)
          g     = dopac_gscat(1)
          return
      endif
c
      if(temp.lt.dopac_temp(1)) then 
          cabs  = dopac_kappa(1)
          cscat = dopac_kscat(1)
          g     = dopac_gscat(1)
      elseif(temp.gt.dopac_temp(dopac_nt)) then
          cabs  = dopac_kappa(dopac_nt)
          cscat = dopac_kscat(dopac_nt)
          g     = dopac_gscat(dopac_nt)
      else
          call hunt(dopac_temp,dopac_nt,temp,it)
          dt = (temp-dopac_temp(it))/
     %            (dopac_temp(it+1)-dopac_temp(it))
          if((dt.lt.0.d0).or.(dt.gt.1.d0)) stop 552
          cabs  = (1.d0-dt)*dopac_kappa(it)+
     %       dt*dopac_kappa(it+1)
          cscat = (1.d0-dt)*dopac_kscat(it)+
     %       dt*dopac_kscat(it+1)
          g     = (1.d0-dt)*dopac_gscat(it)+
     %       dt*dopac_gscat(it+1)
      endif
c
      return
      end


c     --------------------------------------------------------------
c                    READ THE ROSSELAND MEAN OPACITY
c     --------------------------------------------------------------
      subroutine read_rossopac
      implicit none
c
#     include "common_grid.h"
#     include "common_vertrt.h"
c
      doubleprecision dum1,dum2,dum3,dum4
      integer it,nt
      logical fex
c
      inquire(file='rossopac.inp',exist=fex)
      if(.not.fex) then
          write(*,*) 'ERROR:'
          write(*,*) '  Also need file rossopac.inp!!'
          stop 13
      endif
      open(unit=1,file='rossopac.inp')
      read(1,*) ropac_nt
      if(ropac_nt.gt.NR_TEMPERATURE) then
          write(*,*) 'Too many temperatures in rossopac.inp'
          write(*,*) ' Increase NR_TEMPERATURE and recompile'
          stop 13
      endif
      do it=1,ropac_nt
          read(1,*) dum1,dum2,dum3,dum4
          ropac_temp(it)=dum1
          ropac_kappa(it)=dum2
          ropac_kscat(it)=dum3
          ropac_gscat(it)=dum4
      enddo
      close(1)
c
      end



c     --------------------------------------------------------------
c                  INTERPOLATE THE PLANCK MEAN OPACITY
c     --------------------------------------------------------------
      function find_rossopac(temp)
      implicit none
      doubleprecision find_rossopac,temp
c
#     include "common_grid.h"
#     include "common_vertrt.h"
c
      doubleprecision dum1,dum2,dt
      integer it
c
      if(ropac_nt.le.1) then 
          find_rossopac = ropac_kappa(1)
          return
      endif
c
      if(temp.lt.ropac_temp(1)) then 
          find_rossopac = ropac_kappa(1)
      elseif(temp.gt.ropac_temp(ropac_nt)) then
          find_rossopac = ropac_kappa(ropac_nt)
      else
          call hunt(ropac_temp,ropac_nt,temp,it)
          dt = (temp-ropac_temp(it))/
     %            (ropac_temp(it+1)-ropac_temp(it))
          if((dt.lt.0.d0).or.(dt.gt.1.d0)) stop 552
          find_rossopac = (1.d0-dt)*ropac_kappa(it)+
     %       dt*ropac_kappa(it+1)
      endif
c
      return
      end



c     --------------------------------------------------------------
c                     SOLVE TEMPERATURE FROM J
c     --------------------------------------------------------------
      function find_temperature(intmean,ts,tau,qfpi,kappaav)
      implicit none
      doubleprecision intmean,ts,tau,qfpi,kappaav
      doubleprecision find_temperature,find_planckopac
c
      doubleprecision xp,plkaps,ti,kappap,tiold
      doubleprecision planck_mean_dust_kappa
      integer i
c
#     include "common_grid.h"
c
      xp      = exp(-tau)
cc###############
c      xp = 0.d0
c      write(*,*) 'MODIFIED'
cc###############
      if(freq_nr.gt.1) then
          plkaps = planck_mean_dust_kappa(ts,1,1,1,0)
      else
          plkaps = find_planckopac(ts)
      endif
c
      ti = ts
      do i=1,60
          if(freq_nr.gt.1) then
              kappap = planck_mean_dust_kappa(ti,1,1,1,0)
          else
              kappap = find_planckopac(ti)
          endif
          kappaav = 0.5*xp*plkaps+(1.d0-0.5*xp)*kappap
          tiold   = ti
          ti      = 15.3421d0 * ( ( intmean*kappaav + qfpi ) 
     %                  / kappap )**0.25
          if(abs(ti-tiold)/(ti+tiold).lt.TEMPACCUR) goto 10
      enddo
      write(*,*) 'No convergence in temperature'
      stop 733
 10   continue
      find_temperature = ti
c
      return
      end
