c     ==============================================================
c           PROGRAM THAT COMPUTES THE VERTICAL DENSITY STRUCTURE
c
c     This program reads the dust temperatures as computed by the
c     2D Radtrans program, and reads a file "sigma.inp" which contains
c     the surface density as it is supposed to be. Then it will
c     recompute the density structure according to vertical 
c     pressure balance. This will then be written to the file
c     "dustdens.dat". Note that this program needs a file named
c     "starinfo.inp", which contains the stellar mass. 
c
c     NOTE: A better way of implementing multiple species/sizes is
c           to scale the abun() array with the rescaled density
c           structure. This is straightforward to implement, but
c           I have had no time yet (02-03-03) to do this.
c     ==============================================================

#include "configure.h"
#define ZFINE 1


c     --------------------------------------------------------------
c                             MAIN ROUTINE
c     --------------------------------------------------------------
      program main
      implicit none
c
#include "common_struct.h"
c
      doubleprecision zcover,dum
      integer ir,iz,idum
      logical f_ex
c
c     Read the input file
c
c      inquire(file='zcover.inp',exist=f_ex)
c      if(f_ex) then 
c          open(unit=1,file='zcover.inp')
c          read(1,*) zcover    ! Zmax = zcover * Hp
c          close(1)
c      else
c          write(*,*) 'No zcover.inp found. Taking zcover=5.0'
          zcover=5.d0
c      endif
c
      call read_all_data()
      call solve_all_structs(vs_mstar,vs_r,vs_z,vs_nr,vs_nz,
     %                      zcover,vs_sigma,vs_temp,vs_rho)
c
      if(global_ir_start.eq.2) then
          do iz=1,vs_nz
              vs_rho(iz,1) = vs_rhoold(iz,1)
          enddo
      endif
c
      call write_density_struct()
c
      end




c     --------------------------------------------------------------
c                 READ THE TEMPERATURE STRUCTURE IN
c
c     NOTE: The input files have theta in the order from 0 to pi/2
c           i.e. from pole to equator. We are going to read this in
c           in opposite order.
c     --------------------------------------------------------------
      subroutine read_all_data()
      implicit none
c
#include "common_struct.h"
c
      doubleprecision rstar,a,b,c,d,e
      integer nr,nz,ir,iz,nrp,nzp,isize,ispec
      integer iformat
      logical multi,fex
c
c     Read the vertstruct.inp
c
      inquire(file='vertstruct.inp',exist=fex)
      if(fex) then
          open(unit=1,file='vertstruct.inp',status='old')
          read(1,*) iformat
          read(1,*) isp_str
          read(1,*) isz_str
          close(1)
      else
          isp_str = 1
          isz_str = 1
      endif
c
c     Read the starinfo file
c
      open(unit=1,file='starinfo.inp',status='old')
      read(1,*) iformat
      read(1,*) rstar
      read(1,*) vs_mstar
      close(1)      
c
c     Default value for starting parameter
c
cccccc  >>>>> MODIFIED ON 12 DEC 2002 <<<<<     global_ir_start = 2
      global_ir_start = 1
c
c     Read the current temperature structure
c
      open(unit=1,file='diskstruct_2.dat',status='old')
      read(1,*) nr,nz
      nrspec=1
      nrsize(1)=1
      if(nr.lt.1) then
c
c         New format: multiple sizes/species
c
          read(1,*) nr,nz
          read(1,*) nrspec
          if(nrspec.gt.DUST_SPECIES_MAX) then
              write(*,*) 'ERROR: Too many species specified.'
              write(*,*) '       Recompile with larger max specs.'
              stop
          endif
          do ispec=1,nrspec
              read(1,*) nrsize(ispec)
          enddo
          multi=.true.
      else
          nrspec=1
          nrsize(1)=1
          multi=.false.
      endif
      if(nr.gt.FRSIZE_R) then
          write(*,*) 'ERROR in vertstruct:'
          write(*,*) 'R grid too large'
          stop 13
      endif
      if(nz.gt.FRSIZE_Z) then
          write(*,*) 'ERROR in vertstruct:'
          write(*,*) 'Z grid too large'
          stop 13
      endif
      do ir=1,nr
          do iz=1,nz
              if(.not.multi) then
                  read(1,*) a,b,c,d,e
                  vs_r(ir) = a
                  vs_z(iz,ir) = b
                  vs_rhoold(iz,ir) = c
                  vs_temp(iz,ir) = d
                  vs_vz(iz,ir) = e
              else
                  read(1,*) a,b,c,e
                  vs_r(ir) = a
                  vs_z(iz,ir) = b
                  vs_rhoold(iz,ir) = c
                  vs_vz(iz,ir) = e
                  do ispec=1,nrspec
                      read(1,*) a
                      abun(ispec,iz,ir)=a
                  enddo
                  do ispec=1,nrspec
                      do isize=1,nrsize(ispec)
                          read(1,*) d
                          if((ispec.eq.isp_str).and.
     %                       (isize.eq.isz_str)) then
                              vs_temp(iz,ir) = d
                          endif
                      enddo
                  enddo
              endif
          enddo
      enddo
      close(1)
      if(vs_z(1,1).ne.0.d0) then
          write(*,*) '>>> Warning: z(iz=1).ne.0.d0 <<<'
      endif
c
c     Put global_ir_start to 1 if only 1 or 2 shells
c
      if(nr.le.2) then
          global_ir_start = 1
      endif
c
c     Read the Sigma
c
      open(unit=1,file='sigma.inp',status='old')
      read(1,*) nrp
      if(nrp.ne.nr) then
          write(*,*) 'ERROR in vertstruct:'
          write(*,*) 'file sigma.inp: nr neq nr radius.dat'
          stop 13
      endif
      do ir=1,nr
          read(1,*) vs_sigma(ir)
      enddo      
      close(1)
c
      vs_nr = nr
      vs_nz = nz
c
      end



c     --------------------------------------------------------------
c                 WRITE THE DENSITY STRUCTURE TO FILE
c     --------------------------------------------------------------
      subroutine write_density_struct()
      implicit none
c
#include "common_struct.h"
c
      integer ir,iz,ispec,isize
      logical multi
c
      multi=.false.
      if((nrspec.gt.1).or.(nrsize(1).gt.1)) then
          multi=.true.
          write(*,*) 'WARNING: Multiple species/size not yet ',
     %                'self-consistently implemented'
          write(*,*) '         in the vertstruct code!!!'
          write(*,*) '         Using ad-hoc method...'
      endif
c
      open(unit=1,file='diskstruct_3.dat',status='unknown')
      if(.not.multi) then
          write(1,*) vs_nr,vs_nz
      else
          write(1,*) -1,-1
          write(1,*) vs_nr,vs_nz
          write(1,*) nrspec
          do ispec=1,nrspec
              write(1,*) nrsize(ispec)
          enddo
      endif
      write(1,*) 
      do ir=1,vs_nr
          do iz=1,vs_nz
              if(.not.multi) then
                  write(1,10) vs_r(ir),vs_z(iz,ir),
     %                  vs_rho(iz,ir)+1d-95,
     %                  vs_temp(iz,ir),vs_vz(iz,ir)
 10               format(5(E13.6,1X))
              else
                  write(1,20) vs_r(ir),vs_z(iz,ir),
     %                  vs_rho(iz,ir)+1d-95,vs_vz(iz,ir)
 20               format(3(E13.6,1X))
                  do ispec=1,nrspec
                      write(1,30) abun(ispec,iz,ir)
                  enddo
                  do ispec=1,nrspec
                      do isize=1,nrsize(ispec)
                          write(1,30) 0.d0
                      enddo
                  enddo
 30               format(1(E13.6,1X))
                  
              endif
          enddo
      enddo
      close(1)
c
      end




c     --------------------------------------------------------------
c             PRODUCE THE OVERALL STRUCTURE OF THE DISK
c     --------------------------------------------------------------
      subroutine solve_all_structs(mstar,r,z,nr,nz,
     %                             zcover,sigma,temp,rho)
      implicit none
      doubleprecision r(FRSIZE_R),z(FRSIZE_Z,FRSIZE_R),mstar
      doubleprecision sigma(FRSIZE_R),temp(FRSIZE_Z,FRSIZE_R)
      doubleprecision rho(FRSIZE_Z,FRSIZE_R),zcover
      integer nr,nz
c
#include "common_struct.h"
c
      doubleprecision pi
      integer iwarn,ir,iz
      parameter(pi=3.14159265359)
c
      if(global_ir_start.eq.0) stop 982
      do ir=global_ir_start,nr
c
c         First rescale the z grid at this location to fit with the
c         pressure scale height at the equator. Also put all quantities
c         on this new grid
c
ccc          call rescale_zgrid(mstar,r(ir),zcover,nz,z(1,ir),temp(1,ir))
c
c         Now solve the vertical structure here
c
          call vertstruct_integrate(mstar,r(ir),z(1,ir),nz,
     %                 temp(1,ir),sigma(ir),rho(1,ir),iwarn)
      enddo
c
      end




c     --------------------------------------------------------------
c                     RESCALE THE Z GRID
c     --------------------------------------------------------------
      subroutine rescale_zgrid(mstar,r,cover,nz,z,temp)
      implicit none
      integer nz
      doubleprecision mstar,r,z(nz),temp(nz),cover
c
#include "common_natconst.h"
c
      doubleprecision zmax,zold(nz),told(nz),fact
      integer iz
c
      write(*,*) 'Rescaling contains bug!!'
      stop 55
c
c     Store the z grid and the temperature
c
      do iz=1,nz
          zold(iz) = z(iz)
          told(iz) = temp(iz)
      enddo
c     
c     Find the scaling factor
c
      zmax = cover*sqrt(kk*temp(1)*r**3/(muh2*mp*GG*mstar))
      fact = zmax/z(nz)
c
c     Scale the z grid
c
      do iz=1,nz
          z(iz) = z(iz)*fact
      enddo
c
c     Regrid the temperature to the new z grid
c
      call regridlin(nz,zold,told,nz,z,temp)
c     
      end


c     --------------------------------------------------------------
c                 INTEGRATE THE VERTICAL STRUCTURE 
c
c     The equation to solve is:
c
c      dlg(rho)     / mu mp G M  z    dlg(T) \
c      -------- = - | ---------  -  + ------ |
c        dz         \   k R^3    T      dz   /
c     
c     --------------------------------------------------------------
      subroutine vertstruct_integrate(mstar,r,z,nz,temp,sigma,rho,iwarn)
      implicit none
      doubleprecision mstar,r,z(1:FRSIZE_Z),sigma
      doubleprecision temp(1:FRSIZE_Z),rho(1:FRSIZE_Z)
      integer iwarn,nz
c
#include "common_natconst.h"
c
      doubleprecision zfine(1:ZFINE*(FRSIZE_Z))
      doubleprecision tempfine(1:ZFINE*(FRSIZE_Z))
      doubleprecision rhofine(1:ZFINE*(FRSIZE_Z))
      doubleprecision mugas,dx,dlgt,grv,gravc,sig0,dz
      integer izfine,iz,nzfine
c
c     Check
c
      if(nz.gt.FRSIZE_Z) then
          write(*,*) 'ERROR in vertstruct_integrate()'
          stop 13
      endif
c
      mugas = muh2
c
c     Compute some constants
c
      gravc = (GG*mstar/r**3)*mugas*mp/kk
c      write(*,*) gravc
c
c     First make the zfine grid and put the temp on it
c
      do iz=1,nz-1
          do izfine=1,ZFINE
              zfine((iz-1)*ZFINE+izfine) = 
     %            z(iz)+(izfine-1)*(z(iz+1)-z(iz))/ZFINE
              tempfine((iz-1)*ZFINE+izfine) = 
     %            temp(iz)+(izfine-1)*(temp(iz+1)-temp(iz))/ZFINE
          enddo
      enddo
      zfine((nz-1)*ZFINE+1) = z(nz)
      tempfine((nz-1)*ZFINE+1) = temp(nz)
      nzfine=(nz-1)*ZFINE+1
c
c     Now perform the integration 
c
      rhofine(1) = 1.d0
      do iz=1,nzfine-1
          dz   = (zfine(iz+1)-zfine(iz))
          dlgt = (log(tempfine(iz+1))-log(tempfine(iz))) / dz
          grv  = 0.5 * gravc * ( zfine(iz+1)/tempfine(iz+1) +
     %               zfine(iz)/tempfine(iz) )
          rhofine(iz+1) = rhofine(iz) * exp ( -dz*(grv+dlgt) )
      enddo
c
c     Now find the Sigma for this rho
c
      sig0 = 0.d0
      do iz=1,nzfine-1
          dz   = (zfine(iz+1)-zfine(iz))
          sig0 = sig0 + 0.5*(rhofine(iz+1)+rhofine(iz))*dz
      enddo
      sig0 = sig0*2       ! Both disk sides
c
c     Normalize the rho
c
      do iz=1,nzfine
          rhofine(iz) = rhofine(iz) * sigma / sig0
      enddo
c
c     Now put onto coarse grid
c
      do iz=1,nz
          rho(iz) = rhofine((iz-1)*ZFINE+1)
      enddo
c
c     Check
c
      if(abs(rho(2)-rho(1))/(rho(2)+rho(1)).gt.0.3) then
          iwarn=1
      else 
          iwarn=0
      endif
c
      end


c     --------------------------------------------------------------
c                     REGRID A DISCRETIZED FUNCTION 
c                           (LINEAR VERSION)
c     --------------------------------------------------------------
      subroutine regridlin(n,x,y,nnew,xnew,ynew)
      implicit none
      integer n,nnew
      doubleprecision x(n),y(n),xnew(nnew),ynew(nnew),d
      integer ix,ixnew
      do ixnew=1,nnew
          call hunt(x,n,xnew(ixnew),ix)
          if(ix.le.0) then
              ynew(ixnew) = y(1)
          elseif(ix.ge.n) then
              ynew(ixnew) = y(n)
          else
              d = (xnew(ixnew)-x(ix))/(x(ix+1)-x(ix))
              ynew(ixnew) = (1.d0-d)*y(ix)+d*y(ix+1)
          endif
      enddo
      end


c     -------------------------------------------------------------
c               HUNT FOR LOCATION OF A NUMBER IN A TABLE
c     -------------------------------------------------------------
      subroutine hunt(xx,n,x,jlo)
      integer jlo,n
      doubleprecision x,xx(n)
      integer inc,jhi,jm
      logical ascnd
      ascnd=xx(n).gt.xx(1)
      if(jlo.le.0.or.jlo.gt.n)then
        jlo=0
        jhi=n+1
        goto 3
      endif
      inc=1
      if(x.ge.xx(jlo).eqv.ascnd)then
1       jhi=jlo+inc
        if(jhi.gt.n)then
          jhi=n+1
        else if(x.ge.xx(jhi).eqv.ascnd)then
          jlo=jhi
          inc=inc+inc
          goto 1
        endif
      else
        jhi=jlo
2       jlo=jhi-inc
        if(jlo.lt.1)then
          jlo=0
        else if(x.lt.xx(jlo).eqv.ascnd)then
          jhi=jlo
          inc=inc+inc
          goto 2
        endif
      endif
3     if(jhi-jlo.eq.1)return
      jm=(jhi+jlo)/2
      if(x.gt.xx(jm).eqv.ascnd)then
        jlo=jm
      else
        jhi=jm
      endif
      goto 3
      end
C  (C) Copr. 1986-92 Numerical Recipes Software =v1.9"217..
